(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * encoding_example.thy - 
 * Copyright (C) 2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This File is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


header {* Encoding Recursive Data Structures *}

theory digraph 
imports 
  OCL
begin


text {* The invariant proofs in this theory work only with smashed collection types. *}
text {*


In the following we describe which definition and theorems are
generated during the import of this simple class diagram into \holocl.

\begin{figure}
  \begin{center}
    \includegraphics[scale=.6]{figures/digraph}
  \end{center}
  \caption{A Simple Class Diagram for Digraphs\label{fig:digraph}}
\end{figure}
\begin{table}
  \begin{small}
    \lstinputlisting[style=ocl]{digraph.ocl}
  \end{small}
  \caption{OCL specification\label{tab:digraph}} 
\end{table}




*}


ML {* rep_encoder_common.log_level := 10 *}

text_raw{* \label{thy:digraph} *} 

load_xmi "digraph_ocl.xmi"

 text {* \verb|UML_OCL.univ.digraph.Node_base_type|: @{typ [display] UML_OCL.univ.digraph.Node_base_type} *} 
 text {* \verb|UML_OCL.univ.digraph.Node_type|: @{typ [display] UML_OCL.univ.digraph.Node_type} *}

 text {* \verb|UML_OCL.univ.digraph.Node_ext|: @{typ [display] "'a UML_OCL.univ.digraph.Node_ext"} *} 
 text {* \verb|UML_OCL.univ.digraph.Cnode_base_type|: @{typ [display] UML_OCL.univ.digraph.Cnode_base_type} *} 
 text {* \verb|UML_OCL.univ.digraph.Cnode_type|: @{typ [display] UML_OCL.univ.digraph.Cnode_type} *} 
 text {* \verb|UML_OCL.univ.digraph.Cnode_ext|: @{typ [display] "'a UML_OCL.univ.digraph.Cnode_ext"} *} 
 text {* \verb|U_digraph|: @{typ [display] "('a,'b,'c) U_digraph"} *} 
 text {* \verb|UML_OCL.univ.digraph.Node|: @{typ [display] "('a,'b) UML_OCL.univ.digraph.Node"} *} 
 text {* \verb|UML_OCL.univ.digraph.Cnode|: @{typ [display] "('a) UML_OCL.univ.digraph.Cnode"} *} 
 text {* \verb|UML_OCL.level0.digraph.mk_Node|: @{const [display,show_types] UML_OCL.level0.digraph.mk_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.get_Node|: @{const [display,show_types] UML_OCL.level0.digraph.get_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.is_Node|: @{const [display,show_types] UML_OCL.level0.digraph.is_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.isType_Node|: @{const [display,show_types] UML_OCL.level0.digraph.isType_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.is_Node_univ|: @{const [display,show_types] UML_OCL.level0.digraph.is_Node_univ} *} 
 text {* \verb|UML_OCL.level0.digraph.Node|: @{const [display,show_types] UML_OCL.level0.digraph.Node} *} 
 text {* \verb|UML_OCL.level0.digraph.Node_2_OclAny|: @{const [display,show_types] UML_OCL.level0.digraph.Node_2_OclAny} *} 
 text {* \verb|UML_OCL.level0.digraph.OclAny_2_Node|: @{const [display,show_types] UML_OCL.level0.digraph.OclAny_2_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.is_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.is_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.is_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.isType_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.isType_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.isType_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.is_Node_univ_def|: @{thm [display,show_types] UML_OCL.level0.digraph.is_Node_univ_def} *} 
      (*<*) thm UML_OCL.level0.digraph.is_Node_univ_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.get_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.get_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.get_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.mk_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.mk_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.mk_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Node_2_OclAny_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node_2_OclAny_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node_2_OclAny_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.OclAny_2_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.OclAny_2_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.OclAny_2_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.mk_Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.mk_Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.get_Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.get_Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.is_Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.is_Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.isType_Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.isType_Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.is_Cnode_univ|: @{const [display,show_types] UML_OCL.level0.digraph.is_Cnode_univ} *} 
 text {* \verb|UML_OCL.level0.digraph.Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.Cnode_2_Node|: @{const [display,show_types] UML_OCL.level0.digraph.Cnode_2_Node} *} 
 text {* \verb|UML_OCL.level0.digraph.Node_2_Cnode|: @{const [display,show_types] UML_OCL.level0.digraph.Node_2_Cnode} *} 
 text {* \verb|UML_OCL.level0.digraph.Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.is_Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.is_Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.is_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.isType_Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.isType_Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.isType_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.is_Cnode_univ_def|: @{thm [display,show_types] UML_OCL.level0.digraph.is_Cnode_univ_def} *} 
      (*<*) thm UML_OCL.level0.digraph.is_Cnode_univ_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.get_Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.get_Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.get_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.mk_Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.mk_Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.mk_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Cnode_2_Node_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Cnode_2_Node_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Cnode_2_Node_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Node_2_Cnode_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node_2_Cnode_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node_2_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Node.i|: @{const [display,show_types] UML_OCL.level0.digraph.Node.i} *} 
 text {* \verb|UML_OCL.level0.digraph.Node.left|: @{const [display,show_types] UML_OCL.level0.digraph.Node.left} *} 
 text {* \verb|UML_OCL.level0.digraph.Node.right|: @{const [display,show_types] UML_OCL.level0.digraph.Node.right} *} 
 text {* \verb|UML_OCL.level0.digraph.Node.i_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node.i_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node.i_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Node.left_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node.left_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node.left_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Node.right_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Node.right_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Node.right_def (*>*) 
 text {* \verb|UML_OCL.level0.digraph.Cnode.color|: @{const [display,show_types] UML_OCL.level0.digraph.Cnode.color} *} 
 text {* \verb|UML_OCL.level0.digraph.Cnode.color_def|: @{thm [display,show_types] UML_OCL.level0.digraph.Cnode.color_def} *} 
      (*<*) thm UML_OCL.level0.digraph.Cnode.color_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.mk_Node|: @{const [display,show_types] UML_OCL.level1.digraph.mk_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.get_Node|: @{const [display,show_types] UML_OCL.level1.digraph.get_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.is_Node|: @{const [display,show_types] UML_OCL.level1.digraph.is_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.isType_Node|: @{const [display,show_types] UML_OCL.level1.digraph.isType_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.is_Node_univ|: @{const [display,show_types] UML_OCL.level1.digraph.is_Node_univ} *} 
 text {* \verb|UML_OCL.level1.digraph.Node_2_OclAny|: @{const [display,show_types] UML_OCL.level1.digraph.Node_2_OclAny} *} 
 text {* \verb|UML_OCL.level1.digraph.OclAny_2_Node|: @{const [display,show_types] UML_OCL.level1.digraph.OclAny_2_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.Node_type|: @{const [display,show_types] UML_OCL.level1.digraph.Node_type} *} 
 text {* \verb|UML_OCL.level1.digraph.Node|: @{const [display,show_types] UML_OCL.level1.digraph.Node} *} 
 text {* \verb|UML_OCL.level1.digraph.Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.get_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.get_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.get_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.mk_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.mk_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.mk_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.is_Node_univ_def|: @{thm [display,show_types] UML_OCL.level1.digraph.is_Node_univ_def} *} 
      (*<*) thm UML_OCL.level1.digraph.is_Node_univ_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node_type_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node_type_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node_type_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node_2_OclAny_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node_2_OclAny_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node_2_OclAny_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.OclAny_2_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.OclAny_2_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.OclAny_2_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.is_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.is_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.is_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.isType_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.isType_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.isType_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.mk_Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.mk_Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.get_Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.get_Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.is_Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.is_Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.isType_Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.isType_Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.is_Cnode_univ|: @{const [display,show_types] UML_OCL.level1.digraph.is_Cnode_univ} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode_2_Node|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode_2_Node} *} 
 text {* \verb|UML_OCL.level1.digraph.Node_2_Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.Node_2_Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode_type|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode_type} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.get_Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.get_Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.get_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.mk_Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.mk_Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.mk_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.is_Cnode_univ_def|: @{thm [display,show_types] UML_OCL.level1.digraph.is_Cnode_univ_def} *} 
      (*<*) thm UML_OCL.level1.digraph.is_Cnode_univ_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode_type_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode_type_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode_type_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode_2_Node_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode_2_Node_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode_2_Node_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node_2_Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node_2_Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node_2_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.is_Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.is_Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.is_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.isType_Cnode_def|: @{thm [display,show_types] UML_OCL.level1.digraph.isType_Cnode_def} *} 
      (*<*) thm UML_OCL.level1.digraph.isType_Cnode_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.i|: @{const [display,show_types] UML_OCL.level1.digraph.Node.i} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.i_pre|: @{const [display,show_types] UML_OCL.level1.digraph.Node.i_pre} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.left_pre|: @{const [display,show_types] UML_OCL.level1.digraph.Node.left_pre} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.right_pre|: @{const [display,show_types] UML_OCL.level1.digraph.Node.right_pre} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.i_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.i_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.i_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.i_pre_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.i_pre_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.i_pre_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.left_pre_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.left_pre_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.left_pre_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.right_pre_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.right_pre_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.right_pre_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.color|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.color} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.color_pre|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.color_pre} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.color_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.color_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.color_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.color_pre_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.color_pre_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.color_pre_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_i|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_i} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.inv_3|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.inv_3} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_i_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_i_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.init_i_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.init_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.init_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.init_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.init_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.inv_3_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.inv_3_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.inv_3_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.typeconstraint_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_i|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_i} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.inv_3|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.inv_3} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_i_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_i_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_i_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.init_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.inv_3_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.inv_3_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.inv_3_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_i|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_i} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.inv_3|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.inv_3} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_left|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_left} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_right|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_right} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_i_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_i_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_i_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.init_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.inv_3_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.inv_3_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.inv_3_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_left_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_left_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_left_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_right_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_right_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.typeconstraint_right_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.model.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.model.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.model.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.model.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.model.inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_ref.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_ref.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_ref.inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Node.relax_refval.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Node.relax_refval.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Node.relax_refval.inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.init_color|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.init_color} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.inv_1|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.inv_1} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.init_color_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.init_color_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.hol_invariant.init_color_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.inv_1_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.inv_1_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.hol_invariant.inv_1_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.init_color|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.init_color} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.inv_1|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.inv_1} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.init_color_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.init_color_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.init_color_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.inv_1_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.inv_1_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.inv_1_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.init_color|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.init_color} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.inv_1|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.inv_1} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.check_parent_inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.check_parent_inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.init_color_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.init_color_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.init_color_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.inv_1_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.inv_1_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.inv_1_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.check_parent_inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.check_parent_inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant.check_parent_inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.hol_invariant_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.set_fun|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.set_fun} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.set_fun_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.set_fun_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.set_fun_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.set_fun.mono|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.set_fun.mono} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.set_fun.mono (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.characteristic_set|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.characteristic_set} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.characteristic_set_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.characteristic_set_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.characteristic_set_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.model.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.model.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.model.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.model.inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_ref.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_ref.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_ref.inv_def (*>*) 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.inv|: @{const [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.inv} *} 
 text {* \verb|UML_OCL.level1.digraph.Cnode.relax_refval.inv_def|: @{thm [display,show_types] UML_OCL.level1.digraph.Cnode.relax_refval.inv_def} *} 
      (*<*) thm UML_OCL.level1.digraph.Cnode.relax_refval.inv_def (*>*) 
 text {* \verb|digraph.mk_Node|: @{const [display,show_types] digraph.mk_Node} *} 
 text {* \verb|digraph.get_Node|: @{const [display,show_types] digraph.get_Node} *} 
 text {* \verb|digraph.is_Node|: @{const [display,show_types] digraph.is_Node} *} 
 text {* \verb|digraph.isType_Node|: @{const [display,show_types] digraph.isType_Node} *} 
 text {* \verb|digraph.is_Node_univ|: @{const [display,show_types] digraph.is_Node_univ} *} 
 text {* \verb|digraph.Node_2_OclAny|: @{const [display,show_types] digraph.Node_2_OclAny} *} 
 text {* \verb|digraph.OclAny_2_Node|: @{const [display,show_types] digraph.OclAny_2_Node} *} 
 text {* \verb|digraph.Node|: @{const [display,show_types] digraph.Node} *} 
 text {* \verb|digraph.Node_type|: @{const [display,show_types] digraph.Node_type} *} 
 text {* \verb|digraph.get_Node_def|: @{thm [display,show_types] digraph.get_Node_def} *} 
      (*<*) thm digraph.get_Node_def (*>*) 
 text {* \verb|digraph.mk_Node_def|: @{thm [display,show_types] digraph.mk_Node_def} *} 
      (*<*) thm digraph.mk_Node_def (*>*) 
 text {* \verb|digraph.is_Node_def|: @{thm [display,show_types] digraph.is_Node_def} *} 
      (*<*) thm digraph.is_Node_def (*>*) 
 text {* \verb|digraph.isType_Node_def|: @{thm [display,show_types] digraph.isType_Node_def} *} 
      (*<*) thm digraph.isType_Node_def (*>*) 
 text {* \verb|digraph.is_Node_univ_def|: @{thm [display,show_types] digraph.is_Node_univ_def} *} 
      (*<*) thm digraph.is_Node_univ_def (*>*) 
 text {* \verb|digraph.Node_2_OclAny_def|: @{thm [display,show_types] digraph.Node_2_OclAny_def} *} 
      (*<*) thm digraph.Node_2_OclAny_def (*>*) 
 text {* \verb|digraph.OclAny_2_Node_def|: @{thm [display,show_types] digraph.OclAny_2_Node_def} *} 
      (*<*) thm digraph.OclAny_2_Node_def (*>*) 
 text {* \verb|digraph.Node_def|: @{thm [display,show_types] digraph.Node_def} *} 
      (*<*) thm digraph.Node_def (*>*) 
 text {* \verb|digraph.Node_type_def|: @{thm [display,show_types] digraph.Node_type_def} *} 
      (*<*) thm digraph.Node_type_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_OclAny_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_OclAny_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_OclAny_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_OclAny_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_OclAny_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_OclAny_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Node_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Node_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Node_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_Integer_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_Integer_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_Integer_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_Real_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_Real_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_Real_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_String_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_String_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_String_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_Set_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_Set_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_Set_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_OrderedSet_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_OrderedSet_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_OrderedSet_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_Sequence_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_Sequence_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_Sequence_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.from_Bag_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.from_Bag_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.from_Bag_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Integer_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Integer_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Integer_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Real_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Real_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Real_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_String_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_String_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_String_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Set_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Set_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Set_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_OrderedSet_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_OrderedSet_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_OrderedSet_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Sequence_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Sequence_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Sequence_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.to_Bag_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.to_Bag_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.to_Bag_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_Node_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_Node_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_Node_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_NodeAtpre_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_NodeAtpre_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Node.OclAllInstancesdigraph_NodeAtpre_def (*>*) 
 text {* \verb|digraph.mk_Cnode|: @{const [display,show_types] digraph.mk_Cnode} *} 
 text {* \verb|digraph.get_Cnode|: @{const [display,show_types] digraph.get_Cnode} *} 
 text {* \verb|digraph.is_Cnode|: @{const [display,show_types] digraph.is_Cnode} *} 
 text {* \verb|digraph.isType_Cnode|: @{const [display,show_types] digraph.isType_Cnode} *} 
 text {* \verb|digraph.is_Cnode_univ|: @{const [display,show_types] digraph.is_Cnode_univ} *} 
 text {* \verb|digraph.Cnode_2_Node|: @{const [display,show_types] digraph.Cnode_2_Node} *} 
 text {* \verb|digraph.Node_2_Cnode|: @{const [display,show_types] digraph.Node_2_Cnode} *} 
 text {* \verb|digraph.Cnode|: @{const [display,show_types] digraph.Cnode} *} 
 text {* \verb|digraph.Cnode_type|: @{const [display,show_types] digraph.Cnode_type} *} 
 text {* \verb|digraph.get_Cnode_def|: @{thm [display,show_types] digraph.get_Cnode_def} *} 
      (*<*) thm digraph.get_Cnode_def (*>*) 
 text {* \verb|digraph.mk_Cnode_def|: @{thm [display,show_types] digraph.mk_Cnode_def} *} 
      (*<*) thm digraph.mk_Cnode_def (*>*) 
 text {* \verb|digraph.is_Cnode_def|: @{thm [display,show_types] digraph.is_Cnode_def} *} 
      (*<*) thm digraph.is_Cnode_def (*>*) 
 text {* \verb|digraph.isType_Cnode_def|: @{thm [display,show_types] digraph.isType_Cnode_def} *} 
      (*<*) thm digraph.isType_Cnode_def (*>*) 
 text {* \verb|digraph.is_Cnode_univ_def|: @{thm [display,show_types] digraph.is_Cnode_univ_def} *} 
      (*<*) thm digraph.is_Cnode_univ_def (*>*) 
 text {* \verb|digraph.Cnode_2_Node_def|: @{thm [display,show_types] digraph.Cnode_2_Node_def} *} 
      (*<*) thm digraph.Cnode_2_Node_def (*>*) 
 text {* \verb|digraph.Node_2_Cnode_def|: @{thm [display,show_types] digraph.Node_2_Cnode_def} *} 
      (*<*) thm digraph.Node_2_Cnode_def (*>*) 
 text {* \verb|digraph.Cnode_def|: @{thm [display,show_types] digraph.Cnode_def} *} 
      (*<*) thm digraph.Cnode_def (*>*) 
 text {* \verb|digraph.Cnode_type_def|: @{thm [display,show_types] digraph.Cnode_type_def} *} 
      (*<*) thm digraph.Cnode_type_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Node_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Node_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Node_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_OclAny_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_OclAny_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_OclAny_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_OclAny_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_OclAny_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_OclAny_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Node_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Node_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Node_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Cnode_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Cnode_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Cnode_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Integer_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Integer_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Integer_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Real_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Real_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Real_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_String_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_String_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_String_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Set_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Set_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Set_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_OrderedSet_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_OrderedSet_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_OrderedSet_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Sequence_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Sequence_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Sequence_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.from_Bag_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.from_Bag_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.from_Bag_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Integer_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Integer_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Integer_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Real_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Real_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Real_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_String_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_String_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_String_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Set_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Set_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Set_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_OrderedSet_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_OrderedSet_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_OrderedSet_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Sequence_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Sequence_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Sequence_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.to_Bag_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.to_Bag_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.to_Bag_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_Cnode_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_Cnode_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_Cnode_def (*>*) 
 text {* \verb|UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_CnodeAtpre_def|: @{thm [display,show_types] UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_CnodeAtpre_def} *} 
      (*<*) thm UML_OCL.univ.digraph.Cnode.OclAllInstancesdigraph_CnodeAtpre_def (*>*) 
 text {* \verb|digraph.Node.i|: @{const [display,show_types] digraph.Node.i} *} 
 text {* \verb|digraph.Node.i_pre|: @{const [display,show_types] digraph.Node.i_pre} *} 
 text {* \verb|digraph.Node.left|: @{const [display,show_types] digraph.Node.left} *} 
 text {* \verb|digraph.Node.left_pre|: @{const [display,show_types] digraph.Node.left_pre} *} 
 text {* \verb|digraph.Node.right|: @{const [display,show_types] digraph.Node.right} *} 
 text {* \verb|digraph.Node.right_pre|: @{const [display,show_types] digraph.Node.right_pre} *} 
 text {* \verb|digraph.Node.i_def|: @{thm [display,show_types] digraph.Node.i_def} *} 
      (*<*) thm digraph.Node.i_def (*>*) 
 text {* \verb|digraph.Node.i_pre_def|: @{thm [display,show_types] digraph.Node.i_pre_def} *} 
      (*<*) thm digraph.Node.i_pre_def (*>*) 
 text {* \verb|digraph.Node.left_def|: @{thm [display,show_types] digraph.Node.left_def} *} 
      (*<*) thm digraph.Node.left_def (*>*) 
 text {* \verb|digraph.Node.left_pre_def|: @{thm [display,show_types] digraph.Node.left_pre_def} *} 
      (*<*) thm digraph.Node.left_pre_def (*>*) 
 text {* \verb|digraph.Node.right_def|: @{thm [display,show_types] digraph.Node.right_def} *} 
      (*<*) thm digraph.Node.right_def (*>*) 
 text {* \verb|digraph.Node.right_pre_def|: @{thm [display,show_types] digraph.Node.right_pre_def} *} 
      (*<*) thm digraph.Node.right_pre_def (*>*) 
 text {* \verb|digraph.Cnode.color|: @{const [display,show_types] digraph.Cnode.color} *} 
 text {* \verb|digraph.Cnode.color_pre|: @{const [display,show_types] digraph.Cnode.color_pre} *} 
 text {* \verb|digraph.Cnode.color_def|: @{thm [display,show_types] digraph.Cnode.color_def} *} 
      (*<*) thm digraph.Cnode.color_def (*>*) 
 text {* \verb|digraph.Cnode.color_pre_def|: @{thm [display,show_types] digraph.Cnode.color_pre_def} *} 
      (*<*) thm digraph.Cnode.color_pre_def (*>*) 
 text {* \verb|digraph.Node.strong_sve_def|: @{thm [display,show_types] digraph.Node.strong_sve_def} *} 
      (*<*) thm digraph.Node.strong_sve_def (*>*) 
 text {* \verb|digraph.Node.strict_sve_def|: @{thm [display,show_types] digraph.Node.strict_sve_def} *} 
      (*<*) thm digraph.Node.strict_sve_def (*>*) 
 text {* \verb|digraph.Cnode.strong_sve_def|: @{thm [display,show_types] digraph.Cnode.strong_sve_def} *} 
      (*<*) thm digraph.Cnode.strong_sve_def (*>*) 
 text {* \verb|digraph.Cnode.strict_sve_def|: @{thm [display,show_types] digraph.Cnode.strict_sve_def} *} 
      (*<*) thm digraph.Cnode.strict_sve_def (*>*) 
 text {* \verb|digraph.Node.inv.init_i|: @{const [display,show_types] digraph.Node.inv.init_i} *} 
 text {* \verb|digraph.Node.inv.init_left|: @{const [display,show_types] digraph.Node.inv.init_left} *} 
 text {* \verb|digraph.Node.inv.init_right|: @{const [display,show_types] digraph.Node.inv.init_right} *} 
 text {* \verb|digraph.Node.inv.inv_3|: @{const [display,show_types] digraph.Node.inv.inv_3} *} 
 text {* \verb|digraph.Node.inv.init_i_def|: @{thm [display,show_types] digraph.Node.inv.init_i_def} *} 
      (*<*) thm digraph.Node.inv.init_i_def (*>*) 
 text {* \verb|digraph.Node.inv.init_left_def|: @{thm [display,show_types] digraph.Node.inv.init_left_def} *} 
      (*<*) thm digraph.Node.inv.init_left_def (*>*) 
 text {* \verb|digraph.Node.inv.init_right_def|: @{thm [display,show_types] digraph.Node.inv.init_right_def} *} 
      (*<*) thm digraph.Node.inv.init_right_def (*>*) 
 text {* \verb|digraph.Node.inv.inv_3_def|: @{thm [display,show_types] digraph.Node.inv.inv_3_def} *} 
      (*<*) thm digraph.Node.inv.inv_3_def (*>*) 
 text {* \verb|digraph.Node.inv.inv|: @{const [display,show_types] digraph.Node.inv.inv} *} 
 text {* \verb|digraph.Node.inv.inv_def|: @{thm [display,show_types] digraph.Node.inv.inv_def} *} 
      (*<*) thm digraph.Node.inv.inv_def (*>*) 
 text {* \verb|digraph.Cnode.inv.init_color|: @{const [display,show_types] digraph.Cnode.inv.init_color} *} 
 text {* \verb|digraph.Cnode.inv.inv_1|: @{const [display,show_types] digraph.Cnode.inv.inv_1} *} 
 text {* \verb|digraph.Cnode.inv.init_color_def|: @{thm [display,show_types] digraph.Cnode.inv.init_color_def} *} 
      (*<*) thm digraph.Cnode.inv.init_color_def (*>*) 
 text {* \verb|digraph.Cnode.inv.inv_1_def|: @{thm [display,show_types] digraph.Cnode.inv.inv_1_def} *} 
      (*<*) thm digraph.Cnode.inv.inv_1_def (*>*) 
 text {* \verb|digraph.Cnode.inv.inv|: @{const [display,show_types] digraph.Cnode.inv.inv} *} 
 text {* \verb|digraph.Cnode.inv.inv_def|: @{thm [display,show_types] digraph.Cnode.inv.inv_def} *} 
      (*<*) thm digraph.Cnode.inv.inv_def (*>*) 




section {* Playing with Equalities *}

lemma sand2and: "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x; \<tau> \<MathOclValid> \<MathOclIsDefined> y \<rbrakk> \<Longrightarrow> (\<tau> \<MathOclValid> (x \<MathOclSand>  y)) = (\<tau> \<MathOclValid> (x \<MathOclAnd> y))"
  apply(simp add: OclAnd_def OclSand_def OclIsDefined_def localValid2sem
                  DEF_def lift1_def lift2_def strictify_def)
done

thm digraph.Node.strict_sve_def
thm digraph.Cnode.strict_sve_def
(* 
   lemma "((x::(('a,'b,'c) U_digraph St \<Rightarrow>  'a Cnode)) \<MathOclStrictValueEq> (y::(('a,'b,'c) U_digraph St \<Rightarrow>  'a Cnode)))  
       =  (((x ->oclAsType(level1.digraph.Node)) \<MathOclStrictValueEq> (y ->oclAsType(level1.digraph.Node)))
           \<MathOclSand>  
           (level1.digraph.Cnode.color x)  \<MathOclStrictEq>  (level1.digraph.Cnode.color y))"
*)
lemma "((x::(('a,'b,'c) U_digraph St \<Rightarrow>  'a Cnode)) \<MathOclStrictValueEq> (y::(('a,'b,'c) U_digraph St \<Rightarrow>  'a Cnode)))  
       =  (((digraph.Cnode_2_Node x) \<MathOclStrictValueEq> (digraph.Cnode_2_Node y))
           \<MathOclSand>  
           (level1.digraph.Cnode.color x)  \<MathOclStrictEq>  (level1.digraph.Cnode.color y))"
apply(simp add: digraph.Cnode.strict_sve_def)
oops

thm digraph.Node.strong_sve_def
thm digraph.Cnode.strong_sve_def

section {* Characterstic Sets for Recursive Object Structures *}

lemmas  Node_model_inv_simps = level1.digraph.Node.model.hol_invariant_def
    level1.digraph.Node.model.hol_invariant.init_i_def
    level1.digraph.Node.model.hol_invariant.init_left_def
    level1.digraph.Node.model.hol_invariant.init_right_def
    level1.digraph.Node.model.hol_invariant.typeconstraint_left_def
    level1.digraph.Node.model.hol_invariant.typeconstraint_right_def
    level1.digraph.Node.model.hol_invariant.inv_3_def
    level1.digraph.Node.model.hol_invariant.check_parent_inv_def

lemmas  Node_relax_ref_inv_simps = level1.digraph.Node.relax_ref.hol_invariant_def
    level1.digraph.Node.relax_ref.hol_invariant.init_i_def
    level1.digraph.Node.relax_ref.hol_invariant.init_left_def
    level1.digraph.Node.relax_ref.hol_invariant.init_right_def
    level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_left_def
    level1.digraph.Node.relax_ref.hol_invariant.typeconstraint_right_def
    level1.digraph.Node.relax_ref.hol_invariant.inv_3_def
    level1.digraph.Node.relax_ref.hol_invariant.check_parent_inv_def

lemmas  Cnode_model_inv_simps = level1.digraph.Cnode.model.hol_invariant_def
    level1.digraph.Cnode.model.hol_invariant.init_color_def
    level1.digraph.Cnode.model.hol_invariant.inv_1_def
    level1.digraph.Cnode.model.hol_invariant.check_parent_inv_def

lemmas  Cnode_relax_ref_inv_simps = level1.digraph.Cnode.relax_ref.hol_invariant_def
    level1.digraph.Cnode.relax_ref.hol_invariant.init_color_def
    level1.digraph.Cnode.relax_ref.hol_invariant.inv_1_def
    level1.digraph.Cnode.relax_ref.hol_invariant.check_parent_inv_def


(* level_1 relations between liberal and non-liberal invariants *)

lemma Node_set_L_includes_Node_set : 
  "level1.digraph.Node.model.set  \<tau>  \<subseteq> level1.digraph.Node.relax_ref.set  \<tau>"
  apply(simp add: level1.digraph.Node.model.set_def     level1.digraph.Node.relax_ref.set_def)
  apply(rule gfp_mono)
  apply(auto simp: level1.digraph.Node.model.set_fun_def level1.digraph.Node.relax_ref.set_fun_def)
  apply(simp add: Node_model_inv_simps   Node_relax_ref_inv_simps)
  done


lemma Cnode_set_L_includes_Cnode_set : 
  "level1.digraph.Cnode.model.set  \<tau>  \<subseteq> level1.digraph.Cnode.relax_ref.set  \<tau>"
  apply(simp add: level1.digraph.Cnode.model.set_def    level1.digraph.Cnode.relax_ref.set_def)
  apply(rule gfp_mono)
  apply(auto simp: level1.digraph.Cnode.model.set_fun_def level1.digraph.Cnode.relax_ref.set_fun_def)
  apply(simp add: Cnode_model_inv_simps Cnode_relax_ref_inv_simps  
                  Node_model_inv_simps  Node_relax_ref_inv_simps  
  )
  done


lemma "(\<tau> \<MathOclValid> level1.digraph.Node.model.inv obj) \<Longrightarrow> (\<tau> \<MathOclValid> level1.digraph.Node.relax_ref.inv obj) "
oops
   
lemma Node_set_includes_Cnode_set : 
  "((\<lambda> self. self ->oclAsType( level1.digraph.Node)) ` (level1.digraph.Cnode.model.set \<tau>))  
                        \<subseteq> level1.digraph.Node.model.set \<tau>"
  apply(simp add: level1.digraph.Node.model.set_def level1.digraph.Cnode.model.set_def)  
  apply auto
  apply(rule_tac X="(\<lambda> self. self ->oclAsType( level1.digraph.Node)) 
                                                         ` (gfp (Cnode.model.set_fun \<tau>))" 
    in coinduct [OF level1.digraph.Node.model.set_fun.mono])
  apply(auto) 
  apply(subst Node.model.set_fun_def)
  apply simp
  apply(thin_tac "self \<in> gfp (Cnode.model.set_fun \<tau>)")
  apply(subst (asm) gfp_unfold [OF level1.digraph.Cnode.model.set_fun.mono])
  apply(simp)
  apply(subst (asm) Cnode.model.set_fun_def)
  apply(simp add: Cnode_model_inv_simps Node_model_inv_simps)
  apply(auto)
  done

(* Generelle Funktion von "invariant1":
   Akkumulation der constraints auf den Attributen.
   Eigentlich mte auch die Definiertheit *aller*
   Attribute bei uns so erfat werden !!!
   Sonst gelingt nicht der Nachweis, da Node_set \<tau> kein undef
   enthlt !!! *) 
  

(* Bemerkung: Paulson schlgt fr gegenseitige
   Rekursion die Summenkonstruktion vor.
   Das geht bei uns genauso - nur nehmen wir
   gleich OclAny. Siehe Paulsons Papier.
 *)


lemma TheLemmaOneNonRec: "
  (\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow>
(
   model.hol_invariant.init_i \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_left \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_right \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.inv_3 \<tau> (Node.model.set \<tau>) obj 
   \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> (Node.model.set \<tau>) obj
)"
apply(simp_all add:Node_model_inv_simps)
apply(simp add: Node.model.inv_def Node.model.characteristic_set_def
               Node.model.set_def image_def)
apply(subst (asm) gfp_unfold [OF level1.digraph.Node.model.set_fun.mono],simp)
apply(subst (asm) Node.model.set_fun_def)
apply(simp add: Node.model.hol_invariant_def)
apply(simp_all add:Node_model_inv_simps)
apply(drule OCL_Set.includes_charn1, simp)
apply(erule conjE)+
apply(erule exE)
apply(erule conjE)+
apply(erule_tac A1="obj" and \<tau>1="\<tau>" in cp_charn [THEN iffD2])
apply(simp)
apply(assumption)
done


lemma TheLemmaOneRecLeft: "
  (\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow>
(
     ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.left obj))     
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.left obj)) 
)"
apply(unfold Node.model.inv_def Node.model.characteristic_set_def)
apply(drule OCL_Set.includes_charn1)
apply(unfold Node.model.set_def)
apply(subst (asm) gfp_unfold [OF level1.digraph.Node.model.set_fun.mono],simp)
apply(subst (asm) Node.model.set_fun_def)
apply(simp add: Node.model.hol_invariant_def)
apply(subst (asm) image_def)
apply(simp add: model.hol_invariant.typeconstraint_left_def)
apply(erule conjE)+
apply(erule  exE)
apply(erule conjE)+
apply(erule_tac \<tau>1="\<tau>" in cp_charn [THEN iffD2], simp)
apply(erule disjE)
apply(simp)
apply(safe)
apply(rule  OCL_Set.includes_charn2)
apply(simp_all)
 apply(simp add: Node.model.characteristic_set_def OclIsDefined_def OclTrue_def
                  lift0_def lift1_def DEF_def)
apply(drule sym)
apply(simp add:localValidDefined2sem DEF_def)
done

lemma TheLemmaOneRecRight: "
  (\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow>
(
     ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.right obj))     
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.right obj)) 
)"

apply(unfold Node.model.inv_def Node.model.characteristic_set_def)
apply(drule OCL_Set.includes_charn1)
apply(unfold Node.model.set_def)
apply(subst (asm) gfp_unfold [OF level1.digraph.Node.model.set_fun.mono],simp)
apply(subst (asm) Node.model.set_fun_def)
apply(simp add: Node.model.hol_invariant_def)
apply(subst (asm) image_def)
apply(simp add: model.hol_invariant.typeconstraint_right_def)
apply(erule conjE)+
apply(erule  exE)
apply(erule conjE)+
apply(erule_tac \<tau>1="\<tau>" in cp_charn [THEN iffD2], simp)
apply(erule disjE)
apply(simp)
apply(safe)
apply(rule  OCL_Set.includes_charn2)
apply(simp_all)
 apply(simp add: Node.model.characteristic_set_def OclIsDefined_def OclTrue_def
                  lift0_def lift1_def DEF_def)
apply(drule sym)
apply(simp add:localValidDefined2sem DEF_def)
done




lemma TheLemmaOne: "
  (\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow>
(
   model.hol_invariant.init_i \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_left \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_right \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.inv_3 \<tau> (Node.model.set \<tau>) obj 
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.left obj))     
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.left obj))
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.right obj))
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.right obj)) 
   \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> (Node.model.set \<tau>) obj
)"
apply(simp add: TheLemmaOneNonRec TheLemmaOneRecLeft TheLemmaOneRecRight)
done


lemma OCL_is_def_Node_i: " \<tau> \<MathOclValid> \<MathOclIsDefined> (level1.digraph.Node.i obj) \<Longrightarrow>  \<tau> \<MathOclValid> \<MathOclIsDefined> obj"
  by(auto simp: level1.digraph.Node.i_def localValidDefined2sem 
    lift1_def strictify_def DEF_def)

lemma TheLemmaTwo: "
(
   model.hol_invariant.init_i \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_left \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_right \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.inv_3 \<tau> (Node.model.set \<tau>) obj 
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.left obj))     
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.left obj))
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.right obj))
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.right obj)) 
   \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> (Node.model.set \<tau>) obj
)
\<Longrightarrow>
  (\<tau> \<MathOclValid>  Node.model.inv obj) 
"


apply(simp (no_asm) add: Node.model.inv_def Node.model.characteristic_set_def )
apply(rule OCL_Set.includes_charn2) 
apply(simp add: DEF_def)
apply(simp (no_asm) add:  Node.model.set_def)
apply(subst gfp_unfold [OF level1.digraph.Node.model.set_fun.mono], simp)
apply(subst Node.model.set_fun_def)

apply(simp add: Node.model.hol_invariant_def
                Node_model_inv_simps)
apply(simp add: image_def)

apply(simp add: Node.model.inv_def Node.model.characteristic_set_def)
apply(simp add: image_def)
apply(erule conjE)+
apply(rule conjI)
apply(rule_tac x = "obj" in  exI)



apply(rule conjI)
apply(assumption)
apply(rule conjI)

apply(erule disjE)
apply(rule disjI2)
apply(drule OCL_Set.includes_charn1, simp)
apply(unfold Node.model.set_def)
apply(simp add: Bex_def)
apply(simp)

apply(rotate_tac 2)
apply(simp )
apply(erule disjE)

apply(rule disjI2)
apply(drule OCL_Set.includes_charn1, simp)
apply(simp add: Bex_def)

apply(frule isDefined_if_valid)
apply(simp add: OCL_is_def_OclLess)
apply(drule OCL_is_def_Node_i)
apply(simp add: localValidDefined2sem DEF_def)
done


lemma TheLemma: "
(
   model.hol_invariant.init_i \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_left \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_right \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.inv_3 \<tau> (Node.model.set \<tau>) obj 
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.left obj))     
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.left obj))
   \<and> ((\<tau> \<MathOclValid>  Node.model.inv (level1.digraph.Node.right obj))
       \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.digraph.Node.right obj)) 
   \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> (Node.model.set \<tau>) obj
)
=
  (\<tau> \<MathOclValid>  Node.model.inv obj) 
"

apply(rule iffI)
apply(erule TheLemmaTwo)
apply(erule TheLemmaOne)
done


lemma invRepEq1: "(\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow> ( \<tau> \<MathOclValid>  Node.inv.inv obj)"
apply(simp add: Node.inv.inv_def
       inv.init_i_def
       inv.init_left_def
       inv.init_right_def
       inv.inv_3_def
)
apply(simp add: digraph.Node.i_def)
apply(ocl_subst, simp)
apply(drule TheLemmaOne)
apply(simp_all add:Node_model_inv_simps)
done


lemma invRepEq2 : "(\<tau> \<MathOclValid>  Node.inv.inv obj) \<Longrightarrow> ( \<tau> \<MathOclValid>  Node.model.inv obj)"
apply(unfold Node.inv.inv_def)
apply(simp add: inv.init_i_def inv.init_left_def inv.init_right_def inv.inv_3_def)
apply(frule isDefined_if_valid)
apply(simp add: OCL_is_def_OclLess)
apply(unfold digraph.Node.i_def)
apply(simp add: OCL_Logic.strong_definedness(20))
apply(drule and_D2, drule and_D1, simp)
done

thm Node.inv.inv_def

lemma invRepEq : "(\<tau> \<MathOclValid>  Node.inv.inv obj) = ( \<tau> \<MathOclValid>  Node.model.inv obj)"
  apply(rule iffI)
  apply(rule invRepEq2, simp)
  apply(rule invRepEq1, simp)
done



lemma " \<tau> \<MathOclValid> (Cnode.model.inv self ) 
       \<Longrightarrow> \<tau> \<MathOclValid> ( Node.model.inv (self ->oclAsType(level1.digraph.Node)))"
apply(simp add: Cnode.model.inv_def
                 Node.model.inv_def)
apply(simp)

apply(simp add: Cnode.model.characteristic_set_def
                 Node.model.characteristic_set_def)


apply(rule includes_charn2, simp add: DEF_def Node.model.characteristic_set_def)
apply(drule includes_charn1)
apply(simp add: image_def)
thm Node_set_includes_Cnode_set [simp] 
oops


(*
text {* sanity check fuer


package digraph

context Node 
inv: true 
endpackage


*}
lemma invRepEq1: "(\<tau> \<MathOclValid>  Node.model.inv obj) \<Longrightarrow> ( \<tau> \<MathOclValid>  Node.inv.inv obj)"
apply(simp add: Node.inv.inv_def
       inv.init_i_def
       inv.init_left_def
       inv.init_right_def
       inv.inv_3_def
)
done

lemma invRepEq2 : "(\<tau> \<MathOclValid>  Node.inv.inv obj \<MathOclAnd> \<MathOclIsDefined> obj) \<Longrightarrow> ( \<tau> \<MathOclValid>  Node.model.inv obj)"
  apply(unfold Node.inv.inv_def)
  apply(simp add: inv.init_i_def inv.init_left_def inv.init_right_def 
                  inv.inv_3_def)
  apply(simp add: Node.model.inv_def)
  apply(rule OCL_Set.includes_charn2)
  apply(simp add: Node.model.characteristic_set_def Node.model.set_def 
                  Node.model.set_fun_def DEF_def)
  apply(simp add: Node.model.characteristic_set_def image_def Bex_def)
  apply(simp add: OclIsDefined_def OclTrue_def
                  lift0_def lift1_def DEF_def localValid2sem)
  apply(simp add: Node.model.set_def Node.model.set_fun_def)
  apply(simp_all add:Node_model_inv_simps)
  apply(simp add: gfp_def)
  apply(auto)
  done
*)


end
