(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_logic.thy - The Core Logics of OCL.
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Logic *}
 
theory OCL_Logic 
  imports "$HOLOCL_HOME/src/OCL_Logic_core"
begin
 text_raw{* \label{thy:OCL_Logic} *}

section {* Testing and Demonstrations of Main Tactics *}


subsection{* Automatic Generation of \acs{uc} Variants of Global Equivalence 
             Theorems*}
ML{* bind_thm("strictEq_is_strongEq_UC",
              OCL_normal (the_context())(thm"strictEq_is_strongEq_LE")) *}

(*<*)
subsection{* Setting up the currently defined operators *}

text{* The call of $ocl\_setup\_op$ is as follows: *}

ML{* OCL_setup_trace := true *}

(*<*)

ocl_setup_op [OclStrictEq]
(*>*)

ML{* OCL_setup_trace := false *}

ML{* print_ss (!cpR_fold_simpset) *}

text{*  This should correspond to:
        \begin{verbatim}
        ML { * OCL_setup_op [("OclStrictEq",thm"OclStrictEq_def")] * }  
        \end{verbatim}

        This isar command leads to the implicit derivation of the
        following theorems, which are vital for the proper functioning
        of the other tactics \ldots
        
     *}
(*>*)

thm OCL_cp_OclStrictEq
    OCL_undef_1_OclStrictEq 
    OCL_undef_2_OclStrictEq
    OCL_is_def_OclStrictEq
    OCL_is_defopt_OclStrictEq

thm OclIsDefined_def

ocl_setup_op[OclIsDefined]
ocl_setup_op[OclStrongEq]
ocl_setup_op[OclAnd, OclNot]
ocl_setup_op[OclOr_alt, OclXor_alt]
ocl_setup_op[OclImplies_alt, OclImplies1_alt, OclImplies2_alt]
ocl_setup_op [OclSand, OclSor_alt, OclSxor_alt, OclSimplies_alt]
ocl_setup_op [OclIf]


subsection{* Test of \tactic{ocl\_hypsubst} *}


lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot>B; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B"
by(ocl_hypsubst [1], simp)

lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B; (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> C)\<rbrakk> \<Longrightarrow> A \<tau> = B \<tau>"
by(ocl_hypsubst, simp)
 
lemma "!!\<tau> A. \<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> A \<MathOclAnd> B ; \<tau> \<MathOclValid> \<MathOclIsUndefined> B; (\<tau> \<MathOclValid> A) = (\<tau> \<MathOclValid> C)\<rbrakk> \<Longrightarrow> A \<tau> = B \<tau>"
by(ocl_hypsubst, simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> D \<MathOclStrongEq> 1\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(ocl_hypsubst, simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> D \<MathOclStrongEq> E ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(ocl_hypsubst, simp)

lemma "\<lbrakk>  \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> E \<MathOclStrongEq> D ;  \<tau> \<MathOclValid> 1 \<MathOclStrongEq> D\<rbrakk>  \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> E \<MathOclStrongEq> 1"
by(ocl_hypsubst, simp)

lemma "\<lbrakk> \<tau> \<MathOclValid> A; \<tau> \<MathOclValid> \<MathOclNot>B; \<tau> \<MathOclValid> E\<MathOclStrongEq>1; \<tau> \<MathOclValid> 1\<MathOclStrongEq>E; 
         \<tau> \<MathOclValid> E\<MathOclStrictEq>1;\<tau> \<MathOclValid> 1\<MathOclStrictEq>E;\<tau> \<MathOclValid> \<MathOclIsUndefined> B\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B"
by(ocl_hypsubst,simp)

subsection{* Test of \tactic{ocl\_subst} *}

lemma "(\<MathOclIsDefined> X \<MathOclOr> \<MathOclIsUndefined> X) = \<MathOclTrue>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
apply (ocl_subst  (no_asm_simp),simp_all)
apply (ocl_subst  (no_asm_use)(no_asm_simp)(no_concl_simp) | ocl_subst)
apply (simp)
done 

(*<*)
text{* this example is just to test the syntax ... *}
lemma "\<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> A = A' \<rbrakk> \<Longrightarrow>
       (\<MathOclIsDefined> X \<MathOclAnd> A) = (\<MathOclIsDefined> X \<MathOclAnd> A')"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)",simp_all)
apply (ocl_subst [2],simp_all)
apply (ocl_subst,simp)
oops  (* well, surprise, this does not hold \<dots> 
         This makes congruence reasoning of this 
         kind complicated \<dots> *)  
(*>*)

lemma StrongEq_valid_undef :
" \<lbrakk>\<tau> \<MathOclValid> \<MathOclIsUndefined> x; \<tau> \<MathOclValid> \<MathOclIsUndefined> y\<rbrakk> \<Longrightarrow> (x \<MathOclStrongEq> y) \<tau> = \<MathOclTrue> \<tau>"
by(ocl_hypsubst,simp)

text{* 
  The following lemma explains strong equality entirely in terms of
   \OCL in the sense of the standard, i.e.,  by using definedness and
   strict equality, which is the default. Due to @{text "strongEq_charn"}
   (@{thm "strongEq_charn"}), the outstanding importance of strong
   equality becomes obvious.  The following lemma explains, why its
   reformulation in standard operators is possible, but tedious, and
   represents a post-hoc justification for the introduction strong
   equality.
*}

lemma strongEq_charn2: 
"(x \<MathOclStrongEq> y) = ( \<MathOclIf> \<MathOclIsDefined> x \<MathOclAnd> \<MathOclIsDefined> y \<MathOclThen> x \<MathOclStrictEq> y 
             \<MathOclElse> \<MathOclIf> \<MathOclIsUndefined> x \<MathOclAnd> \<MathOclIsUndefined> y \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclFalse> \<MathOclEndif> 
             \<MathOclEndif>)"
apply (rule ext,case_tac "xa \<MathOclValid>  \<MathOclIsDefined>(x)",simp_all)
apply (ocl_subst,simp_all)
apply (case_tac "xa \<MathOclValid>  \<MathOclIsDefined>(y)",simp_all)
apply (ocl_subst,simp_all, rule sym)
apply (auto intro: strictEq_is_strongEq_LE)
prefer 2
apply (case_tac "xa \<MathOclValid>  \<MathOclIsDefined>(y)",simp_all)
apply (ocl_hypsubst [2], simp_all)
apply (rotate_tac -1) (* TODO : Without THIS ocl_subst loops !!! *)
apply (ocl_subst,simp_all)
apply (ocl_hypsubst [1],ocl_hypsubst [2], simp_all)
done

subsection {* Strong and Weak Definedness Calculi *}
lemma isDefined_cong5 : "(\<MathOclIsDefined> X \<MathOclOr> \<MathOclIsUndefined> X) = \<MathOclTrue>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp_all)

lemma isDefined_cong6 : "(\<MathOclIsUndefined> X \<MathOclOr> \<MathOclIsDefined> X ) = \<MathOclTrue>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)


lemma isDefined_cong7 : "(\<MathOclIsDefined> X \<MathOclOr> \<MathOclIsUndefined> X \<MathOclOr> Z) = \<MathOclTrue>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)


lemma isDefined_cong8 : "(\<MathOclIsUndefined> X \<MathOclOr> \<MathOclIsDefined> X \<MathOclOr> Z) = \<MathOclTrue>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)


lemma isDefined_cong9 : "(\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsUndefined> X) = \<MathOclFalse>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)


lemma isDefined_cong10 : "(\<MathOclIsUndefined> X \<MathOclAnd> \<MathOclIsDefined> X) = \<MathOclFalse>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)

lemma isDefined_cong11 : "(\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsUndefined> X \<MathOclAnd> Z) = \<MathOclFalse>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)

lemma isDefined_cong12 : "(\<MathOclIsUndefined> X \<MathOclAnd> \<MathOclIsDefined> X \<MathOclAnd> Z) = \<MathOclFalse>"
apply (rule ext,case_tac "x \<MathOclValid>  \<MathOclIsDefined>(X)")
by    (ocl_subst, simp_all, ocl_subst, simp)



lemmas core_definedness = 
          OclIsDefined__False OclIsDefined__True OclIsDefined__undef
          isDefined_idem isDefined_notD0  
          isDefined_cong4 isDefined_cong3 isDefined_cong2 isDefined_cong1 
          isDefined_cong5 isDefined_cong6 isDefined_cong7 isDefined_cong8 
          isDefined_cong9 isDefined_cong10 isDefined_cong11 isDefined_cong12 
          isDefined_strongEq [THEN is_TRUE_charn[THEN iffD2]]
          isDefined_strictEq


lemmas strong_definedness = core_definedness
          isDefined_ifD0 isDefined_andD0 isDefined_impliesD0 isDefined_orD0


lemmas weak_definedness = core_definedness
          (* the weak rules *)
          isDefined_if isDefined_implies 
          isDefined_or isDefined_and 
          (* isDefined_not - is subsumed by isDefined_idem*)

thm weak_definedness
thm strong_definedness
thm strictEq_is_strongEq_LE

lemmas weak_prop_LJE = 
          OclTrue_is_Valid OclFalse_is_Invalid OclUndefined_is_Invalid
          OclTrue_is_LocalValid absurd6S absurd5S 

          localValidNot2not localValidAnd2conj 
          localValidOr2disj localValidImplies2impl                   

thm isDefined_if [simplified is_TRUE_charn global_vs_local_validity[symmetric]]

thm weak_prop_LJE

(* TODO: xor, if then else, let. *)
          
lemma globalValidAnd2conj:
"\<lbrakk>\<MathOclValid> \<MathOclIsDefined> A; \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> (\<MathOclValid> A \<MathOclAnd> B) = ((\<MathOclValid> A) \<and> (\<MathOclValid> B))"
by (auto simp: global_vs_local_validity[symmetric] localValidAnd2conj)

lemma globalValidNot2localValidNot:
"\<MathOclValid> \<MathOclIsDefined> A \<Longrightarrow> (\<MathOclValid> \<MathOclNot> A) = (\<forall> \<tau>. \<not> (\<tau> \<MathOclValid> A))"
by (auto simp: global_vs_local_validity[symmetric] 
               localValidNot2not[symmetric] )

lemma globalValidOr2localValidDisj:
"\<lbrakk>\<MathOclValid> \<MathOclIsDefined> A; \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> (\<MathOclValid> A \<MathOclOr> B) = (\<forall> \<tau>. (\<tau> \<MathOclValid> A) \<or> (\<tau> \<MathOclValid> B))"
by (auto simp: global_vs_local_validity[symmetric] 
               localValidOr2disj [symmetric] )

lemma globalValidImplies2localValidimpl:
"\<lbrakk>\<MathOclValid> \<MathOclIsDefined> A; \<MathOclValid> \<MathOclIsDefined> B\<rbrakk> \<Longrightarrow> (\<MathOclValid> A \<MathOclImplies> B) = (\<forall> \<tau>. (\<tau> \<MathOclValid> A) \<longrightarrow> (\<tau> \<MathOclValid> B))"
by (auto simp: global_vs_local_validity[symmetric] 
               localValidImplies2impl [symmetric] )


text{* 
  These previous equvalences demonstrate, why a ``global judgement
  equivalence'' calculus does not make sense.  Desirable versions of
  these rules would be:
\begin{gather}
\isasymMathOclValid  \isasymMathOclIsDefined  A
 \isasymLongrightarrow  
(\isasymMathOclValid  \isasymMathOclNot  A) = (\isasymnot  (\isasymMathOclValid  A))\\
\isasymlbrakk \isasymMathOclValid  \isasymMathOclIsDefined  A; \isasymMathOclValid  \isasymMathOclIsDefined  B\isasymrbrakk  
\isasymLongrightarrow  
(\isasymMathOclValid  A \isasymMathOclOr  B) = ((\isasymMathOclValid  A) \isasymor  (\isasymMathOclValid  B))\\
\isasymlbrakk \isasymMathOclValid  \isasymMathOclIsDefined  A; \isasymMathOclValid  \isasymMathOclIsDefined  B\isasymrbrakk
  \isasymLongrightarrow 
 (\isasymMathOclValid  A \isasymMathOclImplies  B) = ((\isasymMathOclValid  A) 
\isasymlongrightarrow  (\isasymMathOclValid> B))
\end{gather}

  This is not achievable since the logical disjunction not distribute
  over the universal quantifier hidden in the global validity
  statement.

  Thus, a ``global judgement equivalence'' is not closed under global
  validity, such that any complete reasoning must be broken down on
  local judgements sooner or later.
*}


(*<*)
subsection {* Test of $ocl\_subst$ *}
text{* \textbf{More to come.} *}

subsection {* Test of $ocl\_auto$ *}
text{* \textbf{More to come.} *}
(*>*)
end 
