(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Operation.thy -
 * Copyright (C) 2004-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {*   Encoding of \OCL Operations *}

theory OCL_Operation 
imports 
  "$HOLOCL_HOME/src/OCL_Calculi" 
begin 
 text_raw{* \label{thy:OCL_Operation} *}

section {* Standard and Strict Method Invocation, and Recursion *}
text {* 
  The core primitives @{text "OCL_choose"} and @{text "OCL_invoke"}, 
  its strict variant @{text "OCL_invokeS"} are used to represent 
  method. 
*}
constdefs OCL_choose :: 
          "(('tau, 'b) VAL \<Rightarrow> ('tau) Boolean) \<Rightarrow> ('tau, 'b) VAL"
          "OCL_choose F \<equiv> (\<lambda> \<tau>. SOME x. F (\<lambda> \<tau>. x) \<tau> = \<MathOclTrue> \<tau>)"

lemma "cp F \<Longrightarrow> cp(OCL_choose F)"
  apply(simp add: cp_def OCL_choose_def)
  apply safe
  apply (rule exI, rule allI, rule allI)
  apply (simp)
  oops

consts OCL_invoke :: " ('e \<Rightarrow> 's)
                       \<Rightarrow> ('s set \<rightharpoonup> ('a, 'b) VAL \<Rightarrow> 'c)
                       \<Rightarrow> ('a, 'b) VAL  \<Rightarrow> 'c"
consts OCL_invokeS ::" ('b \<Rightarrow> 's)
                       \<Rightarrow> ('s set \<rightharpoonup> ('a, 'b) VAL \<Rightarrow> 'c)
                       \<Rightarrow> ('a, 'b) VAL  \<Rightarrow> 'c"
(*
text{* Instance for null-ary invocation with only trivial return value:
   \verb|OCL_invoke (C\<Colon>'c\<Rightarrow>'d) (tab\<Colon>'d set \<rightharpoonup> ('a\<Rightarrow>'c)\<Rightarrow>'a \<Rightarrow> 'b)(self\<Colon>'a \<Rightarrow> 'c)|
 *}

defs OCL_invokeVoid_def:
    "OCL_invoke C tab self \<equiv> \<lambda> St. (case tab (LEAST X. X \<in> dom tab \<and> C(self St) \<in> X) of
                                    None \<Rightarrow>  arbitrary
                                  | Some f \<Rightarrow> f self St )"
     OCL_invokeVoidS_def:
    "OCL_invokeS C tab self \<equiv> (if \<MathOclIsDefined>(self) 
                                 then OCL_invoke C tab self
                                 else OclUndefined
                                 endif)"


lemma OCL_invokeVoidS_undef0[simp] :
     "OCL_invokeS C (tab \<Colon> 'd set \<rightharpoonup> ('a\<Rightarrow>'c::bot)\<Rightarrow>'a\<Rightarrow>'b::bot) OclUndefined 
      = OclUndefined"
   by(simp add: OCL_invokeVoidS_def)


lemma OCL_invokeVoidS_is_def:
     "(\<MathOclValid> \<MathOclIsDefined> (OCL_invokeS C (tab \<Colon> 'd set \<rightharpoonup> ('a \<Rightarrow> 'c\<Colon>bot) \<Rightarrow> 'a \<Rightarrow> 'b::bot) self)) 
       \<Longrightarrow> (\<MathOclValid> \<MathOclIsDefined> (self))"
   apply(simp_all add: OCL_invokeVoidS_def isDefined_ifD0 OclValid_def)
   (* to lower calculus *)
    apply(rule ext)
    apply(drule_tac x = "x" in fun_cong)
   (* plus some OCL-logical reasoning *)
    apply(fold OclLocalValid_def OclValid_def) 
    apply(erule and_TRUE_E)
    by(simp)
*)

text{* 
  Instance for null-ary invocation with return value:
\begin{align}
  \operatorname{OCL\_invoke} &
   \ap(C\isasymColon 'a\isasymRightarrow 'd)\notag\\
   &\ap (\mathit{tab}\isasymColon 'd set\isasymrightharpoonup ('s\isasymRightarrow 'a)\isasymRightarrow ('s\isasymRightarrow 'b)\isasymRightarrow 's\isasymRightarrow 'c)\notag\\
   &\ap (\self\isasymColon 's\isasymRightarrow 'a)(\result\isasymColon 's\isasymRightarrow 'b)\notag
\end{align}
 *}
defs OCL_invoke0_def :
    "OCL_invoke C tab self result \<equiv> 
          \<lambda> St. (case tab (LEAST X. X \<in> dom tab \<and> C(self St) \<in> X) of
                                    None \<Rightarrow>  arbitrary
                                  | Some f \<Rightarrow> f (self\<Colon>'s \<Rightarrow> 'a) 
                                                (result\<Colon>'s \<Rightarrow>'b) St )"

defs
     OCL_invoke0S_def:
         "OCL_invokeS C tab self result \<equiv> 
          (if \<MathOclIsDefined>(self)  then OCL_invoke C tab self result else \<MathOclUndefined> endif)"

lemma OCL_invoke0S_undef0[simp] :
      "OCL_invokeS C (tab\<Colon>'c set\<rightharpoonup>('a\<Rightarrow>'d::bot)\<Rightarrow>('a\<Rightarrow>'e::bot)\<Rightarrow>'a\<Rightarrow>'b::bot) \<MathOclUndefined> X 
       =  \<MathOclUndefined>"
   by(simp add: OCL_invoke0S_def)

lemma OCL_invoke0S_is_def:
      "\<MathOclValid> \<MathOclIsDefined>(OCL_invokeS C
                      (tab\<Colon>'c set \<rightharpoonup> ('a\<Rightarrow>'d::bot)\<Rightarrow>('a\<Rightarrow>'e::bot)\<Rightarrow>'a\<Rightarrow>'b::bot)
                            self result) \<Longrightarrow>
       \<MathOclValid> \<MathOclIsDefined>(self)"
   by(simp add: OCL_invoke0S_def isDefined_ifD0 globalValidAnd2conj)

text {* 
Instance for binary invocation:
\begin{align}
\operatorname{OCL\_invoke}&
\ap (C\isasymColon 'a \isasymRightarrow  'e)\notag\\
&\ap (tab\isasymColon 'e set \isasymrightharpoonup  ('s\isasymRightarrow 'a)\isasymRightarrow ('s\isasymRightarrow 'b)\isasymRightarrow ('s\isasymRightarrow 'c)\isasymRightarrow 's\isasymRightarrow 'd)(self\isasymColon 's\isasymRightarrow 'a)\notag\\
&\ap (a\isasymColon 's \isasymRightarrow  'b) (b\isasymColon 's \isasymRightarrow  'c)\notag
\end{align}
*}

defs OCL_invoke1_def :
         "OCL_invoke C tab self a result 
          \<equiv> \<lambda> \<tau> . (case tab (LEAST X. X \<in> dom tab \<and> C(self \<tau>) \<in> X) of
                     None \<Rightarrow>  arbitrary
                   | Some f \<Rightarrow> f (self\<Colon>'tau \<Rightarrow> 'a) 
                                (a\<Colon>'tau \<Rightarrow> 'b) 
                                (result\<Colon>'tau \<Rightarrow> 'c) \<tau>)"

defs OCL_invoke1S_def:
     "OCL_invokeS C tab self a result 
      \<equiv> (if \<MathOclIsDefined>(self) and \<MathOclIsDefined>(a) 
         then OCL_invoke C tab self (a\<Colon>'tau \<Rightarrow> 'b::bot)(result\<Colon>'tau \<Rightarrow> 'c)
         else \<MathOclUndefined> endif)"


lemma OCL_invoke1S_undef0[simp] :
      "OCL_invokeS C 
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)\<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>'s\<Rightarrow>'d::bot)
                   \<MathOclUndefined> X Y
       = \<MathOclUndefined> "
   by(simp add: OCL_invoke1S_def)

lemma OCL_invoke1S_undef1[simp] :
      "OCL_invokeS C 
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)\<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>'s\<Rightarrow>'d::bot)
                   self \<MathOclUndefined> Y
       =  \<MathOclUndefined>"
   by(simp add: OCL_invoke1S_def)

lemma OCL_invoke2S_is_def':
      "\<MathOclValid> \<MathOclIsDefined>(OCL_invokeS C
                       (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)\<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>'s\<Rightarrow>'d::bot)
                       self X Y) \<Longrightarrow>
       \<MathOclValid> \<MathOclIsDefined>(self) and \<MathOclIsDefined>(X)"
   by(simp add: OCL_invoke1S_def isDefined_ifD0 globalValidAnd2conj)

text{* Instance for Binary Invocation: *}
defs OCL_invoke2_def :
         "OCL_invoke C tab self a b result  
          \<equiv> \<lambda> St. (case tab (LEAST X. X \<in> dom tab \<and> C(self St) \<in> X) of
                     None \<Rightarrow>  arbitrary
                   | Some f \<Rightarrow> f (self\<Colon>'s\<Rightarrow>'a)(a\<Colon>'s\<Rightarrow>'b)(b\<Colon>'s\<Rightarrow>'c)(result\<Colon>'s\<Rightarrow>'d) St )"

defs
     OCL_invoke2S_def:
         "OCL_invokeS C tab self a b result
          \<equiv> (if \<MathOclIsDefined>(self) and \<MathOclIsDefined>(a) and \<MathOclIsDefined>(b)
             then OCL_invoke C tab (self\<Colon>'s\<Rightarrow>'a::bot)
                                   (a\<Colon>'s\<Rightarrow>'b::bot)(b\<Colon>'s\<Rightarrow>'c::bot)
                                   (result\<Colon>'s\<Rightarrow>'d)
             else \<MathOclUndefined>
             endif)"

lemma OCL_invoke2S_undef0[simp] :
      "OCL_invokeS C 
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)
                                 \<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>('s\<Rightarrow>'d::bot)\<Rightarrow>('s\<Rightarrow>'e::bot))
                    \<MathOclUndefined> X Y Z
       = \<MathOclUndefined>"
   by(simp add: OCL_invoke2S_def)


lemma OCL_invoke2S_undef1[simp] :
      "OCL_invokeS C 
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)
                                 \<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>('s\<Rightarrow>'d::bot)\<Rightarrow>('s\<Rightarrow>'e::bot))
                   X \<MathOclUndefined> Y Z
       = \<MathOclUndefined>"
   by(simp add: OCL_invoke2S_def)

lemma OCL_invoke2S_undef2[simp] :
      "OCL_invokeS C 
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)
                                 \<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>('s\<Rightarrow>'d::bot)\<Rightarrow>('s\<Rightarrow>'e::bot))
                   X Y \<MathOclUndefined> Z
       =  \<MathOclUndefined>"
   by(simp add: OCL_invoke2S_def)


lemma OCL_invoke2S_is_def:
      "\<MathOclValid> \<MathOclIsDefined>(OCL_invokeS C
                   (tab\<Colon>'e set\<rightharpoonup>('s\<Rightarrow>'a::bot)\<Rightarrow>('s\<Rightarrow>'b::bot)
                                 \<Rightarrow>('s\<Rightarrow>'c::bot)\<Rightarrow>('s\<Rightarrow>'d::bot)\<Rightarrow>('s\<Rightarrow>'e::bot))           
                           self X Y result) \<Longrightarrow>
       \<MathOclValid> \<MathOclIsDefined>(self) and \<MathOclIsDefined>(X) and \<MathOclIsDefined>(Y)"
   by(simp add: OCL_invoke2S_def isDefined_ifD0 globalValidAnd2conj)
end  
