(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Sequence.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* Requirements for \OCL Primitive Types*}
theory OCL_Basic_requirements
imports 
  "$HOLOCL_HOME/src/OCL_Library"
begin
text {*
  This theory contains the requirements for the primitive types defined in 
  the \OCL standard library (Integer, Real, String, and 
  Boolean)~\cite[Chapter 11.5]{omg:ocl:2003}. 
*}

(**************************************************************************)
section {* Real *}
(**************************************************************************)
text_raw {*
\begin{oclrequirement}{11.5.1-a}{+ (r : Real) : Real}
  The value of the addition of self and r.
\end{oclrequirement}

\begin{oclrequirement}{11.5.1-b}{- (r : Real) : Real}
  The value of the subtraction of r from self.
\end{oclrequirement}

\begin{oclrequirement}{11.5.1-c}{ * (r : Real) : Real}
  The value of the multiplication of self and r.
\end{oclrequirement}

\begin{oclrequirement}{11.5.1-d}{- : Real}
  The negative value of self.
\end{oclrequirement}

\begin{oclrequirement}{11.5.1-e}{/ (r : Real) : Real}
  The value of self divided by r.
\end{oclrequirement}
*}

text_raw {* \begin{oclrequirement}{11.5.1-f}{abs() : Real}
  The absolute value of self.
\begin{ocl}
post: if self < 0 then result = - self 
                  else result = self 
                  endif
\end{ocl}*}
lemma "req_11_5_1-f": 
"( self ->abs()) = (\<MathOclIf> (self \<MathOclLess> 0) \<MathOclThen> OclNegative self \<MathOclElse> self \<MathOclEndif>)"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-g}{floor() : Integer}
  The largest integer which is less than or equal to self.
\begin{ocl}
post: (result <= self) and (result + 1 > self)
\end{ocl}*}
lemma "req_11_5_1-g": 
"\<MathOclValid> ((self ->floor())->toReal() \<MathOclLe>  self) \<MathOclAnd> ((self ->floor() + 1) ->toReal() \<MathOclGreater>  self)"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-h}{round() : Integer}
  The integer which is closest to self. When there are two 
  such integers, the largest one.
\begin{ocl}
post: ((self - result).abs() < 0.5) 
      or ((self - result).abs() = 0.5 and (result > self))
\end{ocl}*}
lemma "req_11_5_1-h": "true"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-i}{max(r : Real) : Real}
  The maximum of self and r.
\begin{ocl}
post: if self >= r then result = self 
                   else result = r 
                   endif
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-j}{min(r : Real) : Real}
  The minimum of self and r.
\begin{ocl}
post: if self <= r then result = self 
                   else result = r 
                   endif
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-k}{< (r : Real) : Boolean}
  True if self is less than r.
\end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-l}{> (r : Real) : Boolean}
  True if self is greater than r.
\begin{ocl}
post: result = not (self <= r)
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-m}{<= (r : Real) : Boolean}
  True if self is less than or equal to r.
\begin{ocl}
post: result = ((self = r) or (self < r))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {*
\begin{oclrequirement}{11.5.1-n}{>= (r : Real) : Boolean}
  True if self is greater than or equal to r.
\begin{ocl}
post: result = ((self = r) or (self > r))
\end{ocl}
\end{oclrequirement}
*}

(**************************************************************************)
section {* Integer *}
(**************************************************************************)
text {*
\begin{oclrequirement}{11.5.2-a}{- : Integer}
The negative value of self.
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-b}{+ (i : Integer) : Integer}
The value of the addition of self and i.
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-c}{- (i : Integer) : Integer}
The value of the subtraction of i from self.
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-d}{* (i : Integer) : Integer}
The value of the multiplication of self and i.
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-e}{/ (i : Integer) : Real}
The value of self divided by i.
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-f}{abs() : Integer}
The absolute value of self.
\begin{ocl}
post: if self < 0 then result = - self
                  else result = self
                  endif
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-g}{div( i : Integer) : Integer}
The number of times that i fits completely within self.
\begin{ocl}
pre : i <> 0
post: if self / i >= 0 then result = (self / i).floor()
                       else result = -((-self/i).floor())
                       endif
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-h}{mod( i : Integer) : Integer}
The result is self modulo i.
\begin{ocl}
post: result = self - (self.div(i) * i)
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-i}{max(i : Integer) : Integer}
The maximum of self an i.
\begin{ocl}
post: if self >= i then result = self
                   else result = i
                   endif
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.2-j}{min(i : Integer) : Integer}
The minimum of self an i.
\begin{ocl}
post: if self <= i then result = self
                   else result = i
                   endif
\end{ocl}
\end{oclrequirement}
*}

(**************************************************************************)
section {* String *}
(**************************************************************************)
text {*
\begin{oclrequirement}{11.5.3-a}{size() : Integer}
The number of characters in self.
\end{oclrequirement}

\begin{oclrequirement}{11.5.3-b}{concat(s : String) : String}
The concatenation of self and s.
\begin{ocl}
post: result.size() = self.size() + string.size()
post: result.substring(1, self.size() ) = self
post: result.substring(self.size() + 1, result.size() ) = s
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.3-c}{substring(lower : Integer, upper : Integer) : String}
The sub-string of self starting at character number lower, up 
to and including character number upper. Character numbers
run from 1 to self.size().
\begin{ocl}
pre: 1 <= lower
pre: lower <= upper
pre: upper <= self.size()
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.3-d}{toInteger() : Integer}
Converts self to an Integer value.
\end{oclrequirement}

\begin{oclrequirement}{11.5.3-e}{toReal() : Real}
Converts self to a Real value.
\end{oclrequirement}

*}
(**************************************************************************)
section {* Boolean *}
(**************************************************************************)
text {*
\begin{oclrequirement}{11.5.4-a}{or (b : Boolean) : Boolean}
True if either self or b is true.
\end{oclrequirement}

\begin{oclrequirement}{11.5.4-b}{xor (b : Boolean) : Boolean}
True if either self or b is true, but not both.
\begin{ocl}
post: (self or b) and not (self = b)
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.4-c}{and (b : Boolean) : Boolean}
True if both b1 and b are true.
\end{oclrequirement}

\begin{oclrequirement}{11.5.4-d}{not : Boolean}
True if self is false.
\begin{ocl}
post: if self then result = false
              else result = true
              endif
\end{ocl}
\end{oclrequirement}

\begin{oclrequirement}{11.5.4-e}{implies (b : Boolean) : Boolean}
True if self is false, or if self is true and b is true.
\begin{ocl}
post: (not self) or (self and b)
\end{ocl}
\end{oclrequirement}
*}
end