(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Real.ML -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Real *}

theory OCL_Real 
imports 
  "$HOLOCL_HOME/src/library/basic/OCL_Numerals" 
begin
text_raw{* \label{thy:OCL_Real} *}

constdefs
  is_Real_0       :: "'a U \<Rightarrow> bool"
  "is_Real_0      \<equiv>  sum_case  (\<lambda>x. False) 
                           (sum_case (\<lambda>x. True) (\<lambda>x. False))"
  get_Real_0      :: "'a U \<Rightarrow> Real_0"
  "get_Real_0     \<equiv>   sum_case (\<lambda>x. \<some>  x. True)
                             (sum_case (\<lambda>x. x) (\<lambda>x. \<some> x. True))"
  mk_Real_0       :: "Real_0 \<Rightarrow> 'a U"
  "mk_Real_0      \<equiv>  Inr o Inl"

lemma get_mk_Real_id_0 : "get_Real_0(mk_Real_0 x) = x"
  apply (simp add: get_Real_0_def mk_Real_0_def)
  done
lemma is_mk_Real_0 : "is_Real_0(mk_Real_0 x) = True"
  apply (simp add: is_Real_0_def mk_Real_0_def)
  done
lemma mk_get_Real_id_0 : "is_Real_0 x \<Longrightarrow> mk_Real_0(get_Real_0 x) = x"
  apply (simp add: get_Real_0_def mk_Real_0_def is_Real_0_def )
  apply (case_tac x, simp, simp)
  apply (case_tac b, simp, simp)
  done

constdefs
   Real_0    :: "('st, Real_0 Set_0)    VAL"
  "Real_0    \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>)"

defs (*of overloaded constants *)
  OclZero_def: "0 \<equiv> lift\<^sub>0( \<HolOclLiftLeft>0\<Colon>real\<HolOclLiftRight> )"
  OclOne_def : "1 \<equiv> lift\<^sub>0( \<HolOclLiftLeft>1\<Colon>real\<HolOclLiftRight> )"

defs 
  plus_def   : "op + \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                               \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )+( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  minus_def  : "op - \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                               \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )-( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  times_def  : "op * \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                               \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )*( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"

ocl_setup_op [plus, minus, times]

consts
  OclRound   :: "'a Real => 'a Integer"         ("_ ->round'( ')" [66])
  OclFloor   :: "'a Real => 'a Integer"         ("_ ->floor'( ')" [66])
                 
  OclMin     :: "['a Real, 'a Real] => 'a Real"  ("_ ->min'( _ ')" [66,65]65) 
  OclMax     :: "['a Real, 'a Real] => 'a Real"  ("_ ->max'( _ ')" [66,65]65) 

  OclDivide  :: "['a Real, 'a Real] => 'a Real" 
                                                 ("_ ->divide'( _ ')"[66,65]65)

  OclAbs     :: "'a Real => 'a Real"         ("_ ->abs'( ')" [66])
  OclNegative:: "'a Real => 'a Real"         ("- _" [66])


defs
  OclDivide_def: 
   "OclDivide \<equiv> lift\<^sub>2 (strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y\<Colon>Real_0. 
                                      ( \<HolOclLiftLeft> (\<HolOclDropLeft>x\<HolOclDropRight> / \<HolOclDropLeft>y\<HolOclDropRight>) \<HolOclLiftRight> ))))"

defs
  OclLe_def: 
  "OclLe     \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                                       \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) <= ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclLess_def: 
  "OclLess   \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                                       \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) < ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclGe_def: 
  "OclGe     \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                                       \<HolOclLiftLeft> \<not>( \<HolOclDropLeft>x\<HolOclDropRight> ) < ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclGreater_def: 
  "OclGreater\<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Real_0. strictify(\<lambda> y. 
                                       \<HolOclLiftLeft> \<not>( \<HolOclDropLeft>x\<HolOclDropRight> ) <= ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"

  OclMin_def: 
  "SELF ->min(I) \<equiv> if (SELF `<=` I) then SELF else I endif"
 
  OclMax_def: 
  "SELF ->max(I) \<equiv> if (SELF `>=` I) then SELF else I endif"

  OclAbs_def: 
  "OclAbs        \<equiv> lift\<^sub>1(strictify(\<lambda> x\<Colon>Real_0. 
                                       \<HolOclLiftLeft> abs ( \<HolOclDropLeft>x\<HolOclDropRight> ) \<HolOclLiftRight> ))"

  OclNegative_def: 
  "OclNegative   \<equiv> lift\<^sub>1(strictify(\<lambda> x\<Colon>Real_0. 
                                       \<HolOclLiftLeft> ( - ( \<HolOclDropLeft>x\<HolOclDropRight> )) \<HolOclLiftRight> ))"

ocl_setup_op [OclMax, OclMin, OclNegative]
ocl_setup_op [OclLe, OclLess, OclGe, OclGreater]

end
