(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Bag.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Bag *}

theory OCL_Bag 
imports 
  "$HOLOCL_HOME/src/library/collection/nonsmashed/OCL_Sequence" 
  "Library/Permutation"
begin

section{* Redefinition of the overwritten theorems in the Permutation theory *}

lemmas trans = HOL.trans
lemmas swap = HOL.swap

section{* Properties of the Congruence inside the Datatype Adaption *}

text{* The following rules are transformed versions of the automatically
generated rules for datatype conversions. *}


lemma Abs_Bag_0_inject_charn:
  "\<lbrakk> count x (\<bottom>::'a::bot) = 0; count y (\<bottom>::'a::bot) = 0 \<rbrakk> \<Longrightarrow> 
   (((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Bag_0_inject, 
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

(* TODO: which form is to be preferred *)

(* as an elimination rule:


lemma Abs_Bag_inject_absurd1:
  "\<lbrakk> (Abs_Bag_0 down) = (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>); (\<bottom>::'a::bot) \<notin> set x \<rbrakk> \<Longrightarrow> P"
  by(subst (asm) Abs_Bag_0_inject,
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)


  or as a simplifier rule:

  - are both of them needed?

*)

lemma Abs_Bag_0_inject_absurd11 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 down) = (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Bag_0_inject,
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inject_absurd12 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Bag_0 down)) = False"
  by(subst Abs_Bag_0_inject,
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inject_absurd21 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ( \<bottom> = (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Bag_def)

lemma Abs_Bag_0_inject_absurd22 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Bag_def)

lemma Abs_Bag_0_inject_charn:
  "\<lbrakk> count x (\<bottom>::'a::bot) = 0; count y (\<bottom>::'a::bot) = 0 \<rbrakk> \<Longrightarrow> 
   (((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Bag_0_inject, 
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)


lemma Abs_Bag_0_inverse_charn [simp]:
  "(count x \<bottom> = 0) \<Longrightarrow> Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Bag_0_inverse, 
      simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inverse_charn2 [simp]:
  "(count x \<bottom> = 0) \<Longrightarrow> \<HolOclDropLeft>Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Bag_0_inverse, 
      simp_all add: OCL_Bag_type.Bag_0_def smash_def)
(*
lemma Abs_Bag_0_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; count y \<bottom> = 0 \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_Bag_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_Bag_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(simp add: OCL_Bag_type.Bag_0_def smash_def DEF_def)
  apply(case_tac "(count \<HolOclDropLeft>y\<HolOclDropRight> \<bottom>) = (0\<Colon>nat)")
  apply(auto simp: OCL_Bag_type.Bag_0_def smash_def DEF_def)
  done

lemma Abs_Bag_0_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> count y \<bottom> = 0 \<rbrakk> \<Longrightarrow> P (Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_Bag_0_induct)
  apply(rule_tac  x="Abs_Bag_0 y" in Abs_Bag_0_cases_charn)
  apply(auto intro!: bottomCase stepCase)
  done

lemma inj_on_Abs_Bag_0: "inj_on Abs_Bag_0 Bag_0";
  by(rule inj_on_inverseI,rule Bag_0.Abs_Bag_0_inverse,assumption)

lemma inj_Rep_Bag_0 : "inj Rep_Bag_0"
  by(rule inj_on_inverseI, rule Bag_0.Rep_Bag_0_inverse)

lemma smashed_bag_charn: 
  "(count X \<bottom> = 0) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> Bag_0)"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 1 *)

lemma UU_in_smashed_bag [simp]: 
  "\<bottom> \<in> Bag_0"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 2a *)

lemma down_in_smashed_bag [simp]: 
  "down \<in> Bag_0"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 2b *)

lemma mt_in_smashed_bag[simp]: 
  "(\<HolOclLiftLeft>{#}\<HolOclLiftRight>) \<in> Bag_0"
  by (unfold smash_def Bag_0_def,auto)

lemma DEF_Abs_Bag: "\<And>X. (count X \<bottom> = 0) \<Longrightarrow> DEF (Abs_Bag_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  apply(unfold DEF_def UU_Bag_def, simp)
  done

lemma DEF_Rep_Bag: 
  "\<And>X. DEF X \<Longrightarrow> DEF (Rep_Bag_0 X)"
  apply(unfold DEF_def UU_Bag_def,auto)
  apply(drule_tac f = "Abs_Bag_0" in arg_cong)
  apply(simp add: Rep_Bag_0_inverse)
  done

lemma not_DEF_Rep_Bag: 
  "\<And>X. \<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Bag_0 X)"
  apply(unfold DEF_def UU_Bag_def,auto)
  apply(simp add: Abs_Bag_0_inverse)
  done


lemma exists_lift_Bag: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  by (drule DEF_Rep_Bag,simp add: DEF_X_up)

lemma exists_lift_Bag2: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  apply(frule exists_lift_Bag)
  apply(auto simp: UU_Bag_def DEF_def)
  apply(drule sym, rule swap)
  apply assumption
  apply(unfold not_def)
  apply(subst (asm) smashed_bag_charn)
  apply(simp add: Rep_Bag_0)
  done

lemma Rep_Bag_cases: 
  "Rep_Bag_0 X = \<bottom> \<or> (\<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(case_tac "DEF X")
  apply(drule exists_lift_Bag2)
  apply(drule_tac [2] not_DEF_Rep_Bag)
  apply(simp_all add: DEF_def)
  done


lemma DEF_X_Bag_0 : "DEF(X) = (\<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(insert  Rep_Bag_cases [of "X"], auto)
  apply(drule exists_lift_Bag,auto)
  apply(rule swap)
  prefer 2
  apply(rule not_DEF_Rep_Bag, auto)
  done


text{* The following lemma is crucial for thy\_morpher: *} 
lemma DEF_X_Bag: "DEF X = (\<exists>c.  count c \<bottom> = 0 \<and> X = Abs_Bag_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"

  apply (auto simp: DEF_Abs_Bag)
  apply (simp add:DEF_X_Bag_0)
  apply (erule exE, rule exI, auto)
  apply (rule injD [OF inj_Rep_Bag_0])
  apply (auto simp: smashed_bag_charn Abs_Bag_0_inverse)
  done 

text{* These four lemmas are used by ocl\_setup\_op to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_Bag_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="(Y St)" in  inv_g)
  prefer 2
  apply(drule DEF_Abs_Bag, simp add: DEF_def, auto)
  done

lemma lift2_strict_is_isdef_fw_Bag_Bag:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Bag_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; count b \<bottom> = 0 \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(case_tac "DEF (Y St)")
  apply(drule DEF_X_Bag_0[THEN iffD1])+
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="ca" in inv_g, assumption)
  apply(rotate_tac -1)
  apply(drule DEF_Abs_Bag, simp add: DEF_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Bag_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Bag_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Bag_Bag:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Bag_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; count a \<bottom> = 0 \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Bag_Bag)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

section{* Characteristic Set of Bags (parametric): *}

(* TODO: What Bags should be included in this set? *)
*)
(*
constdefs
   Bag:: "('st, ('b::bot) Bag_0) VAL \<Rightarrow> ('st, 'b Bag_0 Set_0) VAL"
  "Bag \<equiv> lift1(strictify (\<lambda>X.
                   Abs_Set_0 \<HolOclLiftLeft>(Abs_Bag_0 o lift) ` (lists (set  \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>))\<HolOclLiftRight> ))"

ocl_setup_op [Bag]

lemma Bag_defined:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> X) = (\<tau> \<MathOclValid> \<MathOclIsDefined> (Bag X))"
  apply(simp add: Bag_def localValidDefined2sem
                  strictify_def lift1_def o_def image_def DEF_def)
  apply(rule impI, fold DEF_def)
  apply(frule DEF_X_Bag[THEN iffD1])
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(rule DEF_Abs_Set, auto simp: UU_Bag_def)
  done 

*)

constdefs
    OclMtBag   :: "('a, 'b::bot Bag_0) VAL"    
   "OclMtBag  \<equiv> lift0(Abs_Bag_0(\<HolOclLiftLeft>{#}\<HolOclLiftRight>))"

    syntax  (* standard syntax *)
    "_OclMtBag_std"   :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<OclMtBag>")
    syntax  (* ascii syntax *)
    "_OclMtBag_ascii" :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("mtBag")
    syntax  (xsymbols)
    "_OclMtBag_math"  :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<MathOclMtBag>")
    parse_translation {* flat(map (emb trans_const) ["OclMtBag"]) *}
    print_translation {* map emb_print ["OclMtBag"] *}




section{* Foundational Properties of MtBag *}

lemma RepAbs_Mt[simp]:
  "Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>{#}\<HolOclLiftRight>) = \<HolOclLiftLeft>{#}\<HolOclLiftRight>"
  by(simp)
(*
lemma Abs_Bag_Mt_DEF[simp]:
  "DEF(Abs_Bag_0 \<HolOclLiftLeft>{#}\<HolOclLiftRight>)"
  by(simp add: DEF_X_Bag_0)

(* TODO: there are more general properties that the empty
         bag is defined in every context. Do we need that?
*)

lemma OCL_is_isdef_OclMtBag [simp]:
  "\<MathOclValid> \<MathOclIsDefined> \<MathOclMtBag>"
  by(simp add: sym[OF global_vs_local_validity] Abs_Bag_Mt_DEF 
               localValidDefined2sem OclMtBag_def lift0_def)

lemma OCL_is_defopt_OclMtBag [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> \<MathOclMtBag>"
  by(rule OCL_is_isdef_OclMtBag[THEN valid_elim])

*)
section{* Foundational Properties of includes and excludes *}

defs 
   (* includes     :: "[('a,'b::bot) Bag, ('a,'b) VAL] => 'a Boolean" *)
   OclIncludes_def : "OclIncludes \<equiv> lift2(strictify (\<lambda>X. strictify (\<lambda>x.
                                      \<HolOclLiftLeft>0 < count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x\<HolOclLiftRight>)))"

   (* excludes     :: "[('a,'b::bot) Bag, ('a,'b) VAL] => 'a Boolean" *)
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                      \<HolOclLiftLeft>count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x = 0\<HolOclLiftRight>)))"

ocl_setup_op [OclIncludes, OclExcludes]


lemma includes_charn1 :
  "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> (0 < count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight> (x \<tau>))" 
  apply(frule isDefined_if_valid)
  apply(drule OCL_is_defopt_OclIncludes[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 
  done

(*
lemma includes_charn2 :
  "\<lbrakk> DEF(S \<tau>); 0 < count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight> (x \<tau>) \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
  apply(frule DEF_X_Bag [THEN iffD1], erule exE)
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
  apply(auto)
  done


section{* Higher Properties of includes and excludes *}


(* We prefer a normal form where excludes is eliminated. *)
lemma exludes_not_includes[simp]: 
  "x \<MathOclExcludes> X = \<MathOclNot> (x::('a, 'b)VAL) \<MathOclIncludes> (X::('a, 'b::bot) Bag)"
  apply (rule ext)
  apply (simp add:  OclExcludes_def OclIncludes_def)
  apply (simp only: OclNot_def)
  apply (simp only: OclIncludes_def OclExcludes_def lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                               split add: split_if up.split)
  done


lemma mtBag_Includes_charn :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> 
    \<tau> \<MathOclValid> (x::('a, 'b)VAL) \<MathOclExcludes> (\<MathOclMtBag>::('a, 'b::bot) Bag)"
  apply(simp add: localValidDefined2sem)
  apply(simp only: exludes_not_includes OclMtBag_def OclLocalValid_def
                  OclIncludes_def  lift0_def lift1_def lift2_def 
                  strictify_def o_def OclNot_def OclTrue_def DEF_def) 
  apply(simp add: Abs_Bag_Mt_DEF [simplified DEF_def] strictify_def)
  done


(*

TODO: Fix this, when the Characteristic Set of Bags is done..

lemma mtBag_in_Bag :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Bag) \<Longrightarrow> 
    \<tau> \<MathOclValid> (\<MathOclMtBag>::('a,'b::bot)Bag) \<MathOclIncludes> (Bag  S)"
  apply(frule Bag_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Bag_def OCL_Set.OclIncludes_def OclMtBag_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Bag_def UU_Set_def)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Bag_def)
  done


lemma X_in_Bag_X :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Bag) \<Longrightarrow> 
    \<tau> \<MathOclValid> S \<MathOclIncludes> (Bag S)"
  apply(frule Bag_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Bag_def OCL_Set.OclIncludes_def OclMtBag_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Bag_def UU_Set_def)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Bag_def)
  done

*)

subsection{* Consequences of Smashedness of Bags *}

text{* This is the characteristic property of smashed bags:
       if \verb+X ->includes(x)+ is valid, we know that x must be defined. *}

lemma smashed_bags_charn:
  "\<tau> \<MathOclValid> ((x::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Bag)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> x"
  apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                   strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
  apply(case_tac "X \<tau> = \<bottom>", auto)
  apply(simp add: strictify_def DEF_def)
  apply(case_tac "x \<tau> = \<bottom>", auto)
  done

*)
section{* The conversion to sequences *}

text{* Because there is no ordering on the elements of the bags the AsSequence operator
choses some sequence which has exactly the same elements as the bag.

Because this is just some sequence the proof about statements containing the AsSequence
operator, have to show the same properties as one has to show for the general Hilber Choice
operator.*}

defs
   (* asSequence     :: "('a,'b::bot) Bag \<Rightarrow> ('a,'b::bot) Sequence" *)
   OclAsSequence_def  : 
   "OclAsSequence     \<equiv> lift1 (strictify (\<lambda>X. 
                          Abs_Sequence_0 \<HolOclLiftLeft>SOME xs. \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> = multiset_of xs\<HolOclLiftRight>))"

ocl_setup_op [OclAsSequence]

(* NOTE: quick hack to be able to remove assumptions *)
lemma rem_assum: "P \<Longrightarrow> True" by (simp)
(*

lemma OclAsSequence_Bag_charn_0:
  assumes undefCase: "P \<MathOclUndefined> \<bottom> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Bag_0 B\<HolOclDropRight> = multiset_of xs; DEF B \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) (\<lambda> s. B) \<tau>"
  shows              "P ((OclAsSequence (\<lambda> s. (B::'b::bot Bag_0)))::('a,'b::bot)Sequence) (\<lambda> s. B) \<tau>"
  apply(insert undefCase)
  apply(simp add: localValidDefined2sem OclIsDefined_def OclAsSequence_def
                  OclUndefined_def
                  lift0_def lift1_def lift2_def o_def strictify_def DEF_def)
  apply(rule impI, fold DEF_def)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto)
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD])
(* TODO: how can I simplify an expression before inserting it?
         thus enabling me to use lift0 instead of lamda-abstraction *)
  apply(rule defCase)
  apply(simp_all add: localValidDefined2sem)
  done

lemma OclAsSequence_Bag_charn:
  assumes undefCase: "P \<MathOclUndefined> \<MathOclUndefined> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Bag_0 (B \<tau>)\<HolOclDropRight> = multiset_of xs; \<tau> \<MathOclValid> \<MathOclIsDefined>B \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) B \<tau>"
  and     cp2_P    : "cp2 P"
  shows              "P ((OclAsSequence (B::('a,'b::bot)Bag))::('a,'b::bot)Sequence) B \<tau>"
  apply(rule subst, rule_tac B="B" and A="lift0 (B \<tau>)" in cp_charn)
  apply(auto simp: lift0_def cp2_P[THEN cp2_imp_cp] intro!: cp2_P[THEN cp2_imp_cp])
  apply(rule_tac P="P" in OclAsSequence_Bag_charn_0)
  apply(insert undefCase, simp add: OclUndefined_def lift0_def)
  apply(drule defCase, simp add: localValidDefined2sem)
  apply(rule subst)
  apply(rule_tac P="\<lambda> x St. (P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) x) St" and A="B" in cp_charn) 
  apply(auto simp: lift0_def cp2_P[THEN cp2_imp_cp] intro!: cp2_P[THEN cp2_imp_cp])
  done

(* NOTE: thm cp2_P[THEN cp2_imp_cp] is the one that can be used to apply cp2 properties 
         cp2_by_cp is the theorem used to introduce cp2*)



lemma OCL_is_defopt_OclAsSequence: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSequence())::(('a, 'b::bot)Sequence))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: localValidDefined2sem)
  apply(rule_tac B="self" in OclAsSequence_Bag_charn)
  prefer 3
  apply(rule cp2_by_cp)
  apply(simp_all add: OclUndefined_def DEF_def lift0_def)
  apply(simp add: localValidDefined2sem)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: mem_set_multiset_eq)
  done

lemma OCL_is_def_OclAsSequence:
  "(\<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSequence())::(('a, 'b::bot)Sequence))) = (\<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule ext)
  apply(rule arg_cong) back
  apply(rule_tac B="self" in OclAsSequence_Bag_charn)
  prefer 3
  apply(rule cp2_by_cp)
  apply(simp_all add: OclUndefined_def DEF_def lift0_def)
  apply(simp add: localValidDefined2sem)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: mem_set_multiset_eq)
  done

(* TODO: shorten these two proofs by converting one result in the other one *)

lemma OclAsSequence_mtBag_mtSequence_conv[simp]:
  "((OclAsSequence (\<MathOclMtBag>::('a,'b::bot)Bag))) = (\<MathOclMtSequence> ::('a,'b::bot)Sequence)"
  apply(insert Abs_Bag_Mt_DEF)
  apply(simp add: OclAsSequence_def OclMtBag_def OclMtSequence_def
                  lift0_def lift1_def strictify_def DEF_def)
  done

subsection{* Relating includes and excludes on Bags to their counterparts on Sequences *}

lemma includes_bag2seq:
  "(x::('a,'b)VAL) \<MathOclIncludes> (B::('a,'b::bot)Bag) =
   x \<MathOclIncludes> ((OclAsSequence B)::('a,'b::bot)Sequence)" 
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3
  apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp_all add: OclIncludes_def OCL_Sequence.OclIncludes_def
                  OclAsSequence_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto simp: mem_set_multiset_eq)
  done

lemma excludes_bag2seq:
  "(x::('a,'b)VAL) \<MathOclExcludes> (S::('a,'b::bot)Bag) =
   x \<MathOclExcludes> ((OclAsSequence S)::('a,'b::bot)Sequence)"
  by(simp add: includes_bag2seq)
*)
section{* The conversion to sets *}

defs
   (* asSet     :: "('a,'b::bot) Bag \<Rightarrow> ('a,'b::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> lift1 (strictify (\<lambda>X. Abs_Set_0 \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

subsection{* Helper theorems about the relations of the Set and Bag datatypes *}

(* TODO: needed anymore?

lemma set_Bag_eq_Set_set: 
  "(count x \<bottom> = 0) \<Longrightarrow>
   set_of \<HolOclDropLeft>(Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>))\<HolOclDropRight> = 
   \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set_of x \<HolOclLiftRight>))\<HolOclDropRight>  "
  by(simp)

*)
(*
lemma DEF_Set_if_DEF_Bag:
  "(DEF X) \<Longrightarrow> (DEF (Abs_Set_0 \<HolOclLiftLeft>(set_of \<HolOclDropLeft>(Rep_Bag_0 X)\<HolOclDropRight>)\<HolOclLiftRight>))"
  apply(frule DEF_X_Bag_0[THEN iffD1], auto)
  apply(subgoal_tac "\<bottom> \<notin> set_of c")
  apply(erule DEF_Abs_Set, simp)
  done

lemma in_over_RepAbs:
"\<bottom> \<notin> S \<Longrightarrow> (x \<in> \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>S\<HolOclLiftRight>))\<HolOclDropRight>)= (x \<in> S)"
by(simp)


subsection{* Basic properties of OclAsSet *}

ocl_setup_op [OclAsSet]
 

(* TODO: take proof techniques from OclAsSequence\<dots> when they are properly finished *)

lemma OCL_is_defopt_OclAsSet: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSet())::(('a, 'b::bot)Set))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI) prefer 2
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp add: OclAsSet_def strictify_def lift1_def DEF_def)
  apply(fold DEF_def)
  apply(erule DEF_Set_if_DEF_Bag)
  apply(auto simp: OclAsSet_def strictify_def lift1_def DEF_def)
done

lemma OCL_is_def_OclAsSet:
  "(\<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSet())::(('a, 'b::bot)Set))) = (\<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule ext)
  apply(rule arg_cong) back
  apply(subgoal_tac "((St \<MathOclValid> (\<MathOclIsDefined> \<MathOclAsSet> self)) = (St \<MathOclValid> (\<MathOclIsDefined> self)))")
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule OCL_is_defopt_OclAsSet)
  done

lemma OclAsSet_finite:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,'b::bot)Bag)) =
   (\<tau> \<MathOclValid> \<MathOclIsDefined>(((OclAsSet S)::('a,'b::bot)Set)->size()))"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp_all add: OclAsSet_def OCL_Set.OclSize_def
                      strictify_def lift1_def DEF_def)
  apply(auto)
  done

lemma OclAsSet_mtBag_mtSet_conv[simp]:
  "((OclAsSet (\<MathOclMtBag>::('a,'b::bot)Bag))) = (\<MathOclMtSet>::('a,'b::bot)Set)"
  apply(insert Abs_Bag_Mt_DEF)
  apply(simp add: OclAsSet_def OclMtBag_def OclMtSet_def
                  lift0_def lift1_def strictify_def DEF_def)
  done

subsection{* Fundamental relations between OclAsSequence and OclAsSet *}

lemma bag2seq2set_eq_bag2set[simp]:
  "((OclAsSet ((OclAsSequence (B::('a,'b::bot)Bag))::('a,'b::bot)Sequence))::('a,'b::bot)Set) = 
   ((OclAsSet B)::('a,'b::bot)Set)"
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  apply(simp_all add: cp2_by_cp)
  apply(drule localValidDefined2sem[THEN iffD1])
  apply(simp add: OclAsSet_def OCL_Sequence.OclAsSet_def
                  lift0_def lift1_def strictify_def)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: mem_set_multiset_eq)
  done

 
subsection{* Relating includes and excludes on Bags to their counterparts on Sets *}

lemma includes_bag2set:
  "(x::('a,'b)VAL) \<MathOclIncludes> (S::('a,'b::bot)Bag) =
   x \<MathOclIncludes> ((OclAsSet S)::('a,'b::bot)Set)" 
  by(simp add: includes_seq2set includes_bag2seq)
 

(* TODO: decide how strongly the theories should be intertwined *)
(* This is the old proof for this theorem:

  apply(rule ext)
  apply(case_tac "DEF (S xa)")
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp_all add: OclIncludes_def OCL_Set.OclIncludes_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done
*)

lemma excludes_bag2set:
  "(x::('a,'b)VAL) \<MathOclExcludes> (S::('a,'b::bot)Bag) =
   x \<MathOclExcludes> ((OclAsSet S)::('a,'b::bot)Set)"
  by(simp add: includes_bag2set)

*)
section{* Semantic Definitions of the Standard Bag Operations *}


defs 
   (* size         :: "('a,'b::bot Bag_0)VAL=>'a Integer" *)
   OclSize_def     :"(OclSize::('a, 'b::bot Bag_0) VAL => 'a Integer) \<equiv> 
                      lift1(strictify(\<lambda>X. \<HolOclLiftLeft>int (size(\<HolOclDropLeft> Rep_Bag_0 X \<HolOclDropRight>))\<HolOclLiftRight> ))"

   (* count        :: "[('a, 'b Bag_0) VAL, ('a, 'b::bot) VAL] => 'a Integer" *)
   OclCount_def    : "OclCount \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>int (count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x)\<HolOclLiftRight>)))"

defs
  
   (* includesAll  :: "[('a, 'b::bot Bag_0) VAL, ('a, 'b Bag_0) VAL] => 'a Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight> \<subseteq> set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

   (* excludesAll  :: "[('a, 'b::bot Bag_0) VAL, ('a, 'b Bag_0) VAL] => 'a Boolean" *)
   OclExcludesAll_def: "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> \<inter> set_of \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"
defs

 
   (* OclIsEmpty      ::"('a, 'b::bot Bag_0) VAL => 'a Boolean" *)
   OclIsEmpty_def  : "OclIsEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{#} =  \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"
   OclNotEmpty_def : "OclNotEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{#}  \<noteq>  \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"


defs
(* PROBLEM: wrongly defined in the standard: Idempotence!.
   TODO:    A definition that captures the intention of the standard. *)
   OclSum_def    : "OclSum (self::('a, 'b::bot Bag_0) VAL)  \<equiv> \<MathOclUndefined>"
text{* 
  \OCLglitch[\inlineocl{sum}]{\inlineocl{sum} wrongly defined in the standard: 
  idempotence} 
*}


(* TODO: OclFlatten *)

defs
   OclAsBag_def  : "OclAsBag(self::('a, 'b::bot Bag_0) VAL) \<equiv> self"  


ocl_setup_op [OclSize, OclCount, OclIncludesAll, OclExcludesAll,
              OclIsEmpty, OclNotEmpty]

(* 

defs
   OclFlatten_def: "(OclFlatten ::('a,'b::bot Bag_0 Bag_0)VAL \<Rightarrow> ('a,'b Bag_0)VAL) \<equiv> 
                   lift1 (strictify (\<lambda>X. Abs_Bag_0 
                                           (\<HolOclLiftLeft>\<Union> (Lifting.drop \<circ> Rep_Bag_0) ` \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>)))"

ocl_setup_op [OclFlatten]

*)




section{*The iterators on Sequences*}

(* TODO: Correct this note *)
text{* Note: Quantifiers were not defined on the basis of iterate in \holocl. This
       is since iterate is undefined for all infinite collections, in contrast
       to quantifiers, which do not necessarily have computational content. *}

defs


   (* forall    :: "[('a,'b::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclForAll_def:  "OclForAll S P \<equiv> \<lambda> \<tau>.  
                    if DEF (S \<tau>)
                    then if  \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0(S \<tau>)\<HolOclDropRight>.  P (lift0 x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>) 
                         then \<HolOclLiftLeft>True\<HolOclLiftRight>
                         else if \<exists>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0(S \<tau>)\<HolOclDropRight>. P (lift0 x) \<tau> = (\<HolOclLiftLeft>False\<HolOclLiftRight>) 
                              then \<HolOclLiftLeft>False\<HolOclLiftRight>
                              else \<bottom>
                    else \<bottom>"

defs
   (* select    :: "[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] *)
   OclSelect_def:  "OclSelect S P \<equiv>  \<lambda> \<tau>. 
                    if DEF (S \<tau>)
                    then if    \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight>. DEF (P (lift0 x) \<tau>)
                         then  Abs_Bag_0(\<HolOclLiftLeft> MCollect \<HolOclDropLeft>Rep_Bag_0(S \<tau>)\<HolOclDropRight> (\<lambda>x. P (lift0 x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>) \<HolOclLiftRight>)
                         else \<bottom>
                    else \<bottom>"

   (* collect    ::"[('a,'d::bot Bag) VAL, ('a,'b::bot)VAL => ('a,'c::bot)VAL] *)
   OclCollect_def: "OclCollect S P \<equiv>  \<lambda> \<tau>. 
                    if DEF (S \<tau>)
                    then if \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                         then Abs_Bag_0 \<HolOclLiftLeft>Abs_multiset ((count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight>) o (inv (\<lambda> x. P (lift0 x) \<tau>)))\<HolOclLiftRight> 
                         else \<bottom>
                    else \<bottom>"
defs

   (* iterate     ::"[ ('a,'d::bot)Bag, 
                       [('a,'b::bot)VAL,('a,'b::bot)VAL] => ('a,'c::bot)VAL,
                       ('a,'c::bot)Bag
                     ] 
                       => ('a,'c::bot) Bag" *)
   OclIterate_def : 
   "OclIterate S P A \<equiv> (\<lambda> \<tau>. if DEF (S \<tau>)
                             then  if (\<forall> a b c \<tau>. (P a (lift0 (P b c \<tau>)) \<tau>) =
                                                  (P b (lift0 (P a c \<tau>))) \<tau>)
                                   then (foldl (\<lambda> x y. (P (lift0 y) (lift0 x)) \<tau>) 
                                        (A \<tau>)
                                        (SOME xs. \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight> = multiset_of xs))
                                   else \<bottom>
                             else \<bottom>)"

defs

   (* isUnique    ::"[('a,'c::bot)Bag, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   (* TODO: define OclIsUnique *)

   (*   any       ::"[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] 
                     => ('a,'b::bot)VAL" *)
   OclAny_def   : "OclAny S P \<equiv>  \<lambda> \<tau>. 
                   if DEF (S \<tau>)
                   then if   \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0(S \<tau>)\<HolOclDropRight>. DEF(P (lift0 x) \<tau>) 
                        then SOME x.(x  \<in> set_of \<HolOclDropLeft>Rep_Bag_0(S \<tau>)\<HolOclDropRight> \<and> 
                                    P (lift0 x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>))
                        else \<bottom>
                   else \<bottom>" 

   (*   OclAny_def: "OclAny (S::('a,'b::bot)Bag) P \<equiv> 
                     OclAny ((OclAsSet_of S)::('a,'b::bot)Set_of) P" *)


   (* one         ::"[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclOne_def     : "OclOne (S::('a,('b::bot) Bag_0)VAL) P  \<equiv>
                     \<MathOclSizeLeft>(\<MathOclSelectLeft>x : S | P (x\<Colon>'a\<Colon>type \<Rightarrow> 'b\<Colon>bot)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrongEq> 1"

(* TODO: finish definition

   (* collectionRange :: "[('a,'b::{bot,plus}) VAL,('a,'b) VAL] \<Rightarrow> ('a,'b) Bag" *)
   OclCollectionRange_def :
   "OclCollectionRange \<equiv> lift2 (\<lambda> x\<Colon>Integer. (\<lambda> y. 
                           (Abs_Bag_0 (\<HolOclLiftLeft>(map (\<lambda> z::nat. ((int z) - (\<HolOclDropLeft>x\<HolOclDropRight>))) [0..<((nat \<HolOclDropLeft>y\<HolOclDropRight>) - (nat \<HolOclDropLeft>x\<HolOclDropRight>))])\<HolOclLiftRight>))))"

*)

section{* Proving the basic properties of the iterators *}

ocl_setup_op [OclForAll, OclExists, OclSelect, OclCollect,
              OclIterate, OclAny, OclOne]

(* TODO: prove them *)

section{* Relating the operations for which it is possible to the corresponding sequence operations *}
(*
lemma isEmpty_bag2seq:
  "OclIsEmpty(B::('a, 'b::bot) Bag) = OclIsEmpty ((OclAsSequence B)::('a,('b::bot))Sequence)"
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply((erule exE)+, clarify)
  apply(simp_all add: OclIsEmpty_def OCL_Sequence.OclIsEmpty_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma notEmpty_eq_not_isEmpty:
  "OclNotEmpty(X::('a, 'b::bot) Bag) =  \<MathOclNot> (OclIsEmpty(X))" 
  apply (rule ext)
  apply (simp add:  OclNotEmpty_def OclIsEmpty_def OclNot_def)
  apply (simp only: lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: DEF_def strictify_def 
                                    split add: split_if up.split)
  done

lemma isNotEmpty_bag2seq:
  "OclNotEmpty(S::('a, 'b::bot) Bag) = OclNotEmpty ((OclAsSequence S)::('a,('b::bot))Sequence)"
  by(simp only: notEmpty_eq_not_isEmpty OCL_Sequence.notEmpty_eq_not_isEmpty isEmpty_bag2seq)

(* TODO: Reorder *)

lemma cp_forall1 [simp,intro!]: 
  "\<lbrakk> cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Bag)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'a \<Rightarrow> 'b))"
  apply(simp only: cp_def OclForAll_def) 
  apply(erule exE, rule exI, rule allI, rule allI)
  apply(erule_tac x=X in allE)
  apply(erule_tac x=St in allE)
  apply(simp only:)
  done

lemma forall_bag2seq: 
  "(\<MathOclForAll> x \<in> (B :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclForAll> x \<in> ((OclAsSequence B)::('a,('b::bot))Sequence)  \<bullet> P x)"
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(simp add: OclUndefined_def OclForAll_def OCL_Sequence.OclForAll_def
                  lift0_def)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply((erule exE)+, clarify)
  apply(simp_all add: OclForAll_def OCL_Sequence.OclForAll_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  done

lemma exists_bag2seq:
  "(\<MathOclExists> x \<in> (B :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclExists> x \<in> ((OclAsSequence B)::('a,('b::bot))Sequence)  \<bullet> P x)"
  by(simp add: OclExists_def forall_bag2seq)

lemma includesAll_bag2seq:
  "OclIncludesAll (B1::('a, 'b::bot)Bag) (B2::('a, 'b::bot)Bag) =
   OclIncludesAll ((OclAsSequence B1)::('a, 'b::bot)Sequence) ((OclAsSequence B2)::('a, 'b::bot)Sequence)"
  apply(rule ext)
  apply(rule_tac B="B1" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(rule_tac B="B2" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(rotate_tac 3)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(erule exE)+
  apply(clarify)
  apply(simp_all add: OclIncludesAll_def OCL_Sequence.OclIncludesAll_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  done

lemma excludesAll_bag2seq:
  "OclExcludesAll (B1::('a, 'b::bot)Bag) (B2::('a, 'b::bot)Bag) =
   OclExcludesAll ((OclAsSequence B1)::('a, 'b::bot)Sequence) ((OclAsSequence B2)::('a, 'b::bot)Sequence)"
  apply(rule ext)
  apply(rule_tac B="B1" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(rule_tac B="B2" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(rotate_tac 3)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(erule exE)+
  apply(clarify)
  apply(simp_all add: OclExcludesAll_def OCL_Sequence.OclExcludesAll_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  done

(* TODO: Reorder *)

lemma cp_any1 [simp,intro!]: 
  "\<lbrakk> cp (B :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Bag)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. OclAny (B X) P::('a \<Rightarrow> 'b))"
  apply(simp only: cp_def OclAny_def) 
  apply(erule exE, rule exI, rule allI, rule allI)
  apply(erule_tac x=X in allE)
  apply(erule_tac x=St in allE)
  apply(simp only:)
  done

(* TODO: put into sequence theory *)
lemma cp_any1_sequence [simp,intro!]: 
  "\<lbrakk> cp (B :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. OclAny (B X) P::('a \<Rightarrow> 'b))"
  apply(simp only: cp_def OCL_Sequence.OclAny_def) 
  apply(erule exE, rule exI, rule allI, rule allI)
  apply(erule_tac x=X in allE)
  apply(erule_tac x=St in allE)
  apply(simp only:)
  done

(*
lemma any_bag2seq:
  "(OclAny (B :: (('a,('b::bot))Bag)) (P::('a,'b::bot)VAL \<Rightarrow> ('a,Boolean_0)VAL)) = 
   (OclAny ((OclAsSequence B)::('a,('b::bot))Sequence) P)" 
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(simp add: OclUndefined_def OclAny_def OCL_Sequence.OclAny_def
                  lift0_def cp_def DEF_def)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply((erule exE)+, clarify)
  apply(simp_all add: OclAny_def OCL_Sequence.OclAny_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  done

*)
subsection{* Now come the operations which are only relatible to sequences and not to sets *}

lemma size_bag2seq:
  "(OclSize (B :: (('a,('b::bot))Bag))) =
   (OclSize ((OclAsSequence B)::('a,('b::bot))Sequence))" 
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply((erule exE)+, clarify)
  apply(simp_all add: OclSize_def OCL_Sequence.OclSize_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  apply(rotate_tac -1, erule rev_mp)
  apply(induct_tac xs, auto)
  done

lemma count_bag2seq:
  "(OclCount (B :: (('a,('b::bot))Bag)) (x::('a,'b)VAL)) =
   (OclCount ((OclAsSequence B)::('a,('b::bot))Sequence) x)" 
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_Bag_charn)
  prefer 3; apply(rule cp2_by_cp)
  apply(simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply((erule exE)+, clarify)
  apply(simp_all add: OclCount_def OCL_Sequence.OclCount_def
                  OclAsSequence_def DEF_def mem_set_multiset_eq
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  apply(rotate_tac -1, erule rev_mp)
  apply(induct_tac xs, auto)
  done


(* TODOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO

   REORDER

*)

lemmas ss_lift_strict = lift0_def lift1_def lift2_def lift3_def strictify_def
lemmas ss_lift_strict_DEF = ss_lift_strict DEF_def

lemma iterate_bag2seq:
  assumes ordIndep: "\<forall> a b c \<tau>. (P a (lift0 (P b c \<tau>)) \<tau>) = (P b (lift0 (P a c \<tau>)) \<tau>)"
  shows             "(OclIterate (B :: (('a,('b::bot))Bag)) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
                                  A) =
                     (OclIterate ((OclAsSequence B)::('a,('b::bot))Sequence) P A)" 
  apply(insert ordIndep, rule ext)
  apply(simp add: OclAsSequence_def OclIterate_def OCL_Sequence.OclIterate_def
                  ss_lift_strict)
  apply(case_tac "DEF(B x)")
  apply(frule DEF_X_Bag[THEN iffD1])
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(clarify)
  apply(subgoal_tac "\<bottom> \<notin> (set (SOME xs. (ca = (multiset_of xs))))")
  prefer 2
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD])
  apply(simp add: mem_set_multiset_eq)
  apply(auto simp: DEF_def)
  done

lemma iterate_bag2seq_cp:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and        cpP_2: "\<And> x. cp (P x)"
  shows             "(OclIterate (B :: (('a,('b::bot))Bag)) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
                                  A) =
                     (OclIterate ((OclAsSequence B)::('a,('b::bot))Sequence) P A)"
  by(insert cpP_2 ordIndep, simp add: iterate_bag2seq cp_by_cpify) 

lemma fw_eq_seq2bag:
  assumes eq: "\<And> (S::('a,'b::bot)Sequence). f S = g S"
  and    lhs: "f' (B::('a,'b::bot)Bag) = f (OclAsSequence B)"
  and    rhs: "g' B = g (OclAsSequence B)"
  shows       "f' B = g' B"
  by(subst lhs, subst rhs, rule eq)

(* TODO: how could one use such a theorem\<dots> not unique unification\<dots>
thm fw_eq_seq2bag
thm fw_eq_seq2bag[OF size_by_iterate, of _ _ "id", simplified]
thm size_by_iterate
thm fw_eq_seq2bag[OF forall_by_iterate, simplified]

thm bla
thm size_bag2seq
*)

lemma REQ_11_7_1_1:
  "((self::('a,'b::bot)Bag) ->size()) = 
    OclIterate self (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer). y + 1) 0"
  by(simp add: size_bag2seq iterate_bag2seq size_by_iterate)

lemmas size_by_iterate = REQ_11_7_1_1

lemma REQ_11_7_1_4:
  "((self::('a,'b::bot)Bag) ->count((obj::('a,'b)VAL))) = 
   (OclIterate self 
               (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer).\<MathOclIf>  (x \<MathOclStrictEq> obj)\<MathOclThen> (y+1) \<MathOclElse> (y) \<MathOclEndif> )   
               (\<MathOclIf>    (\<MathOclIsDefined> obj) \<MathOclThen> 0 \<MathOclElse>\<MathOclUndefined> \<MathOclEndif> ) )"
  by(simp add: count_bag2seq iterate_bag2seq count_by_iterate
               OclIf_def plus_def ss_lift_strict)
(* TODO: are there rules about If that make this reasoning simpler? *)

lemmas count_by_iterate = REQ_11_7_1_4


(* TODO: both of the following theorems are false as it was proven in Set *)
(* NOTE: this will be proved later by referring to the corresponding theorems on Set *)
lemma REQ_11_7_1_5: "(self ->includesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->includes(elem))))"
oops


(* NOTE: this will be proved later by referring to the corresponding theorems on Set *)
lemma REQ_11_7_1_6:"(self ->excludesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->excludes(elem))))" 
oops


lemma REQ_11_7_1_7: 
  "self ->isEmpty() = (self::('a,'b::bot)Bag) ->size() \<MathOclStrictEq> 0"
  apply (rule ext)
  apply (simp add:  OclIsEmpty_def OclSize_def OclStrictEq_def 
                    OCL_Integer.Zero_ocl_int_def)
  apply (simp only: lift0_def lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                               split add: split_if up.split)
  apply (simp only: lift0_def lift1_def lift2_def 
                    strictify_def o_def not_def OclIsDefined_def DEF_def OclTrue_def)
  apply auto
  done


lemma REQ_11_7_1_8: 
  "self ->notEmpty() = (\<MathOclNot> ((self::('a,'b::bot)Bag) ->size() \<MathOclStrictEq> 0))"
  by(simp add: notEmpty_eq_not_isEmpty REQ_11_7_1_7)

(* TODO: define Sum for Bags

lemma REQ_11_7_1_9:
  "OclSum S = OclIterate (S::('a,'b::{bot,plus,zero})Bag) (\<lambda> x y. x + y) 0"
  by(simp add: OclSum_def)

*)

(* TODO: define product? Or do we not need it? *)
(*
lemma REQ_11_7_1_10:
  OclProduct S1 S2 = \<dots> \<dots> \<dots>

*)


subsection{* Requirements of the bag functions *}



subsection{* Requirements of the iterator functions on collections *}

lemma REQ_11_9_1_2:
  "(\<MathOclForAll> x \<in> (source::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  apply(subst iterate_bag2seq_cp)
  apply(simp_all add: forall_by_iterate forall_bag2seq and_assoc)
  apply(subst and_commute, simp)
  done

lemmas forall_by_iterate = REQ_11_9_1_2
(* TODO: how can this be automated? *)


lemma REQ_11_9_1_1:
  "(\<MathOclExists> x \<in> (source::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  apply(subst iterate_bag2seq_cp)
  apply(simp_all add: exists_by_iterate exists_bag2seq)
  apply(subst or_commute, subst or_assoc, simp add: subst or_commute)
  oops 

(* @BU: cp_or is missing 

lemmas exists_by_iterate = REQ_11_9_1_1

*)


section{* Relating the operations for which it is possible to the corresponding set operations *}

lemma isEmpty_bag2set:
  "OclIsEmpty(S::('a, 'b::bot) Bag) = OclIsEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  by(simp add: isEmpty_bag2seq isEmpty_seq2set)

lemma isNotEmpty_bag2set:
  "OclNotEmpty(S::('a, 'b::bot) Bag) = OclNotEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  by(simp only: notEmpty_eq_not_isEmpty OCL_Set.not_empty isEmpty_bag2set)

lemma forall_bag2set: 
  "(\<MathOclForAll> x \<in> (S :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclForAll> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  by(simp add: forall_bag2seq forall_seq2set)

lemma exists_bag2set:
  "(\<MathOclExists> x \<in> (S :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclExists> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  by(simp add: OclExists_def forall_bag2set)

lemma includesAll_bag2set:
  "OclIncludesAll (S1::('a, 'b::bot)Bag) (S2::('a, 'b::bot)Bag) =
   OclIncludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  by(simp add: includesAll_bag2seq includesAll_seq2set)

lemma excludesAll_bag2set:
  "OclExcludesAll (S1::('a, 'b::bot)Bag) (S2::('a, 'b::bot)Bag) =
   OclExcludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  by(simp add: excludesAll_bag2seq excludesAll_seq2set)

(* TODO: correct

lemma any_bag2set:
  "(OclAny (S :: (('a,('b::bot))Bag)) (P::('a,'b::bot)VAL \<Rightarrow> ('a,Boolean_0)VAL)) = 
   (OclAny ((OclAsSet S)::('a,('b::bot))Set) P)" 
  by(simp add: any_bag2seq any_seq2set)

*)
*)

(* Bag operations *)
defs 

   (* including     :: "[('a,('b::bot) Bag_0) VAL,('a,'b) VAL] => ('a,'b Bag_0) VAL" *)
   OclIncluding_def  :
   "OclIncluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft>{#Y#} + \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))))"

   (* excluding     :: "[('a,('b::bot) Bag_0) VAL,('a,'b) VAL] => ('a,'b Bag_0) VAL" *)
   OclExcluding_def  :
   "OclExcluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> MCollect \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> (\<lambda> a. a \<noteq> Y) \<HolOclLiftRight>))))"

defs

(* OclComplement :: "('a, 'b::bot Bag_0) VAL => ('a, 'b Bag_0) VAL"  *)
   OclComplement_def:"OclComplement   \<equiv> lift1 (strictify(\<lambda>X. Abs_Bag_0 
                                           (\<HolOclLiftLeft> -(\<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> - {#\<bottom>#})\<HolOclLiftRight>)))"


(* OclUnion      :: "[('a,('b::bot)Bag_0)VAL, ('a,'b Bag_0)VAL] => ('a,'b Bag_0)VAL" *)
   OclUnion_def  :  "OclUnion      \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> + \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"

(* TODO: why can't it by defined twice? once for Bag,Bag and once for Bag,Set *)
(*
(* OclUnion      :: "[('a,('b::bot)Bag_0)VAL, ('a,'b Set_0)VAL] => ('a,'b Bag_0)VAL" *)
   OclUnion_def  :  "OclUnion      \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> + 
                                              Abs_multiset (\<lambda> a. if a \<in> \<HolOclDropLeft>Rep_Set_0 Y\<HolOclDropRight> then 1 else 0)\<HolOclLiftRight>))))"
*)

defs
(* OclIntersection:: "[('a,('b::bot) Bag_0)VAL,('a,'b Bag_0)VAL]\<Rightarrow>('a,'b Bag_0)VAL" *)
   OclIntersection_def: "OclIntersection \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> #\<inter> \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"

(* TODO: the same holds for intersections\<dots> here we need as well two definitions *)



ocl_setup_op [OclIncluding, OclExcluding,
             OclComplement, OclUnion]

ocl_setup_op [OclIntersection]

(* TODO: Add missing theorems

   and is this necessary? what rules are automatically added to the simpsets
   by ocl_setup_op

declare OCL_cp_OclIntersection[simp,intro!]

*)

text{* Note that also sequencing of declarations are possible. *}


syntax
   "@OclFinBag"      :: "args \<Rightarrow> ('a,'b Bag_0)VAL"           ("mkBag{(_)}")

translations
  "mkBag{x, xs}"     == "OclIncluding (mkBag{xs}) x"
  "mkBag{x}"         == "OclIncluding OclMtBag x"

lemma "mkBag{1,1,1,0,1} = ?X" oops (*test *)


section{* Properties of the Operations *}

subsection{* Useful Lemmas (necessary in this theory) *}


(* TODO: reorder *)


lemma OCL_cp_OclIterate [simp,intro!]:
  "\<lbrakk>(cp F); (cp F')\<rbrakk> \<Longrightarrow> 
   (cp (\<lambda>X. (OclIterate ((F X)::(('a,'b::bot)Bag)) (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) (F' X))))"
  apply(subst cp_by_cpify)
  apply(drule cp_by_cpify[THEN iffD1])+
  apply(auto simp: OclIterate_def lift0_def lift1_def)
  done

lemma OCL_undef_1_OclIterate [simp]:
  "OclIterate (\<MathOclUndefined>::('a,'b::bot)Bag) 
              (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
              A 
   = (\<MathOclUndefined>::('a,'c::bot)VAL)"
  by(simp add: OclIterate_def OclUndefined_def lift0_def DEF_def)


text{* Its characterisation on the empty sequence *}
(*
lemma iterate_charn_mtBag[simp]:
  "\<lbrakk> \<forall> a b c. (P a (P b c)) = (P b (P a c));
    \<And> x. cp (P x) \<rbrakk> \<Longrightarrow> 
  (OclIterate (\<MathOclMtBag>::('a,'b::bot)Bag) 
              (P::(('a \<Rightarrow> 'b) \<Rightarrow> (('a \<Rightarrow> 'c::bot) \<Rightarrow> ('a \<Rightarrow> 'c)))) 
               A) = A"
  apply(rule ext)
  apply(simp_all add: localValidDefined2sem strictify_def
                      OclIterate_def OclIncluding_def OclMtBag_def
                      lift0_def lift1_def lift2_def cp_by_cpify)
 done

lemma foldr_multiset_perm:
  shows   "\<And> xa xs. \<lbrakk> multiset_of xa = multiset_of xs; 
                      \<forall> a b c. (P a (P b c)) = (P b (P a c)) \<rbrakk> 
           \<Longrightarrow> foldr P xa e = foldr P xs e"
  apply(subst (asm) multiset_of_eq_perm)
  apply(frule rev_mp) prefer 2
  apply(assumption)
  apply(rule_tac xb="xa" and xa="xs" in perm.induct)
  apply(simp_all)
  done

lemma foldl_multiset_perm:
  shows   "\<And> xa xs. \<lbrakk> multiset_of xa = multiset_of xs; 
                      \<forall> a b c. (P (P a b) c) = (P (P a c) b) \<rbrakk> 
           \<Longrightarrow> foldl P e xa = foldl P e xs"
  apply(subst foldl_foldr)+
  apply(rule foldr_multiset_perm)
  apply(subst (asm) multiset_of_eq_perm)
  apply(subst multiset_of_eq_perm)
  apply(rule Permutation.trans, rule perm_rev)
  apply(rule perm_sym, rule Permutation.trans, rule perm_rev)
  apply(rule perm_sym, simp_all)
  done

lemma iterate_charn_including_0:
  assumes ordIndep: "\<forall> a b c \<tau>. (P a (lift0 (P b c \<tau>)) \<tau>) = (P b (lift0 (P a c \<tau>)) \<tau>)"
  and     defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclIncluding B x) P A) \<tau>) =
         (P (lift0 (x \<tau>)) (lift0 ((OclIterate (B::(('a,'b::bot)Bag)) P A) \<tau>)) \<tau>)"
  apply(insert defx defB ordIndep)
  apply(simp add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(clarify)
  apply(simp add: OclIncluding_def OclIterate_def ss_lift_strict_DEF)
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD])
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD])
  apply(subgoal_tac "multiset_of (xa @ [x \<tau>]) = multiset_of xaa")
  apply(drule_tac P="\<lambda> x y. (P (\<lambda>s. y) (\<lambda>s. x) \<tau>)" and e="A \<tau>" in foldl_multiset_perm)
  apply(auto simp: union_ac)
  done

lemma iterate_charn_including:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and     cpP_1:    "\<And> y. cp (\<lambda> x. P x y)"
  and     cpP_2:    "\<And> x. cp (P x)"
  and     defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclIncluding B x) P A) \<tau> ) =
         (P x (OclIterate (B::('a,'b::bot)Bag) P A) \<tau>)"
  apply(rule sym[OF trans]) 
  apply(rule_tac x="x" in cp_subst, simp add: cpP_1)
  apply(rule trans)
  apply(rule_tac x="(B \<MathOclIterate>(u;ua=A | (P u ua)))" in cp_subst)
  apply(insert cpP_2, simp add: cp_by_cpify lift0_def)
  apply(rule sym[OF iterate_charn_including_0])
  apply(simp add: ordIndep cp_by_cpify)
  apply(rule defB, rule defx)
  done

lemma iterate_opt_charn_including:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and    P_undef_1: "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and    P_undef_2: "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and        cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and        cpP_2: "\<And> x. cp ((P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) x)"
  shows              "(OclIterate (OclIncluding B x) P A) =
                      (P x (OclIterate (B::(('a,'b::bot)Bag)) P A))"
  apply(insert P_undef_1 P_undef_2 cpP_1 cpP_2 ordIndep)
  apply(rule ext)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> B")
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> x")
  apply(rule iterate_charn_including, simp_all)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="x" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="x" in cp_charn, simp, simp)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="B" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="B" in cp_charn)
  apply(rule_tac P="P x" in cp_compose2, simp_all)
  done


subsection{* Forward rules for functions defined by iterate *}

lemma f_by_iterate_undef_1:
  "\<lbrakk> \<And> B. f (B::('a,'b::bot)Bag) = 
     OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclUndefined> = \<MathOclUndefined>"
  by(simp)
  
lemma f_by_iterate_mtBag:
  "\<lbrakk> \<And> B. f (B::('a,'b::bot)Bag) = 
     OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A;
     \<forall> a b c. (P a (P b c)) = (P b (P a c));
     \<And> x. cp (P x) \<rbrakk>
   \<Longrightarrow>
   f \<MathOclMtBag> = A"
  by(simp)

lemma f_by_iterate_including: 
  assumes f_by_it: "\<And> B. (f (B::('a,'b::bot)Bag)) = 
                         (OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  and  ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and      defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and      defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  and     cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and     cpP_2: "\<And> x. cp (P x)"
  shows "(f (OclIncluding B x) \<tau>) = (P x (f B) \<tau>)"
  by(simp add: f_by_it iterate_charn_including ordIndep defB defx cpP_1 cpP_2)


lemma f_by_iterate_opt_including:
  assumes f_by_it: "\<And> S. (f (S::('a,'b::bot)Bag)) = 
                         (OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  and    ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and   P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and   P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and       cpP_1:  "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:  "\<And> x. cp (P x)"
  shows "f (OclIncluding S x) = P x (f S)"
  by(simp add: f_by_it iterate_opt_charn_including ordIndep
               P_undef_1 P_undef_2 cpP_1 cpP_2)


thm f_by_iterate_undef_1[OF size_by_iterate]
thm f_by_iterate_mtBag[OF size_by_iterate, simplified]
thm f_by_iterate_including[OF size_by_iterate, simplified]
thm f_by_iterate_opt_including[OF size_by_iterate, simplified]

thm f_by_iterate_undef_1[OF count_by_iterate]
thm f_by_iterate_mtBag[OF count_by_iterate, simplified]
thm f_by_iterate_including[OF count_by_iterate, simplified]
thm f_by_iterate_opt_including[OF count_by_iterate, simplified]
(* TODO: add simplification in the last rule to remove the last assumption *)

thm forall_by_iterate

thm f_by_iterate_undef_1[OF forall_by_iterate]
thm f_by_iterate_mtBag[OF forall_by_iterate, simplified]
thm f_by_iterate_including[OF forall_by_iterate, simplified]
thm f_by_iterate_opt_including[OF forall_by_iterate, simplified]
(* TODO: fix cp-ness *)

(*
thm f_by_iterate_undef_1[OF exists_by_iterate]
thm f_by_iterate_mtBag[OF exists_by_iterate]
thm f_by_iterate_including[OF exists_by_iterate, simplified]
thm f_by_iterate_opt_including[OF exists_by_iterate, simplified]
(* TODO: fix cp-ness *)

*)

text{* Old but still a little bit noteworthy stuff *}

lemma unzip_zip_id:
  "\<lbrakk> length xs = length ys; unzip = (\<lambda> x. (map fst x, map snd x)); curried = (\<lambda> x. P (fst x) (snd x))  \<rbrakk> \<Longrightarrow>
   P xs ys = curried (unzip (zip xs ys))"
  by(simp)


lemma length_induct2:
  "\<lbrakk> length as = length bs;
     (\<forall> xs ys. length xs = length ys \<and>
         (\<forall> ws vs. 
            ((length ws < length xs) \<and> (length ws = length vs)) \<longrightarrow> P ws vs) 
         \<longrightarrow> P xs ys) \<rbrakk>
   \<Longrightarrow> P as bs"
  apply(rule_tac t="P as bs" in subst, rule sym[OF unzip_zip_id])
  apply(simp) prefer 3
  apply(rule_tac xs="(zip as bs)" in length_induct)
  prefer 3 prefer 3
  apply(rule refl)+
  apply(erule_tac x="(fst ((map fst xs), (map snd xs)))" in allE)
  apply(rotate_tac -1)
  apply(erule_tac x="(snd ((map fst xs), (map snd xs)))" in allE)
  apply(erule mp)
  apply(auto)
  apply(erule_tac x="zip ws vs" in allE)
  apply(simp)
  done

lemma rev_list_induct2:
  "\<lbrakk>length xs = length ys; P [] [];
    \<And> x xs y ys. \<lbrakk> length xs = length ys; P xs ys \<rbrakk> \<Longrightarrow> P (xs @ [x]) (ys @ [y])\<rbrakk>
   \<Longrightarrow> P xs ys"
  apply(simplesubst rev_rev_ident[symmetric])
  apply(subst rev_rev_ident[symmetric])
  apply(rule_tac xs="rev xs" and ys="rev ys" in list_induct2)
  apply(simp_all)
  done


*)
end


