(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * holocl_term.sml - 
 * Copyright (C) 2002-2006  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

signature HOLOCL_TERM =
sig  
    (** Constructor for not. *)
    val OclNot          : typ -> term -> term
    (** Constructor for oclIsDefined. *)
    val OclIsDefined    : typ -> term -> term
    (** Constructor for oclIsUndefined. *)
    val OclIsUndefined  : typ -> term -> term
    (** Constructor for true. *)
    val OclTrue         : typ -> term 
    (** Constructor for false. *)
    val OclFalse        : typ -> term 
    (** Constructor for and. *)
    val OclAnd          : typ -> term -> term -> term
    (** Constructor for or. *)
    val OclOr           : typ -> term -> term -> term
    (** Constructor for xor. *)
    val OclXor          : typ -> term -> term -> term
    (** Constructor for if _ then _ else _ endif. *)
    val OclIf           : typ -> term -> term -> term -> term
    (** Constructor for implies. *)
    val OclImplies      : typ -> term -> term -> term
    (** Constructor for implies1. *)
    val OclImplies1     : typ -> term -> term -> term
    (** Constructor for implies2. *)
    val OclImplies2     : typ -> term -> term -> term
    (** Constructor for strict equality. *)
    val OclStrictEq     : typ -> term -> term -> term
    (** Constructor for referential equality. *)
    val OclRefEq        : typ -> term -> term -> term
    (** Constructor for strong equality. *)
    val OclStrongEq     : typ -> term -> term -> term
    (** Constructor for not equal. *)
    val OclNeq          : typ -> term -> term -> term
    (** Convert (HOL-)OCL formulae in abstract OCL syntax to Isabelle
      * term representation. 
      *)
    val Ocl2holocl      : rep_encoder_common.Level -> theory -> Rep_Core.Classifier list -> typ 
			  -> Rep_HolOclTerm.HolOclTerm -> term
end

structure holocl_term  : HOLOCL_TERM  =
struct

open Rep_OclType
open Rep_OclTerm
open Rep_Core
open rep_encoder_common

fun charset_of Integer           u   = error ("charset_of 'Integer' not yet supported") (* IntegerT *)
  | charset_of String            u   = error ("charset_of 'String' not yet supported" )(*StringT *)
  | charset_of Real              u   = error ("charset_of 'Real' not yet supported") (*RealT *)
  | charset_of Boolean           u   = error ("charset_of 'Boolean' not yet supported") (*BooleanT *)
  | charset_of (Set(t))          u   = error ("charset_of '"^(string_of_OclType (Set t))^"' not yet supported" )
  | charset_of (OrderedSet(t))   u   = error ("charset_of '"^(string_of_OclType (OrderedSet t))^"' not yet supported") 
  | charset_of (Bag(t))          u   = error ("charset_of '"^(string_of_OclType (Bag t))^"' not yet supported") 
  | charset_of (Sequence(t))     u   = error ("charset_of '"^(string_of_OclType (Sequence t))^"' not yet supported")
  | charset_of (Classifier(s))   u   = Const((string_of_path (hc_path l1 s)),dummyT)
  | charset_of DummyT            u   = error ("Error in charset_of(DummyT)")
  | charset_of OclAny            u   = Const("UML_OCL.OclAny",SETT(OclAnyT u,St u))
  | charset_of t                 u   = error ("Error in charset_of '"^(string_of_OclType t)^"'")



fun type_charset_of Integer         u  = error ("type_charset_of 'Integer' not yet supported") (* IntegerT *)
  | type_charset_of String          u  = error ("type_charset_of 'String' not yet supported") (*StringT *)
  | type_charset_of Real            u  = error ("type_charset_of 'Real' not yet supported") (*RealT *)
  | type_charset_of Boolean         u  = error ("type_charset_of 'Boolean' not yet supported") (*BooleanT *)
  | type_charset_of (Set(t))        u  = error ("type_charset_of '"^(string_of_OclType (Set t))^"' not yet supported") 
  | type_charset_of (OrderedSet(t)) u  = error ("type_charset_of '"^(string_of_OclType (OrderedSet t))^"' not yet supported")
  | type_charset_of (Bag(t))        u  = error ("type_charset_of '"^(string_of_OclType (Bag t))^"' not yet supported")
  | type_charset_of (Sequence(t))   u  = error ("type_charset_of '"^(string_of_OclType (Sequence t))^"' not yet supported") 
  | type_charset_of (Classifier(s)) u  = Const((string_of_path (hc_path l1 (type_charset_name_of s))),dummyT)
  | type_charset_of DummyT          u  = error ("Error in type_charset_of(DummyT)")
  | type_charset_of OclAny          u  = Const("OCL_OclAny_type.OclAny_type",dummyT)
  | type_charset_of t               u  = error ("Error in type_charset_of '"^(string_of_OclType t)^"'")


(**************************************)
(* OCL_Let                            *)
(**************************************)
fun OclLet u v e  = Const("OCL_Let.OclLet",dummyT) $ v $ e
(**************************************)
(* OclAny                             *)
(**************************************)
fun OclAny u = Const("OclAny.oid",dummyT)
fun OclAsType u self C = Const("OclAsType",dummyT) $ self $ C
fun OclIsTypeOf u self C = Const("OclIsTypeOf",dummyT) $ self $ C
fun OclIsKindOf u self C = Const("OclIsKindOf",dummyT) $ self $ C

fun HolNot  x          = Const("Not",boolT --> boolT) $ x 
fun HolAnd  x y        = Const("op &",boolT --> boolT --> boolT)$ x $ y 
fun HolOr  x y         = Const("op |",boolT --> boolT --> boolT) $ x $ y
fun HolXor  x y        = Const("op |",boolT --> boolT --> boolT)
			      $ (Const("op &",boolT --> boolT --> boolT)
				      $ x 
				      $ ( Const("Not",boolT --> boolT) $ y)) 
			      $ (Const("op &",boolT --> boolT --> boolT)
				      $ ( Const("Not",boolT --> boolT) $ x) 
				      $ y) 
fun HolImplies  a i    = Const("op -->",boolT --> boolT --> boolT) $ a $ i
fun OclLocalValid u tau expr    = Const("OCL_Logic_core.OclLocalValid",dummyT) $ tau $ expr

(**************************************)
(* OCL_Boolean                        *)
(**************************************)
fun OclNot u x          = Const("OCL_Boolean.OclNot",mk_funT(BOOLT (St u), BOOLT (St u))) $ x 
fun OclIsDefined u x    = Const("OCL_Boolean.OclIsDefined",dummyT) $ x
fun OclIsUndefined u x  = OclNot u (Const("OCL_Boolean.OclIsDefined",dummyT) $ x)
fun OclTrue  u          = Const("OCL_Boolean.OclTrue", (BOOLT (St u)))
fun OclFalse u          = Const("OCL_Boolean.OclFalse", (BOOLT (St u)))
fun OclAnd u x y        = Const("OCL_Boolean.OclAnd",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u))))$ x $ y 
fun OclOr u x y         = Const("OCL_Boolean.OclOr",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u)))) $ x $ y
fun OclXor u x y        = Const("OCL_Boolean.OclXor",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u)))) $ x $ y
fun OclIf u c t e       = Const("OCL_Boolean.OclIf",dummyT) $ c $ t $ e
fun OclImplies u a i     = Const("OCL_Boolean.OclImplies",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u)))) $ a $ i
fun OclImplies1 u a i   = Const("OCL_Boolean.OclImplies1",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u)))) $ a $ i
fun OclImplies2 u a i   = Const("OCL_Boolean.OclImplies2",mk_funT(BOOLT (St u),mk_funT(BOOLT (St u), BOOLT (St u)))) $ a $ i
fun OclStrictEq u e1 e2   = Const("OCL_Boolean.OclStrictEq",dummyT) $ e1 $ e2
fun OclRefEq u e1 e2    = Const("OCL_Boolean.OclRefEq",dummyT) $ e1 $ e2
fun OclStrongEq u e1 e2 = Const("OCL_Boolean.OclStrongEq",dummyT) $ e1 $ e2
fun OclNeq u e1 e2      = Const("OCL_Boolean.OclNot",dummyT) $ (OclStrictEq u e1 e2)
(**************************************)
(* OCL_Numeral;s                      *)
(**************************************)
fun OclLe u x y         = Const("OCL_Numerals.OclLe",dummyT) $ x $ y
fun OclLess u x y       = Const("OCL_Numerals.OclLess",dummyT) $ x $ y
fun OclGe u x y         = Const("OCL_Numerals.OclGe",dummyT) $ x $ y
fun OclGreater u x y    = Const("OCL_Numerals.OclGreater",dummyT) $ x $ y
(**************************************)
(* OCL_Real                           *)
(**************************************)
fun OclRealRound u x    = Const("OCL_Real.OclRound",mk_funT(REALT (St u), INTEGERT (St u))) $ x
fun OclRealFloor u x    = Const("OCL_Real.OclFloor",mk_funT(REALT (St u), INTEGERT (St u))) $ x
fun OclRealMin u x y    = Const("OCL_Real.OclMin",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u)))) $ x $ y
fun OclRealMax u x y    = Const("OCL_Real.OclMax",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u)))) $ x $ y
fun OclRealDivide u x y = Const("OCL_Real.OclDivide",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u))))$ x $ y
fun OclRealAbs u x      = Const("OCL_Real.OclAbs",mk_funT(REALT (St u), REALT (St u))) $ x 
fun OclRealNegative u x = Const("OCL_Real.OclNegative",mk_funT(REALT (St u), REALT (St u))) $ x 
fun OclRealMinus u x y  = Const("op -",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u))))$ x $ y 
fun OclRealPlus u x y   = Const("op +",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u))))$ x $ y 
fun OclRealTimes u x y  = Const("op *",mk_funT(REALT (St u),mk_funT(REALT (St u), REALT (St u))))$ x $ y 
fun HolOclToReal u x    = (Const("OCL_Library.HolOclToReal",mk_funT(INTEGERT (St u),REALT (St u))) $ x)
(**************************************)
(* OCL_Integer                        *)
(**************************************)
fun OclIntMod u x y    = Const("OCL_Integer.OclMod",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u)) $ x $ y
fun OclIntDiv u x y    = Const("OCL_Integer.OclDiv",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u)) $ x $ y
fun OclIntRound u x    = Const("OCL_Integer.OclRound",dummyT) $ x 
fun OclIntFloor u x    = Const("OCL_Integer.OclFloor",dummyT) $ x 
fun OclIntMin u x y    = Const("OCL_Integer.OclMin",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u)) $ x $ y
fun OclIntMax u x y    = Const("OCL_Integer.OclMax",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u))$ x $ y 
fun OclIntAbs u x      = Const("OCL_Integer.OclAbs",dummyT) $ x 
fun OclIntNegative u x = Const("OCL_Integer.OclNegative",dummyT) $ x 
fun OclIntMinus u x y  = Const("op -",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u))$ x $ y 
fun OclIntPlus u x y   = Const("op +",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u))$ x $ y 
fun OclIntTimes u x y  = Const("op *",INTEGERT (St u) --> INTEGERT (St u) -->  INTEGERT (St u))$ x $ y 
(**************************************)
(* OCL_Undefined                      *)
(**************************************)
fun OclUndefined u = Const("OCL_Undefined.OclUndefined",(BOOLT (St u)))
(**************************************)
(* OCL_String                         *)
(**************************************)
fun OclConcat u s1 s2    = Const("OCL_String.OclConcat",STRINGT (St u) -->STRINGT (St u) --> STRINGT (St u)) $ s1 $ s2
fun OclStringSize u s    = Const("OCL_String.stringsize",dummyT) $ s
fun OclSubString u s b e = Const("OCL_String.OclSubstring",dummyT) $ s $ b $ e
fun OclToLower u s  = Const("OCL_String.OclToLower",STRINGT (St u) --> STRINGT (St u)) $ s 
fun OclToUpper u s  = Const("OCL_String.OclToUpper",STRINGT (St u) --> STRINGT (St u)) $ s 
(**************************************)
(* OCL_Collection                    *)
(**************************************)
(*
fun OclSize styp u C          = Const("OCL_Collection.OclSize",mk_funT(V(u, tilde u styp), 
								    INTEGERT (St u))) $ C
*)
fun OclSize styp u C          = Const("OCL_Collection.OclSize",dummyT) $ C


fun OclCount u C e       = Const("OCL_Collection.OclCount",dummyT) $ C $ e
fun OclIncludes styp etyp u C e    =  
   let 
       val intyp = case styp of 
		      Set et        => mk_funT(V(u, tilde u (Set et)), mk_funT(V(u, tilde u ( et)),BOOLT(St u)))
		    | Sequence et   => mk_funT(V(u, tilde u (Sequence et)),mk_funT( V(u, tilde u ( et)),BOOLT(St u)))
		    | OrderedSet et => mk_funT(V(u, tilde u (OrderedSet et)),mk_funT( V(u, tilde u ( et)),BOOLT(St u)))
		    | Bag et        => mk_funT(V(u, tilde u (Bag et)), mk_funT(V(u, tilde u ( et)),BOOLT(St u)))
		    | t => error ("unkown collection source '"^(string_of_OclType t)^"' in OclIncludes")
  in
    Const("OCL_Collection.OclIncludes",intyp)$ C  $ e
end
fun OclExcludes styp u  C e  =
   let 
       val intyp = case styp of 
		      Set et        => mk_funT(V(u, tilde u (Set et)), mk_funT(V(u, tilde u ( et)),BOOLT(St u)))
		    | Sequence et   => mk_funT(V(u, tilde u (Sequence et)),mk_funT( V(u, tilde u ( et)),BOOLT(St u)))
		    | OrderedSet et => mk_funT(V(u, tilde u (OrderedSet et)),mk_funT( V(u, tilde u ( et)),BOOLT(St u)))
		    | Bag et        => mk_funT(V(u, tilde u (Bag et)), mk_funT(V(u, tilde u ( et)),BOOLT(St u)))
		    | t => error ("unkown collection source '"^(string_of_OclType t)^"' in OclExcludes")
  in
    Const("OCL_Collection.OclExcludes",intyp)$ C  $ e
end


fun OclFlatten u C       = Const("OCL_Collection.OclFlatten",dummyT) $ C 
fun OclSum styp rtyp u C  = Const("OCL_Collection.OclSum",mk_funT(V(u, tilde u styp),V(u, tilde u rtyp))) $ C
fun OclAsSet styp u C    = 
    let
	val _   = trace 99 ("OclAsSet got type "^(string_of_OclType styp)) 
	val typ = case styp of 
		      Set et        => mk_funT(V(u, tilde u (Set et)), V(u, tilde u (Set et)))
		    | Sequence et   => mk_funT(V(u, tilde u (Sequence et)), V(u, tilde u (Set et)))
		    | OrderedSet et => mk_funT(V(u, tilde u (OrderedSet et)), V(u, tilde u (Set et)))
		    | Bag et        => mk_funT(V(u, tilde u (Bag et)), V(u, tilde u (Set et)))
		    | t => error ("unkown collection source '"^(string_of_OclType t)^"' in OclAsSet")
    in 
	Const("OCL_Collection.OclAsSet",typ) $ C
    end
    
fun mkSingularCollection  etyp ctyp u C =  
    let
	val _ = trace 99 ("mkSingularCollection for "^(string_of_OclType etyp)
			 ^" got result type "
			 ^(string_of_OclType ctyp))        
	val coltyp = case ctyp of 
			Set et        => V(u, tilde u (Set etyp))
		      | Sequence et   => V(u, tilde u (Sequence etyp))
		      | OrderedSet et => V(u, tilde u (OrderedSet etyp))
		      | Bag et        => V(u, tilde u (Bag etyp))
		      | t => error ("unkown collection source '"^(string_of_OclType t)^"' in mkSingularCollection")
	val eltyp = V(u, tilde u etyp)
    in 
	Const("OCL_Collection.OclIncluding",mk_funT(coltyp,mk_funT(eltyp,coltyp)))
	     $ Const ("OCL_Set.OclMtSet",coltyp) $ C
    end

fun OclIncluding  styp  u C e    =  
    let 
	val intyp = case styp of 
			Set et        => mk_funT(V(u, tilde u (Set et)), 
					      mk_funT(V(u, tilde u (et)), V(u, tilde u (Set et))  ))
		      | Sequence et   => mk_funT(V(u, tilde u (Sequence et)),
					      mk_funT( V(u, tilde u ( et)),V(u, tilde u (Sequence et))   ))
		      | OrderedSet et => mk_funT(V(u, tilde u (OrderedSet et)),
					      mk_funT( V(u, tilde u ( et)), V(u, tilde u (OrderedSet et))  ))
		      | Bag et        => mk_funT(V(u, tilde u (Bag et)), 
					      mk_funT(V(u, tilde u ( et)), V(u, tilde u (Bag et))  ))
		      | t => error ("unkown collection source '"^(string_of_OclType t)^"' in OclIncluding")
    in
	Const("OCL_Collection.OclIncluding",intyp)$ C  $ e
    end
    

fun OclAsSequence u C    = Const("OCL_Collection.OclAsSequence",dummyT)$ C 
fun OclAsBag u C         = Const("OCL_Collection.OclAsBag",dummyT) $ C
fun OclAsOrderedSet u C  = Const("OCL_Collection.OclAsOrderedSet",dummyT) $ C
fun OclIncludesAll u C e = Const("OCL_Collection.OclIncludesAll",dummyT) $ C $ e
fun OclExcludesAll u C e = Const("OCL_Collection.OclExcludesAll",dummyT) $ C $ e
fun OclIsEmpty u C       = Const("OCL_Collection.OclIsEmpty",dummyT) $ C
fun OclNotEmpty u C      = Const("OCL_Collection.OclNotEmpty",dummyT) $ C
fun OclExists u C be     = Const("OCL_Collection.OclExists",dummyT) $ C $ be
fun OclForAll u C be     = Const("OCL_Collection.OclForAll",dummyT) $ C $ be
fun OclSelect u C be     = Const("OCL_Collection.OclSelect",dummyT) $ C $ be
fun OclReject u C be     = Const("OCL_Collection.OclReject",dummyT) $ C $ be

fun OclCollect styp btyp rtyp u C be    = 
    let
 	val range = case styp of 
			Set et        => V(u, tilde u (et))
		      | Sequence et   => V(u, tilde u (et))
		      | OrderedSet et => V(u, tilde u (et))
		      | Bag et        => V(u, tilde u (et))
		      | t => error ("unkown collection result '"^(string_of_OclType t)^"' in OclCollect")
    in
	Const("OCL_Collection.OclCollect",mk_funT(V(u, tilde u (styp)), 
					       mk_funT(mk_funT(range,V(u, tilde u btyp)), V(u, tilde u (rtyp)))))  $ C $ be 
    end

fun OclIterate u C be e  = Const("OCL_Collection.OclIterate",dummyT) $ C $ be $ e
fun OclIsUnique  u C be  = Const("OCL_Collection.OclIsUnique",dummyT) $ C $ be
fun OclOne  u C be       = Const("OCL_Collection.OclCollOne",dummyT) $ C $ be
fun OclAny  u C be       = Const("OCL_Collection.OclCollAny",dummyT) $ C $ be
fun OclUnion u S1 S2     = Const("OCL_Collection.OclUnion",dummyT) $ S1 $ S2
fun OclIntersection u S1 S2     = Const("OCL_Collection.OclIntersection",dummyT) $ S1 $ S2
fun OclCollectionRange styp u S low up =  Const("OCL_Collection.OclRange",dummyT) $ S $ low $ up
(**************************************)
(* OCL_OrderedSet                     *)
(**************************************)
fun OclOSetSize u S          = Const("OCL_OrderedSet.sosetsize",dummyT) $ S 
fun OclOSetCount u S e       = Const("OCL_OrderedSet.sosetcount",dummyT) $ S $ e
fun OclOSetIncludesAll u S e = Const("OCL_OrderedSet.sosetincludesall",dummyT) $ S $ e
fun OclOSetExcludesAll u S e = Const("OCL_OrderedSet.sosetexcludesall",dummyT) $ S $ e
fun OclOSetIsEmpty u S       = Const("OCL_OrderedSet.sosetisempty",dummyT) $ S 
fun OclOSetNotEmpty u S      = Const("OCL_OrderedSet.sosetnotempty",dummyT) $ S
fun OclOSetInsert u S e      = Const("OCL_OrderedSet.insert'",dummyT) $ S $ e
fun OclOsetUnion u S1 S2     = Const("OCL_OrderedSet.union",dummyT) $ S1 $ S2
(**************************************)
(* OCL_Sequence                       *)
(**************************************)
fun OclSeqUnion u S1 S2       = Const("OCL_Sequence.union",dummyT) $ S1 $ S2
fun OclSeqSubSequence u S b e = Const("OCL_Sequence.OclSubSequence",dummyT) $ S $ b $ e

fun OclSeqFirst u styp S          = 
    let 
	val range = case styp of 
			Set et        => V(u, tilde u (et))
		      | Sequence et   => V(u, tilde u (et))
		      | OrderedSet et => V(u, tilde u (et))
		      | Bag et        => V(u, tilde u (et))
		      | t => error ("unkown collection result '"^(string_of_OclType t)^"' in OclCollect")
    in
	Const("OCL_Collection.OclFirst",mk_funT(V(u, tilde u (styp)),range)) $ S 
    end
fun OclSeqLast u styp S = 
    let 
	val range = case styp of 
			Set et        => V(u, tilde u (et))
		      | Sequence et   => V(u, tilde u (et))
		      | OrderedSet et => V(u, tilde u (et))
		      | Bag et        => V(u, tilde u (et))
		      | t => error ("unkown collection result '"^(string_of_OclType t)^"' in OclCollect")
    in
	Const("OCL_Collection.OclLast",mk_funT(V(u, tilde u (styp)),range)) $ S 
    end

fun OclSeqAppennd u S1 S2     = Const("OCL_Collection.OclAppend",dummyT) $ S1 $ S2
fun OclSeqPrepend u S1 S2     = Const("OCL_Collection.OclPrepend",dummyT) $ S1 $ S2
fun OclSeqIncluding u S e     = Const("OCL_Collection.OclIncluding",dummyT) $ S $ e
fun OclSeqExcluding u S e     = Const("OCL_Collection.OclExcluding",dummyT) $ S $ e
fun OclSeqAt u S p            = Const("OCL_Collection.OclAt",dummyT) $ S $ p
fun OclSeqInsertAt u S e p    = Const("OCL_Collection.OclInsertAt",dummyT) $ S $ e $ p
(**************************************)
(* OCL_Bag                            *)
(**************************************)
fun OclBagSize u B          = Const("OCL_Bag.bagsize",dummyT) $ B
fun OclBagIncludes u B e    = Const("OCL_Bag.OCLincludes",dummyT) $ B $ e
fun OclBagExcludes u B e    = Const("OCL_Bag.excludes",dummyT) $ B $ e
fun OclBagCount u B e       = Const("OCL_Bag.bagcount",dummyT) $ B $ e
fun OclBagIncludesAll u B e = Const("OCL_Bag.bagincludesall",dummyT) $ B $ e
fun OclBagExcludesAll u B e = Const("OCL_Bag.bagexcludesall",dummyT) $ B $ e
fun OclBagIsEmpty u B       = Const("OCL_Bag.bagisempty",dummyT) $ B
fun OclBagNotEmpty u B      = Const("OCL_Bag.bagnotempty",dummyT) $ B
fun OclBagComplement B      = Const("OCL_Bag.complement",dummyT) $ B 
fun OclBagInsert u B e      = Const("OCL_Bag.insert'",dummyT) $ B $ e
fun OclBagUnion u B1 B2     = Const("OCL_Bag.union",dummyT) $ B1 $ B2
(**************************************)
(* OCL_Set                            *)
(**************************************)
fun OclSetComplement u S    = Const("OCL_Set.OclComplement",dummyT) $ S
fun OclSetIncluding u S e   = Const("OCL_Set.OclIncluding",dummyT) $ S $ e
fun OclSetExcluding u S e   = Const("OCL_Set.OclExcluding",dummyT) $ S $ e
fun OclSetUnion u S1 S2     = Const("OCL_Set.OclUnion",dummyT) $ S1 $ S2
fun OclSetExcludes u S e    = Const("OCL_Set.OclExcludes",dummyT) $ S $ e
fun OclSetIncludesAll u S e = Const("OCL_Set.OclIncludesAll",dummyT) $ S $ e
fun OclSetEcludesAll u S e  = Const("OCL_Set.OclExcludesAll",dummyT) $ S $ e
(**************************************)
(* OCL_Tuple                          *)
(**************************************)

(**************************************)
(* OCL_logic                          *)
(**************************************)

(**************************************)
(* OCL_StateAccess                    *)
(**************************************)
fun OclIsNew u e              = Const("OCL_State.OclIsNew",dummyT) $ e
fun OclAllInstances u e       = Const("OCL_State.OclAllInstances",dummyT) $ e
fun OclAllInstanecesAtpre u e = Const("OCL_State.OclAllInstancesAtpre",dummyT) $ e
fun OclModifiedOnly styp u C  = Const("OCL_State.OclModifiedOnly",dummyT) $ C
(**************************************)
(* OCL_Object                         *)
(**************************************)
(* fun OclIsTypeOf u e = Const("OCL_Object.oclIsTypeOf",dummyT) $ e
fun OclIsKindOf u e = Const("OCL_Object.oclIsKindOf",dummyT) $ e *)
(*************************************)

fun Ocl2holocl level thy cl u oclterm =  let

fun ocl2holocl u oclterm =  let
    fun TimedAttributeCall (src,ts,p,t) time = let 
	    val cname = case level of 
			    l1 => string_of_path (("UML_OCL"::["level1"])@p)
			  | l2 => string_of_path p
	    val cname = if time then cname^"_pre" else cname
	 
	    

	    val ctype = (case ts of
			    (Classifier cn) =>  (let
				val C = class_of cn cl
				val thy_name = Context.theory_name thy
				val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
				val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
			    in
				mk_funT(V(u,class_u),V(u, (tilde u t)))
			    end)
			  | _ => dummyT)
					       
	in    
	    (Const(cname ,ctype) $ (ocl2holocl u src))
	end

   fun ATpre_of u (AttributeCall(src,ts,p,t))      =  TimedAttributeCall (src,ts,p,t) true
      | ATpre_of u (AssociationEndCall(src,ts,p,t)) =  TimedAttributeCall (src,ts,p,t) true
      | ATpre_of u (Literal _) = error "at_pre: found Literal"
      | ATpre_of u   (OperationCall (src,srctype,op_name,args,t)) = 
	ocl2holocl u (OperationCall(OperationCall (src,srctype,["oclLib","OclAny","atPre"],[],t),
		      srctype,op_name,args,t))
      | ATpre_of u (Iterator _) = error "at_pre: iterator"
      | ATpre_of u (Iterate _) = error "at_pre: iterate"
      | ATpre_of u (If _) = error "at_pre: if"
      | ATpre_of u (Let _) = error "at_pre: let"
      | ATpre_of u (Variable(s,t))   = let val _ = warning ("ignoring @pre for  variable "^s)
				       in 
					   	ocl2holocl u (Variable(s,t))
				       end


in 
    case oclterm of 
	(**************************************)
	(* Literal                            *)
	(**************************************)
	(* OCL Boolean                        *)
	Literal ("true", Boolean)          => OclTrue u  
      | Literal ("false",Boolean)          => OclFalse u
      (* OCL_Undefined                      *)
      | Literal ("OclUndefined",Boolean)   => OclUndefined u				     
      (* OCL_Real                           *)
      | Literal ("0",Real)                 => Const("0", REALT (St u) )
      | Literal ("-0",Real)                => Const("0", REALT (St u) )
      | Literal ("1",Real)                 => Const("1", REALT (St u))
      | Literal ("0.0",Real)               => Const("0", REALT (St u) )
      | Literal ("-0.0",Real)              => Const("0", REALT (St u) )
      | Literal ("1.0",Real)               => Const("1", REALT (St u))
      | Literal (rstr,Real)                   => (* this is *not* a generic solution! *)
	let 
	    val expand = 100000000
	    fun intOf i = (HolOclToReal u (Const("Numeral.number_of",mk_funT(Type("Numeral.bin",[]),INTEGERT (St u)))
				$ (HOLogic.mk_bin (IntInf.fromInt i))))
	    		 
	    val r =  (case Real.fromString rstr of
			  SOME (num) => num
			| NONE       => error ("Real literal ("^rstr^")in ocl2holocl is not an real."))
	    val rval = Real.split r
	    val wholeI  = (round (#whole rval))
	    val fracI   = (round ((#frac rval) * Real.fromInt(expand)))
	    val check   = (Real.fromInt(wholeI)) + Real.fromInt(fracI) /  Real.fromInt(expand)
	    val _ = if(Real.compare(check,r) = EQUAL) then () 
		  else  error ("Conversion of real literal "^rstr^" is incorrect!")
	in
		  OclRealPlus u  (intOf wholeI) (OclRealDivide u (intOf fracI) (intOf expand))
	end
	    
     (* OCL_Integer                        *)
      | Literal ("0",Integer)               => Const("0",  INTEGERT (St u))
      | Literal ("1",Integer)               => Const("1", INTEGERT (St u)) 
      | Literal (i,Integer)                 => 
	(case IntInf.fromString i of
	     SOME (num) =>
	     (Const("Numeral.number_of",mk_funT(Type("Numeral.bin",[]),mk_funT(dummyT,IntegerT)))
		   $ (HOLogic.mk_bin num))
	   | NONE       => error ("Integer literal ("^i^")in ocl2holocl is not an integer."))
      (* OCL_String                         *)
      | Literal (s,String)                 => let
	    val term_of_list = HOLogic.mk_list;		
	    val nibbleT = Type ("List.nibble", []);
	    fun term_of_char c = 
		Const ("List.char.Char", nibbleT --> nibbleT --> Type ("List.char", [])) $
		      Const ("List.nibble.Nibble" ^ nibble_of_int (ord c div 16), nibbleT) $
		      Const ("List.nibble.Nibble" ^ nibble_of_int (ord c mod 16), nibbleT);
	in 
	    Const ("Lifting.lift0",dummyT) 
		  $( Const ("Lifting.up.lift",dummyT) 
		  $ (term_of_list term_of_char (Type("List.char", []))  (String.explode s)))
	end
      (* Error                              *)
      | Literal (s,t)                      => error ("error: unknown Literal '"^s^"::"^(string_of_OclType t)^"'  in ocl2holocl") 
      (**************************************)
      (* Collection Literal                 *)
      (**************************************)					      
      | CollectionLiteral (cpl,t)        =>  let
	    fun mkMtCollection u (Set styp)        = (Const ("OCL_Set.OclMtSet",SETT(V(u,tilde u t),St u)))
	      | mkMtCollection u (Sequence styp)   = (Const ("OCL_Sequence.OclMtSequence",SEQUENCET(V(u,tilde u t),St u)))
	      | mkMtCollection u (OrderedSet styp) = (Const ("OCL_OrderedSet.OclMtOrderedSet",ORDEREDSETT(V(u,tilde u t),St u)))
	      | mkMtCollection u (Bag styp)        = (Const ("OCL_Bag.OclMtBag",BAGT(V(u,tilde u t),St u)))
	      | mkMtCollection _ typ  =  error ("error: unknown Collection Literal of type '"^(string_of_OclType typ)^"'")
	    fun insert_elem styp u ((CollectionItem(i,itype)),S)     = OclIncluding styp u S (ocl2holocl u i)
	      | insert_elem styp u ((CollectionRange(i,j,etype)),S)  = OclUnion u S 
								       (OclCollectionRange styp u S (ocl2holocl u i)
											            (ocl2holocl u j))
	in 
	    foldl (insert_elem (t) u ) (mkMtCollection u t) cpl
	end
      
      
      (**************************************)
      (* if                                 *)
      (**************************************)
      (* OCL Boolean *)
      | If (cterm,ctype, tterm,ttype,eterm,etype,iftype) => OclIf u (ocl2holocl u cterm)  
								  (ocl2holocl u tterm) 
								  (ocl2holocl u eterm)
      (**************************************)
      (* AssociationEndCall                 *)
      (**************************************)
      | AssociationEndCall(src,ts,p,t)    => (case src of 
						 OperationCall (tsrc,_,["oclLib","OclAny","atPre"],[],_) 
						 =>  TimedAttributeCall (tsrc,ts,p,t) true
					       | _ => TimedAttributeCall (src,ts,p,t) false)
      (**************************************)
      (* AttributeCall                      *)
      (**************************************)
      | AttributeCall(src,ts,p,t)             => (case src of 
						 OperationCall (tsrc,_,["oclLib","OclAny","atPre"],[],_) 
						 =>  TimedAttributeCall (tsrc,ts,p,t) true
					       | _ => TimedAttributeCall (src,ts,p,t) false)
      (**************************************)
      (* OperationCall                      *)
      (**************************************)
      | OperationCall (src,Boolean,["holOclLib","Boolean","implies"],[(arg,_)],rtyp) => HolImplies  (ocl2holocl u src) 
								      (ocl2holocl u arg)
      | OperationCall (src,Boolean,["holOclLib","Boolean","and"],[(arg,_)],rtyp) => HolAnd (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["holOclLib","Boolean","xor"],[(arg,_)],rtyp) => HolXor (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["holOclLib","Boolean","or"],[(arg,_)],rtyp)  => HolOr  (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["holOclLib","Boolean","not"],[],rtyp)        => HolNot (ocl2holocl u src)
      | OperationCall (src,Boolean,["holOclLib","Boolean","OclLocalValid"],[(Literal(arg,_),_)],rtyp)        => OclLocalValid u 
													      (Free(arg,St(u))) 
													      (ocl2holocl u src)
      (* OCL Boolean *)
      | OperationCall (src,Boolean,["oclLib","Boolean","implies"],[(arg,_)],rtyp) => OclImplies u (ocl2holocl u src) 
								      (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","implies1"],[(arg,_)],rtyp) => OclImplies1 u (ocl2holocl u src) 
												  (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","implies2"],[(arg,_)],rtyp) => OclImplies2 u (ocl2holocl u src) 
												  (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","and"],[(arg,_)],rtyp) => OclAnd u (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","xor"],[(arg,_)],rtyp) => OclXor u (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","or"],[(arg,_)],rtyp) => OclOr u (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,Boolean,["oclLib","Boolean","not"],[],rtyp) => OclNot u (ocl2holocl u src)
      | OperationCall (src,styp,["oclIsDefined"],[],rtyp) => OclIsDefined u (ocl2holocl u src)
      | OperationCall (src,styp,["oclIsUndefined"],[],rtyp) => OclIsUndefined u (ocl2holocl u src)
      | OperationCall (src,styp,["oclLib",_,"oclIsUndefined"],[],rtyp) => OclIsUndefined u (ocl2holocl u src)

      | OperationCall (src,Integer,["oclLib",classifier,"="],[(arg,Real)],_) => OclStrongEq u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,"<>"],[(arg,Real)],_) => OclNeq u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,"=="],[(arg,Real)],_) => OclStrictEq u (HolOclToReal u (ocl2holocl u src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,"~="],[(arg,Real)],_) => OclRefEq u (HolOclToReal u (ocl2holocl u src)) (ocl2holocl u arg)
      (* OCL Numerals                       *)
      | OperationCall (src,Integer,["oclLib",classifier,"<"],[(arg,Real)],_) => OclLess u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,"<="],[(arg,Real)],_) => OclLe u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,">"],[(arg,Real)],_) => OclGreater u (HolOclToReal u (ocl2holocl u src))  (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",classifier,">="],[(arg,Real)],_) => OclGe u (HolOclToReal u (ocl2holocl u src))  (ocl2holocl u arg)


      | OperationCall (src,Real,["oclLib",classifier,"="],[(arg,Integer)],_) => OclStrongEq u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,"<>"],[(arg,Integer)],_) => OclNeq u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,"=="],[(arg,Integer)],_) => OclStrictEq u (ocl2holocl u src) (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,"~="],[(arg,Integer)],_) => OclRefEq u (ocl2holocl u src) (HolOclToReal u (ocl2holocl u arg))
      (* OCL Numerals                       *)
      | OperationCall (src,Real,["oclLib",classifier,"<"],[(arg,Integer)],_) => OclLess u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,"<="],[(arg,Integer)],_) => OclLe u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,">"],[(arg,Integer)],_) => OclGreater u (ocl2holocl u src)  (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib",classifier,">="],[(arg,Integer)],_) => OclGe u (ocl2holocl u src)  (HolOclToReal u (ocl2holocl u arg))

  
      | OperationCall (src,styp,["oclLib",classifier,"="],[(arg,_)],_) => OclStrongEq u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"<>"],[(arg,_)],_) => OclNeq u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"=="],[(arg,_)],_) => OclStrictEq u (ocl2holocl u src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"~="],[(arg,_)],_) => OclRefEq u (ocl2holocl u src) (ocl2holocl u arg)
      (* OCL Numerals                       *)
      | OperationCall (src,styp,["oclLib",classifier,"<"],[(arg,_)],_) => OclLess u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"<="],[(arg,_)],_) => OclLe u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,">"],[(arg,_)],_) => OclGreater u (ocl2holocl u src)  (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,">="],[(arg,_)],_) => OclGe u (ocl2holocl u src)  (ocl2holocl u arg)

      (* OCL Real                          *)
      | OperationCall (src,Real,["oclLib",_,"round"],[],Integer) => OclRealRound u (ocl2holocl u  src)   
      | OperationCall (src,Real,["oclLib",_,"floor"],[],Integer) => OclRealFloor u (ocl2holocl u  src) 
      | OperationCall (src,Real,["oclLib",_,"abs"],[],Real) => OclRealAbs u (ocl2holocl u src)   
      | OperationCall (src,Real,["oclLib",_,"-"],[],Real) => OclRealNegative u  (ocl2holocl u  src)
	
      | OperationCall (src,Real,["oclLib","Real","min"],[(arg,Real)],Real) => OclRealMin u (ocl2holocl u  src) (ocl2holocl u arg) 
      | OperationCall (src,Real,["oclLib","Real","max"],[(arg,Real)],Real) => OclRealMax u (ocl2holocl u  src) (ocl2holocl u arg)  
 
      | OperationCall (src,Integer,["oclLib","Real","min"],[(arg,Real)],Real) => OclRealMin u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg) 
      | OperationCall (src,Integer,["oclLib","Real","max"],[(arg,Real)],Real) => OclRealMax u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)  
									      
      | OperationCall (src,Real,["oclLib","Real","min"],[(arg,Integer)],Real) => OclRealMin u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg)) 
      | OperationCall (src,Real,["oclLib","Real","max"],[(arg,Integer)],Real) => OclRealMax u (ocl2holocl u  src) (HolOclToReal u (ocl2holocl u arg))  
    
							  
      | OperationCall (src,Real,["oclLib","Real","/"],[(arg,Real)],Real) => OclRealDivide u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,Real,["oclLib","Real","-"],[(arg,Real)],Real) => OclRealMinus u  (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,Real,["oclLib","Real","+"],[(arg,Real)],Real) => OclRealPlus u  (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,Real,["oclLib","Real","*"],[(arg,Real)],Real) => OclRealTimes u  (ocl2holocl u  src) (ocl2holocl u arg)

      | OperationCall (src,Real,["oclLib","Real","/"],[(arg,Integer)],Real) => OclRealDivide u (ocl2holocl u  src)  (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib","Real","-"],[(arg,Integer)],Real) => OclRealMinus u  (ocl2holocl u  src)  (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib","Real","+"],[(arg,Integer)],Real) => OclRealPlus u  (ocl2holocl u  src)  (HolOclToReal u (ocl2holocl u arg))
      | OperationCall (src,Real,["oclLib","Real","*"],[(arg,Integer)],Real) => OclRealTimes u  (ocl2holocl u  src) (HolOclToReal u  (ocl2holocl u arg))

      | OperationCall (src,Integer,["oclLib","Real","/"],[(arg,Real)],Real) => OclRealDivide u (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib","Real","-"],[(arg,Real)],Real) => OclRealMinus u  (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib","Real","+"],[(arg,Real)],Real) => OclRealPlus u  (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib","Real","*"],[(arg,Real)],Real) => OclRealTimes u  (HolOclToReal u (ocl2holocl u  src)) (ocl2holocl u arg)
      | OperationCall (src,Integer,["oclLib",_,"/"],[(arg,Integer)],Real) => OclRealDivide u (HolOclToReal u (ocl2holocl u  src)) (HolOclToReal u (ocl2holocl u arg))
  


      (* OCL Integer                        *)
      | OperationCall (src,styp,["oclLib",classifier,"mod"],[(arg,_)],Integer) => OclIntMod u (ocl2holocl u  src)   
											       (ocl2holocl u arg)    
      | OperationCall (src,styp,["oclLib",classifier,"div"],[(arg,_)],Integer) => OclIntDiv u (ocl2holocl u  src)   
											       (ocl2holocl u arg)   
      | OperationCall (src,styp,["oclLib",classifier,"round"],[],Integer) => OclIntRound u (ocl2holocl u  src) 
      | OperationCall (src,Integer,["oclLib",classifier,"floor"],[],Integer) => OclIntFloor u (ocl2holocl u  src)    
      | OperationCall (src,styp,["oclLib",classifier,"min"],[(arg,_)],Integer) => OclIntMin u   (ocl2holocl u  src)   
											       (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"max"],[(arg,_)],Integer) => OclIntMax u  (ocl2holocl u  src)   
											       (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"abs"],[],Integer) => OclIntAbs u   (ocl2holocl u  src)    
      | OperationCall (src,styp,["oclLib",classifier,"-"],[],Integer) => OclIntNegative u  (ocl2holocl u  src)
      | OperationCall (src,styp,["oclLib",classifier,"-"],[(arg,_)],Integer) => OclIntMinus u  (ocl2holocl u  src) 
											       (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"+"],[(arg,_)],Integer) => OclIntPlus u  (ocl2holocl u  src) 
											      (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",classifier,"*"],[(arg,_)],Integer) => OclIntTimes u  (ocl2holocl u  src) 
											       (ocl2holocl u arg)
      (* OCL String                          *)
      | OperationCall (src,styp,["oclLib",classifier,"concat"],[(arg,_)],String) => OclConcat u  (ocl2holocl u  src) 
												 (ocl2holocl u arg)
      | OperationCall (src,stype,["oclLib","String","size"],[],rtyp) => OclStringSize u   (ocl2holocl u  src) 
      | OperationCall (src,styp,["oclLib",classifier,"substring"],[(b,Integer),(e,Integer)],String) 
	=> OclSubString u (ocl2holocl u  src) (ocl2holocl u  b) (ocl2holocl u  e)		
      | OperationCall (src,styp,["oclLib",classifier,"toLower"],[],String) => OclToLower u (ocl2holocl u  src)	
      | OperationCall (src,styp,["oclLib",classifier,"toUpper"],[],String) => OclToUpper u (ocl2holocl u  src)	
 
     (* OCL Collection                    *)
      | OperationCall (src,styp,["oclLib",_,"modifiedOnly"],[],_) => OclModifiedOnly styp u (ocl2holocl u  src)         
      | OperationCall (src,styp,["oclLib",_,"size"],[],_) => OclSize styp u (ocl2holocl u  src)         
      | OperationCall (src,styp,["oclLib",_,"count"],[(arg,_)],_) => OclCount u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",_,"includes"],[(arg,atyp)],_) => OclIncludes styp atyp u (ocl2holocl u  src)   
											  (ocl2holocl u arg) 
      | OperationCall (src,styp,["oclLib",_,"excludes"],[(arg,atyp)],_) => OclExcludes atyp u (ocl2holocl u  src)   
											  (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",_,"flatten"],[(arg,_)],_) => OclFlatten u (ocl2holocl u  src)      
      | OperationCall (src,styp,["oclLib",_,"sum"],[],rtyp) => OclSum styp rtyp u (ocl2holocl u  src)          
(*      | OperationCall (src,stype,["oclLib",_,"asSet"],[],_) => mkSingularSet stype u (ocl2holocl u  src)      
*)
      | OperationCall (src,styp,["oclLib",_,"including"],[(arg,atyp)],_) => OclIncluding styp  u (ocl2holocl u  src) 
											   (ocl2holocl u arg)      
      | OperationCall (src,styp,["oclLib",_,"excludes"],[(arg,atyp)],_) => OclExcludes atyp u (ocl2holocl u  src)   
											 (ocl2holocl u arg)    

      | OperationCall (src,Real,["oclLib",_,"asSet"],[],rtyp) =>  mkSingularCollection Real rtyp u (ocl2holocl u  src)   
      | OperationCall (src,Integer,["oclLib",_,"asSet"],[],rtyp) =>  mkSingularCollection Integer rtyp u (ocl2holocl u  src)   
      | OperationCall (src,String,["oclLib",_,"asSet"],[],rtyp) => mkSingularCollection String rtyp u (ocl2holocl u  src)  
      | OperationCall (src,Boolean,["oclLib",_,"asSet"],[],rtyp) => mkSingularCollection Boolean rtyp u (ocl2holocl u  src)
      | OperationCall (src,Classifier c,["oclLib",_,"asSet"],[],rtyp) =>  mkSingularCollection (Classifier c) rtyp u (ocl2holocl u  src)    
      | OperationCall (src,OclVoid,["oclLib",_,"asSet"],[],rtyp) =>  mkSingularCollection OclVoid rtyp u (ocl2holocl u  src) 
  
      | OperationCall (src,styp,["oclLib",_,"asSet"],[],_) => OclAsSet styp u (ocl2holocl u  src)        
      | OperationCall (src,styp,["oclLib",_,"asSequence"],[],_) => OclAsSequence u (ocl2holocl u  src)   
      | OperationCall (src,styp,["oclLib",_,"asBag"],[],_) => OclAsBag u (ocl2holocl u  src)        
      | OperationCall (src,styp,["oclLib",_,"asOrderedSet"],[],_) => OclAsOrderedSet u (ocl2holocl u  src) 
      | OperationCall (src,styp,["oclLib",_,"includesAll"],[(arg,_)],_) => OclIncludesAll u (ocl2holocl u  src) 
											     (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",_,"excludesAll"],[(arg,_)],_) => OclExcludesAll u (ocl2holocl u  src) 
											     (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",_,"isEmpty"],[],rtyp) => OclIsEmpty u (ocl2holocl u  src)      
      | OperationCall (src,styp,["oclLib",_,"notEmpty"],[],rtyp) => OclNotEmpty u (ocl2holocl u  src)     
      (* OCL OrderedSet                     *)
      (* OCL Sequence                       *)
      | OperationCall (src,styp,["oclLib","Sequence","union"],[(arg,_)], _) => OclSeqUnion u (ocl2holocl u  src) 
											      (ocl2holocl u arg)   
      | OperationCall (src,styp,["oclLib",_,"subSequence"],[(b,_),(e,_)],_) 
	=> OclSeqSubSequence u  (ocl2holocl u  src) (ocl2holocl u  b) (ocl2holocl u  e)
      | OperationCall (src,styp,["oclLib",_,"first"],[],_) =>OclSeqFirst u  styp (ocl2holocl u  src)     
      | OperationCall (src,styp,["oclLib",_,"last"],[],_) => OclSeqLast u   styp (ocl2holocl u  src)         
      | OperationCall (src,styp,["oclLib",_,"append"],[(arg,_)],_) => OclSeqAppennd u  (ocl2holocl u  src) 
										       (ocl2holocl u arg)    
      | OperationCall (src,styp,["oclLib",_,"prepend"],[(arg,_)], _) => OclSeqPrepend u  (ocl2holocl u  src) (ocl2holocl u arg)    
      | OperationCall (src,styp,["oclLib",_,"at"],[(arg,_)], _) => OclSeqAt u   (ocl2holocl u  src) (ocl2holocl u arg)    
      | OperationCall (src,styp,["oclLib",_,"insertAt"],[(b,Integer),(p,_)], _) 	
	=> OclSeqSubSequence u  (ocl2holocl u  src) (ocl2holocl u  b) (ocl2holocl u p) 
      | OperationCall (src,styp,["oclLib",_,"union"],[(arg,_)], _) => OclUnion u (ocl2holocl u  src) (ocl2holocl u arg)
      | OperationCall (src,styp,["oclLib",_,"intersection"],[(arg,_)], _) => OclIntersection u (ocl2holocl u  src) (ocl2holocl u arg)
      (* OCL Bag                            *)
      | OperationCall (src,styp,["oclLib","Bag","size"],[],Integer) => OclBagSize u   (ocl2holocl u  src)       
      | OperationCall (src,styp,["oclLib","Bag","count"],[(arg,_)], _) => OclBagCount u   (ocl2holocl u  src) (ocl2holocl u arg)    
      | OperationCall (src,styp,["oclLib","Bag","isEmpty"],[], _) => OclBagIsEmpty u    (ocl2holocl u  src)  
      | OperationCall (src,styp,["oclLib","Bag","notEmpty"],[], _) => OclBagNotEmpty u   (ocl2holocl u  src)   
      | OperationCall (src,styp,["oclLib","Bag","complement"],[], _) => OclBagComplement  (ocl2holocl u  src)    
      | OperationCall (src,styp,["oclLib","Bag","insert"],[(arg,_)], _) => OclBagInsert u (ocl2holocl u  src) (ocl2holocl u arg)      
      | OperationCall (src,styp,["oclLib","Bag","union"],[(arg,_)], _) => OclBagUnion u (ocl2holocl u  src) (ocl2holocl u arg)     
      (* OCL Set                            *)
      | OperationCall (src,styp,["oclLib","Set","complement"],[], _) => OclSetComplement u  (ocl2holocl u  src)    
      (* OCL Tuple                          *)
      (* OCL StateAccess                    *)
      | OperationCall (src,styp,["oclLib",classifier,"oclIsNew"],[],_)  => OclIsNew u (ocl2holocl u  src)                 
      | OperationCall (src,styp,["oclLib",classifier,"allInstances"],[],_)  =>  OclAllInstances u (ocl2holocl u  src)         
      (* @pre                               *)
       | OperationCall (src,styp,["oclLib","OclAny","atPre"],[],_) =>  ATpre_of u src  
 

      (* conversions                        *)

      (* OperationCalls to modell and Error *)
      | OperationCall (src,styp,op_name,args,t) => let 
	    val _ = trace 3 ("Found Op ("^(Int.toString(length args))^"):"^(string_of_path op_name))
      	    val Op = case (operation_of cl op_name) of
			 SOME(Op) => Op
		       | NONE     => error ("error: unknown OperationCall '"
					    ^(string_of_path op_name)^":("^(string_of_OclType t)
					    ^")'  in ocl2holocl")
	    val m_name = mangled_name_of_op Op
	in 
	    (foldl (fn (t0,t1) => t1$t0) (Const((string_of_path (hc_path operation [m_name]))^".Op",dummyT))
			   (map (fn a => ocl2holocl u (fst a)) ((src,styp)::args)))
	end
 
      (**************************************)
      (* Variable                           *) 
      (**************************************)
      | Variable (vname,t) => Free(vname,V(u, tilde u (t)))
      (**************************************)
      (* Let                                *)
      (**************************************) 
      (* Error                              *)
      | Let (var,vtyp,rhs,rhstyp,interm,intyp) => OclLet u (OclStrongEq u (Free(var,V(u, tilde u (vtyp)))) (ocl2holocl u rhs)) 
      							   (ocl2holocl u interm)
      | Let (s,_,_,_,_,_) => error ("error: unknown Let '"^(s)^"'  in ocl2holocl") 
      (**************************************)
      (* OperationWithType                  *)
      (**************************************)
      (* Error                              *)
      | OperationWithType (src,srcT,"oclIsKindOf",oclType,_) => OclIsKindOf  u (ocl2holocl u src)  (charset_of oclType u)
      | OperationWithType (src,srcT,"oclIsTypeOf",oclType,_) => OclIsTypeOf  u (ocl2holocl u src)  (type_charset_of oclType u) 
      | OperationWithType (src,srcT,"oclAsType",oclType,_) => OclAsType  u (ocl2holocl u src)  (charset_of oclType u)
      | OperationWithType (src,srcT,name,oclType,_) =>  error ("error: unknown OperationWithName '"^(name)^"' in in ocl2holocl") 
      (**************************************)
      (* Iterate                            *)
      (**************************************)
      (* Error                              *)
      | Iterate (vars,rvar,rvtyp,rterm,sterm,styp,iterm,ityp,rtyp) =>  
	error ("error: unknown Iterate '"^(rvar)^"' in in ocl2holocl") 
      (**************************************)
      (* Iterator                           *)
      (**************************************)
      (* forAll                              *)
      | Iterator ("forAll",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun forall S [v] P = OclForAll u S (Term.absfree (fst v,snd v,P))
	      | forall S vs P  = OclForAll u S (Term.absfree (fst (hd vs),snd (hd vs), forall S (tl vs) P))
	      | forall S []  P = error "no vars in forAll"
	in
	    forall (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end
      | Iterator ("exists",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun exists S [v] P = OclExists u S (Term.absfree (fst v,snd v,P))
	      | exists S vs P  = OclExists u S (Term.absfree (fst (hd vs),snd (hd vs), exists S (tl vs) P))
	      | exists S []  P = error "no vars in exists"
	in
	    exists (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end

      | Iterator ("select",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun select S [v] P = OclSelect u S (Term.absfree (fst v,snd v,P))
	      | select S vs P  = OclSelect u S (Term.absfree (fst (hd vs),snd (hd vs), select S (tl vs) P))
	      | select S []  P = error "no vars in select"
	in
	    select (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end

      | Iterator ("reject",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun reject S [v] P = OclReject u S (Term.absfree (fst v,snd v,P))
	      | reject S vs P  = OclReject u S (Term.absfree (fst (hd vs),snd (hd vs), reject S (tl vs) P))
	      | reject S []  P = error "no vars in reject"
	in
	    reject (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end


      | Iterator ("collect",vars,src,styp,c,btyp,rtyp) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun collect S [v] P = OclCollect styp btyp rtyp u S (Term.absfree (fst v,snd v,P))
	      | collect S vs P  = error "wellconformed-violation: collect may have at most one iterator variable"
	      | collect S []  P = error "no vars in collect"
	in
	    collect (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end

      (* OCL Collection                     *) 

      | Iterator ("isUnique",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun isUnique S [v] P = OclIsUnique u S (Term.absfree (fst v,snd v,P))
	      | isUnique S vs P  = error "wellconformed-violation: isUnique may have at most one iterator variable"
	      | isUnique S []  P = error "no vars in isUnique"
	in
	    isUnique (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end



      | Iterator ("one",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun one S [v] P = OclOne u S (Term.absfree (fst v,snd v,P))
	      | one S vs P  = error "wellconformed-violation: one may have at most one iterator variable"
	      | one S []  P = error "no vars in one"
	in
	    one (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end


      | Iterator ("any",vars,src,styp,c,_,_) => 
	let
	    val hc_vars = map (fn a => (fst a,V(u, tilde u (snd a)))) vars
	    fun any S [v] P = OclAny u S (Term.absfree (fst v,snd v,P))
	      | any S vs P  = error "wellconformed-violation: any may have at most one iterator variable"
	      | any S []  P = error "no vars in any"
	in
	    any (ocl2holocl u src) hc_vars (ocl2holocl u c)
	end
   (* 
      | Iterate (src,styp,["oclLib",classifier,"iterate"],args,Collection _) => OclIterate u C be e  
      (* OCL OrderedSet                     *) 
      | Iterate (src,styp,["oclLib",classifier,"count"],[(arg,_)],OrderedSet _) => OclOSetCount u S e       *)
      (* Error                              *)
      | Iterator (s,_,_,_,_,_,_) => error ("error: unknown Iterator '"^(s)^"' in in ocl2holocl") 
      (**************************************)
      (* Catch out                          *)
      (**************************************)
      (* Error                              *)
      | _ => error ("error: unknown OCL-term in in ocl2holocl") 
end
in 

ocl2holocl u oclterm 

end
end
