(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * invariant_encoder.sml - 
 * Copyright (C) 2006  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

signature INVARIANT_ENCODER =
sig
    val transformations : (string * (Rep_Core.Classifier list -> Rep_Core.Classifier 
				     -> Rep_HolOclTerm.HolOclTerm -> Rep_HolOclTerm.HolOclTerm))  list ref
    (**
     * map OCL formulae to a two-valued HOL/OCL one, as described in
     * HOL-OCL book.
     *)
    val mk_twovalued     : Rep_OclTerm.OclTerm -> Rep_HolOclTerm.HolOclTerm
    val add_invariant : Rep_Core.Classifier -> typ -> Rep_Core.Classifier list -> theory -> theory  
    val add_user_invariant : Rep_Core.Classifier -> typ -> Rep_Core.Classifier list -> theory -> theory  
end  

structure invariant_encoder : INVARIANT_ENCODER =
struct

val transformations = ref ([]: (string * (Rep_Core.Classifier list -> Rep_Core.Classifier ->
					  Rep_HolOclTerm.HolOclTerm -> Rep_HolOclTerm.HolOclTerm)) list)
open Rep_HolOclTerm
open Rep_Core
open rep_encoder_common
open holocl_term
fun path_of trans C = if (fst trans) = "" 
		      then  (hc_path l1 ((name_of C)))
		      else  (hc_path l1 ((name_of C)))@[fst trans]
			    
			    
fun OclLocalValid u tau expr    = Const("OCL_Logic_core.OclLocalValid", St(u) --> BOOLT (St u) --> boolT) $ tau $ expr
				  
fun HolOr  x y         = Const("op |",boolT --> boolT --> boolT) $ x $ y
fun HolAnd  x y        = Const("op &",boolT --> boolT --> boolT) $ x $ y
			 
fun mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","implies"],[(arg,argt)],rtyp)) 
  = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","implies"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","implies1"],[(arg,argt)],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","implies"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","implies2"],[(arg,argt)],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","implies"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","and"],[(arg,argt)],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","and"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","xor"],[(arg,argt)],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","xor"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","or"],[(arg,argt)],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","or"],[(mk_twovalued arg,argt)],rtyp))
  | mk_twovalued (OperationCall (src,Boolean,["oclLib","Boolean","not"],[],rtyp)) 
    = (OperationCall (mk_twovalued src,Boolean,["holOclLib","Boolean","not"],[],rtyp))  
  | mk_twovalued t = (OperationCall (t,Boolean,["holOclLib","Boolean","OclLocalValid"],
                                     [(Literal("\\<tau>",DummyT),DummyT)],Boolean))
		     



fun ref_attribs C = filter (fn a:attribute => case (#attr_type a) of (Classifier _ ) => true 
								| _               => false )  
			   (attributes_of C)



fun rec_attribs C = filter (fn a:attribute => (#attr_type a = (Classifier (name_of C))))  
			   (attributes_of C)
	


fun mk_relax_ref (cl:(Rep_Core.Classifier list)) (C:Rep_Core.Classifier ) (inv:HolOclTerm) = 
  let
      fun isdef attribute attr_type = Rep_OclTerm.OperationCall (attribute,attr_type,
                                                          ["oclIsUndefined"],[],
                                                          Rep_OclType.Boolean)
      fun getAttr (a:Rep_Core.attribute) = 
	  let
              val attr_name = (name_of C)@[(#name a)]
              val attr_type = #attr_type a
              val cls       = Classifier (name_of C)
              val self      = Variable ("self",cls)
          in 
	      isdef (AttributeCall (self,cls,attr_name,attr_type)) attr_type
	  end	      

      val attribs = map (fn a => getAttr a) (ref_attribs C) 
  
      fun mkOr (a,b) = (OperationCall (a, Rep_OclType.Boolean,
						  ["oclLib","Boolean","or"],
						  [(b,Rep_OclType.Boolean)],
						  Rep_OclType.Boolean)) 		
  in      
      foldl (mkOr)  (inv:(HolOclTerm)) (attribs:(HolOclTerm list)) 
  end


fun mk_relax_refval (cl:(Rep_Core.Classifier list)) (C:Rep_Core.Classifier ) (inv:Rep_HolOclTerm.HolOclTerm) = 
  let
      fun isdef attribute attr_type = Rep_OclTerm.OperationCall (attribute,attr_type,
                                                          ["oclIsUndefined"],[],
                                                          Rep_OclType.Boolean)
      fun getAttr (a:Rep_Core.attribute) = 
	  let
              val attr_name = (name_of C)@[(#name a)]
              val attr_type = #attr_type a
              val cls       = Classifier (name_of C)
              val self      = Variable ("self",cls)
          in 
	      isdef (AttributeCall (self,cls,attr_name,attr_type)) attr_type
	  end	      

      val attribs = map (fn a => getAttr a) (attributes_of C) 
  
      fun mkOr (a,b) = (OperationCall (a, Rep_OclType.Boolean,
						  ["oclLib","Boolean","or"],
						  [(b,Rep_OclType.Boolean)],
						  Rep_OclType.Boolean)) 		
  in      
      foldl (mkOr)  (inv:(HolOclTerm)) (attribs:(HolOclTerm list)) 
  end

	

val transformations = ref ([("relax_ref",mk_relax_ref),("relax_refval",mk_relax_refval)])



fun hol_inv_of trans thy cl u C  = map (fn (n,c) => (n, Ocl2holocl l1 thy cl u (mk_twovalued ((snd trans) cl C c)))) (invariant_of C) 


fun hol_invariant_of trans path   = if (fst trans) = "" 
				    then path@["hol_invariant"]
				    else path@[fst trans,"hol_invariant"]
fun set_fun_of trans path         = if (fst trans) = "" 
				    then path@["set_fun"]
				    else path@[fst trans,"set_fun"]
fun cset_of trans path  = if (fst trans) = "" 
					then path@["characteristic_set"]
					else path@[fst trans,"characteristic_set"]
fun inv_of trans path  = if (fst trans) = "" 
					then path@["inv"]
					else path@[fst trans,"inv"]

fun user_inv_of path  = path@["inv"]

fun set_of trans path             = if (fst trans) = ""
				    then path@["set"]
				    else path@[fst trans,"set"]
					 
fun hol_invariant_of trans path   = if (fst trans) = "" 
				    then path@["hol_invariant"]
				    else path@[fst trans,"hol_invariant"]



fun typeconstraint_of n = "typeconstraint_"^n

fun add_hol_invariant trans C u cl  thy =
    let

	val thy_name = Context.theory_name thy

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 



	fun rec_type_check C u cl =
	    let
		fun type_constraint (attr:attribute) = 
		    let
			val attr_type = mk_funT(V(u,class_u),V(u, (tilde u (#attr_type attr))))
			val name = #name attr
			val name_of_get =  string_of_path (("UML_OCL"::["level1"])@(name_of C)@ [name])
			val getAttr = (Const(name_of_get,attr_type)$(Free("self",V(u,class_u))))
				      
			val et  = (Classifier (name_of C))
			val inType = mk_funT(V(u, tilde u (Set et)), mk_funT(V(u, tilde u ( et)),BOOLT(St u)))
			val constraint = 
			    HolOr (OclLocalValid u 
						 (Free("\\<tau>", St(u)))
						 (OclIsUndefined u getAttr ))
			    	  
				  (Const("op :",dummyT)$(getAttr$(Free("\\<tau>", St(u))))
					$(Const("image",dummyT)
					       $Term.absfree("f",dummyT,
							     Free("f",dummyT)$(Free("\\<tau>", St(u)))  )  
					       $(Free("C",mk_setT (V(u,class_u))))))
		    in 
			(typeconstraint_of name ,constraint)
		    end
	    in
		map (type_constraint) (rec_attribs C) 
	    end


	fun parent_invariant C u cl = 
	    let
		val checkParentInv = if (parent_name_of C) = (name_of OclAnyC)
				     then Const("True",boolT)  
				     else
					 let
					     val pholes = map (fn x => TFree(x,typeS)) (holes_of_type u (parent_of C cl))
					     val pclass_u = Type((string_of_path (hc_path univ (parent_name_of C))),pholes) 
					     val phol_inv_type      = mk_funT(St(u),mk_funT(mk_setT(V(u,pclass_u)),
											    mk_funT(V(u,pclass_u),boolT))) 
					 in
					     (Const(string_of_path(hol_invariant_of trans (parent_name_of C)),phol_inv_type))
						 $(Free("\\<tau>", St(u)))
						 $(Const ("image", (V(u,class_u) --> V(u,pclass_u)) 
								       --> mk_setT (V(u,class_u)) 
								       --> mk_setT (V(u,pclass_u))) 
							 $Term.absfree("obj",V(u,class_u),
								       (Const("OclAsType",V(u, class_u) 
											   -->V(u,SetT(pclass_u))
											   -->V(u,pclass_u)) 
									     $(Free("obj",V(u,class_u))) 
									     $Const((string_of_path 
											 (hc_path l1 
												  (parent_name_of C))),
										    V(u,SetT(pclass_u)))))
							 $(Free("C",mk_setT (V(u,class_u)))))
						 $(Const("OclAsType",V(u, class_u) 
								      -->V(u,SetT(pclass_u))
								      -->V(u,pclass_u)) 
							$(Free("self",V(u,class_u))) 
							$Const((string_of_path (hc_path l1 (parent_name_of C))),
							       V(u,SetT(pclass_u))))
					 end
	    in 
		[("check_parent_inv",checkParentInv)]
	    end

	val name = name_of C	
	val _ = trace 3  ("        "^(string_of_path name))
		
	val invariants_hol  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => ((rm_whitespace s), snd c)
							   | NONE => ("inv_"^Int.toString(n), snd c )
					   )) (0,(hol_inv_of trans thy cl u C))) 


 	val invariants_hol = invariants_hol@(rec_type_check C u cl)@(parent_invariant C u cl)




	val hol_inv_type      = mk_funT(St(u),mk_funT(mk_setT (V(u,class_u)),mk_funT(V(u,class_u),boolT)))  

	val thy = foldl (fn (constraint,thy) 
			    =>  add_consts_i (SOME (hc_path l1 (hol_invariant_of trans  (name_of C))))
				    [((fst constraint),hol_inv_type,NoSyn)] thy) 
		  thy invariants_hol
	val  timer    =  Timer.startRealTimer ()
        val startTime = Timer.checkRealTimer timer 


	 val invs_hol = map (fn (n,t) =>  (n,#1(Sign.infer_types (Sign.pp  thy) thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", hol_inv_type --> hol_inv_type --> propT) $
                               Const ((string_of_path ((hc_path l1 (hol_invariant_of trans  (name_of C)))))^"."^n,hol_inv_type) $ 
			       Term.absfree("\\<tau>", St(u),Term.absfree("C",mk_setT (V(u,class_u)),Term.absfree ("self",V(u,class_u), t))))
			      
		      ],propT)))) invariants_hol 
      
	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for infering types for invs: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 


	val thy = foldl (fn (constraint,thy) 
			    =>  add_definitions_i false (SOME ((hc_path l1 (hol_invariant_of trans  (name_of C)))))
				    [((fst constraint)^"_def",(snd constraint))] thy) 
		  thy invs_hol 

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for adding defs: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 


	val thy = add_consts_i (SOME (path_of trans C)) 
		  [((local_name_of (hc_path l1 (hol_invariant_of trans  (name_of C) ))),hol_inv_type,NoSyn)] thy 

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for adding final consts: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 


	val c_str_i = map (fn s => Const((string_of_path  (hc_path l1 (hol_invariant_of trans  (name_of C))))
					 ^"."^(fst s),hol_inv_type)
				      $Free("\\<tau>", St(u))
				      $Free("C",mk_setT (V(u,class_u)))
				      $Free("obj",V(u,class_u))) invariants_hol



	val hol_inv_def_i =  (#1(Sign.infer_types (Sign.pp thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", hol_inv_type --> hol_inv_type --> propT) 
				      $ (Const(string_of_path (hc_path l1 (hol_invariant_of trans  (name_of C))),hol_inv_type)) 
				      $ (Term.absfree ("\\<tau>", St(u),
						       (Term.absfree ("C",mk_setT (V(u,class_u)),
								      Term.absfree ("obj",V(u,class_u),
										    (foldr1 (fn (a,b) => HolAnd a b) c_str_i)))))))
			       ],propT)))


	val thy = add_definitions_i false  (SOME (path_of trans C))
				  [((local_name_of (hc_path l1 (hol_invariant_of trans  (name_of C)))),hol_inv_def_i)] thy 
	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for adding final def: "
				 ^(Time.toString(Time.-(endTime,startTime))))


     in 
	thy
    end
	
fun add_set trans C u cl  thy =  
    let
	val thy_name = Context.theory_name thy
		       
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 
		      
	val set_type     = mk_funT(St(u),mk_setT (V(u,class_u)))  
	val set_fun_type = mk_funT(St(u),mk_funT(mk_setT (V(u,class_u)),mk_setT (V(u,class_u))))  


	val set_fun = Const(string_of_path(hc_path l1 (set_fun_of trans  (name_of C))), set_fun_type)
		     
	val  timer    =  Timer.startRealTimer ()
        val startTime = Timer.checkRealTimer timer 

	val thy = add_consts_i (SOME (path_of trans C)) 
		  [((local_name_of (hc_path l1 (set_of trans  (name_of C) ))),set_type,NoSyn)] thy 

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for add const : "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 


	val set_i = (Const  ("==", set_type --> set_type --> propT) 
				    $ (Const(string_of_path (hc_path l1 (set_of trans  (name_of C))),set_type)) 
				    $ (Term.absfree("\\<tau>",St(u), 
						    Const("FixedPoint.gfp",(mk_setT (V(u,class_u)) 
										    --> mk_setT (V(u,class_u))) 
									       --> mk_setT (V(u,class_u))) 
							 $ (set_fun $ Free("\\<tau>",St(u))))) 
			    )


	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for constructing definition : "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 


	val thy = add_definitions_i false  (SOME (path_of trans C))  
				    [((local_name_of (hc_path l1 (set_of trans  (name_of C)))),set_i)] thy 


	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for add def : "
				 ^(Time.toString(Time.-(endTime,startTime))))

    in
	thy
    end


fun add_set_fun trans C u cl  thy =  
    let
	val thy_name = Context.theory_name thy
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 
	val hol_inv_type = mk_funT(St(u),mk_funT(mk_setT (V(u,class_u)),mk_funT(V(u,class_u),boolT)))  
	val set_fun_type = mk_funT(St(u),mk_funT(mk_setT (V(u,class_u)),mk_setT (V(u,class_u))))  
			   



	fun prove_set_fun_mono thy C  = 
	    let
		fun unfold_inv_fast_tac thy n = let 
		    fun f thy topthm = 
			let
			    val unfold  = let 
				val (_,_,t,_) = (dest_state (topthm,n))
						
				fun const_defs (Const ("==>",_) $ _  $ r)  = const_defs r 
				  | const_defs (Const ("all",_) $ r  )     = const_defs r 
				  | const_defs (Const ("Trueprop",_) $ r ) = const_defs r 
				  | const_defs (Abs (_,_,r))               = const_defs r 
				  | const_defs (l $ r )                    = (const_defs l) @ (const_defs r)
				  | const_defs (Const (t,_))               = [(t^"_def")] 
				  | const_defs (Free  (_,_))               = [] 
				  | const_defs _                           = [] 
			    in	
				case (filter (String.isSubstring "inv" ) (const_defs t)) of 
				    [] => []
				  | l => map (fn n => get_thm (thy) (Name n)) l
			    end
			in
			    ((asm_full_simp_tac (HOL_ss addsimps unfold) n)
				 THEN (fast_tac HOL_cla n ORELSE all_tac))  topthm
			end
		in
		    f thy 
		end
						
		fun hol_inv_names C = if ((name_of C) = (name_of OclAnyC)) 
				      then [] 
				      else
					  let
					      val pinv =     ["check_parent_inv"]
					  in
					      [(string_of_path ( (hc_path l1 (hol_invariant_of trans  (name_of C)))))^"_def"]
					      @(
					      (map (fn i => (string_of_path ( (hc_path l1 (hol_invariant_of trans 
													  (name_of C)))))
							  ^"."^i^"_def")
						   pinv))
					  end
		 			  
		val pl =  (name_of C)::(filter (fn c => not (c = name_of OclAnyC)) (parents_of C cl))
		val l = ((map (fn n => class_of n cl) pl))
			
		val invariants_hol_names = List.concat (map (hol_inv_names) l)
		val invariants_defs = map (fn i => get_thm thy (Name(i))) 
					  (invariants_hol_names)
		val set_funP =  (hc_path l1 (set_fun_of trans  (name_of C)))
		val set_fun = string_of_path set_funP
		val set_fun_def = get_thm thy (Name (set_fun^"_def"))
		val hol_inv_def = get_thm thy (Name ((string_of_path (hc_path l1 (hol_invariant_of trans  (name_of C))) ^"_def")))
		val mono_def = get_def thy ( "mono")
			       
		fun mkGoal_i prems goal = Const ("==>",propT --> propT --> propT) $
						(Const ("Trueprop",boolT --> propT) $ prems) $
						(Const ("Trueprop",boolT --> propT) $ goal)
		(* mono (set_fun \<tau>) *)
		val goal = mkGoal_i (Const("True",boolT)) 
				    (Const("Orderings.mono",(mk_funT(mk_setT (V(u,class_u)),mk_setT (V(u,class_u)))) 
								--> boolT)
					  $(Const(set_fun,set_fun_type)$Free("\\<tau>",St(u))))
			   
		val goal_ci = (cterm_of (sign_of thy) (goal))
		val thy = let
		    fun script p = if (!quick_and_dirty) then []
				   else 
				        [cut_facts_tac p 1,
					 asm_full_simp_tac(HOL_ss 
							       addsimps 
							       (invariants_defs
								@[mono_def,set_fun_def])) 1,
					 safe_tac OCL_cla,
					 (ALLGOALS(unfold_inv_fast_tac thy)),
					 TRY(auto_tac (OCL_cla,OCL_ss))
					]
		    val thm = quick_and_dirty_prove_goalw_cterm thy [] goal_ci script
		in
		    add_thm (SOME(set_funP)) [(("mono",thm),[])] (thy)
		end     
	    in
		thy 
	    end
	val inv = Const(string_of_path(hc_path l1 (hol_invariant_of trans  (name_of C))), hol_inv_type)
	val thy = add_consts_i (SOME (path_of trans C)) 
		  [((local_name_of (hc_path l1 (set_fun_of trans  (name_of C) ))),set_fun_type,NoSyn)] thy 
	val set_fun  =  (Term.absfree ("\\<tau>",St(u),
				      (Term.absfree ("C",mk_setT (V(u,class_u)),
						     (Const ("Collect",( V(u,class_u) --> boolT) 
									   --> (mk_setT (V(u,class_u)))) $
							    (Term.absfree("self",V(u,class_u),  
								   (inv $ Free("\\<tau>",St(u)) $ 
									Free("C",mk_setT (V(u,class_u)))$ 
									Free("self",V(u,class_u)))
 		       )))))))
       	
          val set_fun_i =   (Const  ("==", set_fun_type --> set_fun_type --> propT) 
				    $ (Const(string_of_path (hc_path l1 (set_fun_of trans  (name_of C))),set_fun_type)) 
				    $ set_fun )
	val thy = add_definitions_i false  (SOME (path_of trans C))  
				    [((local_name_of (hc_path l1 (set_fun_of trans  (name_of C)))),set_fun_i)] thy 
	 val thy = prove_set_fun_mono thy C  
    in
	thy
    end 

fun add_cset trans C u cl  thy =  
    let
	fun prove_def path defs thy = 
	    let
		val path = case path of
			       SOME p => p
			     | NONE => []
		val local_ss = simpset_of (thy)
		fun def_goal thy t = (cterm_of (sign_of thy)
						  (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						      [ (Const ("Trueprop",boolT --> propT) 
								    $ (Const ("OCL_Logic_core.OclValid",
									      BOOLT (St u) --> boolT) $
									     (Const ("OCL_Boolean.OclIsDefined",dummyT)
										    $ t)))
						      ],propT))))
		fun prove thy def = let 
		    val _ = trace 3 ("          proving def of "^def)
		in
		    SOME(prove_goalw_cterm_nocheck [] (def_goal thy (Const(string_of_path (path@[def]),dummyT)))
						   (fn p => [cut_facts_tac p 1,
							   rewrite_goals_tac [get_thm thy (Name(def^"_def"))],
							   simp_tac (OCL_DEF_ss) 1
			]))
		    handle _ => let 
			       val _ = trace 1 ("Warning: could not prove def for "
						^(string_of_path (path@[def]))^"!")
			   in 
			       NONE
			   end
		end
				    
		val thms = filter (fn (n, t) => isSome t) (map (fn n => ("def_"^n, prove thy n)) defs)
		val rsim = simpset_ref_of thy
	    in
		rsim := !rsim addsimps (map (valOf o snd) thms);
		thy
	    end


	val thy_name = Context.theory_name thy
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 
	val cset_type  = mk_funT(St(u),SetT (class_u))
	val thy = add_consts_i (SOME (path_of trans C)) 
		  [((local_name_of (hc_path l1 (cset_of trans (name_of C) ))),
		    cset_type,NoSyn)] thy 
	val set_name  = string_of_path (set_of trans  (name_of C))
	val  timer    =  Timer.startRealTimer ()
        val startTime = Timer.checkRealTimer timer 
	val cset_i    = (#1(Sign.infer_types (Sign.pp thy)  thy
					  (K NONE) (K NONE) [] true ([
			 (Const  ("==", cset_type --> cset_type --> propT)
				 $(Const(string_of_path (hc_path l1 (cset_of trans  (name_of C))),
					 cset_type)) 		 
				 $(Term.absfree("\\<tau>",St(u), 
				 Const("OCL_Set_type.Abs_Set_0",dummyT) $ 
				      (Const("Lifting.up.lift",dummyT)
					   $ (Const("op -",dummyT)
						   $ (Const("image",dummyT)
							   $ (Term.absfree("f",dummyT,
									   Free("f",dummyT)$Free("\\<tau>", St(u))))
							   $ (Const(set_name,dummyT)
								   $Free("\\<tau>",St(u)))
						     )
						   $ (Const("insert",dummyT)
							   $Const("Lifting.UU",dummyT)
							   $Const("{}",dummyT)
						     )
					     )
				      )
				 
				 ))
			 )
			 ],propT))) 	    

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for infer types:: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 

	val thy = add_definitions_i false  (SOME (path_of trans C))  
				    [((local_name_of (hc_path l1 (cset_of trans  (name_of C))))^"_def",cset_i)] thy 

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for add_def types:: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	val startTime = Timer.checkRealTimer timer 

	val thy = prove_def (SOME (path_of trans C))
					 [local_name_of (hc_path l1 (cset_of trans (name_of C)))]  thy

	val endTime = Timer.checkRealTimer timer
	val _ = trace 2 ("        -> time for prove_def types:: "
				 ^(Time.toString(Time.-(endTime,startTime))))

    in
	thy
    end
fun add_inv trans C u cl  thy =  
    let
	val thy_name = Context.theory_name thy
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 
	val inv_type  = mk_funT(V(u,class_u),BOOLT(St(u)))

	val thy = add_consts_i (SOME (path_of trans C)) 
		  [((local_name_of (hc_path l1 (inv_of trans (name_of C) ))),
		    inv_type,NoSyn)] thy 
	val cset_type  = mk_funT(St(u),SetT (class_u))
	val cset_name = string_of_path (cset_of trans  (name_of C))
	val inv_i = (#1(Sign.infer_types (Sign.pp thy)  thy
					  (K NONE) (K NONE) [] true ([
			 (Const  ("==", inv_type --> inv_type --> propT)
				 $(Const(string_of_path (hc_path l1 (inv_of trans  (name_of C))),
					 inv_type)) 		 
				 $(Term.absfree("self",V(u,class_u),
						Const("OCL_Collection.OclIncludes",dummyT)
						$(Const(cset_name,cset_type))
						$(Free("self",V(u,class_u)))

					       )
				  )
			 )
			 ],propT))) 	    

	val thy = add_definitions_i false  (SOME (path_of trans C))  
				    [((local_name_of (hc_path l1 (inv_of trans (name_of C))))^"_def",inv_i)]
				    thy 


	val thy = prove_basic_properties (SOME (path_of trans C))
			   [local_name_of (hc_path l1 (inv_of trans (name_of C)))]  thy


    in
	thy
    end






fun add_user_invariant C u cl  thy =
    let

	fun prove_uinv_valid_if_model_inv cname thy = 
	    let
		val uinv      = hc_path l2 (user_inv_of cname)
		val userinv = ("model",let fun f (cl:(Rep_Core.Classifier list)) (C:(Rep_Core.Classifier)) 
						 (inv:(Rep_HolOclTerm.HolOclTerm)) =inv in f end)
		val model_inv = local_name_of (hc_path l1 (inv_of userinv cname))
		val goal =  (#1(Sign.infer_types (Sign.pp thy)  thy
						 (K NONE) (K NONE) [] true ([
				Const("Trueprop",boolT --> propT) 
				     $(Const("op =",dummyT) 
					    $ (OclLocalValid u 
							     (Free("\\<tau>", St(u)))
							     (Const(string_of_path uinv,dummyT)$Free("self",dummyT)))
					    $ (OclLocalValid u 
							     (Free("\\<tau>", St(u)))
							     (Const(model_inv,dummyT)$Free("self",dummyT)))
				      )
				],propT)))

			    
	    in
		(* TODO: this should be proven in the next version ... *)
		add_axioms_i (SOME (hc_path l2 (name_of C)) ) [("inv2sem",goal)]  thy
	    end

	val thy_name = Context.theory_name thy

	val name = name_of C	

	fun oclterm_of_inv thy cl u C  = map (fn (n,c) => (n, Ocl2holocl l2 thy cl u c)) (invariant_of C)
	val _ = trace 3  ("        "^(string_of_path name))
	val invariants  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("inv_"^Int.toString(n), snd c )
					   )) (0,(oclterm_of_inv thy cl u C))) 


	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 


        (* add INV constant *)
	val inv_type      = mk_funT(V(u,class_u),BOOLT(St(u)))  

	val thy = foldl (fn (constraint,thy) 
			    =>  add_consts_i (SOME (hc_path l2 (user_inv_of  (name_of C))))
				    [((fst constraint),inv_type,NoSyn)] thy) 
		  thy invariants

	 val invs = map (fn (n,t) =>  (n,#1(Sign.infer_types (Sign.pp  thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", inv_type --> inv_type --> propT) $
                               Const ((string_of_path ((hc_path l2 (user_inv_of  (name_of C)))))^"."^n,inv_type) $ 
			       Term.absfree ("self",V(u,class_u), t)
			      )
			      ],propT)))) invariants 

	val thy = foldl (fn (constraint,thy) 
			    =>  add_definitions_i false (SOME ((hc_path l2 (user_inv_of  (name_of C)))))
				    [((fst constraint)^"_def",(snd constraint))] thy) 
		  thy invs 

	val thy = prove_basic_properties (SOME(hc_path l2 (user_inv_of  (name_of C))))
			       (map fst invs)  thy


	(* add top-level invariant *)
	val thy = add_consts_i (SOME ((hc_path l2 (user_inv_of  (name_of C))))) 
		  [((local_name_of (hc_path l2 (user_inv_of  (name_of C) ))),inv_type,NoSyn)] thy 



	  

	val c_str_i = map (fn s => Const((string_of_path  (hc_path l2 (user_inv_of  (name_of C))))
					 ^"."^(fst s),inv_type)$Free("obj",V(u,class_u))) invariants


	val inv_def_i =  (#1(Sign.infer_types (Sign.pp thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", inv_type --> inv_type --> propT) 
				      $ (Const(string_of_path (hc_path l2 (user_inv_of  (name_of C))),inv_type)) 
				      $ (Term.absfree ("obj",V(u,class_u),(foldr1 (fn (a,b) => OclAnd u a b) c_str_i))))
			       ],propT)))


	val thy = add_definitions_i false (SOME ((hc_path l2 (user_inv_of  (name_of C)))))  
				  [((local_name_of (hc_path l2 (user_inv_of  (name_of C)))),inv_def_i)] thy 

	val thy = prove_basic_properties (SOME(hc_path l2 (user_inv_of  (name_of C))))
			       [local_name_of (hc_path l2 (user_inv_of  (name_of C)))] thy


	val thy = prove_uinv_valid_if_model_inv (name_of C) thy;
     in 
	thy
    end








fun add_invariant C u cl  thy = 
    let


	val timer =  Timer.startRealTimer ()

	fun printTiming startTime = 
	    let		
		val endTime = Timer.checkRealTimer timer
		val _ = trace 2 ("        ==> time used: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	    in 
		Timer.checkRealTimer timer
	    end



	val userinv = ("model",let fun f (cl:(Rep_Core.Classifier list)) (C:(Rep_Core.Classifier)) 
					 (inv:(Rep_HolOclTerm.HolOclTerm)) =inv in f end)
	fun invtrans () = userinv::(!transformations)

		val startTime = Timer.checkRealTimer timer 
                val _ = trace 2 "       adding hol_invariant"
	val thy = fst (foldl_map (fn (thy, t) => (add_hol_invariant t C u cl thy,()))   (thy, invtrans ()))
	        val startTime = printTiming startTime
                val _ = trace 2 "       adding set_fun"
	val thy = fst (foldl_map (fn (thy, t) => (add_set_fun t C u cl thy,()))   (thy, invtrans ()))
	        val startTime = printTiming startTime
                val _ = trace 2 "       adding set"
	val thy = fst (foldl_map (fn (thy, t) => (add_set t C u cl thy,()))   (thy, invtrans ()))
	        val startTime = printTiming startTime
                val _ = trace 2 "       adding cset"
	val thy = fst (foldl_map (fn (thy, t) => (add_cset t C u cl thy,()))   (thy, invtrans ()))
	        val startTime = printTiming startTime
                val _ = trace 2 "       adding inv"
	val thy = fst (foldl_map (fn (thy, t) => (add_inv t C u cl thy,()))   (thy, invtrans ()))
	        val startTime = printTiming startTime
    in
	thy
    end

end
