(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * mdr_encoder_common.sml - generic mdr import interface for HOL-OCL
 * Copyright (C) 2001-2005  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

signature REP_ENCODER_COMMON =
sig
    datatype Level = univ    (* universe toolbox *)
		   | operation
		   | holocl  (* HOL-OCL toplevel *)
		   | l0      (* Level 0          *)
		   | l1      (* Level 1          *)
		   | l2      (* Level 2          *)
	
    (* log level, etc *)
    val log_level : int ref
    val isa_packages_quiet : unit->unit
    val isa_packages_verbose : unit->unit

    val trace : int -> string -> unit

    (* useful simplfier- and cla-sets *)	     
    val HOL_cla : claset
    val OCL_cla : claset
    val OCL_ss  : simpset
    val HOL_DEF_ss : simpset
    val OCL_DEF_ss : simpset

    (* how picky are we? *)
    val strict_checking : bool ref

    (* generic operations on path expressions; move to su4sml? *)
    val string_of_path :Rep_OclType.Path -> string

    (* useful stuff; move to su4sml? *)
    val rm_whitespace : string -> string
    val listin : ''a  -> ''a list -> bool

    (* "overloaeded" Isabelle add_*, etc *)
    val add_tyabbrs_i : Rep_OclType.Path option -> (string * string list * typ * mixfix) list -> theory -> theory
    val add_consts_i : Rep_OclType.Path option -> (string * typ * mixfix) list -> theory -> theory
    val add_definitions :  Rep_OclType.Path option -> (string * string) list -> theory -> theory
    val add_definitions_i : bool -> Rep_OclType.Path option -> (string * term) list -> theory -> theory
    val add_axioms_i : Rep_OclType.Path option -> (string * term) list -> theory -> theory
    val add_thm :  Rep_OclType.Path option -> ((string * thm) * theory attribute list) list -> theory -> theory
  
    (* unclassified stuff *)
    val post_1_of : Rep_OclType.Path -> Rep_OclType.Path 
    val pre_1_of : Rep_OclType.Path -> Rep_OclType.Path 
    val local_name_of : Rep_OclType.Path -> string

    val hc_path : Level -> Rep_OclType.Path -> Rep_OclType.Path

    val get_class_of : Rep_OclType.Path -> Rep_OclType.Path 
    val mk_class_of :  Rep_OclType.Path -> Rep_OclType.Path
    val mk_def_of :  Rep_OclType.Path -> Rep_OclType.Path
					     
    val is_hc_pathstr : Level-> string -> bool
    val class2get_parent : Rep_OclType.Path -> string -> Rep_OclType.Path
    val parent2class_of : Rep_OclType.Path -> string -> Rep_OclType.Path

    val is_univ_class_of : Rep_OclType.Path -> Rep_OclType.Path
    val is_class_of : Rep_OclType.Path -> Rep_OclType.Path
    val set_thy_path : Rep_OclType.Path option -> theory -> (Rep_OclType.Path option * theory)

    val boolT : typ
    val mk_funT : (typ * typ) -> typ
    val mk_setT : typ -> typ
    val LiftT : typ -> typ
    val typeS : sort

    val holes_of_type : typ -> Rep_Core.Classifier -> string list
    val holes_of_typestr : typ -> Rep_OclType.Path -> string list
    val get_type_universe : typ -> Rep_OclType.Path -> typ option
    val key_name_of :  Rep_OclType.Path -> string list

    val collect_holes :  typ -> string list
    val base_type_name_of : Rep_OclType.Path -> Rep_OclType.Path
    val type_name_of : Rep_OclType.Path -> Rep_OclType.Path
    val attr_of : Rep_OclType.Path -> string -> Rep_OclType.Path
    val is_classType_of : Rep_OclType.Path -> Rep_OclType.Path

    val extT : typ * typ * typ -> typ
    val mk_prodT : (typ * typ) -> typ
    val unitT : typ
    val ext_name_of : Rep_OclType.Path -> string list
    val dest_LiftT : typ -> typ

    val StringT : typ
    val IntegerT : typ
    val RealT : typ
    val BooleanT : typ
    val SetT : typ -> typ
    val OrderedSetT : typ -> typ
    val BagT : typ -> typ
    val SequenceT : typ -> typ
    val RefT : typ
    val cl_extT : (typ * typ) -> typ

    val prove_basic_properties : Rep_OclType.Path  option -> string list -> theory -> theory
    val unfold_level_1_tac : theory -> thm -> thm Seq.seq

    val tilde :  typ -> Rep_OclType.OclType -> typ
    val V : (typ * typ) -> typ
    val attr_pre_of : Rep_OclType.Path -> string -> Rep_OclType.Path 

    val type_charset_name_of : Rep_OclType.Path -> Rep_OclType.Path
    val OclAnyT : typ -> typ
    val BOOLT : typ -> typ
    val REALT : typ -> typ
    val BAGT : (typ * typ) -> typ
    val SETT : (typ * typ) -> typ
    val ORDEREDSETT : (typ * typ) -> typ
    val SEQUENCET : (typ * typ) -> typ
    val INTEGERT : typ -> typ
    val STRINGT : typ -> typ
    val St : typ -> typ
    val get_def : theory -> string -> thm
    val invariant_1_of : Rep_OclType.Path -> Rep_OclType.Path
    val invariant_2_of : Rep_OclType.Path -> Rep_OclType.Path
    val strong_sve_2_of :Rep_OclType.Path -> Rep_OclType.Path
    val strict_sve_2_of :Rep_OclType.Path -> Rep_OclType.Path
    val isValueType : Rep_OclType.OclType -> bool
    val UT : typ -> typ
end


structure rep_encoder_common : REP_ENCODER_COMMON = 
struct
open Rep_OclType;
open Rep_Core;

val strict_checking = ref true
     
val log_level = ref 2 (* 0 -> no output *)
                      (* 1 -> level information (sections) *)
                      (* 2 -> steps on each level (subsection) *)
                      (* 3 -> class name on each step *)


(* defining simpsets *)

val OCL_cla     = claset_of (theory "OCL")
val HOL_cla     = claset_of (theory "HOL")

val OCL_ss     = simpset_of (theory "OCL")
val HOL_DEF_ss = (HOL_ss addsimps [
		  get_thm (theory "OCL") (Name("OclIsDefined_def")),
		  get_thm (theory "OCL") (Name("OclTrue_def")),
		  get_thm (theory "OCL") (Name("lift0_def")),
		  get_thm (theory "OCL") (Name("lift1_def")),
		  get_thm (theory "OCL") (Name("DEF_def")),
		  get_thm (theory "OCL") (Name("OclValid_def")),
		  get_thm (theory "OCL") (Name("OclLocalValid_def")),
		  get_thm (theory "OCL") (Name("OclTrue_def")),
		  get_thm (theory "OCL") (Name("strictify_def"))

		 ])
val OCL_DEF_ss = (OCL_ss addsimps [
		  get_thm (theory "OCL") (Name("OclIsDefined_def")),
		  get_thm (theory "OCL") (Name("OclTrue_def")),
		  get_thm (theory "OCL") (Name("lift0_def")),
		  get_thm (theory "OCL") (Name("lift1_def")),
		  get_thm (theory "OCL") (Name("DEF_def")),
		  get_thm (theory "OCL") (Name("OclValid_def")),
		  get_thm (theory "OCL") (Name("OclLocalValid_def")),
		  get_thm (theory "OCL") (Name("OclTrue_def")),
		  get_thm (theory "OCL") (Name("strictify_def"))
		 ])
		 

		

fun get_def thy name = get_thm thy (Name(name^"_def")) 


fun listin _ []  = false
  | listin e (x::xs) = if e=x then true else listin e xs


fun isa_packages_quiet() = let
    val _ = (DatatypePackage.quiet_mode :=  true)
    val _ = (InductivePackage.quiet_mode:= true)
    val _ = (Syntax.ambiguity_level:= 100000)
in () end

fun isa_packages_verbose() = let
    val _ = (DatatypePackage.quiet_mode :=  false)
    val _ = (InductivePackage.quiet_mode:= false)
    val _ = (Syntax.ambiguity_level:= 1)
in () end

fun trace lev  s = if(lev  <= !log_level ) then tracing("### rep_encoder: "^s) else ()

fun path_of_thy thy = (NameSpace.unpack o NameSpace.path_of) (#naming (Sign.rep_sg (sign_of thy)))

fun set_thy_path path thy = let
    val old_path = SOME(path_of_thy thy)
    val thy      = Theory.add_path "/" thy
    val thy      = case path of
                       SOME (p) => Theory.add_path (string_of_path p) thy
                     | NONE => thy
in
    (old_path,thy)
end


fun antiquot_trace_thm l path n = let
    val pkg_name = case path of
		       SOME (p) => p
		     | NONE => []
    val name = (string_of_path (pkg_name@n))
    val _ = trace l ("LaTeX: text {* \verb|"^name^"|: @{thm [display,show_types] "^name^"} *} ")       
    val _ = trace l ("LaTeX:      (*<*) thm "^name^" (*>*) ")
in
    ()
end

fun antiquot_trace_typ l path n = let
    val pkg_name = case path of
		       SOME (p) => p
		     | NONE => []
    val name = (string_of_path (pkg_name@n))
    val _ = trace l ("LaTeX: text {* \verb|"^name^"|: @{typ [display] "^name^"} *} ")       
in
    ()
end

fun antiquot_trace_const l path n = let
    val pkg_name = case path of
		       SOME (p) => p
		     | NONE => []
    val name = (string_of_path (pkg_name@n))
    val _ = trace l ("LaTeX: text {* \verb|"^name^"|: @{const [display,show_types] "^name^"} *} ")       
in
    ()
end


fun add_thm path thms thy =
    let
        val _ = case path of
                    SOME (pkg_name) => trace 9  ("Adding theorem (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 9  ("Adding theorem:")

	val _ = map (fn ta => antiquot_trace_thm 10 path [(fst o fst) ta]) thms

        val (oldpath,thy) = set_thy_path path thy

        val thy = (fst(PureThy.add_thms thms (thy)))

        val (_,thy) = set_thy_path oldpath thy

    in
        thy
    end


fun add_tyabbrs_i path tas thy =
    let
        val _ = case path of
                    SOME (pkg_name) => trace 9  ("Adding type abbrevs (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 9 ("Adding type abbrevs:")
        val _ = let val pstr = (case path of
                               SOME (pkg_name) => (string_of_path pkg_name)^"."
                             | NONE => "")
                in
                    map (fn (d,_,_,t) => trace 9 ("   "
                                                 ^pstr
                                                 ^(d)^" ("
                                                  ^(Syntax.string_of_mixfix(t))^")")
                                           ) tas
                end
	val _ = map (fn (d,_,_,_) => antiquot_trace_typ 10 path [d]) tas
        val (oldpath,thy) = set_thy_path path thy

        val thy =Theory.add_tyabbrs_i tas thy
        val thy = Theory.add_types [] thy

        val (_,thy) = set_thy_path oldpath thy

    in
        thy
    end 

fun add_consts_i path consts thy =
    let
        val _ = case path of
                    SOME (pkg_name) => trace 4  ("          "^"Adding consts (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 4  ("          "^"Adding consts:")
        val (oldpath,thy) = set_thy_path path thy
        val _ = map (fn (d,_,t) => trace 4 ("          "^"  "^(d)
					    ^" ("^(Syntax.string_of_mixfix(t))^")") ) 
		    consts
	val _ = map (fn (d,_,_) => antiquot_trace_const 10 path [d]) consts


        val thy = Theory.add_consts_i consts  thy
        val (_,thy) = set_thy_path oldpath thy
   in
        thy
    end

val boolT      = (Type("bool",[]))

fun normalize_def_name name = if String.isSuffix "_def" name 
			      then name 
			      else (name^"_def")

fun prove_basic_properties path defs thy = 
    let 

	
	fun prove_def path defs thy = 
	    let
		val path = case path of
			       SOME p => p
			     | NONE => []
		val local_ss = simpset_of (thy)
		fun def_goal thy t = (cterm_of (sign_of thy)
						  (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						      [ 
						       Const ("==>",propT --> propT --> propT) $
							     (Const ("Trueprop",boolT --> propT) 
								    $ (Const ("OCL_Logic_core.OclValid",dummyT) $
									     (Const ("OCL_Boolean.OclIsDefined",dummyT)
										    $ Free ("obj",dummyT)))) 
							     $(Const ("Trueprop",boolT --> propT) 
								    $ (Const ("OCL_Logic_core.OclValid",dummyT) $
									     (Const ("OCL_Boolean.OclIsDefined",dummyT)
										    $( t $ Free ("obj",dummyT)))))
						      ],propT))))
		fun prove thy def = let 
		    val _ = trace 3 ("          proving def of "^def)
		in
		    SOME(prove_goalw_cterm_nocheck [] (def_goal thy (Const(string_of_path (path@[def]),dummyT)))
						   (fn p => [cut_facts_tac p 1,
							   rewrite_goal_tac [get_thm thy (Name(def^"_def"))] 1,
							   simp_tac (local_ss) 1,
							   TRY(simp_tac (HOL_DEF_ss) 1)
			]))
		    handle _ => let 
			       val _ = trace 2 ("Warning: could not prove def for "
						^(string_of_path (path@[def]))^"!")
			   in 
			       NONE
			   end
		end
				    
		val thms = filter (fn (n, t) => isSome t) (map (fn n => ("def_"^n, prove thy n)) defs)
	    in
		((map (valOf o snd) thms),[],thy)
	    end

	fun prove_cp path defs thy = 
	    let
		val path = case path of
			       SOME p => p
			     | NONE => []
		val local_ss = simpset_of (thy)
		val local_cla = claset_of (thy)
		fun cp_goal thy t = (cterm_of (sign_of thy)
					      (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						  [(Const ("Trueprop",boolT --> propT) $
							  (Const("Lifting.cp",dummyT)$t))
						  ],propT))))
		fun prove thy def = let 
		    val _ = trace 3 ("          proving cp of "^def)
		in
		    SOME(prove_goalw_cterm_nocheck [] (cp_goal thy (Const(string_of_path (path@[def]),dummyT)))
						   (fn p => [cut_facts_tac p 1,
							   rewrite_goal_tac [get_thm thy (Name(def^"_def"))] 1,
							   asm_full_simp_tac (local_ss) 1
			]))
		    handle _ => let 
			       val _ = trace 2 ("Warning: could not prove cp for "
						^(string_of_path (path@[def]))^"!")
			   in 
			       NONE
			   end
		end
				    
		val thms = filter (fn (n, t) => isSome t) (map (fn n => ("cp_"^n, prove thy n)) defs)
	    in
		((map (valOf o snd) thms),(map (valOf o snd) thms),thy)
	    end
	    
	    
	    
	fun prove_strict path defs thy = 
	    let
		val path = case path of
			       SOME p => p
			     | NONE => []
		val local_ss = simpset_of (thy)
		fun strict_goal thy t = (cterm_of (sign_of thy)
						  (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						      [ Const ("Trueprop",boolT --> propT) 
							      $(Const("op =",dummyT)
								     $(t$(Const("OCL_Undefined.OclUndefined",dummyT)))
								     $(Const("OCL_Undefined.OclUndefined",dummyT))
							       )
						      ],propT))))
		fun prove thy def = let 
		    val _ = trace 3 ("          proving strict of "^def)
		in
		    SOME(prove_goalw_cterm_nocheck [] (strict_goal thy (Const(string_of_path (path@[def]),dummyT)))
						   (fn p => [cut_facts_tac p 1,
							   rewrite_goal_tac [get_thm thy (Name(def^"_def"))] 1,
							   simp_tac (local_ss) 1
			]))
		    handle _ => let 
			       val _ = trace 2 ("Warning: could not prove strict for "
						^(string_of_path (path@[def]))^"!")
			   in 
			       NONE
			   end
		end
				    
		val thms = filter (fn (n, t) => isSome t) (map (fn n => ("strict_"^n, prove thy n)) defs)
	    in
		(map (valOf o snd) thms, [], thy)
	    end

	val (cp_sim, cp_cla, thy)         = prove_cp path defs thy
	val (strict_sim, strict_cla, thy) = prove_strict path defs thy
	val (def_sim, def_cla, thy)       = prove_def path defs thy
	val sim = cp_sim@strict_sim@def_sim
	val cla = cp_cla@strict_cla@def_cla

	val rsim = simpset_ref_of thy
	val rcla = claset_ref_of thy

    in
	rsim := !rsim addsimps sim;
	rcla := !rcla addSIs cla;
	thy
    end




fun add_definitions path defs thy  =
    let
	val defs = map (fn (n,d) => (normalize_def_name n,d)) defs
        val _ = case path of
                    SOME (pkg_name) => trace 4  ("          "^"Adding defs (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 4  ("          "^"Adding defs:")
        val _ = let val pstr = (case path of
                               SOME (pkg_name) => (string_of_path pkg_name)^"."
                             | NONE => "")
                in
                    map (fn (name,def) => trace 4 ("          "^"  "^(pstr^name)^": "^def ) ) defs
                end
        val (oldpath,thy) = set_thy_path path thy
        val _ = map (fn ta => trace 9 ("   "^(#1 ta)) ) defs
	val _ = map (fn ta => antiquot_trace_thm 10 path [#1 ta]) defs
        val thy = IsarThy.add_defs (true,map (fn x => (x,[])) defs) thy
        val (_,thy) = set_thy_path oldpath thy
    in
        thy
   end

fun add_definitions_i overloaded path defs thy  =
    let
        val defs = map (fn (n,d) => (normalize_def_name n,d)) defs
        val _ = case path of
                    SOME (pkg_name) => trace 4  ("          "^"Adding defs (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 4  ("          "^"Adding defs:")
	val pkg_name = case path of
			   SOME (p) => p
			 | NONE => []
        val (oldpath,thy) = set_thy_path path thy
        val _ = map (fn ta => trace 4 ("          "^"  "^(#1 ta)) ) defs
	val _ = map (fn ta => antiquot_trace_thm 10 path [#1 ta])  defs

        val thy = IsarThy.add_defs_i (overloaded,map (fn x => (x,[])) defs) thy
        val (_,thy) = set_thy_path oldpath thy
    in
        thy
   end

fun add_axioms path defs thy  =
    let
        val _ = case path of
                    SOME (pkg_name) => trace 9  ("Adding Axioms (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 9  ("Adding Axioms:")
	val pkg_name = case path of
			   SOME (p) => p
			 | NONE => []
        val _ = map (fn (name,def) => trace 9 ("   "^(name)^": "^def ) ) defs
	val _ = map (fn (name,def) => antiquot_trace_thm 10 path [name]) defs
				     
       val (oldpath,thy) = set_thy_path path thy
        val _ = map (fn ta => trace 9 ("   "^(#1 ta)) ) defs
        val thy = IsarThy.add_axioms (map (fn x => (x,[])) defs) thy
        val (_,thy) = set_thy_path oldpath thy

    in
        thy
   end

fun add_axioms_i path defs thy  =
    let
        val _ = case path of
                    SOME (pkg_name) => trace 9  ("*Caution* Adding axiom (namespace: "
						 ^(string_of_path pkg_name)^"):")
                  | NONE => trace 9  ("*Caution* Adding axioms:")
	val pkg_name = case path of
			   SOME (p) => p
			 | NONE => []
        val _ = map (fn (name,def) => trace 9 ("   "^(name)) ) defs
        val _ = map (fn (name,def) => antiquot_trace_thm 10 path [name]) defs
        val (oldpath,thy) = set_thy_path path thy
        val _ = map (fn ta => trace 9 ("   "^(#1 ta)) ) defs
        val thy = IsarThy.add_axioms_i (map (fn x => (x,[])) defs) thy
        val (_,thy) = set_thy_path oldpath thy

    in
        thy
   end


(***************************************************)

datatype Level = univ    (* universe toolbox *)
	       | operation
	       | holocl  (* HOL-OCL toplevel *)
	       | l0      (* Level 0          *)
	       | l1      (* Level 1          *)
	       | l2      (* Level 2          *)

fun local_name_of path = case path of 
			     [] =>  "" 
			   | p => (hd o rev) p

fun hc_path level p =  if ((String.isSubstring "OclAny" (local_name_of p))
		       andalso not  (String.isSubstring "_2_OclAny" (local_name_of p))
		       andalso not  (String.isSubstring "OclAny_2" (local_name_of p)))
		       (* hack *)
		      then 			  
			   case level of
			      univ      => p  
			    | operation => p
			    | holocl    => p 
			    | l0        => ("UML_OCL"::["level0"])@p 	
			    | l1        => ("UML_OCL"::["level1"])@p 
			    | l2        => p 
		      else
			  case level of
			      univ      => ("UML_OCL"::["univ"])@p   (* universe toolbox *)
			    | operation => ("UML_OCL"::["ops"])@p 
			    | holocl    =>  "UML_OCL"::p             (* HOL-OCL toplevel *)
			    | l0        => ("UML_OCL"::["level0"])@p   
			    | l1        => ("UML_OCL"::["level1"])@p
			    | l2        => p 

fun is_hc_pathstr level pstr = 	if ((String.isSubstring "OclAny" pstr)
		       andalso not  (String.isSubstring "_2_OclAny" pstr)
		       andalso not  (String.isSubstring "OclAny_2" pstr))
		       (* hack *)
		      then 			  
			   case level of
			      univ   => false
			    | operation => false
			    | holocl    => false
			    | l0        => String.isSubstring "UML_OCL.level0." pstr
			    | l1        => String.isSubstring "UML_OCL.level1." pstr 
			    | l2        => false 
		      else
			  case level of
			      univ   => String.isSubstring "UML_OCL.univ." pstr
			    | operation => String.isSubstring "UML_OCL.ops" pstr
			    | holocl    => String.isSubstring "UML_OCL." pstr
			    | l0        => String.isSubstring "UML_OCL.level0." pstr
			    | l1        => String.isSubstring "UML_OCL.level1." pstr
			    | l2        => not (String.isSubstring "UML_OCL." pstr)





(* generating constant names *)

fun key_name_of path = if (path = (name_of OclAnyC)) (* hack *)
		  then ["OCL_OclAny_type","OclAny_key"]
		  else rev (((hd (rev path))^"_key")::(tl (rev path))) (* keyC *)	

fun ext_name_of path = if (path = (name_of OclAnyC)) (* hack *) 
		  then  ["OCL_OclAny_type","OclAny_ext"]
		  else rev (((hd (rev path))^"_ext")::(tl (rev path))) (* extC *)	

fun base_type_name_of path = if (path = (name_of OclAnyC)) (* hack *) 
			     then  ["OCL_OclAny_type","OclAny_key"] (* ??? *)
			     else rev (((hd (rev path))^"_base_type")::(tl (rev path)))

fun type_name_of path = if (path = (name_of OclAnyC)) (* hack *)
			then ["OCL_OclAny_type","OclAny_key"]
			else rev (((hd (rev path))^"_type")::(tl (rev path)))           (* typeC *)	


fun mk_class_of path = rev (("mk_"^(hd (rev path)))::(tl (rev path)))           (* mkC *)	

fun get_class_of path = rev (("get_"^(hd (rev path)))::(tl (rev path)))           (* mkC *)	

fun is_class_of path =  rev (("is_"^(hd (rev path)))::(tl (rev path)))           (* mkC *)	
fun is_classType_of path =  rev (("isType_"^(hd (rev path)))::(tl (rev path)))           (* mkC *)	
fun is_univ_class_of path = rev (("is_"^(hd (rev path))^"_univ")::(tl (rev path)))           (* mkC *)	

fun class2get_parent path superclass = rev (((hd (rev path))^"_2_"^superclass)::(tl (rev path)))

fun parent2class_of path superclass = rev ((superclass^"_2_"^(hd (rev path)))::(tl (rev path)))

fun mk_def_of path = rev (((hd (rev path))^"_def")::(tl (rev path))) 	
fun type_charset_name_of path = rev (((hd (rev path))^"_type")::(tl (rev path)))

fun attr_of path attr=  path @[attr] 	

fun attr_pre_of path attr = path@[attr^"_pre"] 	
fun invariant_1_of  path = path@["invariant1"]
fun invariant_2_of  path = path@["inv"]
fun strong_sve_2_of  path = path@["strong_sve"]
fun strict_sve_2_of  path = path@["strict_sve"]

fun rm_whitespace s = space_implode "_" (String.tokens (fn c => c = (#" ")
								orelse c = (#":")) s)

fun op_tab_of   path = path@["op_tab"]

fun pre_1_of  path = path@["pre1"]
fun post_1_of  path = path@["post1"]

(************************************************)

val mk_prodT = HOLogic.mk_prodT
val typeS    = HOLogic.typeS
	       
	       
fun mk_setT T = Type ("set", [T])
fun mk_funT (T1, T2) =  Type("fun", [T1,T2])  
			
fun mk_sumT (T1, T2) = Type ("+", [T1, T2])
		       
fun dest_sumT (Type ("+", [T1, T2]))=(T1, T2)
  | dest_sumT T = raise TYPE ("dest_sumT", [T], [])
			
fun sum_const T1 T2 = Const ("Sum", [T1, T2] ---> mk_sumT (T1, T2))
		      
fun mk_sum (t1, t2) =
    let val T1 = fastype_of t1 and T2 = fastype_of t2 in
	sum_const T1 T2 $ t1 $ t2
    end
    
fun dest_sum (Const ("Sum", _) $ t1 $ t2) = (t1, t2)
  | dest_sum t = raise TERM ("dest_sum", [t])
		       
fun mk_sum_fst p =
    let val pT = fastype_of p in
	Const ("sum_fst", pT --> fst (dest_sumT pT)) $ p
    end
    
fun mk_sum_snd p =
    let val pT = fastype_of p in
	Const ("sum_snd", pT --> snd (dest_sumT pT)) $ p
    end


fun LiftT       t = Type("Lifting.up",[t])
fun dest_LiftT  (Type("Lifting.up",[t])) = t
fun is_LiftT (Type("Lifting.up",[t])) = true
  | is_LiftT _                        = false

fun mapT       t   = Type("fun",[Type("OCL_OclAny_type.oid",[]),
				 Type("Datatype.option",[t])])
val mk_prodT = HOLogic.mk_prodT

fun VAL        (t,t') = Type("fun",[t,t'])
fun V          (t,t') = VAL((mk_prodT((mapT t),(mapT t))),t') 
(* fun funT        (t,t') = Type("fun",[t,t']) *)
fun St         t      = (mk_prodT((mapT t),(mapT t)))

fun extT  (a,b,T) = mk_sumT((mk_prodT(T,LiftT(a))),b) 
fun cl_extT  (a,T) = (mk_prodT(T,LiftT(a))) 
   
val unitT      = (Type("Product_Type.unit",[]))
val IntegerT      = LiftT(Type("IntDef.int",[]))
val BooleanT      = LiftT(Type("bool",[]))
val RealT         = LiftT(Type("RealDef.real",[]))
val StringT       = LiftT(Type("List.list",[Type("List.char",[])]))
val RefT          = Type("OCL_OclAny_type.oid",[])

fun BOOLT      t =   VAL(t,BooleanT)
fun INTEGERT   t =   VAL(t,IntegerT)
fun REALT   t =   VAL(t,RealT)
fun STRINGT   t =   VAL(t,StringT)

fun SetT        t = Type("OCL_Set_type.Set_0",[t])
fun OrderedSetT t = Type("OCL_OrderedSet_type.OrderedSet_0",[t])
fun SequenceT   t = Type("OCL_Sequence_type.Sequence_0",[t])
fun BagT        t = Type("OCL_Bag_type.Bag_0",[t])
	
fun SETT (et,t)        = VAL(t,SetT et)
fun ORDEREDSETT (et,t) = VAL(t,OrderedSetT et)
fun SEQUENCET (et,t)   = VAL(t,SequenceT et)
fun BAGT (et,t)        = VAL(t,BagT et)
		
val OclAny_keyT  = Type("OCL_OclAny_type.OclAny_key",[])
		   
fun OclAnyT t  = case config.universe of
		     config.referential    => 
		     LiftT(mk_prodT
			       (mk_prodT(OclAny_keyT,RefT),
					 LiftT(t)))
		   | config.nonreferential => LiftT(mk_prodT(OclAny_keyT,
							     LiftT(t)))
			
val valuesT    =    (mk_sumT(RealT, 
			   mk_sumT(IntegerT, 
				   mk_sumT(BooleanT, StringT))))   
fun UT  t      =    mk_sumT(OclAnyT(t),valuesT)


(****************************************************)

(* 
 * Return universe-type representation of an ocl-type 
 *)    
fun  get_type_universe (A as (Type("*",[Type(T,[]),U]))) C   = 
     if (T=(string_of_path C)) then SOME (LiftT(A)) else get_type_universe U C 
   | get_type_universe (A as (Type("*",[Type("*",[Type(T,[]),_]),U]))) C = 
     if (T=(string_of_path C)) then SOME(LiftT(A)) else get_type_universe U C 
   | get_type_universe (Type("Lifting.up",[T])) C = get_type_universe T C 
   | get_type_universe (Type("+",[T1,T2])) C = 
     (case get_type_universe T1 C of
	  NONE => get_type_universe T2 C 
        | A   => A)
   | get_type_universe _ C = NONE
			     
fun collect_holes (Type("+",[T1,T2])) 
  = (collect_holes T1)@(collect_holes T2)
  | collect_holes (Type("*",[T1,T2])) 
    = (collect_holes T1)@(collect_holes T2)
  | collect_holes (Type("Lifting.up",[T]))  = collect_holes T 
  | collect_holes (TFree(hole,_)) = [hole]
  | collect_holes _  = [] 

		  
fun holes_of_typestr u cname =
    let 
	val pkey = ( (hc_path univ (base_type_name_of (cname))))

	val ext = case (get_type_universe u pkey) of 
		      SOME(ut) => ut
		    | NONE     =>  error("no universe for "^(string_of_path cname)
					 ^" (  => "^(string_of_path pkey)^")")	
    in
	collect_holes ext
    end

fun holes_of_type u C = holes_of_typestr u (name_of C)
  
(*
 * local tactics
 * TODO: move in its own ML file and provide Isar interface
 *)
    	      

fun unfold_level_n_tac lev thy =
    let 
	fun collect_level1 thy (t1 $ t2)     =  (collect_level1 thy t1)@(collect_level1 thy t2)  
	  | collect_level1 thy (Const(s,_))  =  if (is_hc_pathstr lev s)  
						then  [(get_def thy s)]
						else []
	  | collect_level1 thy (Abs (_,_,t)) =  collect_level1 thy t
	  | collect_level1 _ _               = []

	fun defs_level1 thy thm = let
	    val terms = (prems_of thm)@[(concl_of thm)]
	    fun mk_unique      [] = []
	      | mk_unique (x::xs) = if (listin x xs) then mk_unique xs else x::(mk_unique xs) 
	in 
	    mk_unique (flat (map (fn t => collect_level1 thy t) terms))
	end

	fun f thy topthm = (rewrite_goals_tac (defs_level1 thy (topthm)) topthm )
    in 
	f thy
    end

(* fun unfold_level_0_tac thy = unfold_level_n_tac l0 thy *) 
fun unfold_level_1_tac thy = unfold_level_n_tac l1 thy 
(* fun unfold_level_2_tac thy = unfold_level_n_tac l2 thy  *)
(* fun unfold_level_univ_tac thy = unfold_level_n_tac univ thy *) 


fun isValueType Integer         = true 
  | isValueType String          = true
  | isValueType Real            = true
  | isValueType Boolean         = true
  | isValueType (Set(t))        = false
  | isValueType (OrderedSet(t)) = false
  | isValueType (Bag(t))        = false
  | isValueType (Sequence(t))   = false
  | isValueType OclAny          = false
  | isValueType (Classifier s)  = false
  | isValueType DummyT          = false
  | isValueType OclVoid         = false
  | isValueType t               = error ("Error in isValueType(_,"^(string_of_OclType t)^")")		   


			       
fun tilde gu Integer         = IntegerT
  | tilde gu String          = StringT
  | tilde gu Real            = RealT
  | tilde gu OclAny          = OclAnyT gu
  | tilde gu Boolean         = BooleanT
  | tilde gu (Set(t))        = SetT (tilde gu t)
  | tilde gu (OrderedSet(t)) = OrderedSetT (tilde gu t)
  | tilde gu (Bag(t))        = BagT (tilde gu t)
  | tilde gu (Sequence(t))   = SequenceT (tilde gu t) 
  | tilde gu (Classifier s)  =   
  let		
      val holes = map (fn x => TFree(x,typeS)) (holes_of_typestr gu (s))
  in
      Type(string_of_path (hc_path univ s),holes)   
  end 
  | tilde _ DummyT              = dummyT
  | tilde _ OclVoid             = unitT
  | tilde _ t              = error ("Error in tilde(_,"^(string_of_OclType t)^")")		   




fun listin _ []  = false
  | listin e (x::xs) = if e=x then true else listin e xs

fun infer_types thy t = 
    let 
	fun copy_types (Const (srcN, srcT)) (Const (destN,_)) =  (Const (destN,srcT))
	  | copy_types (Free (srcN, srcT)) (Free _)           =  (Free (srcN,srcT))
	  | copy_types (Var (srcI, srcT))  (Var _)            =  (Var (srcI,srcT)) 
	  | copy_types (Bound (srcI))  (Bound _)              =  (Bound (srcI))
	  | copy_types (Abs  (srcN, srcT, srcTerm)) (Abs(destN, destT, destTerm))  = 
	    let
		val t = copy_types srcTerm destTerm
	    in
		(Abs  (srcN, srcT, t))
	    end
	  | copy_types (srct1 $ srct2) (destt1 $ destt2)      = ((copy_types srct1 destt1)
								     $(copy_types srct2 destt2))
	  | copy_types _ _  = error "Error in copy types after type inference" 
				    
	val typed_term = #1(Sign.infer_types (Sign.pp  thy)  thy
					     (K NONE) (K NONE) [] true ([ t ],propT))
    in
	typed_term (* copy_types typed_term t *)
    end


end
