(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * mdr_encoder_level2.sml - generic mdr import interface for HOL-OCL
 * Copyright (C) 2001-2005  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

signature REP_ENCODER_LEVEL2 =
sig
    val  add_level2    : typ -> Rep_Core.Classifier list -> theory -> theory
end


structure rep_encoder_level2 : REP_ENCODER_LEVEL2 =
struct
open Rep_Core
open holocl_term
open rep_encoder_common

open Rep_OclType
open Rep_OclTerm


(*hack for using new basic invariant (gfp) *)
fun invariant_1_of path  = path@["model","inv"]
(* end-hack *)


fun pre_of  thy cl u Op = map (fn (n,c) => (n, Ocl2holocl l1 thy  cl u c)) (precondition_of_op Op) 

fun post_of thy cl u Op = map (fn (n,c) => (n, Ocl2holocl l1 thy cl u c)) (postcondition_of_op Op) 

fun mkSndLevel u thy lhs inv prevLevel undef = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				 Const("==", dummyT) 
				      $ (lhs)
				      $ Term.absfree("x",dummyT, 
						     OclIf u (inv $ Free("x",dummyT)) 
							   (prevLevel $ Free("x",dummyT) )
							undef
				 )],propT))



fun add_attributes_2 C u  cl thy =
    let
	val thy_name = Context.theory_name thy
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 

	val inv = hc_path l2 (invariant_1_of  (name_of C))
		
	fun add_aconsts  thy_path thy (a: attribute) = 
	    let
		val attr_name = #name a
		val attr = #attr_type a

		val attr_type = mk_funT(V(u,class_u),V(u, (tilde u attr)))
		val attrn      = (hc_path l2 (attr_of name attr_name))
		val attr_pre  = (hc_path l2 (attr_pre_of name attr_name))
	    in
		add_consts_i (SOME thy_path) [(local_name_of attrn,    attr_type,NoSyn),
					      (local_name_of attr_pre,attr_type,NoSyn)] thy
	    end
	    			

	fun add_defs thy_path thy (a:attribute)   =
	    let
		val attr_n = #name a
		val attr = #attr_type a

                val attr_name    = (hc_path l2 (attr_of name attr_n))
		val attr_pre_name     = (hc_path l2 (attr_pre_of name attr_n))
		val attr_pre_name1     = (hc_path l1 (attr_pre_of name attr_n))
		val attr_name1   = (hc_path l1 (attr_of name attr_n))
 
		val attr_def_name = local_name_of (mk_def_of attr_name)
		val attr_pre_def_name = local_name_of (mk_def_of attr_pre_name)
 
		val attr_def_i = mkSndLevel u thy (Const(string_of_path attr_name, dummyT)) 
					    (Const(string_of_path inv,dummyT))
					    (Const(string_of_path attr_name1,dummyT))
					    (Const("OCL_Undefined.OclUndefined",dummyT))

		val attr_pre_def_i =  mkSndLevel u thy (Const(string_of_path attr_pre_name, dummyT)) 
					    (Const(string_of_path inv,dummyT))
					    (Const(string_of_path attr_pre_name1,dummyT))
					    (Const("OCL_Undefined.OclUndefined",dummyT))


		val thy = add_definitions_i false (SOME thy_path) [(attr_def_name,attr_def_i) ,
								   (attr_pre_def_name,attr_pre_def_i)] 
					    thy 	

		val thy = prove_basic_properties (SOME thy_path)
				   [local_name_of attr_name  
			         ,local_name_of attr_pre_name ]  thy


	    in
		thy
	    end 
	    
	val thy = foldl (fn (a,thy) => add_aconsts ((hc_path l2 (name_of C))) thy a) 
			thy (attributes_of C)
	val thy = foldl (fn (a,thy) => add_defs ((hc_path l2 (name_of C))) thy a) 
			thy (attributes_of C)  

    in 
	thy
    end


fun add_operation_consts C u  cl thy =
    let
	val thy_name = Context.theory_name thy
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
	val operations = operations_of C

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 


	fun const_undefined thy const = case Sign.const_type (sign_of thy) const of 
					    NONE     => true
					  | (SOME t) => false
							
	fun add_op thy_path thy Op = 
	    let
		val args = arguments_of_op Op 
		val arg_types = (V(u,class_u))::(map (fn t => V(u, tilde u (t))) (map snd args))
		val pre_type = foldr mk_funT  (BOOLT(St(u))) arg_types


		val post_type = foldr mk_funT (BOOLT(St(u))) (arg_types@[V(u,tilde u (result_of_op Op))]) 
		val post_type = foldr mk_funT (BOOLT(St(u))) (arg_types@[V(u,tilde u (result_of_op Op))])

		val pres  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("pre_"^Int.toString(n), snd c )
					   )) (0,(pre_of thy cl u Op))) 
			    
		val posts  = snd (foldl_map (fn (n,c) => (n+1,
							  case (fst c) of 
							      (SOME(s)) => (rm_whitespace s, snd c)
							    | NONE => ("post_"^Int.toString(n), snd c )
					    )) (0,(post_of thy cl u Op))) 

		(* add constants for pres *)
		val thy = foldl (fn (constraint,thy) 
			    =>  add_consts_i (SOME (hc_path l2 (pre_1_of(
								(name_of C)@[mangled_name_of_op Op]))))
				    [((fst constraint),pre_type,NoSyn)] thy) 
				thy  pres
		(* add constants for posts *)
		val thy = foldl (fn (constraint,thy) 
			    =>  add_consts_i (SOME (hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))))
				    [((fst constraint),post_type,NoSyn)] thy) 
				thy posts
		(* add final precondition *)
		val thy = add_consts_i   (SOME (rev (tl (rev (hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))))))) 
					 [(local_name_of(hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))),
					   pre_type,NoSyn)] thy 
		(* add final postcondition *)	  
		val thy = add_consts_i   (SOME (rev (tl (rev (hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))))))) 
				       [(local_name_of(hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))),
					 post_type,NoSyn)] thy 



		(* add operation tab ... *)
		val m_path = hc_path operation [mangled_name_of_op Op]
		val arg_types = (V(u,OclAnyT (TFree("'a",typeS))))::(map (fn t => V(u, tilde u (t))) (map snd args))
		val op_type = foldr1 mk_funT (arg_types@[V(u,tilde u (result_of_op Op)),BOOLT(St(u))])
		val thy = ( add_consts_i  (SOME(hc_path l2 ((name_of) C)))
					  [(name_of_op Op ,op_type,NoSyn)] thy )
		val op_tab_type = Type("fun",[(mk_setT (TFree("'a",typeS))),Type("Datatype.option",[op_type])])
		val thy = if const_undefined thy ((string_of_path(hc_path operation [mangled_name_of_op Op]))^"."^("tab"))
			  then 
			      ( add_consts_i (SOME(hc_path operation [mangled_name_of_op Op]))
					     [("tab" ,op_tab_type,NoSyn)] thy )
			  else thy 
			       

		val thy =  if const_undefined thy ((string_of_path (hc_path operation [mangled_name_of_op Op]))^"."^("Op"))
			   then let
				   val meth_type = foldr mk_funT (V(u,tilde u (result_of_op Op))) arg_types
				   val thy =( add_consts_i (SOME(hc_path operation [mangled_name_of_op Op]))
							   [("Op" ,meth_type,NoSyn)] thy )




				   val args_i = rev(("self",(V(u,class_u)))::(map (fn a =>  (fst a,V(u, tilde u (snd a)))) (args)))
			    
				   val Op_t =  Const("OCL_Operation.OCL_choose",dummyT)
					       $ (
					       foldl (fn (t0,t1) => t1$t0) 
							     (Const("OCL_Operation.OCL_invokeS",dummyT))
							      (Const("arbitrary",dummyT)
							      ::(Const((string_of_path(hc_path operation [mangled_name_of_op Op]))^".tab",op_tab_type))
							      ::(map (fn (a,b) => Free(a,dummyT (* b *))) ( rev   args_i))))
					       
				   val Op_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
						     (Const  ("==", dummyT) (* meth_type --> meth_type --> propT)  *)
							     $ Const ((string_of_path (hc_path operation [mangled_name_of_op Op]))
										       ^"."^("Op"), meth_type)
							     $ (foldl (fn (ar,t) => Term.absfree(fst ar,dummyT (* snd ar *),t)) 
								      (Term.absfree (fst (hd args_i) ,dummyT (*snd (hd args_i) *), Op_t)) (tl args_i))
						   )
						   ],propT))
				   val thy = add_definitions_i false   (SOME(hc_path operation [mangled_name_of_op Op]))
							       [("Op",Op_def_i)] thy 

                                (*
				   val argstr = foldr1 (fn (a,b) => a^" "^b) ("self"::(map fst args))
				   val Op_def = "Op "
						^" == (% "^argstr^" . (OCL_choose(OCL_invokeS arbitrary tab "
						^argstr^")))"
				   val thy = add_definitions   (SOME(hc_path operation [mangled_name_of_op Op]))
							       [("Op",Op_def)] thy 
					     *)
			       in 
				   thy			  
			       end
			   else thy
	    in 
		thy
	    end


	val thy = foldl (fn (a,thy) => add_op ((hc_path l2 (name_of C))) thy a) 
			thy (operations_of C)


    in 
	thy
    end



fun add_operation_defs C u  cl thy =
    let
	val thy_name = Context.theory_name thy
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type   u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
	val operations = operations_of C

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 



							
	fun add_op thy_path thy Op = 
	    let
		val args = arguments_of_op Op 
		val arg_types = (V(u,class_u))::(map (fn t => V(u, tilde u (t))) (map snd args))
		val pre_type = foldr mk_funT (BOOLT(St(u))) arg_types



		val post_type = foldr mk_funT (BOOLT (St(u))) (arg_types@[V(u,tilde u (result_of_op Op))])

		val pres  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("pre_"^Int.toString(n), snd c )
					   )) (0,(pre_of thy cl u Op))) 
			    
		val posts  = snd (foldl_map (fn (n,c) => (n+1,
							  case (fst c) of 
							      (SOME(s)) => (rm_whitespace s, snd c)
							    | NONE => ("post_"^Int.toString(n), snd c )
					    )) (0,(post_of thy cl u Op))) 


		(* add definitionss for pres *)
	
		val cargs = rev(("self",(V(u,class_u)))::(map (fn a =>  (fst a,V(u, tilde u (snd a)))) (args)))

		val pre_defs = let 
		    fun mk_abs (ar,t) = Term.absfree(fst ar,snd ar,t)
		    fun mk_term (n,t) = (n,#1(Sign.infer_types (Sign.pp thy) thy 
							       (K NONE) (K NONE) [] true ([
					      (Const  ("==", pre_type --> pre_type --> propT) $
						      Const ((string_of_path ((hc_path l2 (pre_1_of ((name_of C)@[mangled_name_of_op Op])))))
							     ^"."^n,pre_type) $ 
						      (foldl mk_abs  (Term.absfree (fst (hd cargs), snd (hd cargs), t)) (tl cargs))) 
					      ],propT))) 
		    in
			map  mk_term pres
		    end
		    
		val thy = foldl (fn (constraint,thy) 
				    =>  add_definitions_i true (SOME ((hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op])))))
							  [((fst constraint)^"_def",(snd constraint))] thy) 
				thy pre_defs
		(* add definitionss for posts *)

		val cargs = rev((("self",(V(u,class_u)))
				 ::(map (fn a =>  (fst a,V(u, tilde u (snd a)))) 
					(args)))@[("result",V(u,tilde u (result_of_op Op)))] )
	
		val post_defs = map (fn (n,t) => 
					(n,#1(Sign.infer_types (Sign.pp thy)  thy
							       (K NONE) (K NONE) [] true ([
					      (Const  ("==", post_type --> post_type --> propT) $
						      Const ((string_of_path ((hc_path l2 (post_1_of ((name_of C)@[mangled_name_of_op Op])))))^"."^n,post_type) $ 
						      (foldl (fn (ar,t) => Term.absfree(fst ar,snd ar,t)) 
							     (Term.absfree (fst (hd cargs) ,snd (hd cargs), t)) (tl cargs)))
						   
						   ],propT)))) posts
		val thy = foldl (fn (constraint,thy) 
				    =>  add_definitions_i true (SOME ((hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op])))))
						  [((fst constraint)^"_def",(snd constraint))] thy) 
				thy  post_defs 



		val cargs = rev(("self",(V(u,class_u)))
				::(map (fn a =>  (fst a,V(u, tilde u (snd a)))) (args)))
		val cargs_i =  (map (fn (s,t) => Free(s,t)) (rev cargs))
			     
		val cargsres = rev(("self",(V(u,class_u)))
				::(map (fn a =>  (fst a,V(u, tilde u (snd a)))) (args))@[("result",dummyT)] )

		val cargsres_i = (map (fn (s,t) => Free(s,t)) (rev cargsres))

		(* add final precondition *)	  
		val c_str_i = let 
		    fun add_args s =  foldl (fn (t0,t1) => t1$t0) s  cargs_i

		in 
		    map (fn s => add_args (Const((string_of_path  (hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))))^"."^(fst s),dummyT)))
					  pres
		end
		val pre_def_i =  (#1(Sign.infer_types (Sign.pp thy) thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", pre_type --> pre_type --> propT) 
				      $ (Const(string_of_path (hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))),dummyT)) 
				      $ 
				  (foldl (fn (ar,t) => Term.absfree(fst ar,snd ar,t))
						    (Term.absfree (fst (hd cargs) ,snd (hd cargs),(foldr1 (fn (a,b) => OclAnd u a b) c_str_i)))
						     (tl cargs)))

			       ],propT)))
		val thy = add_definitions_i  false (SOME (rev (tl (rev (hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))))))) 
					  [(local_name_of(hc_path l2 (pre_1_of  ((name_of C)@[mangled_name_of_op Op]))),pre_def_i)] thy 



		(* add final postcondition *)	  
		val c_str_i =  let 
		    fun add_args s =  foldl (fn (t0,t1) => t1$t0) s  cargsres_i

		in 
		    map (fn s => add_args (Const((string_of_path  (hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))))^"."^(fst s),dummyT)))
					  posts
		end

		val post_def_i =  (#1(Sign.infer_types (Sign.pp thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", post_type --> post_type --> propT) 
				      $ (Const(string_of_path (hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))),dummyT)) 
				$  (foldl (fn (ar,t) => Term.absfree(fst ar,snd ar,t))
						    (Term.absfree (fst (hd cargsres) ,snd (hd cargsres),(foldr1 (fn (a,b) => OclAnd u a b) c_str_i)))
						     (tl cargsres)))



			       ],propT)))
		val thy = add_definitions_i  false (SOME (rev (tl (rev (hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))))))) 
					  [(local_name_of(hc_path l2 (post_1_of  ((name_of C)@[mangled_name_of_op Op]))),post_def_i)] thy 



	
		val mangOpName = ((name_of C)@[mangled_name_of_op Op])
		val pre_1 = foldl (fn (t0,t1) => t1$t0) (Const(string_of_path(hc_path l2 (pre_1_of  mangOpName)),dummyT))  cargs_i
		val post_1 = foldl  (fn (t0,t1) => t1$t0) (Const(string_of_path(hc_path l2 (post_1_of mangOpName)),dummyT)) cargsres_i
		val mOp_def_i =    #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				      Const("==",dummyT) $ Const(string_of_path(hc_path l2 ((name_of) C)@[(name_of_op Op)]),dummyT)
					   $ (foldl (fn (ar,t) => Term.absfree(fst ar,snd ar,t))
						    (Term.absfree (fst (hd cargsres) ,snd (hd cargsres),(OclAnd u pre_1 post_1)))
						     (tl cargsres))
			 ],propT))
		val thy = add_definitions_i   true (SOME(hc_path l2 ((name_of) C)))
					      [(name_of_op Op,mOp_def_i)] thy 
			       
		(* overloading *) 
		val overwrite_ax = (string_of_path (hc_path operation [mangled_name_of_op Op]))^".tab ("
				   ^(string_of_path (hc_path l2 ((mk_class_of o name_of) C)))
				   ^" ` UNIV) == SOME    ( " 
				   ^(string_of_path (hc_path l2 (((name_of) C)@[name_of_op Op]))) 
				   ^")"


		val overwrite_ax_i =  #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
					 Const("==",dummyT)
					  $ (Const((string_of_path (hc_path operation [mangled_name_of_op Op]))^".tab",dummyT)
						  $(Const("image",dummyT)
						  $Const(string_of_path (hc_path l2 ((mk_class_of o name_of) C)),dummyT)
						  $Const("UNIV",dummyT)))
					  $ (Const ("Datatype.option.Some",dummyT) 
						   $Const(string_of_path (hc_path l2 (((name_of) C)@[name_of_op Op])),dummyT) 
					    )
					 ],propT))

		val thy = add_axioms_i  (SOME( ( ((hc_path operation [mangled_name_of_op Op])))))
					    [("overwrite_"^((local_name_of (name_of C))),overwrite_ax_i)] thy 

	    in 
		thy
	    end


	val thy = foldl (fn (a,thy) => add_op ((hc_path l2 (name_of C))) thy a) 
			thy (operations_of C)
    in 
	thy
    end


fun add_constructors_2 C u  cl thy =
    let
	val thy_name = Context.theory_name thy
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val ctype_name = type_charset_name_of name
	val parent = parent_name_of C

	val mk = hc_path l2 ((mk_class_of o name_of) C)
	val mk1 = hc_path l1 ((mk_class_of o name_of) C)

        
	val mk_parent = hc_path l2 ((mk_class_of o parent_name_of) C)
	val get = hc_path l2 ((get_class_of o name_of) C)
	val get1 = hc_path l1 ((get_class_of o name_of) C)
	val parent_of =  hc_path l2 ( (get_class_of o parent_name_of)  C)
	val is =  hc_path l2 ((is_class_of o name_of) C)
        val is1 =  hc_path l1 ((is_class_of o name_of) C)
	val isType =  hc_path l2 ((is_classType_of o name_of) C)
        val isType1 =  hc_path l1 ((is_classType_of o name_of) C)
        val is_parent = hc_path l2 ((is_class_of o parent_name_of) C)
	val is_univ =  hc_path l2 ((is_univ_class_of o name_of) C)
	val is_univ1 =  hc_path l1 ((is_univ_class_of o name_of) C)
	val is_univ_parent =  hc_path l2 ((is_univ_class_of o parent_name_of) C)
	val class = name

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 

	val holes_par = map (fn x => TFree(x,typeS)) 
			    (holes_of_typestr  u (parent_name_of C))
        val class_u_par = if (parent_name_of C) = (name_of OclAnyC) then 
			      the(get_type_universe u (key_name_of (name_of OclAnyC))) (* TODO *)
			  else Type((string_of_path (hc_path univ (parent_name_of C))),holes_par)


	val class2parent = hc_path l2 (class2get_parent  name (local_name_of parent))
	val parent2class = hc_path l2 (parent2class_of  name (local_name_of parent))
	val class2parent1 = hc_path l1 (class2get_parent  name (local_name_of parent)) 
	val parent2class1 = hc_path l1 (parent2class_of  name (local_name_of parent))
			   
	


	
	fun add_consts_2 pkg_name thy = 
	    let
		val mk_type      = mk_funT(V(u,class_u),V(u,u))
		val get_type     = mk_funT(V(u,u),V(u,class_u)) 
		val is_type      = mk_funT(V(u,class_u),BOOLT(St(u)))
		val is_univ_type = mk_funT(V(u,u),BOOLT(St(u)))
		val parent2class_type   = mk_funT(V(u,class_u_par),V(u,class_u))
		val class2parent_type   = mk_funT(V(u,class_u),V(u,class_u_par))

		val characterstic_type   = (V(u,SetT(class_u)))
	    in
		add_consts_i (SOME (hc_path l2 (package_of C)) )  
			   [(local_name_of mk,mk_type,NoSyn), 
			    (local_name_of get,get_type,NoSyn),
			    (local_name_of is,is_type,NoSyn), 
			    (local_name_of isType,is_type,NoSyn), 
			    (local_name_of is_univ,is_univ_type,NoSyn),
			    (local_name_of class2parent,class2parent_type, NoSyn),
			    (local_name_of parent2class,parent2class_type, NoSyn),
			    (local_name_of name, characterstic_type, NoSyn),
			    (local_name_of ctype_name, characterstic_type, NoSyn)
			    ] thy
	    end
	    
	fun add_defs pkg_name C thy = 
	    let

		val mk_def_name    = (local_name_of o mk_def_of) mk
	        val get_def_name   = (local_name_of o mk_def_of) get
	        val is_def_name    = local_name_of (mk_def_of is)
	        val isType_def_name    = local_name_of (mk_def_of isType)
	        val is_univ_def_name = local_name_of (mk_def_of is_univ)
	        val class_def_name = local_name_of (mk_def_of name)
       		val parent_name =  (short_parent_name_of C)
		val class2parent_def_name = local_name_of (mk_def_of class2parent)
		val parent2class_def_name = local_name_of (mk_def_of parent2class)





		val inv = hc_path l2 (invariant_1_of name) 


		val parentinv = hc_path l2  (invariant_1_of parent)
   
		val get_def_i = mkSndLevel u thy (Const(string_of_path get,dummyT)) 
				(Const("Fun.comp",dummyT)$Const(string_of_path inv,dummyT)$Const(string_of_path get1,dummyT))
				(Const(string_of_path get1,dummyT))
				(Const("OCL_Undefined.OclUndefined",dummyT))

		val mk_def_i   =  mkSndLevel u thy (Const(string_of_path mk,dummyT)) 
					     (Const(string_of_path inv,dummyT))
					     (Const(string_of_path mk1,dummyT))
					     (Const(string_of_path mk1,dummyT)$(Const("OCL_Undefined.OclUndefined",dummyT)))

	
		val is_univ_def_i = mkSndLevel u thy (Const(string_of_path is_univ,dummyT)) 
				(Const("Fun.comp",dummyT)$Const(string_of_path inv,dummyT)$Const(string_of_path get1,dummyT))
				(Const(string_of_path is_univ1,dummyT))
				(Const("OCL_Boolean.OclFalse", (BOOLT (St u))))
		  

		val is_def_i = mkSndLevel u thy (Const(string_of_path is,dummyT)) 
				(Const(string_of_path inv,dummyT))
				(Const(string_of_path is1,dummyT))
				(Const("OCL_Boolean.OclFalse", (BOOLT (St u))))

		val isType_def_i = mkSndLevel u thy (Const(string_of_path isType,dummyT)) 
				(Const(string_of_path inv,dummyT))
				(Const(string_of_path isType1,dummyT))
				(Const("OCL_Boolean.OclFalse", (BOOLT (St u))))


		val class2parent_def_i = mkSndLevel u thy (Const(string_of_path class2parent,dummyT)) 
				(Const(string_of_path inv,dummyT))
				(Const(string_of_path class2parent1,dummyT))
				(Const("OCL_Undefined.OclUndefined",dummyT))


		val parent2class_def_i =  if (parent = (name_of OclAnyC)) 
					  then  
					      #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
						 Const("==", dummyT) 
						      $ ((Const(string_of_path parent2class,dummyT)))
						      $ Term.absfree("x",dummyT, 
						             OclIf u (Const("OCL_Boolean.OclTrue",(BOOLT (St u)))) 
								   (Const(string_of_path parent2class1,dummyT)$ Free("x",dummyT) )
								   (Const("OCL_Undefined.OclUndefined",dummyT))
						 )],propT))
					  else 
					      mkSndLevel u thy (Const(string_of_path parent2class,dummyT)) 
							 (Const(string_of_path parentinv,dummyT))
							 (Const(string_of_path parent2class1,dummyT))
							 (Const("OCL_Undefined.OclUndefined",dummyT))
		val characterstic_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
							       Const("==", dummyT) 
						  $ Const(string_of_path name,dummyT)
						  $ (Term.absfree("St",dummyT, 
								  (Const("OCL_Set_type.Abs_Set_0",dummyT) 
									$ (Const ("Lifting.up.lift",dummyT) 
										 $  (Const ("op -",dummyT)  
											   $ (Const ("image",dummyT) 
												    $ ( Term.absfree("x",dummyT, 
														     Free("x",dummyT)
															 $Free("St",dummyT))) 
												    $ (Const ("image",dummyT)
													     $Const(string_of_path get,
														    dummyT)
													     $Const ("UNIV",dummyT) )) 
											   $Const ("{}",dummyT)
					     )))))],propT))

		val cset_type_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
						 Const("==", dummyT) 
						      $ (Const(string_of_path ctype_name,dummyT))
						      $ (Const("OCL_Collection.OclSelect",dummyT)$Const(string_of_path name,dummyT)
						      $ Term.absfree("obj",dummyT, Const(string_of_path isType,dummyT)$Free("obj",dummyT)))
						 ],propT))
		val thy = add_definitions_i false (SOME (hc_path l2 (package_of C)) ) 
					    [  (get_def_name,get_def_i),
					       (mk_def_name,mk_def_i),
					       (is_def_name,is_def_i),
					       (isType_def_name,isType_def_i),
					       (is_univ_def_name,is_univ_def_i),
					       (class2parent_def_name,class2parent_def_i)  ,   
					       (parent2class_def_name,parent2class_def_i) ,
					       (local_name_of (mk_def_of name),characterstic_def_i), 
					       (local_name_of (mk_def_of ctype_name),cset_type_def_i)  
					    ] thy
		val thy = prove_basic_properties (SOME (hc_path l2 (package_of C)))
				   [ (* (local_name_of get),
				      (local_name_of mk), *)
				      (local_name_of is) (*,
				      (local_name_of isType), 
				      (local_name_of is_univ),
				      (local_name_of class2parent)  ,   
				      (local_name_of parent2class) ,
				      (local_name_of name), 
				      (local_name_of ctype_name) *) 
				   ] thy	    in
		thy 
	    end 
	    
	fun oclAsType_defs pkg_name C cl thy = let 
	    val hierarchylist = [name_of C]@(parents_of C cl)
	    val notrelatedlist = map (fn c => name_of c) 
				     (filter (fn c => (not ((listin (name_of c) hierarchylist)
					     orelse (listin (name_of C) (parents_of c cl) )))) cl)
	    fun c_type c cl = if c = (name_of OclAnyC) 
			      then  V(u, the(get_type_universe u (key_name_of (name_of OclAnyC)))) 
			      else  let
				      val C = class_of c cl
				      val  holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
				      val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes)  
				  in 
				      V(u,class_u)
				  end
				    
	    fun ch_type c cl =  if c = (name_of OclAnyC) 
				then   V(u,SetT(the(get_type_universe u (key_name_of (name_of OclAnyC)))))  
				else let  
					val C = if c = (name_of OclAnyC) then OclAnyC else class_of c cl
					val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
					val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
				    in
					V(u,SetT(class_u))
				    end
				     

	    fun upcast []   (c::p::pl) = upcast ([[Const(string_of_path (hc_path l1 (class2get_parent  c (local_name_of p))),
							 mk_funT(c_type  c cl , c_type p cl)  )]]) (p::pl)
	      | upcast defs (c::p::pl) = upcast ([(Const(string_of_path (hc_path l1 (class2get_parent  c (local_name_of p))),
							 mk_funT(c_type  c cl ,c_type p cl)))
						  ::(hd defs)]@defs) (p::pl) 
	      | upcast defs [c]        = defs
	      | upcast defs []         = defs
					 
	    fun downcast []   (c::p::pl) = downcast ([[Const( string_of_path (hc_path l1 (parent2class_of  c (local_name_of p))),
							      mk_funT(c_type  p cl , c_type c cl)  )]]) (p::pl)
	      | downcast defs (c::p::pl) = downcast ([(Const(string_of_path (hc_path l1 (parent2class_of  c (local_name_of p))), 
							     mk_funT(c_type  p cl , c_type c cl)  ))::(hd defs)]@defs) (p::pl) 
	      | downcast defs [c]        = defs
	      | downcast defs []         = defs

	    val characterstic_type   = (V(u,SetT(class_u)))
	    val class2parent_type   = mk_funT(V(u,class_u),V(u,class_u_par))

	    fun comp [e] = e 
	      | comp (f::fs) =  let 
		    val fts = (comp fs) 
		in 
		    (Const ("Fun.comp",dummyT)$fts$f)
		end

	    fun pairing [] [] = []
	      | pairing (a::AS) (b::BS)  = (a,b)::(pairing AS BS) 

	    val _ = trace 9  "generating def terms for downcasts along hierarchy"
	    (* definitions for downcasts along hierarchy *)
	    val down_defs =  case hierarchylist of 
				 [] => []
			       | hierarchy => let
 				    val fdownlist = map (fn e => comp e) (downcast [] (hierarchy))
				    val chlist    =  (tl  hierarchy)
				in
				    map (fn (c,f) => ("from_"^(local_name_of (hc_path l1 (c))),
						      #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
							 (Const  ("==",dummyT) (* mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u))) 
										      --> mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u))) -->propT)   *)
								 $ Const("OclAsType",dummyT) (* mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u)))) *)
								 $ Term.absfree("self",dummyT ,Term.absfree("cSet",characterstic_type, (f $ Free("self",dummyT))))) 
							 ],propT))))
					(pairing ( chlist) fdownlist)  
				end
 			      
	    val _ = trace 9 "generating def terms for upcasts along hierarchy"		  
	    (* definitions for upcasts along hierarchy *)
	    val up_defs =  case hierarchylist of 
			       [] => []
			     | hierarchy => let
				   val fuplist   = map (fn e => comp (rev e))  (upcast [] (hierarchy)) 
				   val chlist    = ( (rev (tl hierarchy)))
			       in
				   map (fn (c,f) => 
					   ("to_"^(local_name_of ((hc_path l1 c))),
					    #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
					       (Const  ("==",dummyT) (*   mk_funT(V(u,class_u),mk_funT(ch_type c cl,c_type c cl)) --> 
									      mk_funT(V(u,class_u),mk_funT(ch_type c cl,c_type c cl))  -->propT)  *)
						       $ Const("OclAsType",dummyT) (* mk_funT(V(u,class_u),mk_funT(ch_type c cl,c_type c cl)))  *)
						       (*			    	    $ Term.absfree("self",dummyT,Term.absfree("cSet",ch_type c cl, (f $ Free("self",dummyT)))))   *)
						       $ Term.absfree("self",V(u,class_u),Term.absfree("cSet",ch_type c cl, (f $ Free("self",V(u,class_u))))))   
					       ],propT)))) 
				       (pairing ( chlist) fuplist) 
			       end
					    
	    val _ = trace 9 "generating def terms for cast to self"		  
	    (* definitions for self *)
	    val self_def = ("to_"^(short_name_of C),
			    #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
			       (Const  ("==",mk_funT(V(u,class_u),mk_funT(characterstic_type,V(u,class_u))) 
						 --> mk_funT(V(u,class_u),mk_funT(characterstic_type,V(u,class_u))) -->propT)  
				       $ Const("OclAsType",mk_funT(V(u,class_u),mk_funT(characterstic_type,V(u,class_u)))) 
								 $ Term.absfree("self",V(u,class_u),Term.absfree("cSet",characterstic_type, Free("self",V(u,class_u)))))
			       ],propT)))
	    val _ = trace 9 "generating def terms for casts orthotogonal to hierarchy"		  
	    (* definitions for downcasts orthortogonal to hierarchy *)
	    val notrel_down_defs = if !strict_checking 
	    	                   then case notrelatedlist of 
					    [] => []
					  |  notrelated => 
					     map (fn c => ("from_"^(local_name_of (hc_path l1 c)),
							   #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
							      (Const  ("==",mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u))) 
										--> mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u))) -->propT)  
								      $ Const("OclAsType",mk_funT(c_type c cl,mk_funT(characterstic_type,V(u,class_u)))) 
								      $ Term.absfree("self",c_type c cl,
										     Term.absfree("cSet",
												  characterstic_type, 
												  Const ("OCL_Undefined.OclUndefined",dummyT))))
							      ],propT))))
						 notrelated
				   else []
	    val _ = trace 9 "generating def terms for casts to/from OclLib"		  	   
	    val ocllib_types =     [("Integer",IntegerT),("Real",RealT),("String",StringT), 
				    ("Set",SetT(TFree("'a",typeS))),
				    ("OrderedSet",OrderedSetT(TFree("'a",typeS))),
				    ("Sequence",SequenceT(TFree("'a",typeS))),
				    ("Bag",BagT(TFree("'a",typeS)))
			  ]

	    val ocl_from_defs = if !strict_checking 
	    	                then   map (fn (n,t) => ("from_"^n,
						       #1(Sign.infer_types (Sign.pp  thy) thy (K NONE) (K NONE) [] true ([
							  (Const  ("==",mk_funT(V(u,t),mk_funT(characterstic_type,V(u,class_u))) 
									    --> mk_funT(V(u,t),mk_funT(characterstic_type,V(u,class_u))) -->propT)  
								  $ Const("OclAsType",mk_funT(V(u,t),mk_funT(characterstic_type,V(u,class_u)))) 
								  $ Term.absfree("self",V(u,t),
										 Term.absfree("cSet",
											      characterstic_type, 
											      Const ("OCL_Undefined.OclUndefined",dummyT))))
							  ],propT))))
					 ocllib_types
				else []
	    val ocl_to_defs =  if !strict_checking 
	    	               then  map (fn (n,t) => ("to_"^n,
						       #1(Sign.infer_types (Sign.pp  thy) thy (K NONE) (K NONE) [] true ([
							  (Const  ("==",mk_funT(V(u,class_u),mk_funT(V(u,SetT(t)),V(u,t))) 
									    --> mk_funT(V(u,class_u),mk_funT(V(u,SetT(t)),V(u,t))) --> propT) 
								  $ Const("OclAsType", mk_funT(V(u,class_u),mk_funT(V(u,SetT(t)),V(u,t))))
								  $ Term.absfree("self",V(u,class_u),
										 Term.absfree("cSet",V(u,SetT(t)), 
											      Const ("OCL_Undefined.OclUndefined",dummyT))))
							  ],propT))))
					 ocllib_types
			       else []

	    (* only down_defs@up_defs@[self_def] are non-empty in non-strict checking mode: *)
	    val asType_defs   = down_defs@up_defs@[self_def]@notrel_down_defs@ocl_from_defs@ocl_to_defs		
	in	
	    if !quick_and_dirty 
	    then  add_axioms_i  (SOME (hc_path univ (name_of C)) ) asType_defs thy
	    else  add_definitions_i true (SOME (hc_path univ (name_of C)) ) asType_defs thy
	end 

	fun oclAllInstances_defs pkg_name C cl thy = let 
	    fun mangled_string_of_path (path:Rep_OclType.Path) = case path of
							     [] => ""
							   | p  => foldr1 (fn (a,b) => a^"_"^b) p

	    val characterstic_type   = (V(u,(class_u)))
	    val allInstances_defs = [   
		("OclAllInstances"^(mangled_string_of_path name),
		 #1(Sign.infer_types (Sign.pp  thy) thy (K NONE) (K NONE) [] true ([
		    (Const  ("==",dummyT)
			    $ (Const("OclAllInstances", dummyT)$ Free("self", characterstic_type )) 
			    $ (Const("split",dummyT) $ Term.absfree("s",dummyT,Term.absfree("s'",dummyT,
			     Const("OCL_Set_type.Abs_Set_0",dummyT)$(
				Const("Lifting.up.lift",dummyT)$(Const("image", dummyT)
			            $Const((string_of_path (hc_path l0 (get_class_of name))),dummyT)
			               $(Const("Map.ran",dummyT)$Free("s'",dummyT))		     
		)	      )))))
	    ],propT))),
		("OclAllInstances"^(mangled_string_of_path name)^"Atpre",
		 #1(Sign.infer_types (Sign.pp  thy) thy (K NONE) (K NONE) [] true ([
		    (Const  ("==",dummyT)
			    $ (Const("OclAllInstancesAtpre", dummyT)$ Free("self",  characterstic_type)) 
			    $ (Const("split",dummyT) $ Term.absfree("s",dummyT,Term.absfree("s'",dummyT,
			     Const("OCL_Set_type.Abs_Set_0",dummyT)$(
				Const("Lifting.up.lift",dummyT)$(Const("image", dummyT)
			            $Const((string_of_path (hc_path l0 (get_class_of name))),dummyT)
			               $(Const("Map.ran",dummyT)$Free("s",dummyT))		     
			      ))))))
	    ],propT)))
	    ]
	in
	    if !quick_and_dirty 
	    then  add_axioms_i  (SOME (hc_path univ (name_of C)) ) allInstances_defs thy
	    else  add_definitions_i true (SOME (hc_path univ (name_of C)) ) allInstances_defs thy  
	end

				       
	val thy = add_consts_2 ((string_of_path o package_of) C) thy
	val thy = add_defs ((string_of_path o package_of) C) C thy   
	val thy = oclAsType_defs ((string_of_path o package_of) C) C cl thy  
	val thy = oclAllInstances_defs ((string_of_path o package_of) C) C cl thy  
    in
	thy
    end



fun add_sve_2  C u cl  thy = 
    let
	val thy_name = Context.theory_name thy
	val name = name_of C
	val ctype_name = type_charset_name_of name
	val parent = parent_name_of C

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
	val class2parent = string_of_path (hc_path l2 (class2get_parent  name (local_name_of parent)))
	val C_type = V(u,class_u)

	val holes_par = map (fn x => TFree(x,typeS)) 
			    (holes_of_typestr  u (parent_name_of C))
        val class_u_par = if (parent_name_of C) = (name_of OclAnyC) then 
			      the(get_type_universe u (key_name_of (name_of OclAnyC))) (* TODO *)
			  else Type((string_of_path (hc_path univ (parent_name_of C))),holes_par)
	
	val parentT   = V(u,class_u_par)
	val class2parentT  = mk_funT(C_type,parentT)
	
	val val_attributes = filter (fn  (a: attribute) => isValueType (#attr_type a)) (attributes_of C)

	fun sve_attrib eq attr =  map (fn (a: attribute) => eq
						 $(Const(string_of_path (hc_path l2 (attr_of name (#name a))),dummyT)$Free("x",C_type))
						 $(Const(string_of_path (hc_path l2 (attr_of name (#name a))),dummyT)$Free("y",C_type))
				      ) attr
	fun add_strong_sve C thy =  
	    let val sve_defs = if (parent_name_of C) = (name_of OclAnyC) then 
			      sve_attrib (Const("OCL_Boolean.OclStrongValueEq",dummyT)) 
					 val_attributes
			   else 
			      [(Const("OCL_Boolean.OclStrongValueEq",dummyT)
				      $(Const(class2parent,class2parentT)$Free("x",C_type))$(Const(class2parent,class2parentT)$Free("y",C_type)))]@ 
			      (sve_attrib (Const("OCL_Boolean.OclStrongValueEq",dummyT)) 
					  val_attributes)
		val sve_defs = (case sve_defs of
			[] => [OclTrue u]	
			| l  => l)

			
	    in
		(#1(Sign.infer_types (Sign.pp thy) thy 
				     (K NONE) (K NONE) [] true ([
		    (Const  ("==", (C_type --> C_type --> (BOOLT (St(u)))) 
									--> (C_type --> C_type --> (BOOLT (St(u)))) 
									--> propT)) 
			    $ (Const("OCL_Boolean.OclStrongValueEq",(C_type --> C_type --> (BOOLT (St(u)))))) 
			    $ (Term.absfree ("x",C_type,(Term.absfree ("y",C_type,(foldr1 (fn (a,b) => OclAnd u a b) sve_defs)))))
		    ],propT)))
	    end

	fun add_strict_sve C thy = 
	    let 
		val deflist = (map (fn (a: attribute) => Const("OCL_Boolean.OclIsDefined",dummyT)
							    $(Const(string_of_path (hc_path l2 (attr_of name (#name a))),dummyT) 
								   $Free("x",C_type))) val_attributes)
			       @(map (fn (a: attribute) => Const("OCL_Boolean.OclIsDefined",dummyT)
							      $(Const(string_of_path (hc_path l2 (attr_of name (#name a))),dummyT) 
								     $Free("y",C_type))) val_attributes)
		val deflist = deflist@[Const("OCL_Boolean.OclIsDefined",dummyT)$Free("y",C_type),
				       Const("OCL_Boolean.OclIsDefined",dummyT)$Free("y",C_type)]

		val deflist =  if (parent_name_of C) = (name_of OclAnyC) then 
				   deflist
			       else 
				   [Const("OCL_Boolean.OclIsDefined",dummyT)
					 $(Const("OCL_Boolean.OclStrictValueEq",dummyT)
						$(Const(class2parent,class2parentT)$Free("x",C_type))
						$(Const(class2parent,class2parentT)$Free("y",C_type)))]@ 
				   deflist
		val deflist = ( case deflist of
			[] => [OclTrue u]	
			| l  => l )

		val def_def =  (foldr1 (fn (a,b) => OclOr u a b) deflist)


		val eq_defs = if (parent_name_of C) = (name_of OclAnyC) then 
			      sve_attrib (Const("OCL_Boolean.OclStrictValueEq",dummyT)) 
					 val_attributes
			   else 
			      [(Const("OCL_Boolean.OclStrictValueEq",dummyT)
				      $(Const(class2parent,class2parentT)$Free("x",C_type))$(Const(class2parent,class2parentT)$Free("y",C_type)))]@ 
			      (sve_attrib (Const("OCL_Boolean.OclStrictValueEq",dummyT)) 
					  val_attributes)
		
		val eq_defs = (case eq_defs of
			[] => [OclTrue u]	
			| l  => l )
	    in
		(#1(Sign.infer_types (Sign.pp thy) thy 
				     (K NONE) (K NONE) [] true ([
		    (Const  ("==", (C_type --> C_type --> (BOOLT (St(u)))) 
									--> (C_type --> C_type --> (BOOLT (St(u)))) 
									--> propT)) 
			    $ (Const("OCL_Boolean.OclStrictValueEq",(C_type --> C_type --> (BOOLT (St(u)))))) 
			    $ (Term.absfree ("x",C_type,(Term.absfree ("y",C_type,
								       OclIf u def_def 
									     (Const("OCL_Undefined.OclUndefined",dummyT))
									     (foldr1 (fn (a,b) => OclAnd u a b) eq_defs)
			      ))))



		    ],propT)))
	    end

	val strong_sve_def_i = add_strong_sve C thy  
	val strict_sve_def_i = add_strict_sve C thy 

		  
    	val thy = add_definitions_i true  (SOME ((hc_path l2 ( (name_of C))))) 
				    [((local_name_of (hc_path l2 (strong_sve_2_of (name_of C)))),strong_sve_def_i),
				     ((local_name_of (hc_path l2 (strict_sve_2_of (name_of C)))),strict_sve_def_i)
				    ] thy 
    in
	thy
    end
    
fun add_level2  universe cl thy = 
    let
	val timer =  Timer.startRealTimer ()

	fun printTiming startTime = 
	    let		
		val endTime = Timer.checkRealTimer timer
		val _ = trace 2 ("      -> time used: "
				 ^(Time.toString(Time.-(endTime,startTime))))
	    in 
		Timer.checkRealTimer timer
	    end
	    
	val startTime = Timer.checkRealTimer timer 
	val _ = trace 2 "    Adding constants for operations" 
	val thy = foldl (fn (c,thy) => add_operation_consts
					   c universe cl  thy) 
			thy cl   	      
	val startTime = printTiming startTime
	val _ = trace 2 "    Adding core OCL formulae (invariants)" 
(*	val thy = foldl (fn (c,thy) => classic_invariant_encoder.add_invariants_1
					   c universe cl thy) 
		thy cl   	
*)
	val thy = foldl (fn (c,thy) => invariant_encoder.add_invariant
					   c universe cl thy) 
		thy cl   	

	val startTime = printTiming startTime
	val _ = trace 2 "    Adding constructors" 
	val thy = foldl (fn (c,thy) => add_constructors_2
					   c universe  cl  thy) 
			thy cl   	               

	val startTime = printTiming startTime
	val _ = trace 2 "    Adding attributes" 
	val thy = foldl (fn (c,thy) => add_attributes_2
					   c universe cl  thy) 
			thy cl   	      


	val startTime = printTiming startTime
	val _ = trace 2 "    Adding definition for operations" 
	val thy = foldl (fn (c,thy) => add_operation_defs
					   c universe cl  thy) 
			thy cl   	      

	val startTime = printTiming startTime
	val _ = trace 2 "    Adding definitions for shallow value equality" 
	val thy = foldl (fn (c,thy) => add_sve_2
					   c universe  cl  thy) 
			thy cl   	
	
	val startTime = printTiming startTime
        val _ = trace 2 "    Adding OLC formulae (invariants)" 
(*
	val thy = foldl (fn (c,thy) => classic_invariant_encoder.add_invariants_2
					   c universe  cl  thy) 
			thy cl   	
*)
	val thy = foldl (fn (c,thy) => invariant_encoder.add_user_invariant
					   c universe cl thy) 
		thy cl   	
        
	val startTime = printTiming startTime
        (* proof properties for second level class constructors *) 	
	(* val _ = trace 2 "    Proving ..." 
           val thy = foldl (fn (thy,c) => prove_2
					   c universe  cl thy) fun add_invariants_2 C u cl  thy =

			(thy,cl)   	      
         *)
	

    in 
	thy
    end
end
