(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Sequence.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Sequence *}
theory OCL_Sequence 
imports 
  "$HOLOCL_HOME/src/library/collection/OCL_Collection" 
  "$HOLOCL_HOME/src/library/collection/HOL_Collections_ext"
begin
 text_raw{* \label{thy:OCL_Sequence} *}


section{* Theorems needed in this theory but not belonging to it *}

subsection{* Very often used simplifier sets *}

lemma DEF_def_both:
  "DEF X \<equiv> (\<bottom> \<noteq> X \<and> X \<noteq> \<bottom>)"
  by(simp add: DEF_def neq_commute)


subsection{* Lifting *}

lemma not_down_lift_drop [simp]:
  "x \<noteq> down \<Longrightarrow> \<HolOclLiftLeft>\<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight> = x"
  by(drule not_down_exists_lift[THEN iffD1], auto)


subsection{* Logic core *}

lemma and_equiv_sand:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> Y \<rbrakk> \<Longrightarrow>  (X \<MathOclAnd> Y) \<tau> =  (X \<MathOclSand> Y) \<tau>"
  by(simp add: OclSand_def OclAnd_def localValidDefined2sem ss_lifting)



lemma ocl_if_totally_undefined [simp]:
  "(\<MathOclIf> P \<MathOclThen> \<MathOclUndefined> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif>) = \<MathOclUndefined>"
  by(rule ext, simp add: OclIf_def OclUndefined_def ss_lifting')

lemma if_distrib_defined:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; cp (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) \<rbrakk> \<Longrightarrow>
   P(\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) \<tau> =  (\<MathOclIf> X \<MathOclThen> P Y \<MathOclElse> P Z \<MathOclEndif>) \<tau>"
  apply(auto simp: OclIf_def localValidDefined2sem ss_lifting')
  apply(rule_tac P=P in cp_charn, simp_all)+
  done

lemma if_distrib_strict_alt:
  "\<lbrakk> cp P; P \<MathOclUndefined> = \<MathOclUndefined> \<rbrakk> \<Longrightarrow> P (\<MathOclIf> X \<MathOclThen> Y\<MathOclElse> Z\<MathOclEndif>) = (\<MathOclIf> X \<MathOclThen> (P Y)\<MathOclElse> (P Z)\<MathOclEndif>)"
  by(simp add: if_distrib_strict isStrict_def OclUndefined_def lift0_def)

lemma if_contract:
  "(\<MathOclIf> P \<MathOclThen> (\<MathOclIf> Q \<MathOclThen> X \<MathOclElse> Y \<MathOclEndif>)\<MathOclElse> Y \<MathOclEndif>) = (\<MathOclIf> (P \<MathOclSand> (P \<MathOclImplies> Q)) \<MathOclThen> X \<MathOclElse> Y \<MathOclEndif>)"
  by(rule ext, simp add: OclIf_def OclSand_def OclImplies_def o_def
                         OclOr_def OclNot_def OclAnd_def ss_lifting)

lemma if_not: "(\<MathOclIf> (\<MathOclNot> X) \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = (\<MathOclIf> X \<MathOclThen> Z \<MathOclElse> Y \<MathOclEndif>)"
  by(rule ext, simp add: OclIf_def OclNot_def ss_lifting)




subsection{* Some cp theory to simplify the later proofwork *}
(*
(* TODO: move this statements *)

text{* NOTE: The combinator cp2 isn't used anymore in any theory. These theorems
are here because they give some insight on the properties of context-passingness.

There are places where some functions must be context-passing in their
first and their second argument. This is now written explicitely. What
remains to do is to decide if the current style is the right one.

What is really needed very much is tool support with relation to the
proofing of properties depending on cp-ness.
*}


lemma cp2_by_cp:
  "\<lbrakk> !!x. cp (P x); !!x. cp (\<lambda> y. P y x) \<rbrakk> \<Longrightarrow> cp2 P"
  by(simp add: cp2_def cp_by_cpify)

lemma cp2_imp_cp:
  "\<lbrakk> cp2 P; cp f; cp g \<rbrakk> \<Longrightarrow> cp (\<lambda> x. P (f x) (g x))"
  apply(simp add: cp2_def cp_by_cpify lift0_def)
  apply(auto)
  apply(rule_tac t="(P (f X) (g X) \<tau>)" and s="(P (\<lambda>s. (f X \<tau>)) (\<lambda>s. (g X \<tau>)) \<tau>)" in subst) 
  apply(simp)
  apply(rule_tac t="(P (f (\<lambda>s. (X \<tau>))) (g (\<lambda>s. (X \<tau>))) \<tau>)" 
             and s="(P (\<lambda>s.(f (\<lambda>s. (X \<tau>))) \<tau>) (\<lambda>s. (g (\<lambda>s. (X \<tau>))) \<tau>) \<tau>)" in subst)
  prefer 2
  apply(simp)
  apply(drule_tac x="f (\<lambda>s. (X \<tau>))" in allE)
  apply(rotate_tac -1)
  apply(drule_tac x="g (\<lambda>s. (X \<tau>))" in allE)
  apply(rotate_tac -1)
  apply(drule_tac x="\<tau>" in allE)
  prefer 2 prefer 3 prefer 4
  apply(rule sym)
  apply(assumption)+
  done

lemma cp2_cp_conv:
  "cp2 P = ((\<forall> x. cp (P x)) \<and> (\<forall> x. cp (\<lambda> y. P y x)))"
  apply(auto)
  apply(rule_tac f="\<lambda> x. x" and P="\<lambda>u ua. (P ua u)" in cp2_imp_cp)
  prefer 4
  apply(rule_tac f="\<lambda> x. x" and P="P" in cp2_imp_cp)
  prefer 7
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done

(* TODO: which should be simplification rules and which form should they
         have.

         This concerns most, the relation of the following three rules
         and cp2_by_cp. Probably a better destruction rule would be nice. *)

lemma cp2_imp_cp_unary:
  "cp2 P \<Longrightarrow> cp (\<lambda> x. P x x)"
  by(simp add: cp2_def cp_by_cpify)

lemma cp2_imp_cp_1:
  "cp2 P \<Longrightarrow> (cp (\<lambda> y. P y z))"
  apply(rule_tac f="\<lambda> x. x" and P="P" in cp2_imp_cp)
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done

lemma cp2_imp_cp_2:
  "cp2 P \<Longrightarrow> (cp (P x))"
  apply(rule_tac f="\<lambda> x. x" and P="\<lambda>u ua. (P ua u)" in cp2_imp_cp)
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done


text{* a nice consequence of the above reasoning *}


lemma cp_compose_binary [simp,intro!]:
  "\<lbrakk> \<And> y. cp (\<lambda> x. P x y); \<And> x. cp (P x); cp F; cp F' \<rbrakk> \<Longrightarrow> 
   cp (\<lambda> x. P (F x) (F' x))"
  by(rule_tac P="P" in cp2_imp_cp, simp_all add: cp2_by_cp) 

lemma cp_abstr [simp,intro!]:
  "\<lbrakk> \<And> y. cp (\<lambda> x. F x y) \<rbrakk> \<Longrightarrow> cp (\<lambda> x \<tau> y. F x y \<tau>)"
  by(simp add: cp_by_cpify lift0_def)

*)

text{* Adding currently missing cp-ness results to the simpset *}

lemmas cp_hol_imp [simp,intro!] = cp_lift2[simplified lift2_def, of _ _ "op \<longrightarrow>"]
lemmas cp_hol_and [simp,intro!] = cp_lift2[simplified lift2_def, of _ _ "op \<and>"]
lemmas cp_hol_or  [simp,intro!] = cp_lift2[simplified lift2_def, of _ _ "op \<or>"]
lemmas cp_hol_not [simp,intro!] = cp_lift1[simplified lift1_def, of _ "\<lambda> x. \<not> x"]

(* TODO: according to the warning this lemma exists but is only in the intro! set *)
lemma cp_or [simp,intro!]: "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp (\<lambda> X. (P X) \<MathOclOr> (P' X))"
  by(simp add: OclOr_def)

(* TODO: the same holds for this lemma\<dots>
   @BU: is there a reason for it?  *)
declare cp_strongEq[simp,intro!]

(* TODO: why is this lemma not in the simpset *)
lemma cp_sor [simp,intro!]: "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp (\<lambda> X. (P X) \<MathOclSor> (P' X))"
  by(simp add: OclSor_def)

lemma cp_sand [simp,intro!]: "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp (\<lambda> X. (P X) \<MathOclSand> (P' X))"
  by(simp add: OclSor_def)

lemma cp_sxor [simp,intro!]: "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp (\<lambda> X. (P X) \<MathOclSxor> (P' X))"
  by(simp add: OclSxor_def)

text{* A very useful substitution rule *}
lemma cp_subst:
  "\<lbrakk> cp P \<rbrakk> \<Longrightarrow> P x \<tau> = P (lift0 (x \<tau>)) \<tau>"
  by(subst (asm) cp_by_cpify, auto)

(*
text{* and one formulation of a destruction rule for cp2 *}
lemma cp2D:
  "cp2 P \<Longrightarrow> (P x y \<tau>) = (P (lift0 (x \<tau>)) (lift0 (y \<tau>)) \<tau>)"
  by(unfold cp2_def, auto simp: lift0_def)
*)
(* ***************** *)
(* End of cp theory *)
(* ***************** *)

section{* Setup of the environment *}

subsection{* Special parameters *}
(*
text{* This should signal the ocl\_setup\_op method that it shall print all proven
  theorems. But unfortunately it doesn't do it currently.*}
ML{* val OCL_setup_trace = ref true; *}
*)
section{* Properties of Datatype Adaption *}

text{* The following rules are transformed versions of the automatically
  generated rules for datatype adaption *}

lemma Abs_Sequence_0_inject_charn:
  "\<lbrakk> (\<bottom>::'\<tau>::bot) \<notin> set x; (\<bottom>::'\<tau>::bot) \<notin> set y \<rbrakk> \<Longrightarrow> 
   (((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Sequence_0_inject, 
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

text{* These are derived rules from the above one that allow the simplifier to
  detect false assumptions which make a goal trivially true.*}

lemma Abs_Sequence_0_inject_absurd11 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 down) = (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Sequence_0_inject,
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_0_inject_absurd12 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Sequence_0 down)) = False"
  by(subst Abs_Sequence_0_inject,
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_0_inject_absurd21 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ( \<bottom> = (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Sequence_def)

lemma Abs_Sequence_0_inject_absurd22 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Sequence_def)

lemma Abs_Sequence_0_inverse_charn1 [simp]:
  "(\<bottom> \<notin> set x) \<Longrightarrow> Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Sequence_0_inverse, 
      simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_0_inverse_charn2 [simp]:
  "(\<bottom> \<notin> set x) \<Longrightarrow> \<HolOclDropLeft>Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Sequence_0_inverse, 
      simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_0_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; \<bottom> \<notin> set y \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_Sequence_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_Sequence_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(auto simp: OCL_Sequence_type.Sequence_0_def smash_def)
  done

lemma Abs_Sequence_0_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> \<bottom> \<notin> set y \<rbrakk> \<Longrightarrow> P (Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_Sequence_0_induct)
  apply(rule_tac  x="Abs_Sequence_0 y" in Abs_Sequence_0_cases_charn)
  apply(auto intro!: bottomCase elim: stepCase)
  done

text{* These rules stem from the set theory, that was the first collection
  theory. It should actually be decided which of them are really needed anymore
  because the above rules cover most cases. *}

lemma inj_on_Abs_Sequence: "inj_on Abs_Sequence_0 Sequence_0";
  by(rule inj_on_inverseI,rule Sequence_0.Abs_Sequence_0_inverse,assumption)

lemma inj_Rep_Sequence : "inj Rep_Sequence_0"
  by(rule inj_on_inverseI, rule Sequence_0.Rep_Sequence_0_inverse)

lemma smashed_sequence_charn: 
  "(\<bottom> \<notin> set X) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> Sequence_0)"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 1 *)

lemma UU_in_smashed_sequence [simp]: 
  "\<bottom> \<in> Sequence_0"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 2a *)

lemma down_in_smashed_sequence [simp]: 
  "down \<in> Sequence_0"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 2b *)

lemma mt_in_smashed_sequence[simp]: 
  "(\<HolOclLiftLeft>[]\<HolOclLiftRight>) \<in> Sequence_0"
  by (unfold smash_def Sequence_0_def,auto)

lemma DEF_Abs_Sequence: "\<And>X. (\<bottom> \<notin> set X) \<Longrightarrow> DEF (Abs_Sequence_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  apply(unfold DEF_def UU_Sequence_def, simp)
  done

lemma DEF_Rep_Sequence: 
  "\<And>X. DEF X \<Longrightarrow> DEF (Rep_Sequence_0 X)"
  apply(unfold DEF_def UU_Sequence_def,auto)
  apply(drule_tac f = "Abs_Sequence_0" in arg_cong)
  apply(simp add: Rep_Sequence_0_inverse)
  done

lemma not_DEF_Rep_Sequence: 
  "\<And>X. \<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Sequence_0 X)"
  apply(unfold DEF_def UU_Sequence_def,auto)
  apply(simp add: Abs_Sequence_0_inverse)
  done


lemma exists_lift_Sequence: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  by (drule DEF_Rep_Sequence,simp add: DEF_X_up)

lemma exists_lift_Sequence2: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  apply(frule exists_lift_Sequence)
  apply(auto)
  apply(drule sym, rule swap)
  prefer 2; apply assumption;back;
  apply(subst smashed_sequence_charn)
  apply(simp add: Rep_Sequence_0)
  done

lemma Rep_Sequence_cases: 
  "Rep_Sequence_0 X = \<bottom> \<or> (\<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(case_tac "DEF X")
  apply(drule exists_lift_Sequence2)
  apply(drule_tac [2] not_DEF_Rep_Sequence)
  apply(simp_all add: DEF_def)
  done

text{* These two @{text "DEF_X_Sequence"}  lemmas are a central part in almost every
  proof that unfolds the \OCL definitions. Furthermore they will also be used
  by the \tactic{thy\_morpher} to automatically lift theorems about \HOL lists to
  theorems about \OCL sequences. *}
lemma DEF_X_Sequence_0 : "DEF(X) = (\<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(insert  Rep_Sequence_cases [of "X"], auto)
  apply(drule exists_lift_Sequence,auto)
  apply(rule swap)
  prefer 2
  apply(rule not_DEF_Rep_Sequence, auto)
  done


lemma DEF_X_Sequence: "DEF X = (\<exists>c.  \<bottom> \<notin> set c \<and> X = Abs_Sequence_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply (auto simp: DEF_Abs_Sequence)
  apply (simp add:DEF_X_Sequence_0)
  apply (erule exE, rule exI, auto)
  apply (rule injD [OF inj_Rep_Sequence])
  apply (auto simp: smashed_sequence_charn Abs_Sequence_0_inverse)
  done

text{* Because in most cases both versions of the above theorems are used they are
  packed conveniently in this lemma. *}
lemma DEF_X_Sequence':
  "\<lbrakk> X \<noteq> \<bottom> \<rbrakk>  \<Longrightarrow> 
   (\<exists>c. (\<bottom> \<notin> set c) \<and> (Rep_Sequence_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>)) \<and> 
   (\<exists>c. (\<bottom> \<notin> set c) \<and> (X = (Abs_Sequence_0 \<HolOclLiftLeft>c\<HolOclLiftRight>)))"
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(drule DEF_X_Sequence_0[THEN iffD1])
  apply(simp)
  done
 
text{* And to enable a fast case splitting over (un)definedness of a specific
  sequence together with its semantic representation these two theorems are very
  useful. Note that there exists a @{text "_0"} version which does not require 
  $\HolOclCp$-ness.
  This @{text "_0"} version exists actually for almost all more complex theorems 
  because not in all cases one has or can show $\HolOclCp$-ness. *}
lemma Sequence_sem_cases_0:
  assumes defC: "\<And> c d. \<lbrakk> X \<noteq> \<bottom>; \<bottom> \<noteq> X; 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          Rep_Sequence_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          X = Abs_Sequence_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X"
  and   undefC: "\<lbrakk> X = \<bottom> \<rbrakk> \<Longrightarrow> P X"
  shows "P X"
  apply(rule Abs_Sequence_0_cases_charn, erule undefC)
  apply(rule defC, auto)
  done

lemma Sequence_sem_cases:
  assumes defC: "\<And> c d. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          Rep_Sequence_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Sequence_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and   undefC: "\<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and      cpP: "cp P"
  shows "P X \<tau>"
  apply(rule_tac P2="P" in subst[OF sym[OF cp_subst]], rule cpP)
  apply(rule Sequence_sem_cases_0)
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule defC) prefer 7
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule undefC, simp_all)
  done

text{* And a version for the case that we're reasoning over equalities of
  functions. *}
lemma Sequence_sem_cases_ext:
  assumes defC: "\<And> c d \<tau>. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          Rep_Sequence_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Sequence_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and   undefC: "\<And> \<tau>. \<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and      cpP: "cp P"
  and      cpQ: "cp Q"
  shows "P X = Q X"
  apply(rule ext)
  apply(rule_tac X="X" in Sequence_sem_cases)
  apply(rule defC) prefer 7
  apply(rule undefC)
  apply(simp_all add: cpP cpQ)
  done

text {* 
  These four lemmas are used by \inlineholocl{ocl_setup_op} 
  to reason about definedness: 
*}

lemma lift2_strict_is_isdef_fw_Sequence_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'\<tau>::bot) \<noteq> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule ext)
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def ss_lifting)
  apply(rule_tac X="X x" in Sequence_sem_cases_0)
  apply(rule impI, drule_tac b="(Y x)" in  inv_g)
  apply(simp_all add: neq_commute)
  done

lemma lift2_strict_is_isdef_fw_Sequence_Sequence:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Sequence_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; \<bottom> \<notin> set b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule ext)
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def ss_lifting)
  apply(rule_tac X="X x" in Sequence_sem_cases_0)
  apply(rule_tac X="Y x" in Sequence_sem_cases_0)
  apply(rule impI, drule_tac b="ca" in  inv_g)
  apply(simp_all add: neq_commute)
  done

lemma lift2_strictify_implies_LocalValid_defined_Sequence_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'\<tau>::bot) \<noteq> b\<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Sequence_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Sequence_Sequence:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Sequence_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; \<bottom> \<notin> set b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Sequence_Sequence)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

section{* Building a canonic representation of sequences *}

subsection{* The empty sequence *}

constdefs
    OclMtSequence   :: "('\<tau>, '\<alpha>::bot Sequence_0) VAL"    
   "OclMtSequence  \<equiv> lift0(Abs_Sequence_0(\<HolOclLiftLeft>[]\<HolOclLiftRight>))"

    syntax  (* standard syntax *)
    "_OclMtSequence_std"   :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("\<OclMtSequence>")
    syntax  (* ascii syntax *)
    "_OclMtSequence_ascii" :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("mtSequence")
    syntax  (xsymbols)
    "_OclMtSequence_math"  :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"     ("\<MathOclMtSequence>")
(*<*)
    parse_translation {* flat(map (emb trans_const) ["OclMtSequence"]) *}
    print_translation {* map emb_print ["OclMtSequence"] *}
(*>*)

lemma OCL_is_isdef_OclMtSequence [simp]:
  "\<MathOclValid> \<MathOclIsDefined> \<MathOclMtSequence>"
  by(simp add: OclValid_def OclIsDefined_def OclTrue_def 
               OclMtSequence_def ss_lifting')

lemma OCL_is_defopt_OclMtSequence [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> \<MathOclMtSequence>"
  by(simp add: valid_elim)



subsection{* And the `cons' operation: including *}
defs
   (* including  :: "[('\<tau>,('\<alpha>::bot) ) Sequence, ('\<tau>,'\<alpha>::bot ) VAL] => ('\<tau>,'\<alpha> ) Sequence" *)
   OclIncluding_def   :
   "OclIncluding  \<equiv> lift2(strictify(\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                       (\<HolOclLiftLeft>concat [\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>, [e] ]\<HolOclLiftRight>))))"

(* TODO: correct setup op to handle this case as well *)
ocl_setup_op [OclIncluding]

lemma OCL_is_def_OclIncluding:
  "\<MathOclIsDefined>(OclIncluding (X::('\<tau>,'\<alpha>::bot)Sequence) (Y::('\<tau>,'\<alpha>::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  by(rule lift2_strict_is_isdef_fw_Sequence_Val[OF OclIncluding_def], simp)
  
lemma OCL_is_defopt_OclIncluding:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIncluding (X::('\<tau>,'\<alpha>::bot)Sequence) (Y::('\<tau>,'\<alpha>::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  by(rule lift2_strictify_implies_LocalValid_defined_Sequence_Val[OF OclIncluding_def], simp)



text {* The syntax translation @{text "mkSequence"} 
              builds now our sequences *}

syntax
   "@OclFinSequence" :: "args \<Rightarrow> ('\<tau>,'\<alpha> Sequence_0)VAL"           ("mkSequence[(_)]")

translations
  "mkSequence[x, xs]"    == "OclIncluding (mkSequence[xs]) x"
  "mkSequence[x]"        == "OclIncluding OclMtSequence x"

lemma test : "mkSequence[1,2,3,4,5] = ?X" oops

subsection{* Relation of mtSequence and including *}


text{* These two theorems should actually suffice to decide inequality of sequences *}
(* TODO: name for not strong equaltiy  . . . and representation *)
lemma including_notstrongeq_mtSequence[simp]:
   "(\<MathOclNot> ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> (a::('\<tau>,'\<alpha>)VAL) \<MathOclStrongEq> \<MathOclMtSequence>)) = \<MathOclTrue>"
  apply(rule_tac X="X" in Sequence_sem_cases_ext)
  apply(simp_all add: cp_strongEq)
  apply(simp_all add: OclIncluding_def OclMtSequence_def OclStrongEq_def
                      OclTrue_def OclNot_def localValid2sem ss_lifting'
                      Abs_Sequence_0_inject_charn neq_commute)
  done


(* TODO: name for not strict equality\<dots> and representation *)
lemma including_notstricteq_mtSequence[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> (X \<MathOclIncluding> a) `<>` \<MathOclMtSequence>"
  apply(rule_tac X="X" in Sequence_sem_cases)
  apply(simp_all add: cp_strongEq cp_strictEq localValidDefined2sem)
  apply(simp_all add: OclIncluding_def OclMtSequence_def OclStrictEq_def
                      OclTrue_def OclNot_def localValid2sem ss_lifting'
                      Abs_Sequence_0_inject_charn neq_commute)
  done


subsection{* Case exhaustion over the canonic representation *}

lemma Sequence_cases_including_0:
  assumes undefCase: "\<lbrakk> (X::'\<tau>::bot Sequence_0) = \<bottom> \<rbrakk> \<Longrightarrow> P"
  and        mtCase: "\<lbrakk> X = (\<MathOclMtSequence> \<tau>) \<rbrakk> \<Longrightarrow> P"
  and      stepCase: "\<And> (Y::'\<tau>::bot Sequence_0) (a::'\<tau>::bot). 
                       \<lbrakk> DEF Y; DEF a; X = ((lift0 Y) \<MathOclIncluding> (lift0 a)) \<tau>\<rbrakk> \<Longrightarrow> P"
  shows              "P"
  apply(rule Abs_Sequence_0_cases_charn, erule undefCase)
  apply(rule_tac xs="y" in rev_cases)
  apply(rule mtCase)
  apply(simp add: OclMtSequence_def ss_lifting')
  apply(rule_tac Y="Abs_Sequence_0 \<HolOclLiftLeft>ys\<HolOclLiftRight>" and a="ya" in stepCase)
  apply(simp_all add: OclIncluding_def ss_lifting' neq_commute)
  done

lemma Sequence_cases_including:
  assumes undefCase : "\<lbrakk> \<tau> \<MathOclValid> (X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclStrongEq> \<MathOclUndefined> \<rbrakk> \<Longrightarrow> P \<tau>"
  and         mtCase: "\<lbrakk> \<tau> \<MathOclValid> X \<MathOclStrongEq> \<MathOclMtSequence> \<rbrakk> \<Longrightarrow> P \<tau>"
  and       stepCase: "\<And> (Y::('\<tau>,'\<alpha>::bot)Sequence) (a::('\<tau>,'\<alpha>::bot)VAL). 
                       \<lbrakk> \<tau> \<MathOclValid> X \<MathOclStrongEq> (Y \<MathOclIncluding> a); \<tau> \<MathOclValid> \<MathOclIsDefined> Y; \<tau> \<MathOclValid> \<MathOclIsDefined> a\<rbrakk> \<Longrightarrow> P \<tau>"
  shows               "P \<tau>"
  apply(rule_tac X="X \<tau>" and \<tau>="\<tau>" in Sequence_cases_including_0)
  apply(rule undefCase) prefer 2
  apply(rule mtCase) prefer 3
  apply(rule stepCase, simp_all add: localValidDefined2sem)
  apply(simp_all add: localValid2sem OclStrongEq_def 
                      OclUndefined_def ss_lifting')
  done



subsection{* Induction over the canonic representation *}


text{* 
  To simplify reasoning there exists a @{text "_0"} version which doesn't pose
  any restrictions on the involved functions with regard to context-passingness.
*}
lemma Sequence_induct_including_0:
  assumes undefCase: "P \<bottom>"
  and        mtCase: "P (\<MathOclMtSequence> \<tau>)"
  and      stepCase: "\<And> (X::'\<tau>::bot Sequence_0) (a::'\<tau>::bot). 
                       \<lbrakk> P X; DEF X; DEF a \<rbrakk> \<Longrightarrow> (P ((OclIncluding (lift0 X) (lift0 a)) \<tau>))"
  shows              "P (X::'\<tau>::bot Sequence_0)"
  apply(rule Abs_Sequence_0_induct_charn, rule undefCase)
  apply(erule rev_mp)
  apply(rule rev_induct)
  apply(insert mtCase, simp add: OclMtSequence_def ss_lifting')
  apply(auto)
  apply(rotate_tac -1, drule_tac a="x" and X="Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>" in stepCase)
  apply(erule DEF_Abs_Sequence)
  apply(simp_all add: OclIncluding_def neq_commute ss_lifting')
  done
  
lemma Sequence_induct_including:
  assumes undefCase : "P \<MathOclUndefined> \<tau>"
  and         mtCase: "P \<MathOclMtSequence> \<tau>"
  and       stepCase: "\<And> (X::('\<tau>,'\<alpha>::bot)Sequence) (a::('\<tau>,'\<alpha>::bot)VAL). 
                       \<lbrakk> P X \<tau>; \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> a\<rbrakk> \<Longrightarrow> P (OclIncluding X a) \<tau>"
  and            cpP: "cp P"
  shows               "P (X::('\<tau>,'\<alpha>::bot)Sequence) \<tau>"
  apply(insert cpP)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(rule subst, rule sym)
  apply(rule_tac x="X" in cp_subst, simp)
  apply(rule_tac \<tau>="\<tau>" in Sequence_induct_including_0)
  apply(insert undefCase)
  apply(auto simp: OclUndefined_def intro!: mtCase)
  apply(drule stepCase)
  apply(simp_all add: localValidDefined2sem lift0_def)
  done


section{* A second sequence constructor on the \OCL level *}
defs
   (* collectionRange :: "[('\<tau>,'\<alpha>::{bot,plus}) VAL,('\<tau>,'\<alpha>) VAL] \<Rightarrow> ('\<tau>,'\<alpha>) Sequence" *)
   OclCollectionRange_def :
   "OclCollectionRange \<equiv> lift2 (strictify (\<lambda> x\<Colon>Integer_0. strictify (\<lambda> y. 
                           Abs_Sequence_0 \<HolOclLiftLeft>map (\<lambda> z::nat. \<HolOclLiftLeft>(int z) + \<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight>) 
                                               [0..<(nat (\<HolOclDropLeft>y\<HolOclDropRight> - \<HolOclDropLeft>x\<HolOclDropRight> + 1))]\<HolOclLiftRight>)))"

ocl_setup_op [OclCollectionRange]

(* TODO: 1. add further rules to simplify use of CollectionRange
            e.g. includes and definedness
         2. is there an induction rule over OCL Integers? *)

lemma OCL_is_defopt_OclCollectionRange [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> ((OclCollectionRange (a::('\<tau> Integer)) b)::('\<tau>, Integer_0)Sequence) = 
   ((\<tau> \<MathOclValid> \<MathOclIsDefined> a) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> b))"
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>) + 1))])")
  apply(auto simp: localValidDefined2sem OclCollectionRange_def ss_lifting)
  done

lemma collectionRange_mtSequence_conv:
  "\<lbrakk> \<tau> \<MathOclValid> (b::'\<tau> Integer) \<MathOclLess> (a::'\<tau> Integer) \<rbrakk> \<Longrightarrow>
   OclCollectionRange a b \<tau> = (\<MathOclMtSequence>::('\<tau>,Integer_0)Sequence) \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> a", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> b", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(simp_all add: DEF_def OclCollectionRange_def OclMtSequence_def 
                      OclLess_def localValid2sem ss_lifting')
  done

lemma collectionRange_singleton_UC[simp]:
  "OclCollectionRange a a = \<MathOclMtSequence>::('\<tau>,Integer_0)Sequence \<MathOclIncluding> (a::('\<tau>,Integer_0)VAL)"
  apply(rule ext, case_tac "x \<MathOclValid> \<MathOclIsUndefined>(a)", ocl_hypsubst, simp)
  apply(simp add: OclCollectionRange_def OclMtSequence_def OclIncluding_def
                  localValidDefined2sem ss_lifting')
  done

lemma collectionRange_expand_including:
  "\<lbrakk> \<tau> \<MathOclValid> (a::'\<tau> Integer) \<MathOclLe> b \<rbrakk> \<Longrightarrow>
   ((OclCollectionRange a b)::('\<tau>,Integer_0)Sequence) \<tau> = 
   (OclCollectionRange a (b - 1)) \<MathOclIncluding> b \<tau>"
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>) + 1))])")
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat (\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>))])")
  apply(auto simp: OclCollectionRange_def OclIncluding_def
               OclLe_def minus_def One_ocl_int_def localValid2sem ss_lifting')
  apply(subst Abs_Sequence_0_inject_charn, auto)
  apply(subgoal_tac "(nat (\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight> + 1)) = (nat ((\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>)) + (nat 1))")
  apply(simp, arith)
  done


section{* The conversion operators *}

subsection{* Sequence to sequence *}
defs
   (* asSequnece     :: "('\<tau>,'\<alpha>::bot) Sequence \<Rightarrow> ('\<tau>,'\<alpha>::bot) Sequence" *)
   OclAsSequence_def : 
   "OclAsSequence    \<equiv> id"

lemma asSequence_identity[simp]:
  "OclAsSequence (S::('\<tau>,'\<alpha>::bot)Sequence) = S"
  by(simp add: OclAsSequence_def)


subsection{* Sequence to bag *}
defs
   (* asBag     :: "('\<tau>,'\<alpha>::bot) Sequence \<Rightarrow> ('\<tau>,'\<alpha>::bot) Bag" *)
   OclAsBag_def  : 
   "OclAsBag     \<equiv> lift1 (strictify (\<lambda>X. 
                          Abs_Bag_0 \<HolOclLiftLeft>multiset_of \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsBag]


subsection{* Sequence to ordered set *}
defs
   (* asSet             :: "('\<tau>,'\<alpha>::bot) Sequence \<Rightarrow> ('\<tau>,'\<alpha>::bot) OrderedSet" *)
   OclAsOrderedSet_def  : 
   "OclAsOrderedSet     \<equiv> lift1 (strictify (\<lambda>X. 
                             Abs_OrderedSet_0 \<HolOclLiftLeft>rev (remdups (rev \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>))\<HolOclLiftRight>))"

ocl_setup_op [OclAsOrderedSet]


subsection{* Sequence to set *}
defs
   (* asSet     :: "('\<tau>,'\<alpha>::bot) Sequence \<Rightarrow> ('\<tau>,'\<alpha>::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> lift1 (strictify (\<lambda>X. Abs_Set_0 \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsSet]



section{* OclIterate *}

defs
   (* iterate     ::"[ ('\<tau>,'d::bot)Sequence, 
                       [('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>::bot)VAL] => ('\<tau>,'c::bot)VAL,
                       ('\<tau>,'c::bot)Sequence
                     ] 
                       => ('\<tau>,'c::bot) Sequence" *)
   OclIterate_def : 
   "OclIterate \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) (strictify (\<lambda> S P A.
                           (foldl (\<lambda> x y. P y x) A (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>))))"

(*
   "OclIterate S P A \<equiv> (\<lambda> \<tau>. if DEF (S \<tau>)
                             then  (foldl (\<lambda> x y. (P (lift0 y) (lift0 x)) \<tau>) 
                                          (A \<tau>)
                                          (\<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>))
                             else \<bottom>)"

*)

ocl_setup_op [OclIterate]

(* outdated: by the new ocl_setup_op plus the new construction

text{* Contextpassingness of iterate. This has to be proved at the very
beginning to enable the use of stronger proof techniques. *}

lemma OCL_cp_OclIterate [simp,intro!]: 
  "\<lbrakk> cp (S::('\<tau>,'c)VAL \<Rightarrow>('\<tau>,'\<alpha>::bot)Sequence); cp A; 
     \<And> a b. cp (\<lambda> x \<tau>. P x a b \<tau>) \<rbrakk> \<Longrightarrow>
     cp (\<lambda> X. OclIterate (S X) ((P X)::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL) (A X))"
thm fw_OCL_cp_chain020_lift3'[OF OclIterate_def]
  apply(rule fw_OCL_cp_chain020_lift3'[OF OclIterate_def])
  apply(simp_all)
  apply(simp add: OclIterate_def cp_by_cpify ss_lifting)

lemma OCL_undef_1_OclIterate [simp]:
  "OclIterate (\<MathOclUndefined>::('\<tau>,'\<alpha>::bot)Sequence) 
              (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) 
              A 
   = (\<MathOclUndefined>::('\<tau>,'c::bot)VAL)"
  by(simp add: OclIterate_def OclUndefined_def ss_lifting')

*)

subsection{* The characteristic behaviour of iterate on the canonical representation*}
lemma iterate_of_mtSequence[simp]:
  "(OclIterate (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) 
               (P::(('\<tau> \<Rightarrow> '\<alpha>) \<Rightarrow> (('\<tau> \<Rightarrow> 'c::bot) \<Rightarrow> ('\<tau> \<Rightarrow> 'c)))) 
                A) = A"
  apply(rule ext)
  apply(subgoal_tac "x \<MathOclValid> (\<MathOclIsDefined> \<MathOclMtSequence>)")
  apply(simp_all add: OclIterate_def OclIncluding_def OclMtSequence_def
                      localValidDefined2sem ss_lifting')
 done

lemma iterate_of_including_0:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau> \<Rightarrow> '\<alpha>::bot))"
  shows "((OclIterate (OclIncluding S x) P A) \<tau>) =
         (P (lift0 (x \<tau>)) (lift0 ((OclIterate (S::(('\<tau>,'\<alpha>::bot)Sequence)) P A) \<tau>)) \<tau>)"
  apply(insert defx defS)
  apply(frule_tac Y1="x" in conjI[THEN OCL_is_defopt_OclIncluding[THEN iffD2]], assumption)
  apply(simp add: localValidDefined2sem DEF_def) 
  apply(frule DEF_X_Sequence', clarify)
  apply(simp add: OclIterate_def OclIncluding_def ss_lifting')
  apply(frule DEF_X_Sequence', clarify, simp)
  apply(subst (asm) Abs_Sequence_0_inject_charn, auto)
  done

lemma iterate_of_including:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau> \<Rightarrow> '\<alpha>::bot))"
  and cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and cpP_2: "\<And> x. cp (P x)"
  shows "((OclIterate (OclIncluding S x) P A) \<tau>) =
         (P x (OclIterate (S::(('\<tau>,'\<alpha>::bot)Sequence)) P A) \<tau>)"
  apply(insert defS defx cpP_1 cpP_2)
  apply(drule_tac x="x" and P="P" and A="A" in iterate_of_including_0)
  apply(simp_all add: cp_by_cpify)
  done

lemma iterate_opt_of_including:
  assumes P_undef_1: "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and     P_undef_2: "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and         cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and         cpP_2: "\<And> x. cp ((P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) x)"
  shows "(OclIterate (OclIncluding S x) P A) =
         (P x (OclIterate (S::(('\<tau>,'\<alpha>::bot)Sequence)) P A))"
  apply(insert P_undef_1 P_undef_2 cpP_1 cpP_2)
  apply(rule ext)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> S")
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> x")
  apply(rule iterate_of_including, simp_all)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="x" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="x" in cp_charn, simp, simp)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="S" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="S" in cp_charn)
  apply(rule_tac P="P x" in cp_compose2, simp_all)
  done


subsection{* The universal and fusion property *}

text{* Fusion and universal theorems, as 
described in~\cite{hutton:fold:1999}. *}

lemma iterate_universal_0:
  assumes g_cp: "cp g"
  and g_undef:  "g (\<MathOclUndefined>::('\<tau>,'\<alpha>::bot)Sequence) \<tau> = (\<MathOclUndefined>::('\<tau>,'c::bot)VAL) \<tau>"
  and g_mtSeq:  "g (OclMtSequence::('\<tau>,'\<alpha>::bot)Sequence) \<tau> = A \<tau>"
  and g_incl:   "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                         g (OclIncluding S x) \<tau> = P (lift0 (x \<tau>)) (lift0 (g S \<tau>)) \<tau>"
  shows "g S \<tau> = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau>"
  apply(rule_tac X="S" in Sequence_induct_including)
  apply(simp_all add: g_cp g_undef g_mtSeq)
  apply(subst iterate_of_including_0, simp_all add: g_incl)
  done

lemma iterate_universal:
  assumes g_cp: "cp g"
  and    cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and    cpP_2: "\<And> x. cp (P x)"
  and  g_undef: "g \<MathOclUndefined> \<tau> = (\<MathOclUndefined>::('\<tau>,'c::bot)VAL) \<tau>"
  and  g_mtSeq: "g (OclMtSequence::('\<tau>,'\<alpha>::bot)Sequence) \<tau> = A \<tau>"
  and   g_incl: "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                         g (OclIncluding S x) \<tau> = P x (g S) \<tau>"
  shows "g S \<tau> = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau>"
  by(insert cpP_1 cpP_2 g_cp, rule iterate_universal_0, 
     simp_all add: g_undef g_mtSeq g_incl cp_by_cpify)


lemma iterate_universal_0_opt:
  assumes g_cp: "cp g"
  and  g_mtSeq: "g (OclMtSequence::('\<tau>,'\<alpha>::bot)Sequence) \<tau> = A \<tau>"
  and   g_incl: "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                         g (OclIncluding S x) \<tau> = P (lift0 (x \<tau>)) (lift0 (g S \<tau>)) \<tau>"
  and     defS: "\<tau> \<MathOclValid> \<MathOclIsDefined> S"
  shows "g S \<tau> = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau>"
  apply(insert defS, erule rev_mp)
  apply(rule_tac X="S" in Sequence_induct_including)
  apply(simp_all add: g_cp g_mtSeq)
  apply(subst iterate_of_including_0, simp_all add: g_incl)
  done

lemma iterate_universal_opt:
  assumes g_cp: "cp g"
  and    cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and    cpP_2: "\<And> x. cp (P x)"
  and  g_mtSeq: "g (OclMtSequence::('\<tau>,'\<alpha>::bot)Sequence) \<tau> = A \<tau>"
  and   g_incl: "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                         g (OclIncluding S x) \<tau> = P x (g S) \<tau>"
  and     defS: "\<tau> \<MathOclValid> \<MathOclIsDefined> S"
  shows "g S \<tau> = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau>"
  by(insert cpP_1 cpP_2 g_cp, rule iterate_universal_0_opt, 
     simp_all add: g_mtSeq g_incl cp_by_cpify)



lemma iterate_fusion_0_opt:
  "\<lbrakk> cp h;
    DEF (S \<tau>);
    h B \<tau> = A \<tau>; 
    !!x y. h (Q (lift0 (x \<tau>)) (lift0 (y \<tau>))) \<tau> = 
           P (lift0 (x \<tau>)) (lift0 (h y \<tau>)) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) 
                   (Q::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL) A \<tau>"
   apply(rule iterate_universal_0_opt)
   apply(rule_tac P="h" in cp_compose2)
   apply(simp_all)
   apply(rule trans, rule_tac P="h" in cp_subst, simp)
   apply(subst iterate_of_including_0)
   apply(simp_all add: cp_by_cpify localValidDefined2sem)
   done

lemma iterate_fusion_opt:
  "\<lbrakk> cp h;
    \<And> x. cp (P x);
    \<tau> \<MathOclValid> \<MathOclIsDefined> S;
    h B \<tau> = A \<tau>; 
    \<And> x y. h (Q x y) \<tau> = P x (h y) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) 
                   (Q::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL) A \<tau>"
  apply(rule iterate_fusion_0_opt)
  apply(assumption)+
  apply(simp add: localValidDefined2sem)
  apply(assumption)
  apply(subgoal_tac "h y \<tau> = h (lift0 (y \<tau>)) \<tau>")
  apply(simp)
  apply(rule trans)
  apply(rule_tac P="\<lambda> x. P ?X x" in cp_subst)
  apply(simp_all add: lift0_def cp_by_cpify)
  done


lemma iterate_fusion_0:
  "\<lbrakk> cp h;
    h (\<MathOclUndefined>::('\<tau>,'c::bot)VAL) \<tau> = (\<MathOclUndefined>::('\<tau>,'d::bot)VAL) \<tau>; 
    h B \<tau> = A \<tau>; 
    !!x y. h (Q (lift0 (x \<tau>)) (lift0 (y \<tau>))) \<tau> = 
           P (lift0 (x \<tau>)) (lift0 (h y \<tau>)) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) 
                   (Q::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL) A \<tau>"
   apply(rule iterate_universal_0)
   apply(rule_tac P="h" in cp_compose2)
   apply(simp_all)
   apply(rule trans, rule_tac P="h" in cp_subst, simp)
   apply(subst iterate_of_including_0)
   apply(simp_all add: cp_by_cpify)
   done

lemma iterate_fusion:
  "\<lbrakk> cp h;
    \<And> x. cp (P x);
    h (\<MathOclUndefined>::('\<tau>,'c::bot)VAL) \<tau> = (\<MathOclUndefined>::('\<tau>,'d::bot)VAL) \<tau>; 
    h B \<tau> = A \<tau>; 
    \<And> x y. h (Q x y) \<tau> = P x (h y) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) 
                   (Q::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL \<Rightarrow> ('\<tau>,'d::bot)VAL) A \<tau>"
  apply(rule iterate_fusion_0)
  apply(assumption)+
  apply(subgoal_tac "h y \<tau> = h (lift0 (y \<tau>)) \<tau>")
  apply(simp)
  apply(rule trans)
  apply(rule_tac P="\<lambda> x. P ?X x" in cp_subst)
  apply(simp_all add: lift0_def cp_by_cpify)
  done

(*TODO: make this cp reasoning more elegant #grrrrrrr# *)


subsection{* Further propierties of iterate *}

text{* Undefinedness the second part *}

lemma OCL_undef_2_OclIterate [simp]:
  assumes P_undef_2:  "\<And> x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  shows          "OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) 
                             (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) 
                             \<MathOclUndefined> 
                  = (\<MathOclUndefined>::('\<tau>,'c::bot)VAL)"
  apply(insert P_undef_2, rule ext, rule sym)
  apply(rule iterate_universal_0)
  apply(simp_all add: lift0_def OclUndefined_def)
  done

lemma OCL_is_defopt_OclIterate:
  assumes defopt_P:  "!!x y.   (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) 
                             = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  and        cpP_1:   "\<And> y. cp (\<lambda> x. P x y)"
  and        cpP_2:   "\<And> x. cp ((P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) x)"
  shows "\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIterate (S::('\<tau>,'\<alpha>::bot)Sequence) P A) = ((\<tau> \<MathOclValid> (\<MathOclIsDefined> S)) \<and> (\<tau> \<MathOclValid> (\<MathOclIsDefined> A)))"
  apply(insert defopt_P cpP_1 cpP_2)
  apply(rule_tac X="S" and \<tau>="\<tau>" in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac x="(X \<MathOclIncluding> a  \<MathOclIterate>(u;ua=A | (P u ua)))" in cp_subst, simp)
  apply(subst iterate_of_including)
  apply(simp_all)
  apply(subst OCL_is_defopt_OclIncluding)
  apply(rule trans, rule sym)
  apply(rule_tac x="(P a (OclIterate X P A))" in cp_subst, simp_all)
  done


subsection{* Forward rules to transfer properties about iterate to functions definable by iterate *}

subsubsection{* For \emph{unary} functions *}

lemma f1_by_iterate_cp:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and         cpF: "cp F"
  and        cpF': "cp F'"
  shows "cp (\<lambda>X. (f ((F X)::(('\<tau>,'\<alpha>::bot)Sequence))))"
  by(simp add: f_by_it cpF)

lemma f1_by_iterate_undef_1:
  "\<lbrakk> \<And> S. f (S::('\<tau>,'\<alpha>::bot)Sequence) = 
     OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclUndefined> = \<MathOclUndefined>"
  by(simp)


lemma f1_by_iterate_defopt:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and    defopt_P:  "!!x y.   (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) 
                             = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  and       cpP_1:   "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:   "\<And> x. cp ((P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) x)"
  shows "\<tau> \<MathOclValid> \<MathOclIsDefined>(f (S::('\<tau>,'\<alpha>::bot)Sequence)) = ((\<tau> \<MathOclValid> (\<MathOclIsDefined> S)) \<and> (\<tau> \<MathOclValid> (\<MathOclIsDefined> A)))"
  by(simp add: f_by_it OCL_is_defopt_OclIterate defopt_P cpP_1 cpP_2)
  
  
lemma f1_by_iterate_mtSequence:
  "\<lbrakk> \<And> S. f (S::('\<tau>,'\<alpha>::bot)Sequence) = 
     OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclMtSequence> = A"
  by(simp)
  

lemma f1_by_iterate_including:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and        defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and        defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau> \<Rightarrow> '\<alpha>::bot))"
  and       cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2: "\<And> x. cp (P x)"
  shows "(f (OclIncluding S x) \<tau>) = (P x (f S) \<tau>)"
  by(simp add: f_by_it iterate_of_including defS defx cpP_1 cpP_2)

lemma f1_by_iterate_opt_including:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and   P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and   P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and       cpP_1:  "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:  "\<And> x. cp (P x)"
  shows "f (OclIncluding S x) = P x (f S)"
  by(simp add: f_by_it iterate_opt_of_including
               P_undef_1 P_undef_2 cpP_1 cpP_2)



subsubsection{* For \emph{binary} fuctions *}

text{* This $\HolOclCp$-ness result holds only for binary functions whose second argument
is the neutral element of iterate. Currently the only function of this style is
union.*}

lemma f2_by_iterate_cp:
  assumes f_by_it: "\<And> S A. (f (S::('\<tau>,'\<alpha>::bot)Sequence) A) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and         cpF: "cp F"
  and        cpF': "cp F'"
  shows "cp (\<lambda>X. (f ((F X)::(('\<tau>,'\<alpha>::bot)Sequence)) (F' A)))"
  by(simp add: f_by_it cpF)

lemma f2_by_iterate_undef_1:
  "\<lbrakk> \<And> S A. f (S::('\<tau>,'\<alpha>::bot)Sequence) A = 
     OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclUndefined> A = \<MathOclUndefined>"
  by(simp)

lemma f2_by_iterate_undef_2:
  assumes f_by_it: "\<And> S A. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) A = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and   undef_2_P: "\<And> x. P x \<MathOclUndefined> = \<MathOclUndefined>"
  shows            "f S \<MathOclUndefined> = \<MathOclUndefined>"
  by(simp add: f_by_it undef_2_P)

lemma f2_by_iterate_defopt:
  assumes f_by_it: "\<And> S A. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) A = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and    defopt_P:  "!!x y.   (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) 
                             = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  and       cpP_1:   "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:   "\<And> x. cp ((P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) x)"
  shows "\<tau> \<MathOclValid> \<MathOclIsDefined>(f (S::('\<tau>,'\<alpha>::bot)Sequence) A) = ((\<tau> \<MathOclValid> (\<MathOclIsDefined> S)) \<and> (\<tau> \<MathOclValid> (\<MathOclIsDefined> A)))"
  by(simp add: f_by_it OCL_is_defopt_OclIterate defopt_P cpP_1 cpP_2)
  
  
lemma f2_by_iterate_mtSequence:
  "\<lbrakk> \<And> S A. f (S::('\<tau>,'\<alpha>::bot)Sequence) A = 
     OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclMtSequence> A = A"
  by(simp)
  

lemma f2_by_iterate_including:
  assumes f_by_it: "\<And> S A. (f (S::('\<tau>,'\<alpha>::bot)Sequence) A) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and        defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and        defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau> \<Rightarrow> '\<alpha>::bot))"
  and       cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2: "\<And> x. cp (P x)"
  shows "(f (OclIncluding S x) A \<tau>) = (P x (f S A) \<tau>)"
  by(simp add: f_by_it iterate_of_including defS defx cpP_1 cpP_2)


lemma f2_by_iterate_opt_including:
  assumes f_by_it: "\<And> S A. (f (S::('\<tau>,'\<alpha>::bot)Sequence) A) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and   P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and   P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and       cpP_1:  "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:  "\<And> x. cp (P x)"
  shows "f (OclIncluding S x) A = P x (f S A)"
  by(simp add: f_by_it iterate_opt_of_including
               P_undef_1 P_undef_2 cpP_1 cpP_2)

  

section{* Union *}

defs
   (* union        :: "[('\<tau>,'\<alpha>::bot)Sequence, ('\<tau>,'\<alpha>)Sequence] => ('\<tau>,'\<alpha>)Sequence" *)
   OclUnion_def    :
   "OclUnion       \<equiv> lift2(strictify(\<lambda>X. strictify(\<lambda>Y. Abs_Sequence_0
                       (\<HolOclLiftLeft>(\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> @ \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight>)\<HolOclLiftRight>))))"

ocl_setup_op [OclUnion]

lemma OCL_is_def_OclUnion:
  "\<MathOclIsDefined>(OclUnion (X::('\<tau>,'\<alpha>::bot)Sequence) Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Sequence_Sequence[OF OclUnion_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclUnion:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclUnion (X::('\<tau>,'\<alpha>::bot)Sequence) Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Sequence_Sequence[OF OclUnion_def])
  apply(simp)
  done
  

subsection{* Its computational characterisation using iterate *}

lemma union_by_iterate:
  "((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> (Y::('\<tau>,'\<alpha>::bot)Sequence)) = 
   (Y \<MathOclIterate>(x;y=X | y \<MathOclIncluding> (x::('\<tau>,'\<alpha>::bot)VAL) ))"
  apply(rule ext)
  apply(rule iterate_universal, simp_all)
  apply(rule_tac X="X" in Sequence_sem_cases)
  apply(simp add: OclUnion_def OclMtSequence_def ss_lifting')
  apply(simp add: OclUnion_def OclMtSequence_def ss_lifting')
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(rule_tac X="X" in Sequence_sem_cases, simp_all)
  apply(simp_all add: OclUnion_def OclIncluding_def ss_lifting' neq_commute)
  done

thm f2_by_iterate_including[OF union_by_iterate, simplified]

lemmas union_of_mtSequence1 [simp] = f2_by_iterate_mtSequence[OF union_by_iterate]
 
lemma union_of_mtSequence2 [simp]:
  "\<MathOclMtSequence> \<MathOclUnion> A = (A::('\<tau>,'\<alpha>::bot)Sequence)"
  apply(rule_tac X="A" in Sequence_sem_cases_ext)
  apply(simp_all)
  apply(simp_all add: OclMtSequence_def OclUnion_def ss_lifting')
  done

lemmas union_of_including[simp] = f2_by_iterate_opt_including[OF union_by_iterate, simplified]

(* TODO: really simp? in List.thy it is *)
lemma union_assoc [simp]:
  "(A \<MathOclUnion> (B \<MathOclUnion> C)) = ((A \<MathOclUnion> B) \<MathOclUnion> (C::('\<tau>,'\<alpha>::bot)Sequence))"
  apply(rule_tac X="A" in Sequence_sem_cases_ext)
  apply(rule_tac X="B" in Sequence_sem_cases)
  apply(rule_tac X="C" in Sequence_sem_cases, simp_all)
  apply(simp_all add: OclUnion_def ss_lifting')
  done

subsection{* Properties of iterate over a union of two sequences *}

lemma iterate_of_union:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X \<rbrakk> \<Longrightarrow> 
   OclIterate ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) 
              (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau> =
   OclIterate Y P (OclIterate X P A) \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> Y")
  apply(ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence')
  apply(rotate_tac 1, frule DEF_X_Sequence', clarify)
  apply(simp add:  OclUnion_def OclIterate_def ss_lifting')
  done

(* TODO: BU why does ocl_hypsubst not work????? *)
lemma iterate_of_union_opt:
  "\<lbrakk> \<And> x. P x \<MathOclUndefined> = \<MathOclUndefined> \<rbrakk> \<Longrightarrow> 
   OclIterate ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) 
              (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A=
   OclIterate Y P (OclIterate X P A)"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsDefined> X")
  apply(simp add: iterate_of_union)
  apply(simp add: sym[OF isUndefined_charn_local])
  apply(rule_tac A1="X" in trans[OF cp_charn], simp_all)
  apply(rule_tac A2="X" in sym[OF trans[OF cp_charn]], simp_all)
  apply(simp)
  done


lemma iterate_distrib_union:
  assumes    mtC: "\<And> B. g B A \<tau> = B \<tau>"
  and      stepC: "\<And> B x y. g B (P x y) \<tau> = P x (g B y) \<tau>"
  and      cp2_g: "\<And> x. cp (g x)"
  and      cp2_P: "\<And> x. cp (P x)"
  and       defX: "\<tau> \<MathOclValid> \<MathOclIsDefined> X"
  and       defY: "\<tau> \<MathOclValid> \<MathOclIsDefined> Y"
  shows "OclIterate ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) 
                    (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A \<tau> =
         g (OclIterate X P A) (OclIterate Y P A) \<tau>"
  by(simp add: iterate_of_union defX defY cp2_g cp2_P
               sym[OF iterate_fusion_opt] mtC stepC)


lemma iterate_distrib_union_opt:
  assumes undefC: "\<And> B. g B \<MathOclUndefined> = \<MathOclUndefined>"
  and        mtC: "\<And> B. g B A = B"
  and      stepC: "\<And> B x y. g B (P x y) = P x (g B y)"
  and  undef_2_P: "\<And> x. P x \<MathOclUndefined> = \<MathOclUndefined>"
  and      cp2_g: "\<And> x. cp (g x)"
  and      cp2_P: "\<And> x. cp (P x)"
  shows "OclIterate ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) 
                    (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A =
         g (OclIterate X P A) (OclIterate Y P A)"
  by(rule ext, simp add: iterate_of_union_opt undef_2_P cp2_g cp2_P
                         sym[OF iterate_fusion] undefC mtC stepC)


subsection{* And the corresponding forward rules for \emph{unary} functions*}

lemma f1_by_iterate_distrib_union:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and         mtC: "\<And> B. g B A \<tau> = B \<tau>"
  and       stepC: "\<And> B x y. g B (P x y) \<tau> = P x (g B y) \<tau>"
  and       cp2_g: "\<And> x. cp (g x)"
  and       cp2_P: "\<And> x. cp (P x)"
  and        defX: "\<tau> \<MathOclValid> \<MathOclIsDefined> X"
  and        defY: "\<tau> \<MathOclValid> \<MathOclIsDefined> Y"
  shows "f (X \<MathOclUnion> Y) \<tau> = g (f X) (f Y) \<tau>"
  by(simp add: f_by_it iterate_distrib_union defX defY cp2_g cp2_P mtC stepC)

lemma f1_by_iterate_distrib_union_opt:
  assumes f_by_it: "\<And> S. (f (S::('\<tau>,'\<alpha>::bot)Sequence)) = 
                         (OclIterate S (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) A)"
  and      undefC: "\<And> B. g B \<MathOclUndefined> = \<MathOclUndefined>"
  and         mtC: "\<And> B. g B A = B"
  and       stepC: "\<And> B x y. g B (P x y) = P x (g B y)"
  and   undef_2_P: "\<And> x. P x \<MathOclUndefined> = \<MathOclUndefined>"
  and       cp2_g: "\<And> x. cp (g x)"
  and       cp2_P: "\<And> x. cp (P x)"
  shows "f (X \<MathOclUnion> Y) = g (f X) (f Y)"
  by(simp add: f_by_it iterate_distrib_union_opt
               undef_2_P cp2_g cp2_P undefC mtC stepC)


section{* Append *}

defs
  (* append     :: "[('\<tau>,('\<alpha>::bot) ) Sequence, ('\<tau>,'\<alpha>::bot ) VAL] => ('\<tau>,'\<alpha> ) Sequence" *)
  OclAppend_def :
  "OclAppend    \<equiv> lift2 (strictify (\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                  (\<HolOclLiftLeft>(concat [\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>, [e] ])\<HolOclLiftRight>))))"

ocl_setup_op [OclAppend]

lemma append_including_UC [simp]:
  "OclAppend (S::('\<tau>,'\<alpha>::bot)Sequence) (x::('\<tau>,'\<alpha>)VAL) = S \<MathOclIncluding> x"
  by(simp add: OclAppend_def OclIncluding_def)


section{* Prepend *}
defs
  (* prepend     :: "[('\<tau>,('\<alpha>::bot) ) Sequence, ('\<tau>,'\<alpha>::bot ) VAL] => ('\<tau>,'\<alpha> ) Sequence" *)
  OclPrepend_def :
  "OclPrepend    \<equiv> lift2 (strictify (\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                   (\<HolOclLiftLeft>e # \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>\<HolOclLiftRight>))))"

ocl_setup_op [OclPrepend]


(* TODO: really simp? currently it destroys more than it brings\<dots>
         but it converts into the canonical representation and thats what we want *)
lemma prepend_union_UC [simp]:
  "OclPrepend (X::('\<tau>,'\<alpha>::bot)Sequence) (a::('\<tau>,'\<alpha>)VAL) = ((\<MathOclMtSequence> \<MathOclIncluding> a) \<MathOclUnion> X)"
  apply(rule_tac X="X" in Sequence_sem_cases_ext)
  apply(simp_all)
  apply(simp_all add: OclPrepend_def OclIncluding_def OclUnion_def
                      OclMtSequence_def ss_lifting' neq_commute)
  done

lemma test'': "(OclPrepend (OclPrepend (X::('\<tau>,'\<alpha>::bot)Sequence) (a::('\<tau>,'\<alpha>)VAL)) (a::('\<tau>,'\<alpha>)VAL)) = bla"
  apply(simp) oops


(* NOTE: could be possible to derive them automatically from the rules about union *)
lemma OCL_is_def_OclPrepend:
  "\<MathOclIsDefined>(OclPrepend (X::('\<tau>,'\<alpha>::bot)Sequence) (Y::('\<tau>,'\<alpha>::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Sequence_Val[OF OclPrepend_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclPrepend:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclPrepend (X::('\<tau>,'\<alpha>::bot)Sequence) (Y::('\<tau>,'\<alpha>::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Sequence_Val[OF OclPrepend_def])
  apply(simp)
  done

subsection{* Some sort of case exhaustion over prepend *}

lemma prepend_charn1:
  "\<lbrakk>\<bottom> \<notin> set xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> 
   \<exists> (S::'\<alpha>::bot Sequence_0) a. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>) = (OclPrepend (lift0 S) ( (lift0 (a::'\<alpha>)) )) \<tau>"
  apply(cases xs, simp_all)
  apply(rule_tac x="Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>" in exI)
  apply(rule_tac x="a" in exI)
  apply(auto simp: OclMtSequence_def OclIncluding_def OclUnion_def ss_lifting')
  done

lemma prepend_charn3:
  "\<tau> \<MathOclValid> OclNotEmpty (S::('\<tau>,'\<alpha>::bot)Sequence) \<Longrightarrow>  
   \<exists> (S'::('\<tau>,'\<alpha>::bot)Sequence) a. \<tau> \<MathOclValid> (S \<MathOclStrongEq> (OclPrepend S' (a::('\<tau>,'\<alpha>)VAL)))"
   oops

lemma prepend_charn2:
  "\<lbrakk>\<bottom> \<notin> set xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> 
   \<exists> (S::('\<tau>,'\<alpha>::bot)Sequence) a. Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight> = (OclPrepend S (a::('\<tau>,'\<alpha>)VAL)) \<tau>"
  apply(cases xs, simp_all)
  apply(rule_tac x="lift0 (Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>)" in exI)
  apply(rule_tac x="lift0 a" in exI)
  apply(auto simp: OclMtSequence_def OclIncluding_def OclUnion_def ss_lifting')
  done


subsection{* Reverse induction on sequences: induction over prepend *}

lemma Sequence_induct_prepend_0:
  assumes undefCase : "P \<bottom>"
  and     mtCase    : "P (\<MathOclMtSequence> \<tau>)"
  and     stepCase  : "\<And> (S::'\<tau>::bot Sequence_0) (x::'\<tau>::bot). 
                       \<lbrakk>P S; DEF S; DEF x\<rbrakk> \<Longrightarrow> (P ((OclPrepend (lift0 S) (lift0 (x::'\<tau>))) \<tau>))"
  shows               "P (S::'\<tau>::bot Sequence_0)"
  apply(rule Abs_Sequence_0_induct_charn, rule undefCase)
  apply(erule rev_mp)
  apply(induct_tac y)
  apply(insert mtCase, simp add: OclMtSequence_def lift0_def)
  apply(auto)
  apply(rotate_tac -1, drule_tac x="a" and S="Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>" in stepCase)
  apply(erule DEF_Abs_Sequence)
  apply(simp_all add: OclPrepend_def DEF_def neq_commute
                      lift0_def lift2_def strictify_def)
  done

lemma Sequence_induct_prepend:
  assumes undefCase : "P \<MathOclUndefined> \<tau>"
  and mtCase    : "P \<MathOclMtSequence> \<tau>"
  and stepCase  : "\<And> (S::('\<tau>,'\<alpha>::bot)Sequence) (x::('\<tau>,'\<alpha>::bot)VAL). 
                       \<lbrakk> P S \<tau>; \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> P (OclPrepend S (x::('\<tau>,'\<alpha>)VAL)) \<tau>"
  and cpP       : "cp P"
  shows               "P (S::('\<tau>,'\<alpha>::bot)Sequence) \<tau>"
  apply(insert cpP)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(erule_tac x="S" in allE)
  apply(erule_tac x="\<tau>" in allE)
  apply(simp)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(rule_tac \<tau>="\<tau>" in Sequence_induct_prepend_0)
  apply(insert undefCase mtCase)
  apply(simp_all add: OclUndefined_def lift0_def)
  apply(drule stepCase)
  apply(simp_all add: localValidDefined2sem lift0_def)
  done


subsection{* iterate on prepend *}

lemma iterate_of_prepend:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau> \<Rightarrow> '\<alpha>::bot))"
  and    cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and    cpP_2: "\<And> x. cp (P x)"
  shows "((OclIterate (OclPrepend S (x::('\<tau>,'\<alpha>)VAL)) P A) \<tau>) =
         ((OclIterate (S::(('\<tau>,'\<alpha>::bot)Sequence)) P (P x A)) \<tau>)"
  apply(insert defS defx cpP_1 cpP_2, simp)
  apply(subst iterate_of_union) 
  apply(simp add: OCL_is_defopt_OclIncluding)
  apply(rule_tac A="(OclIterate \<MathOclMtSequence> \<MathOclIncluding> x  P A)" in cp_charn)
  apply(simp_all add: iterate_of_including)
  done




section{* Includes and excludes *}

defs 
   (* includes     :: "[('\<tau>,'\<alpha>::bot) Sequence, ('\<tau>,'\<alpha>) VAL] => '\<tau> Boolean" *)
   OclIncludes_def : "OclIncludes \<equiv> lift2(strictify (\<lambda>X. strictify (\<lambda>x.
                                      \<HolOclLiftLeft>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>)))"

   (* excludes     :: "[('\<tau>,'\<alpha>::bot) Sequence, ('\<tau>,'\<alpha>) VAL] => '\<tau> Boolean" *)
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                      \<HolOclLiftLeft>x \<notin> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>)))"

ocl_setup_op [OclIncludes, OclExcludes]


lemma includes_charn1 :
  "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> ((x \<tau>) \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>)"
  apply(case_tac "(S \<tau>) \<noteq> \<bottom>")
  apply(case_tac "(x \<tau>) \<noteq> \<bottom>")
  apply(auto simp: OclIncludes_def localValid2sem ss_lifting')
  done

lemma includes_charn2 :
  "\<lbrakk> DEF(S \<tau>); (x \<tau>) \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight> \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
  apply(unfold DEF_def, frule DEF_X_Sequence')
  apply(auto simp: OclIncludes_def localValid2sem ss_lifting')
  done


subsection{* Higher Properties of includes and excludes *}


(* We prefer a normal form where excludes is eliminated. *)
lemma exludes_not_includes[simp]: 
  "x \<MathOclExcludes> X = \<MathOclNot> (x::('\<tau>, '\<alpha>)VAL) \<MathOclIncludes> (X::('\<tau>, '\<alpha>::bot) Sequence)"
  apply(rule ext)
  apply(simp add: OclExcludes_def OclIncludes_def OclNot_def
                  ss_lifting') 
  done

(* TODO: are really all cases needed? and in what form: 
         Valid or LocalValid? *)


text{* Currently a non-strict body is taken although a strict one
  is possible. This should be corrected as the experience has shown
  that strict functions are much nicer to work with. *}

lemma includes_by_iterate:
  "a \<MathOclIncludes> (X::('\<tau>,'\<alpha>::bot)Sequence) = 
   OclIterate X (\<lambda> (x::('\<tau>,'\<alpha>::bot)VAL) y. (x \<MathOclStrictEq> a) \<MathOclSor> y)
                (\<MathOclIf> (\<MathOclIsDefined> a) \<MathOclThen> \<MathOclFalse> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif>)"
  apply(rule ext)
  apply(rule iterate_universal, simp_all)
  apply(simp add: OclIncludes_def OclMtSequence_def OclFalse_def 
                  OclIf_def OclIsDefined_def OclUndefined_def 
                  ss_lifting')
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(auto simp: OclIncludes_def OclIncluding_def OclTrue_def
                   OclSor_def OclNot_def OclSand_def OclStrictEq_def
                   ss_lifting' neq_commute)
  done

lemmas includes_mtSequence [simp] = f1_by_iterate_mtSequence[OF includes_by_iterate]
thm f1_by_iterate_opt_including[OF includes_by_iterate, simplified]

lemma excludes_by_iterate:
  "a \<MathOclExcludes> (X::('\<tau>,'\<alpha>::bot)Sequence) = 
   OclIterate X (\<lambda> (x::('\<tau>,'\<alpha>::bot)VAL) y. (x `<>` a) \<MathOclSand> y)
                (\<MathOclIf> (\<MathOclIsDefined> a) \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif>)"
  apply(rule ext, simp add: includes_by_iterate)
  apply(rule iterate_fusion, simp_all)
  apply(simp_all add: OclNot_def OclTrue_def OclFalse_def OclIf_def
                      OclSand_def OclSor_def OclUndefined_def ss_lifting')
  done

text{* containment in the empty list can be reduced to definedness *}
lemma includes_of_mtSequence[simp]:
  "\<not> (\<tau> \<MathOclValid> (x::('\<tau>, '\<alpha>)VAL) \<MathOclIncludes> (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence))"
  by(simp add: OclIncludes_def OclMtSequence_def OclNot_def 
               localValid2sem ss_lifting')

(* TODO: would it be better to use an equality as premise?
   rather not: pattern matching should be enough if everything
   is in normal form *)
lemma includes_charn1_including[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> a \<MathOclIncludes> (X \<MathOclIncluding> a)"
  apply(simp only: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(simp add: localValid2sem OclIncludes_def OclIncluding_def
                  ss_lifting' neq_commute)
  done

lemma includes_charn2_including[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (b::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> a \<MathOclIncludes> (X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> a \<MathOclIncludes> (X \<MathOclIncluding> b)"
  apply(case_tac "X \<tau> \<noteq> \<bottom>")
  apply(frule DEF_X_Sequence', clarify)
  apply(simp_all only: localValidDefined2sem DEF_def)
  apply(simp_all add: localValid2sem OclIncludes_def OclIncluding_def
                  ss_lifting' neq_commute)
  apply(drule neq_commute[THEN iffD1], simp)
  done


(* @BU: NOTE: calling ocl_hypsubst with assumptions that don't concern the current context or if
              there is no current context breaks it. *)
lemma includes_of_collectionRange:
  "(x::('\<tau> Integer)) \<MathOclIncludes> ((OclCollectionRange a b)::('\<tau>,Integer_0)Sequence) = 
   ((a::('\<tau> Integer)) \<MathOclLe> x) \<MathOclSand> (x \<MathOclLe> (b::('\<tau> Integer)))"
  apply(rule ext)
  apply(case_tac "DEF (x xa)", case_tac "DEF (a xa)", case_tac "DEF (b xa)")
  apply(simp add: DEF_def_both, clarify)
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a xa)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b xa)\<HolOclDropRight> - \<HolOclDropLeft>(a xa)\<HolOclDropRight>) + 1))])")
  apply(auto simp: OclIncludes_def OclCollectionRange_def OclLe_def
                   OclSand_def ss_lifting', arith)
  apply(rule_tac x="nat(\<HolOclDropLeft>(x xa)\<HolOclDropRight> - \<HolOclDropLeft>(a xa)\<HolOclDropRight>)" in image_eqI, auto)
  done

lemma includes_of_union [simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> Y \<rbrakk> \<Longrightarrow> 
   ((a::('\<tau>,'\<alpha>::bot)VAL) \<MathOclIncludes> ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y)) \<tau> = 
   ((a \<MathOclIncludes> X) \<MathOclOr> (a \<MathOclIncludes> Y)) \<tau>"
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence')
  apply(rotate_tac 1, frule DEF_X_Sequence', clarify)
  apply(simp add: OclIncludes_def OclUnion_def OclOr_def
                  OclAnd_def OclNot_def ss_lifting')
  done


lemma test''': "\<tau> \<MathOclValid> (1::'\<tau> Integer) \<MathOclIncludes> ((mkSequence[1::'\<tau> Integer])::('\<tau>, Integer_0)Sequence)"
 oops
(* TODO: type problems, reasoning problems? what the heck is going wrong *)


lemma not_includes_of_mtSequence:
  "(\<tau> \<MathOclValid> \<MathOclNot> (x::('\<tau>, '\<alpha>)VAL) \<MathOclIncludes> (\<MathOclMtSequence>::('\<tau>, '\<alpha>::bot) Sequence)) =  (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
  by(simp add: localValidNot2sem localValidDefined2sem OclIncludes_def 
               OclMtSequence_def ss_lifting')



(* THIS was the old theorem as it is stated in the set theory *)
lemma mtSequence_Includes_of_OLD :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> 
    \<tau> \<MathOclValid> (x::('\<tau>, '\<alpha>)VAL) \<MathOclExcludes> (\<MathOclMtSequence>::('\<tau>, '\<alpha>::bot) Sequence)"
  by(simp, ocl_subst, simp)

text{* Note: this will be subsumed by OCL\_is\_defopt\_OclIncludes 
  if \texttt{ocl\_setup\_op} is working correctly on collection types  *}

(*

subsection{* Consequences of Smashedness of Sequences *}

text{* This is the characteristic property of smashed sequences:
       if \verb+X ->includes(x)+ is valid, we know that x must be defined. *}

lemma smashed_sequences_charn:
  "\<tau> \<MathOclValid> ((x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (X::('\<tau>,'\<alpha>::bot)Sequence)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> x"
  apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                   strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
  apply(case_tac "X \<tau> = \<bottom>", auto)
  apply(simp add: strictify_def DEF_def)
  apply(case_tac "x \<tau> = \<bottom>", auto)
  done

*)



section{* Emptiness *}

defs
   (* isEmpty      :: "('\<tau>,'\<alpha>::bot)Sequence => '\<tau> Boolean" *)
   OclIsEmpty_def  :  
   "OclIsEmpty     \<equiv> lift1 (strictify (\<lambda>X. \<HolOclLiftLeft>(\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> = [])\<HolOclLiftRight>))"

 
   (* notEmpty     :: "('\<tau>,'\<alpha>::bot)Sequence => '\<tau> Boolean" *)
   OclNotEmpty_def :  
   "OclNotEmpty    \<equiv> lift1 (strictify (\<lambda>X. \<HolOclLiftLeft>(\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<noteq> [])\<HolOclLiftRight>))"

ocl_setup_op [OclIsEmpty, OclNotEmpty]

subsection{* The normal form for emptiness testing: *}

text{* we prefer a canical form were notEmpty is written using isEmpty *}
lemma notEmpty_not_isEmpty_conv[simp]:
  "\<MathOclNotEmpty>(X::('\<tau>, '\<alpha>::bot) Sequence) =  \<MathOclNot> (\<MathOclIsEmpty> X)" 
  apply(rule ext)
  apply(simp add: OclNotEmpty_def OclIsEmpty_def OclNot_def
                  ss_lifting')
  done

text{* further on the emptiness test is reduced to an equality to a constant *}
lemma isEmpty_stricteq_mtSequence_conv[simp]:
  "(\<MathOclIsEmpty> X) = ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclStrictEq> \<MathOclMtSequence>)"
  apply(rule ext)
  apply(case_tac "X x \<noteq> \<bottom>")
  apply(frule DEF_X_Sequence', clarify)
  apply(simp_all add: OclIsEmpty_def OclMtSequence_def OclStrictEq_def
                  Abs_Sequence_0_inject_charn ss_lifting')
  done

subsection{* Properties of emptiness: *}

lemma isEmpty_of_mtSequence: 
  "\<tau> \<MathOclValid> \<MathOclIsEmpty> \<MathOclMtSequence>" 
  by (simp)

lemma isEmpty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclNot> (\<MathOclIsEmpty>(X \<MathOclIncluding> a))"
  by(simp)
  
lemma isEmpty_union [simp]:
  "(((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) \<MathOclStrictEq> \<MathOclMtSequence>) = (X \<MathOclStrictEq> \<MathOclMtSequence>) \<MathOclSand> (Y \<MathOclStrictEq> \<MathOclMtSequence>)"
  apply(rule_tac X=X in Sequence_sem_cases_ext)
  apply(rule_tac X=Y in Sequence_sem_cases, simp_all)
  apply(simp_all add: OclUnion_def OclStrictEq_def OclMtSequence_def OclSand_def 
                      ss_lifting' Abs_Sequence_0_inject_charn)
  done

lemma notEmpty_of_mtSequence:
  "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclNotEmpty> \<MathOclMtSequence>)"
  by(simp)

lemma notEmpty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclNotEmpty> (X \<MathOclIncluding> a)"
  by(simp)


section{* Size *}

defs
   (* size      :: "('\<tau>,'\<alpha>::bot )Sequence => '\<tau> Integer" *)
   OclSize_def  : "OclSize \<equiv> lift1 (strictify (\<lambda>X. 
                             \<HolOclLiftLeft>int (Nat.size \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>)\<HolOclLiftRight>))"


ocl_setup_op [OclSize]



lemma size_by_iterate:
  "((self::('\<tau>,'\<alpha>::bot)Sequence) ->size()) = 
    OclIterate self (\<lambda> (x::('\<tau>,'\<alpha>::bot)VAL) (y::'\<tau> Integer). y + 1) 0"
  apply(rule ext)
  apply(rule iterate_universal)
  apply(simp_all)
  apply(simp_all add: OclSize_def OclIterate_def plus_def OclMtSequence_def
                   OclUndefined_def OclIncluding_def localValidDefined2sem
                   OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def
                   ss_lifting')
  apply(frule DEF_X_Sequence')
  apply(auto simp: neq_commute)
  done

lemmas size_of_mtSequence[simp] = f1_by_iterate_mtSequence[OF size_by_iterate]

lemma size_of_including[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>::bot)VAL) \<rbrakk> \<Longrightarrow> 
   (\<MathOclSizeLeft>(S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> x \<MathOclSizeRight>) \<tau> = (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1) \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S")
  apply(ocl_hypsubst, simp)
  apply(simp add: f1_by_iterate_including[OF size_by_iterate, simplified])
  done

(* Simons Version: 
lemma size_eq_zero_isEmtpy_UC[simp]:
  "(\<tau> \<MathOclValid> \<MathOclSizeLeft>(S::('\<tau>,'\<alpha>::bot)Sequence)\<MathOclSizeRight> \<MathOclStrictEq> 0) = (\<tau> \<MathOclValid> S \<MathOclStrictEq> \<MathOclMtSequence>)"
  apply(case_tac "S \<tau> \<noteq> \<bottom>")
  apply(frule DEF_X_Sequence', clarify)
  apply(simp_all add: OclSize_def OclStrictEq_def OclMtSequence_def
                      OCL_Integer.Zero_ocl_int_def Abs_Sequence_0_inject_charn
                      localValid2sem ss_lifting')
  done
 *)
lemma size_eq_zero_isEmtpy_UC[simp]:
  "(\<MathOclSizeLeft>(S::('\<tau>,'\<alpha>::bot)Sequence)\<MathOclSizeRight> \<MathOclStrictEq> 0) = (S \<MathOclStrictEq> \<MathOclMtSequence>)"
  apply(rule ext)
  apply(case_tac "S x \<noteq> \<bottom>")
  apply(frule DEF_X_Sequence', clarify)
  apply(simp_all add: OclSize_def OclStrictEq_def OclMtSequence_def
                      OCL_Integer.Zero_ocl_int_def Abs_Sequence_0_inject_charn
                      localValid2sem ss_lifting')
  done


lemma size_of_union_UC [simp]:
  "(\<MathOclSizeLeft>((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y)\<MathOclSizeRight>) = (\<MathOclSizeLeft>X\<MathOclSizeRight> + \<MathOclSizeLeft>Y\<MathOclSizeRight>)"
  apply(rule f1_by_iterate_distrib_union_opt[OF size_by_iterate, of "op +", simplified])
  apply(rule ext, simp add: plus_def Zero_ocl_int_def ss_lifting')
  apply(rule ext, simp add: plus_def One_ocl_int_def ss_lifting')
  done



subsection{* Induction over two lists *}

lemma Sequence_induct2_including_0:
  assumes undefCase: "P \<bottom> \<bottom>"
  and     equalSize: "\<MathOclSizeLeft>(lift0 X)\<MathOclSizeRight> \<tau> = \<MathOclSizeLeft>(lift0 Y)\<MathOclSizeRight> \<tau>"
  and        mtCase: "P (\<MathOclMtSequence> \<tau>) (\<MathOclMtSequence> \<tau>)"
  and      stepCase: "\<And> (X::'\<tau>::bot Sequence_0) (a::'\<tau>::bot) (Y::'\<alpha>::bot Sequence_0) (b::'\<alpha>::bot). 
                      \<lbrakk> P X Y; DEF X; DEF a; DEF Y; DEF b\<rbrakk> \<Longrightarrow> 
                      (P ((OclIncluding (lift0 X) (lift0 a)) \<tau>) ((OclIncluding (lift0 Y) (lift0 b)) \<tau>))"
  shows "P (X::'\<tau>::bot Sequence_0) (Y::'\<alpha>::bot Sequence_0)"
  apply(insert equalSize, erule rev_mp)
  apply(rule Abs_Sequence_0_induct_charn)
  apply(simp add: OclSize_def ss_lifting' undefCase)
  apply(erule rev_mp)+
  apply(rule_tac x="X" in Abs_Sequence_0_induct_charn)
  apply(simp add: OclSize_def ss_lifting' undefCase)
  apply(rule impI)+
(* TODO: is there a nicer solution than this? *)
  apply(frule rev_mp) prefer 2 apply(assumption, rotate_tac 1)
  apply(frule rev_mp) prefer 2 apply(assumption, rotate_tac 1)
  apply(frule rev_mp) prefer 2 apply(assumption, rotate_tac 1)
  apply(rule list_rev_induct2)
  apply(insert mtCase)
  apply(auto simp: OclSize_def OclMtSequence_def ss_lifting')
  apply(rotate_tac -1, drule_tac a="x" and X="Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>" and
                                 b="yb" and Y="Abs_Sequence_0 \<HolOclLiftLeft>ys\<HolOclLiftRight>" in stepCase)
  apply(simp_all add: OclIncluding_def neq_commute ss_lifting')
  done
  

lemma Sequence_induct2_including:
  assumes equalSize: "\<tau> \<MathOclValid> \<MathOclSizeLeft>X\<MathOclSizeRight> \<MathOclStrongEq> \<MathOclSizeLeft>Y\<MathOclSizeRight>"
  and     undefCase: "P \<MathOclUndefined> \<MathOclUndefined> \<tau>"
  and        mtCase: "P \<MathOclMtSequence> \<MathOclMtSequence> \<tau>"
  and      stepCase: "\<And> (X::('\<tau>,'\<alpha>::bot)Sequence) (a::('\<tau>,'\<alpha>::bot)VAL) 
                        (Y::('\<tau>,'c::bot)Sequence) (b::('\<tau>,'c::bot)VAL). 
                      \<lbrakk> P X Y \<tau>; \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> a; \<tau> \<MathOclValid> \<MathOclIsDefined> Y; \<tau> \<MathOclValid> \<MathOclIsDefined> b\<rbrakk> \<Longrightarrow> 
                       P (OclIncluding X a) (OclIncluding Y b) \<tau>"
  and         cp1_P: "\<And> y. cp (\<lambda> x. P x y)"
  and         cp2_P: "\<And> x. cp (P x)"
  shows              "P (X::('\<tau>,'\<alpha>::bot)Sequence) (Y::('\<tau>,'c::bot)Sequence) \<tau>"
  apply(insert cp1_P cp2_P, simp add: cp_by_cpify)
  apply(insert cp1_P cp2_P)
  apply(rule subst, rule sym, rule_tac x="X" in cp_subst, simp)
  apply(rule subst, rule sym, rule_tac x="Y" in cp_subst, simp)
  apply(rule_tac \<tau>="\<tau>" in Sequence_induct2_including_0)
  apply(simp add: lift0_def)
  oops
(* solve if ocl_simp funktioniert *)
(*
  apply(
  apply(rule refl)
  apply(insert undefCase mtCase, simp_all add: OclUndefined_def)
  apply(rule subst, rule_tac x="X" in cp_subst, simp)
  apply(rule subst, rule_tac x="Y" in cp_subst, simp)
  apply(insert equalSize, simp add: localValid2sem OclStrongEq_def lift2_def)
  apply(drule stepCase)
  apply(simp_all add: localValidDefined2sem lift0_def)
  done
*)


section{* Count *}

defs
   (* count     :: "[('\<tau>,'\<alpha>::bot )Sequence, ('\<tau>,'\<alpha>)VAL] => '\<tau> Integer" *)
   OclCount_def : "OclCount \<equiv> lift2 (strictify   (\<lambda>X. strictify (\<lambda>x. 
                              \<HolOclLiftLeft>int (Nat.size(List.filter (op = x) \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>))\<HolOclLiftRight>)))"

ocl_setup_op [OclCount]


subsection{* Its computational characterisation by iterate *}

lemma count_by_iterate:
  "((S::('\<tau>,'\<alpha>::bot)Sequence) ->count((a::('\<tau>,'\<alpha>)VAL))) = 
   (OclIterate S
               (\<lambda> (x::('\<tau>,'\<alpha>::bot)VAL) (y::'\<tau> Integer).\<MathOclIf>  (x \<MathOclStrictEq> a)\<MathOclThen> (y+1) \<MathOclElse> (y) \<MathOclEndif> )   
               (\<MathOclIf> (\<MathOclIsDefined> a) \<MathOclThen> 0 \<MathOclElse>\<MathOclUndefined> \<MathOclEndif> ) )"
  apply(rule ext)
  apply(rule iterate_universal)
  apply(simp_all)
  apply(simp_all add: OclMtSequence_def OclCount_def
                  OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def
                  OclIf_def OclIsDefined_def OclUndefined_def OclStrictEq_def
                  OclIncluding_def OclLocalValid_def plus_def OclTrue_def
                  ss_lifting')
  apply(frule DEF_X_Sequence')
  apply(auto simp: neq_commute)  
  done

thm f1_by_iterate_including[OF count_by_iterate, simplified]

lemma count_of_mtSequence [simp]:
  "\<tau> \<MathOclValid> (\<MathOclIsDefined> (x::('\<tau>,'\<alpha>)VAL)) \<Longrightarrow> 
   ((\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclCount> x ) \<tau> = 0 \<tau>"
  by(simp add: count_by_iterate, ocl_subst, simp)

thm f1_by_iterate_including[OF count_by_iterate, simplified]
thm f1_by_iterate_opt_including[OF count_by_iterate, simplified]
thm f1_by_iterate_distrib_union[OF count_by_iterate, of "op +", simplified]
thm f1_by_iterate_distrib_union_opt[OF count_by_iterate, of "op +", simplified]

(* TODO: add simplification in the last rule to remove the last assumption *)

lemma plus_assoc1: "((X::('\<tau> Integer)) + Y) + Z = X + (Y + Z)"
  by(rule ext, simp add: plus_def ss_lifting')

lemma plus_assoc2: "(X::('\<tau> Integer)) + (Y + Z) = Y + (X + Z)"
  by(rule ext, simp add: plus_def ss_lifting')

lemmas plus_AC = plus_commute plus_assoc1 plus_assoc2

lemma plus_zero' [simp]: "(X::('\<tau> Integer)) + 0 = X \<and> 0 + (X::('\<tau> Integer)) = X"
  by(rule conjI, (rule ext, simp add: plus_def ss_lifting' Zero_ocl_int_def)+)



(* TODO: move *)

lemma count_of_union [simp]:
  "(((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y ) \<MathOclCount> (a::('\<tau>,'\<alpha>)VAL)) = ((X \<MathOclCount> a ) + (Y \<MathOclCount> a ))"
  apply(rule ext, case_tac "x  \<MathOclValid> \<MathOclIsUndefined> a", ocl_hypsubst, simp)
  apply(case_tac "x  \<MathOclValid> \<MathOclIsUndefined> X", ocl_hypsubst, simp)
  apply(case_tac "x  \<MathOclValid> \<MathOclIsUndefined> Y", ocl_hypsubst, simp)
  apply(simp add: count_by_iterate)
  apply(ocl_subst)
  apply(simp add: iterate_distrib_union plus_AC if_distrib_strict_alt)
  done


lemma includes_by_count_UC[simp]:
  "S \<MathOclCount> x \<MathOclGreater> 0 = (x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Sequence)"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclIncludes_def OclCount_def OclGreater_def filter_empty_conv
                   OclLess_def OclNot_def Zero_ocl_int_def ss_lifting')
  done

lemma excludes_by_count_UC[simp]:
  "S \<MathOclCount> x \<MathOclStrictEq> 0 = \<MathOclNot> ((x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Sequence))"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclIncludes_def OclCount_def OclStrictEq_def OclNot_def
                   Zero_ocl_int_def ss_lifting' filter_empty_conv)
  done

(* TODO: really simp? because this could destroy assumptions needed in
         an ocl_hypsubst wherever we calculate with OclCount values *)

section{* Subsequence *}

constdefs
   OclSubSequence :: "[('\<tau>,'\<alpha>::bot)Sequence,'\<tau> Integer,'\<tau> Integer]=>('\<tau>,'\<alpha> ) Sequence"     


  "OclSubSequence \<equiv> lift3 (strictify (\<lambda>S. strictify (\<lambda>l. strictify (\<lambda>u. Abs_Sequence_0
                     (\<HolOclLiftLeft>take (nat (\<HolOclDropLeft>u\<HolOclDropRight> - \<HolOclDropLeft>l\<HolOclDropRight> + 1))
                        (List.drop (nat (\<HolOclDropLeft>l\<HolOclDropRight> - 1))
                          (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>))\<HolOclLiftRight>)))))"

ocl_setup_op [OclSubSequence]

(* @BU: will ocl_setup_op work once with lift3 ?? *)
text{* Because the \texttt{ocl\_setup\_op} method doesn't deal with
  \texttt{lift3} we have to setup the simple properties by ourselves. *}

lemma OCL_undef_1_OclSubsequence[simp]:
  "OclSubSequence (\<MathOclUndefined>::('\<tau>,'\<alpha>::bot)Sequence) a b = \<MathOclUndefined>"
  by(simp add: OclUndefined_def OclSubSequence_def ss_lifting')

lemma OCL_undef_2_OclSubsequence[simp]:
  "OclSubSequence (S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUndefined> b = \<MathOclUndefined>"
  by(rule ext, simp add: OclUndefined_def OclSubSequence_def ss_lifting')

lemma OCL_undef_3_OclSubsequence[simp]:
  "OclSubSequence (S::('\<tau>,'\<alpha>::bot)Sequence) a \<MathOclUndefined> = \<MathOclUndefined>"
  by(rule ext, simp add: OclUndefined_def OclSubSequence_def ss_lifting')

lemma OCL_cp_OclSubsequence[simp,intro!]:
  "\<lbrakk> cp P; cp F; cp F' \<rbrakk> \<Longrightarrow>
   cp (\<lambda> X. OclSubSequence ((P X)::('\<tau>,'\<alpha>::bot)Sequence) (F X) (F' X))"
  by(simp add: OclSubSequence_def)

subsection{* Higher properties of subsequence *}

text{* Note: \texttt{subsequence} is in our case not undefined except for 
  undefined arguments. To avoid undefinedness is quite favorable in general as it 
  makes proofs simpler and reduces the side conditions to check. 
  But our implementation still fullfills the requirements of the standard.
 *}

lemma size_of_subsequence:
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> a; \<tau> \<MathOclValid> a \<MathOclLe> b - 1; \<tau> \<MathOclValid> b \<MathOclLe> \<MathOclSizeLeft>S\<MathOclSizeRight> \<rbrakk> \<Longrightarrow> 
   \<MathOclSizeLeft>(OclSubSequence (S::('\<tau>,'\<alpha>::bot)Sequence) a b)\<MathOclSizeRight>  \<tau> = (b - a + 1) \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValid2sem)
  apply(case_tac "b \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "a \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclIncluding_def OclSubSequence_def OclSize_def
                   One_ocl_int_def plus_def minus_def ss_lifting'
                   OclLe_def OclLess_def notin_set_take notin_set_drop)
  apply(arith, case_tac "((b \<tau>) = down)", simp_all)
  done


lemma subsequence_mtSequence_conv:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> b \<MathOclLess> a \<rbrakk> \<Longrightarrow>
   (OclSubSequence (S::('\<tau>,'\<alpha>::bot)Sequence) a b) \<tau> = \<MathOclMtSequence> \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(case_tac "b \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "a \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclIncluding_def OclSubSequence_def OclMtSequence_def localValid2sem
                   OclLess_def notin_set_take notin_set_drop ss_lifting')
  done


section{* InsertAt *}

defs

  (* insertAt      ::  "[('\<tau>,'\<alpha>::bot)Sequence,'\<tau> Integer,('\<tau>,'\<alpha>)VAL] => ('\<tau>,'\<alpha> )Sequence" *)
  OclInsertAt_def  :
  "OclInsertAt     \<equiv> lift3(strictify(\<lambda>S. strictify(\<lambda>i. strictify(\<lambda>x. Abs_Sequence_0(
                     if (1 <= \<HolOclDropLeft>i\<HolOclDropRight> \<and>  \<HolOclDropLeft>i\<HolOclDropRight> <= (1+int(Nat.size\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)))
                     then (\<HolOclLiftLeft>concat [take (nat (op - \<HolOclDropLeft>i\<HolOclDropRight> 1)) (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>), [x],
                                    List.drop (nat (op - \<HolOclDropLeft>i\<HolOclDropRight> 1)) (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)]\<HolOclLiftRight>)
                     else \<bottom> )))))"


ocl_setup_op [OclInsertAt]

(* @BU: what about ocl_setup_op and lift3? *)
text{* Because the \texttt{ocl\_setup\_op} method doesn't deal with
  \texttt{lift3} we have to setup the simple properties by ourselves. *}

lemma OCL_cp_OclInsertAt[simp,intro!]:
  "\<lbrakk> cp P; cp F; cp F' \<rbrakk> \<Longrightarrow>
   cp (\<lambda> X. OclInsertAt ((P X)::('\<tau>,'\<alpha>::bot)Sequence) (F X) ((F' X)::('\<tau>,'\<alpha>)VAL))"
  by(simp add: OclInsertAt_def)

lemma OCL_undef_1_OclInsertAt[simp]:
  "OclInsertAt (\<MathOclUndefined>::('\<tau>,'\<alpha>::bot)Sequence) i (x::('\<tau>,'\<alpha>)VAL) = \<MathOclUndefined>"
  by(simp add: OclUndefined_def OclInsertAt_def ss_lifting')

lemma OCL_undef_2_OclInsertAt[simp]:
  "OclInsertAt (S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUndefined> (x::('\<tau>,'\<alpha>)VAL) = \<MathOclUndefined>"
  by(rule ext, simp add: OclUndefined_def OclInsertAt_def ss_lifting')

lemma OCL_undef_3_OclInsertAt[simp]:
  "OclInsertAt (S::('\<tau>,'\<alpha>::bot)Sequence) i (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL) = \<MathOclUndefined>"
  by(rule ext, simp add: OclUndefined_def OclInsertAt_def ss_lifting')

subsection{* Higher properties of insertAt *}

lemma includes_of_insertAt:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('\<tau>,'\<alpha>::bot)Sequence); \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>)VAL);
     \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> x \<MathOclIncludes> (OclInsertAt S i x)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(case_tac "i \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "x \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclInsertAt_def OclIncludes_def OclLe_def OclSize_def plus_def
                   One_ocl_int_def ss_lifting' localValid2sem
                   notin_set_take notin_set_drop)
  done

lemma size_of_insertAt:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>::bot)VAL); \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   \<MathOclSizeLeft>OclInsertAt S i x\<MathOclSizeRight> \<tau> = (\<MathOclSizeLeft>(S::('\<tau>,'\<alpha>)Sequence)\<MathOclSizeRight> + 1) \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(case_tac "i \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "x \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclInsertAt_def OclIncludes_def OclLe_def OclSize_def plus_def
                   One_ocl_int_def ss_lifting' localValid2sem
                   notin_set_take notin_set_drop, arith)
  done





section{* at *}

defs
   (* at     :: "[('\<tau>,'\<alpha>::bot ) Sequence,'\<tau> Integer] => ('\<tau>, '\<alpha>) VAL" *)
  OclAt_def  :
  "OclAt \<equiv> lift2(strictify(\<lambda>X. strictify(\<lambda>x. if \<HolOclDropLeft>x\<HolOclDropRight> <= 0 \<or>
                                               int(Nat.size\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>) < \<HolOclDropLeft>x\<HolOclDropRight>
                                             then \<bottom> 
                                             else \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> ! nat (\<HolOclDropLeft>x\<HolOclDropRight> - 1))))"

ocl_setup_op [OclAt]

(* TODO: should that be a simprule? *)
lemma at_lower_bound:
  "\<lbrakk> \<tau> \<MathOclValid> i \<MathOclLess> 1 \<rbrakk> \<Longrightarrow> OclAt (S::('\<tau>,'\<alpha>::bot)Sequence) i \<tau> = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL) \<tau>"
  by(auto simp: localValid2sem OclLess_def OclAt_def One_ocl_int_def
                OclUndefined_def ss_lifting')

lemma at_upper_bound:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclSizeLeft>S\<MathOclSizeRight> \<MathOclLess> i \<rbrakk> \<Longrightarrow> OclAt (S::('\<tau>,'\<alpha>::bot)Sequence) i \<tau> = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL) \<tau>"
  by(auto simp: localValid2sem OclLess_def OclAt_def One_ocl_int_def
                OclSize_def OclUndefined_def ss_lifting')

lemma OCL_is_defopt_OclAt:
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> \<MathOclSizeLeft>S\<MathOclSizeRight> \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclIsDefined>((OclAt (S::('\<tau>,'\<alpha>::bot)Sequence) i)::('\<tau>,'\<alpha>)VAL) = (\<tau> \<MathOclValid> \<MathOclIsDefined> S)"
  apply(case_tac "i \<tau> = \<bottom>")
  apply(simp add: localValid2sem OclLe_def One_ocl_int_def ss_lifting')
  apply(rule_tac X="S" in Sequence_sem_cases, simp_all add: localValid2sem)
  apply(auto simp: OclIsDefined_def OclAt_def OclLe_def
                   One_ocl_int_def OclSize_def ss_lifting')
  apply(simp add: in_set_conv_nth)
  apply(erule_tac x="nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)" in allE, arith)
  done

lemma at_of_mtSequence[simp]:
  "OclAt (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) i = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL)"
  by(rule ext, auto simp: OclAt_def OclMtSequence_def 
                          OclUndefined_def ss_lifting')

lemma at_last_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> i \<MathOclStrictEq> (\<MathOclSizeLeft>S\<MathOclSizeRight>+1) \<rbrakk> \<Longrightarrow>
   OclAt ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> x) i \<tau> =
   (x::('\<tau>,'\<alpha>)VAL) \<tau>"
  apply(case_tac "i \<tau> = \<bottom>")
  apply(simp add: localValid2sem OclStrictEq_def ss_lifting')
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> x")
  apply(ocl_hypsubst, simp)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule neq_commute[THEN iffD1])
  apply(rotate_tac 2, frule neq_commute[THEN iffD1])
  apply(auto simp: OclAt_def OclIncluding_def OclSize_def localValid2sem
                   OclStrictEq_def plus_def One_ocl_int_def ss_lifting')
  done

lemma at_inside_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x; \<tau> \<MathOclValid> i \<MathOclLe> \<MathOclSizeLeft>S\<MathOclSizeRight> \<rbrakk> \<Longrightarrow>
   OclAt (S \<MathOclIncluding> (x::('\<tau>,'\<alpha>)VAL)) i \<tau> =
   ((OclAt (S::('\<tau>,'\<alpha>::bot)Sequence) i \<tau>)::'\<alpha>)"
  apply(case_tac "i \<tau> = \<bottom>")
  apply(simp add: localValid2sem OclLe_def ss_lifting')
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule neq_commute[THEN iffD1])
  apply(auto simp: OclAt_def OclIncluding_def OclSize_def localValid2sem
                   OclLe_def plus_def One_ocl_int_def ss_lifting')
  apply(simp add: nth_append, arith)
  done


(* TODO: missing are rules about includes: see in_set_conv_nth *)

lemma at_of_union1:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S2; \<tau> \<MathOclValid> i \<MathOclLe> \<MathOclSizeLeft>S1\<MathOclSizeRight> \<rbrakk> \<Longrightarrow>
   OclAt (S1 \<MathOclUnion> S2) i \<tau> =
   ((OclAt (S1::('\<tau>,'\<alpha>::bot)Sequence) i \<tau>)::'\<alpha>)"
  apply(rule_tac X="S1" in Sequence_sem_cases)
  apply(rule_tac X="S2" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OclAt_def OclUnion_def OclSize_def localValid2sem
                   OclLe_def plus_def ss_lifting')
  apply(simp add: nth_append, arith)
  done

lemma at_of_union2:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclSizeLeft>S1\<MathOclSizeRight> \<MathOclLess> i; \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S1\<MathOclSizeRight>+\<MathOclSizeLeft>S2\<MathOclSizeRight>) \<rbrakk> \<Longrightarrow>
   OclAt (S1 \<MathOclUnion> S2) i \<tau> =
   ((OclAt (S2::('\<tau>,'\<alpha>::bot)Sequence) (i-\<MathOclSizeLeft>S1\<MathOclSizeRight>) \<tau>)::'\<alpha>)"
  apply(rule_tac X="S1" in Sequence_sem_cases)
  apply(rule_tac X="S2" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OclAt_def OclUnion_def OclSize_def minus_def localValid2sem
                   OclLe_def OclLess_def plus_def ss_lifting')
  apply(auto simp: nth_append, arith)
  apply(subgoal_tac "((nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) - (length c)) = (nat ((\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - (int (length c))) - 1))")
  apply(simp, arith)
  done

lemma at_first_of_prepend:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S \<rbrakk> \<Longrightarrow>
   OclAt (OclPrepend (S::('\<tau>,'\<alpha>::bot)Sequence) x) 1 \<tau> = (x::('\<tau>,'\<alpha>)VAL) \<tau> "
(*
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> x")
  apply(ocl_hypsubst, simp_all)
  apply(rule trans, rule at_of_union1)
  apply(simp_all)
(* hmm... would be nice to do it like that\<dots> but without any tool support
   for including context dependent simplifier rules this will be nothing *)
*)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> x")
  apply(ocl_hypsubst, simp)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule neq_commute[THEN iffD1])
  apply(auto simp: OclAt_def OclIncluding_def OclUnion_def OclSize_def localValid2sem
                   OclMtSequence_def One_ocl_int_def ss_lifting')
  done

lemma at_inside_of_prepend:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x; \<tau> \<MathOclValid> 1 \<MathOclLess> i \<rbrakk> \<Longrightarrow>
   OclAt (OclPrepend (S::('\<tau>,'\<alpha>::bot)Sequence) (x::('\<tau>,'\<alpha>)VAL)) i \<tau> = 
   ((OclAt S (i - 1) \<tau>)::'\<alpha>)"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S")
  apply(ocl_hypsubst, simp_all)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> i")
  apply(ocl_hypsubst, simp_all)

(*
  apply(rule trans, rule at_of_union2)
(* NOTE: what is missing is a case distinction on i >  \<MathOclSizeLeft>S\<MathOclSizeRight> + 1 and some
         simplification\<dots> but without a real simplifier this is pain in the ass *)
*)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule neq_commute[THEN iffD1])
  apply(rotate_tac 2, frule neq_commute[THEN iffD1])
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclSizeLeft>S\<MathOclSizeRight> + 1 \<MathOclLess> i")
  apply(auto simp: OclAt_def OclIncluding_def OclUnion_def OclSize_def localValid2sem
                   OclMtSequence_def OclLess_def minus_def plus_def 
                   One_ocl_int_def ss_lifting')
  apply(subgoal_tac "((x \<tau>) # c) = ([x \<tau>] @ c)")
  apply(simp only: nth_append)
  apply(subgoal_tac "((nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) - (Suc 0)) = (nat (-2 + \<HolOclDropLeft>(i \<tau>)\<HolOclDropRight>))")
  apply(auto, arith, arith)
  done


lemma at_of_subsequence:
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> a; \<tau> \<MathOclValid> a \<MathOclLess> b; \<tau> \<MathOclValid> b \<MathOclLe> \<MathOclSizeLeft>S\<MathOclSizeRight>;
     \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> (a-b+1) \<rbrakk> \<Longrightarrow> 
   OclAt (OclSubSequence (S::('\<tau>,'\<alpha>::bot)Sequence) a b) i \<tau> =
   ((OclAt S (a + i - 1) \<tau>)::'\<alpha>)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValid2sem)
  apply(case_tac "i \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "a \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(case_tac "b \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(simp_all add: OclIncluding_def OclAt_def OclSubSequence_def
                      One_ocl_int_def plus_def minus_def ss_lifting'
                      OclLe_def OclLess_def notin_set_take notin_set_drop)
  done


lemma at_of_insertAt_before:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>)VAL); \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1); \<tau> \<MathOclValid> j \<MathOclLess> i \<rbrakk> \<Longrightarrow>
   OclAt (OclInsertAt (S::('\<tau>,'\<alpha>::bot)Sequence) i x) j \<tau>  = ((OclAt S j \<tau>)::'\<alpha>)"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> i", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> j", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both)
  apply(clarify, frule DEF_X_Sequence', clarify)
  apply(case_tac "(1 \<le> \<HolOclDropLeft>(j \<tau>)\<HolOclDropRight>)")
  apply(subgoal_tac "((nat (\<HolOclDropLeft>(j \<tau>)\<HolOclDropRight> - 1)) < (length c))")
  apply(auto simp: OclInsertAt_def OclAt_def OclLess_def OclSize_def plus_def
                   One_ocl_int_def ss_lifting' localValid2sem
                   OclLe_def notin_set_take notin_set_drop nth_append)
  apply(arith, arith)
  done

lemma at_of_insertAt:
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   OclAt (OclInsertAt (S::('\<tau>,'\<alpha>::bot)Sequence) i (x::('\<tau>,'\<alpha>)VAL)) i \<tau>  = x \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S")
  apply(ocl_hypsubst, simp)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(case_tac "i \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "x \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclInsertAt_def OclAt_def OclLe_def OclSize_def plus_def
                   One_ocl_int_def ss_lifting' localValid2sem
                   notin_set_take notin_set_drop, arith)
  apply(subgoal_tac "(nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) = length (take (nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) c)")
  apply(rule_tac t="(nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1))" in subst) back
  apply(rule sym, assumption, rule nth_append_length)
  apply(simp, arith)
  done

lemma at_of_insertAt_after:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>)VAL);  \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> (\<MathOclSizeLeft>S\<MathOclSizeRight> + 1); \<tau> \<MathOclValid> i \<MathOclLess> j \<rbrakk> \<Longrightarrow>
   OclAt (OclInsertAt (S::('\<tau>,'\<alpha>::bot)Sequence) i x) j \<tau>  = ((OclAt S (j - 1) \<tau>)::'\<alpha>)"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> i", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> j", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both)
  apply(clarify, frule DEF_X_Sequence', clarify)
  apply(subgoal_tac "(min (length c) (nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1))) = nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)")
  apply(case_tac "(((int (length c)) + 1) < \<HolOclDropLeft>(j \<tau>)\<HolOclDropRight>)")
  apply(auto simp: OclInsertAt_def OclAt_def OclLess_def OclSize_def plus_def
                   One_ocl_int_def ss_lifting' localValid2sem minus_def
                   OclLe_def notin_set_take notin_set_drop nth_append)
  apply(thin_tac "?P \<noteq> ?Q" | thin_tac "?P = ?Q")+ apply(arith)
  apply(thin_tac "?P \<noteq> ?Q" | thin_tac "?P = ?Q")+ apply(arith)
  apply(thin_tac "?P \<noteq> ?Q" | thin_tac "?P = ?Q" | thin_tac "?P \<notin> ?Q")+
  apply(subgoal_tac "(nat (\<HolOclDropLeft>(j \<tau>)\<HolOclDropRight> - 1)) - (nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) = Suc ((nat (\<HolOclDropLeft>(j \<tau>)\<HolOclDropRight> - 2)) - (nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)))")
  apply(simp)
  apply(subst nth_drop)
  apply(thin_tac "?P = ?Q", arith)
  apply(subgoal_tac "((nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)) + ((nat (\<HolOclDropLeft>(j \<tau>)\<HolOclDropRight> - 2)) - (nat (\<HolOclDropLeft>(i \<tau>)\<HolOclDropRight> - 1)))) = (nat (-2 + \<HolOclDropLeft>(j \<tau>)\<HolOclDropRight>))")
  apply(simp)
  apply(rotate_tac 2, thin_tac "?P", thin_tac "?P", thin_tac "?P", arith)
  apply(rotate_tac 2, thin_tac "?P", thin_tac "?P", arith)
  apply((thin_tac "?P \<noteq> ?Q" | thin_tac "?P = ?Q" | thin_tac "?P \<notin> ?Q")+)
  apply(rotate_tac 2, thin_tac "?P", thin_tac "?P", arith)
  done


section{* IndexOf *}

defs
   (* indexOf    :: "[('\<tau>,'\<alpha>::bot ) Sequence,('\<tau>, '\<alpha>) VAL] => '\<tau> Integer" *)
  OclIndexOf_def :
  "OclIndexOf    \<equiv> lift2(strictify(\<lambda>X. strictify(\<lambda>x. if x \<in> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>
                                             then \<HolOclLiftLeft>int(length (takeWhile (\<lambda> y. y \<noteq> x) \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>) + 1)\<HolOclLiftRight>
                                             else \<bottom>)))"

ocl_setup_op [OclIndexOf]

lemma indexOf_excludes[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> ((x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Sequence)) \<rbrakk> \<Longrightarrow>
   OclIndexOf S x \<tau> = \<MathOclUndefined> \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases, simp_all)
  apply(case_tac "x \<tau> = \<bottom>")
  apply(simp_all add: OclIndexOf_def OclIncludes_def OclNot_def o_def
                  OclUndefined_def ss_lifting' localValid2sem)
  done

lemma notin_conv_all_noteq: "a \<notin> S = (\<forall> x \<in> S. x \<noteq> a)"
  by (auto)

lemma in_set_conv_first_decomp: 
  "(x \<in> set xs) = (\<exists>ys zs. xs = ys @ x # zs \<and> x \<notin> set ys)"
  apply(auto, erule rev_mp)
  apply(rule length_induct, auto)
  apply(frule in_set_conv_decomp[THEN iffD1], clarify)
  apply(case_tac "x \<in> set ys")
  apply(erule_tac x="ys" in allE, auto)
  apply(rule_tac x="ysa" in exI, simp)
  done

text{* \OCLglitch{Typo in requirement 11.7.5.10} In the standard
  \inlineocl{self->at(i) = obj} is written where it should be written
  like \inlineocl{self->at(result) = obj}. 
  \OCLglitch{No specification of the returned index of \texttt{indexOf}}
  Furthermore the standard doesn't specify that the returned index is
  the index of the first object that is equal to the searched one. This
  disables the combined use of @{text "indexOf"} and 
  @{text "subsequence"}
  to visit all occurrences of a specific object. *}

lemma at_indexOf:
  "\<lbrakk> \<tau> \<MathOclValid> (x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow>
   OclAt S (OclIndexOf S x) \<tau> = x \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases, simp_all)
  apply(case_tac "x \<tau> \<noteq> \<bottom>")
  apply(frule neq_commute[THEN iffD1])
  apply(simp_all add: OclIndexOf_def OclIncludes_def OclNot_def OclAt_def 
                      OclUndefined_def ss_lifting' localValid2sem
                      o_def)
  apply(frule in_set_conv_first_decomp[THEN iffD1])
  apply(clarify, simp add: notin_conv_all_noteq takeWhile_tail)
  done



section{* First and Last *}

defs

   (* first    ::  "('\<tau>, '\<alpha>::bot ) Sequence => ('\<tau>,'\<alpha>) VAL" *)
  OclFirst_def :
  "OclFirst    \<equiv> lift1 (strictify (\<lambda>S. if \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> = [] then \<bottom>
                                       else hd (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)))"

   (* last    ::  "[('\<tau>, '\<alpha>::bot) Sequence] => ('\<tau>,'\<alpha>) VAL" *)
  OclLast_def :
  "OclLast    \<equiv> lift1 (strictify (\<lambda>S. if \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> = [] then \<bottom>
                                      else List.last (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)))"

ocl_setup_op [OclFirst, OclLast]

lemma first_at_UC:
  "OclFirst (S::('\<tau>,'\<alpha>::bot)Sequence) = ((OclAt S 1)::('\<tau>,'\<alpha>)VAL)"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclFirst_def OclAt_def Zero_ocl_int_def
                   One_ocl_int_def ss_lifting' neq_Nil_conv)
  done

lemma first_of_mtSequence[simp]:
  "OclFirst (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL)"
  by(simp add: first_at_UC)

lemma first_empty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> (S \<MathOclStrictEq> \<MathOclMtSequence>)\<rbrakk> \<Longrightarrow>
   OclFirst ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> (x::('\<tau>,'\<alpha>)VAL)) \<tau> =
   ((x \<tau>)::'\<alpha>)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValid2sem)
  apply(case_tac c)
  apply(auto simp: OclIncluding_def OclFirst_def OclNot_def OclStrictEq_def
                   OclMtSequence_def ss_lifting' neq_commute)
  done


text{* Note: strange writing because of normal form of emptiness test *}
lemma first_not_empty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x; \<tau> \<MathOclValid> \<MathOclNot> (S \<MathOclStrictEq> \<MathOclMtSequence>)\<rbrakk> \<Longrightarrow>
   OclFirst ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> (x::('\<tau>,'\<alpha>)VAL)) \<tau> =
   ((OclFirst S \<tau>)::'\<alpha>)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(case_tac c)
  apply(auto simp: OclIncluding_def OclFirst_def OclNot_def OclStrictEq_def
                   OclMtSequence_def ss_lifting' neq_commute localValid2sem)
  done


lemma last_at_UC:
  "OclLast (S::('\<tau>,'\<alpha>::bot)Sequence) = ((OclAt S \<MathOclSizeLeft>S\<MathOclSizeRight>)::('\<tau>,'\<alpha>)VAL)"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclLast_def OclAt_def OclSize_def
                   Zero_ocl_int_def One_ocl_int_def 
                   ss_lifting' last_conv_nth neq_Nil_conv)
  done


lemma last_of_mtSequence[simp]:
  "OclLast (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL)"
  by(simp add: last_at_UC)

lemma last_empty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S\<rbrakk> \<Longrightarrow>
   OclLast ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> (x::('\<tau>,'\<alpha>)VAL)) \<tau> =
   ((x \<tau>)::'\<alpha>)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(simp add: OclIncluding_def OclLast_def 
                  ss_lifting' neq_commute)
  done

(* TODO: simplifier setup for these rules *)



section{* Forall *}

defs
   (* forall    :: "[('\<tau>,'\<alpha>::bot Sequence) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean" *)

   OclForAll_def: "OclForAll \<equiv> (lift2' lift_arg0 lift_arg1) (strictify( \<lambda> S. (\<lambda> P.  
                      if  \<forall>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>.  P x = \<HolOclLiftLeft>True\<HolOclLiftRight> 
                      then \<HolOclLiftLeft>True\<HolOclLiftRight>
                      else if \<exists>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>
                           then \<HolOclLiftLeft>False\<HolOclLiftRight>
                           else \<bottom>)))"

(*
   OclForAll_def:  "OclForAll S P \<equiv> \<lambda> \<tau>.  
                    if DEF (S \<tau>)
                    then if  \<forall>x \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>.  P (lift0 x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>) 
                         then \<HolOclLiftLeft>True\<HolOclLiftRight>
                         else if \<exists>x \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>. P (lift0 x) \<tau> = (\<HolOclLiftLeft>False\<HolOclLiftRight>) 
                              then \<HolOclLiftLeft>False\<HolOclLiftRight>
                              else \<bottom>
                    else \<bottom>"
*)

ocl_setup_op [OclForAll]

subsection{* Its computation characterisation by iterate *}

lemma forall_by_iterate:
  "(\<MathOclForAll> x \<in> (S::('\<tau>,'\<alpha>::bot)Sequence) \<bullet> (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> '\<tau> Boolean) x) = 
   (S \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  apply(rule ext)
  apply(rule iterate_universal_0, simp_all)
  apply(simp add: cp_by_cpify OclForAll_def lift0_def)
  apply(simp_all add: OclForAll_def OclUndefined_def OclMtSequence_def
                      OclTrue_def localValidDefined2sem ss_lifting')
  apply(frule DEF_X_Sequence', clarify)
  apply(frule neq_commute[THEN iffD1]) 
  apply(rotate_tac 1, frule neq_commute[THEN iffD1]) 
  apply(rule_tac x="P (\<lambda>s. (xa x)) x" in boolean_cases_sem)
  apply(simp_all add: OclIncluding_def OclAnd_def ss_lifting')
  done

(*
thm f2_func_by_iterate_cp[OF forall_by_iterate, standard]
*)

subsection{* Derived properties from the relation to iterate *}

(*
lemma OCL_cp_OclForAll [simp,intro!]: 
  "\<lbrakk> \<And> x. cp (P x); cp S \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> ((S X)::('\<tau>,'\<alpha>::bot)Sequence) \<bullet> P (Y::('\<tau>,'\<alpha>)VAL) X) "
  by(simp add: forall_by_iterate)

lemmas OCL_undef_1_OclForAll[simp] = f1_by_iterate_undef_1[OF forall_by_iterate]
*)

lemma OCL_undef_2_OclForAll[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclNot> ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclStrictEq> \<MathOclMtSequence>) \<rbrakk> \<Longrightarrow>
    (\<MathOclForAll> x \<in> S \<bullet> (\<lambda> x::('\<tau>,'\<alpha>)VAL. \<MathOclUndefined>) x) \<tau> = \<MathOclUndefined> \<tau>"
  apply(rule_tac X="S" in Sequence_sem_cases, simp_all)
  apply(simp_all add: OclForAll_def OclUndefined_def OclMtSequence_def 
                      OclStrictEq_def OclNot_def localValid2sem   
                      ss_lifting' o_def Abs_Sequence_0_inject_charn)
  done



lemmas forall_of_mtSequence[simp] =  f1_by_iterate_mtSequence[OF forall_by_iterate]

lemma forall_including [simp]:
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence); \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>)VAL); cp P \<rbrakk> \<Longrightarrow>
   (\<MathOclForAll> x \<in> (X \<MathOclIncluding> a) \<bullet> P (x::('\<tau>,'\<alpha>)VAL)) \<tau> =  ((P a) \<MathOclAnd> (\<MathOclForAll> x \<in> X \<bullet> P x)) \<tau>"
  by(simp add: forall_by_iterate iterate_of_including)

(* FIXME: bad interaction of rules for \<MathOclTrue> and OCL_logic_ACI *)
thm f1_by_iterate_distrib_union[OF forall_by_iterate, of "OclAnd", simplified]

lemma forall_union [simp]:
  "\<lbrakk>\<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha>::bot)Sequence); \<tau> \<MathOclValid> \<MathOclIsDefined> Y \<rbrakk> \<Longrightarrow>
   (\<MathOclForAll> x \<in> (X \<MathOclUnion> Y) \<bullet> P (x::('\<tau>,'\<alpha>)VAL)) \<tau> =  ((\<MathOclForAll> x \<in> X \<bullet> P x) \<MathOclAnd> (\<MathOclForAll> x \<in> Y \<bullet> P x)) \<tau>"
  apply(simp add: forall_by_iterate)
  apply(rule iterate_distrib_union)
  apply(simp_all add: OCL_logic_ACI)
  done



subsection{* Introduction rules for ForAll *}

lemma ForAllI:
  assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> P x" 
  shows          "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert defS)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(subgoal_tac "(\<forall>x \<in> set c. P (\<lambda>\<tau>. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>)") 
  apply(auto simp: OclForAll_def localValid2sem ss_lifting')
  apply(rule_tac \<tau>1 ="\<tau>" in localValid2sem[THEN iffD1], rule allP)
  apply(auto simp: localValid2sem OclIncludes_def ss_lifting')
  done


lemma   notForAllI:
  assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     not  : "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP  : "cp P"
  shows          "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert isin not cpP)
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(subgoal_tac "(P x \<tau>) = (P (\<lambda>s. (x \<tau>)) \<tau>)")
  apply(case_tac "(P x) \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(case_tac "x \<tau> \<noteq> \<bottom>", rotate_tac -1, frule neq_commute[THEN iffD1])
  apply(subgoal_tac "(\<exists>x\<in>(set c). ((P (\<lambda>s. x) \<tau>) = \<HolOclLiftLeft>False\<HolOclLiftRight>))")
  apply(simp_all add: OclForAll_def OclIncludes_def OclNot_def
                      ss_lifting' localValid2sem o_def)
  apply(rule_tac x="x \<tau>" in bexI)
  apply(rule_tac x="P (\<lambda>s. (x \<tau>)) \<tau>" in boolean_cases_sem)
  apply(simp_all)
  apply(rule_tac x="x \<tau>" in bexI)
  apply(rule_tac x="P (\<lambda>s. (x \<tau>)) \<tau>" in boolean_cases_sem)
  apply(simp_all)
  apply(rule trans, rule_tac x="x" in cp_subst)
  apply(simp_all add: lift0_def)
  done


lemma undefForAllI2:
  assumes exUndef1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     exUndef2: "\<tau> \<MathOclValid>  \<MathOclIsUndefined> (P x)"
  and     allP    : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> (P x) \<MathOclOr> \<MathOclIsUndefined>(P x)" 
  and     cpP     : "cp P"
  shows             "\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert exUndef1 exUndef2)
  apply(frule isDefined_if_valid)
  apply(drule OCL_is_defopt_OclIncludes[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def OclTrue_def
                  OclForAll_def OclNot_def OclIncludes_def
                  ss_lifting')
  apply(erule conjE)
  apply(frule DEF_X_Sequence', clarify)
  apply(simp, safe)
  apply(rule_tac x = xa in bexI, simp_all) prefer 2
  apply(rule_tac x = "x \<tau>" in bexI, simp_all)
  apply(subgoal_tac "P (\<lambda>\<tau>a. x \<tau>) \<tau> = P x \<tau>", simp)
  apply(rule_tac \<tau> = \<tau> and P = P in cp_charn, simp add: cpP, rule cpP)
  apply(erule_tac Q = "P (\<lambda>\<tau>. xa) \<tau> = \<HolOclLiftLeft>False\<HolOclLiftRight>" in contrapos_pp)
  apply(rule_tac \<tau>1 = \<tau> in localValidToNotFalse2sem [THEN iffD1]) back
  apply(rule allP)
  apply(simp add: includes_charn2 DEF_def)
  done


(* TODO: not needed anymore, as this expression is reduced
         to \<MathOclFalse> automatically 
lemma isdefForAllI1[simp, intro!]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclForAll> x \<in> (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) \<bullet> P(x\<Colon>'\<tau>\<Rightarrow>'\<alpha>))"
by simp
(* Not analytic, but works for rewriting \<dots> *)
*)

lemma isdefForAllI2:
  assumes exNot1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Sequence)"
  and     exNot2: "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP   : "cp P"
  shows           "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  apply(insert exNot1 exNot2 cpP)
  apply(drule_tac P="P" in notForAllI, simp_all add:  isDefined_if_invalid)
  done

lemmas isdefForAllI3 = ForAllI[THEN isDefined_if_valid]

(*
lemma localValidForall2forall :
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)" 
and     cpP  : "cp P"
shows   "(\<tau> \<MathOclValid> \<MathOclForAll> x \<in> S \<bullet> %G????\<cent>%@ P(x::'\<tau> => '\<alpha>::bot)) =
         (\<forall> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<longrightarrow> \<tau> \<MathOclValid> P x)"
apply (safe)
apply (drule  ForAllD, simp_all add: cpP)
apply (rule   ForAllI, simp_all add: defS)
done
*)



section{* Exists *}

subsection{* Alternative definition for the use with ocl\_setup\_op *}

lemma OclExists_alt_def:
  "OclExists \<equiv> (lift2' lift_arg0 lift_arg1) (strictify( \<lambda> S. (\<lambda> P.  
      if  \<exists> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>.  P x = \<HolOclLiftLeft>True\<HolOclLiftRight> 
      then \<HolOclLiftLeft>True\<HolOclLiftRight>
      else if \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>
           then \<HolOclLiftLeft>False\<HolOclLiftRight>
           else \<bottom>)))"
  apply(rule eq_reflection, (rule ext)+)  
  apply(simp add: OclExists_def OclForAll_def OclNot_def ss_lifting')
  apply(auto)
  oops
(* TODO: finish *)

subsection{* The relation to iterate *}

lemma exists_by_iterate:
  "(\<MathOclExists> x \<in> (source::('\<tau>,'\<alpha>::bot)Sequence) \<bullet> (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> '\<tau> Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  apply(simp add: OclExists_def forall_by_iterate)
  apply(rule ext, rule iterate_fusion_0, simp_all)
(* TODO: would be nice to have the calculus laws on the semantic level *)
  apply(simp add: OclNot_def OclOr_def OclAnd_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  done

subsection{* Derived properties from the relation to iterate *}

lemma OCL_cp_OclExists [simp,intro!]: 
  "\<lbrakk> \<And> x. cp (P x); cp S \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. \<MathOclExists> Y \<in> ((S X)::('\<tau>,'\<alpha>::bot)Sequence) \<bullet> P (Y::('\<tau>,'\<alpha>)VAL) X) "
  by(simp add: exists_by_iterate)


lemmas exists_undef [simp] = f1_by_iterate_undef_1[OF exists_by_iterate]
lemmas exists_empty [simp] = f1_by_iterate_mtSequence[OF exists_by_iterate]
thm f1_by_iterate_including[OF exists_by_iterate, simplified]
thm f1_by_iterate_opt_including[OF exists_by_iterate, simplified]


thm f1_by_iterate_undef_1[OF exists_by_iterate]
thm f1_by_iterate_mtSequence[OF exists_by_iterate]
thm f1_by_iterate_including[OF exists_by_iterate, simplified]
thm f1_by_iterate_opt_including[OF exists_by_iterate, simplified]
(* TODO: fix cp-ness *)


subsection{* Introduction rules for exist *}

lemma   ExistsI:
assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('\<tau>,('\<alpha>::bot))Sequence)"
and     not  : "\<tau> \<MathOclValid> (P x)"
and     cpP  : "cp P"
shows          "\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
by (auto simp:OclExists_def intro!: notForAllI isin not cpP)


lemma   NotExistsI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,('\<alpha>::bot))Sequence)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclNot> P x" 
shows          "\<tau> \<MathOclValid>  \<MathOclNot>(\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
by (auto simp: OclExists_def intro!: ForAllI defS allP)


section{* IncludesAll *}

defs
   (* includesAll     :: "[('\<tau>,'\<alpha>::bot)Sequence, ('\<tau>,'\<alpha>)Sequence] => '\<tau> Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight> \<subseteq> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 


ocl_setup_op [OclIncludesAll]

subsection{* Its computational characterisation by iterate *}


text{* This formulation should be replaced by one with a strict body to
  simplify reasoning over it. *}
lemma includesAll_by_iterate:
  "OclIncludesAll X (Y::('\<tau>,'\<alpha>::bot)Sequence) = 
   (Y \<MathOclIterate>(x;y= \<MathOclIf> (\<MathOclIsDefined> X) \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif> | ((x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> X) \<MathOclSand> y))"
  apply(rule ext)
  apply(rule iterate_universal_0, simp_all)
  apply(simp add: OclIncludesAll_def OclMtSequence_def OclIsDefined_def
                  OclIf_def OclTrue_def OclUndefined_def ss_lifting')
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(rotate_tac 1, frule neq_commute[THEN iffD1])
  apply(simp add: OclIncludesAll_def OclIncluding_def OclIncludes_def
                  OclSand_def ss_lifting')
  done

thm f1_by_iterate_undef_1[OF includesAll_by_iterate]
lemmas includesAll_empty [simp] = f1_by_iterate_mtSequence[OF includesAll_by_iterate]
lemmas includesAll_including [simp] = f1_by_iterate_opt_including[OF includesAll_by_iterate, simplified]

thm includesAll_empty includesAll_including

lemma includesAll_union [simp]:
  "(OclIncludesAll Z ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y)) =
   (OclIncludesAll Z X) \<MathOclSand> (OclIncludesAll Z Y)"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> X ", ocl_hypsubst, simp)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> Y", ocl_hypsubst, simp)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> Z", ocl_hypsubst, simp)
  apply(subst f1_by_iterate_distrib_union[OF includesAll_by_iterate])
  apply(simp_all)
  apply(ocl_subst, simp)
(* TODO: Add rules for strict logic *)
  apply(auto simp: OclIncludesAll_def OclSand_def OclIncludes_def
                   OclTrue_def localValidDefined2sem ss_lifting')
  done

lemma includesAll_by_forall: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow>
   (OclIncludesAll X Y) \<tau> = (\<MathOclForAll> a \<in> Y \<bullet> ((a::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> X)) \<tau>"
  apply(rule_tac X="X" in Sequence_sem_cases)
  apply(rule_tac X="Y" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OCL_Sequence.OclForAll_def OCL_Sequence.OclIncludesAll_def OclStrongEq_def 
                   OCL_Sequence.OclIncludes_def ss_lifting' localValid2sem)
  done


section{* ExcludesAll *}

defs
   (* excludesAll     :: "[('\<tau>,'\<alpha>::bot)Sequence, ('\<tau>,'\<alpha>)Sequence] => '\<tau> Boolean" *)
   OclExcludesAll_def: 
   "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                       \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<inter> set \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"


ocl_setup_op [OclExcludesAll]

subsection{* Its computational characterisation by iterate *}


text{* This formulation should be replaced by one with a strict body to
  simplify reasoning over it. *}
lemma excludesAll_by_iterate:
  "OclExcludesAll X (Y::('\<tau>,'\<alpha>::bot)Sequence) = 
   (Y \<MathOclIterate>(x;y= \<MathOclIf> (\<MathOclIsDefined> X) \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif> | ((x::('\<tau>,'\<alpha>)VAL) \<MathOclExcludes> X) \<MathOclSand> y))"
  apply(rule ext)
  apply(rule iterate_universal_0, simp_all)
  apply(simp add: OclExcludesAll_def OclMtSequence_def OclIsDefined_def
                  OclIf_def OclTrue_def OclUndefined_def ss_lifting')
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(rotate_tac 1, frule neq_commute[THEN iffD1])
  apply(simp add: OclExcludesAll_def OclIncluding_def OclIncludes_def
                  OclSand_def OclNot_def ss_lifting')
  done


thm f1_by_iterate_undef_1[OF excludesAll_by_iterate]
lemmas excludesAll_empty [simp] = f1_by_iterate_mtSequence[OF excludesAll_by_iterate]
lemmas excludesAll_including [simp] = f1_by_iterate_opt_including[OF excludesAll_by_iterate, simplified]

thm excludesAll_empty excludesAll_including

lemma excludesAll_union [simp]:
  "(OclExcludesAll Z ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y)) =
   (OclExcludesAll Z X) \<MathOclSand> (OclExcludesAll Z Y)"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> X ", ocl_hypsubst, simp)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> Y", ocl_hypsubst, simp)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> Z", ocl_hypsubst, simp)
  apply(subst f1_by_iterate_distrib_union[OF excludesAll_by_iterate])
  apply(simp_all)
  apply(ocl_subst, simp)
(* TODO: Add rules for strict logic *)
  apply(auto simp: OclIncludesAll_def OclSand_def OclIncludes_def
                   OclTrue_def localValidDefined2sem ss_lifting')
  done

lemma excludesAll_by_forall: 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(X::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow>
   (OclExcludesAll X Y) \<tau> = (\<MathOclForAll> a \<in> Y \<bullet> ((a::('\<tau>,'\<alpha>)VAL) \<MathOclExcludes> X)) \<tau>"
  apply(rule_tac X="X" in Sequence_sem_cases)
  apply(rule_tac X="Y" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OCL_Sequence.OclForAll_def OCL_Sequence.OclExcludesAll_def OclStrongEq_def 
                   OCL_Sequence.OclIncludes_def ss_lifting' localValid2sem OclNot_def)
  done


section{* Select *}

defs
   (* select    :: "[('\<tau>,'c::bot Sequence) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] *)
   OclSelect_def:  "OclSelect \<equiv>  (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                       if    \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)
                       then  Abs_Sequence_0(\<HolOclLiftLeft>List.filter (\<lambda> x. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>) 
                                                         \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> \<HolOclLiftRight>)
                       else  \<bottom>))"

ocl_setup_op [OclSelect]

subsection{* Its computational characterisation by iterate *}

lemma select_by_iterate:
  "(\<MathOclSelectLeft>x : (source::('\<tau>,'\<alpha>::bot)Sequence) | ((P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> '\<tau> Boolean) x)\<MathOclSelectRight>) = 
   (source \<MathOclIterate>(iterator; result = \<MathOclMtSequence> | 
                             \<MathOclIf>  (P iterator)\<MathOclThen> (OclIncluding result iterator) \<MathOclElse> (result) \<MathOclEndif> ))"
  apply(rule ext, rule iterate_universal_0, simp_all)
  apply(simp add: OclSelect_def cp_by_cpify lift0_def)
  apply(simp_all add: OclUndefined_def OclMtSequence_def OclSelect_def
                      localValidDefined2sem ss_lifting')
  apply(frule DEF_X_Sequence', clarify)
  apply(frule neq_commute[THEN iffD1]) 
  apply(rotate_tac 1, frule neq_commute[THEN iffD1]) 
  apply(rule_tac x="P (\<lambda>s. (xa x)) x" in boolean_cases_sem)
  apply(simp_all add: OclIf_def OclIncluding_def ss_lifting')
  done

subsection{* Properties derived from the correspondence to iterate *}

lemmas select_of_mtSequence [simp] = f1_by_iterate_mtSequence[OF select_by_iterate]

lemma select_including [simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>::bot)VAL); cp P \<rbrakk> \<Longrightarrow>
  OclSelect (S::('\<tau>,'\<alpha>)Sequence \<MathOclIncluding> x) P \<tau> = 
  (\<MathOclIf> (P x) \<MathOclThen> (OclSelect S P) \<MathOclIncluding> x \<MathOclElse> (OclSelect S P)\<MathOclEndif>) \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> S", ocl_hypsubst, simp)
  apply(simp add: f1_by_iterate_including[OF select_by_iterate, simplified])
  done

lemma select_including_strict [simp]:
  "\<lbrakk> P \<MathOclUndefined> = \<MathOclUndefined>; cp P \<rbrakk> \<Longrightarrow>
  OclSelect (S::('\<tau>,'\<alpha>)Sequence \<MathOclIncluding> (x::('\<tau>,'\<alpha>::bot)VAL)) P \<tau> = 
  (\<MathOclIf> (P x) \<MathOclThen> (OclSelect S P) \<MathOclIncluding> x \<MathOclElse> (OclSelect S P)\<MathOclEndif>) \<tau>"
  by(simp add: f1_by_iterate_opt_including[OF select_by_iterate, simplified])

lemma select_union [simp]:
  "(OclSelect ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean)) =
   (OclSelect X P) \<MathOclUnion> (OclSelect Y P)"
  apply(subst f1_by_iterate_distrib_union_opt[OF select_by_iterate])
  apply(simp_all)
  apply(rule_tac X="B" in Sequence_sem_cases_ext, simp_all)
  apply(rule_tac X="y" in Sequence_sem_cases, simp_all)
  apply(case_tac "x \<tau> \<noteq> \<bottom>", drule neq_commute[THEN iffD1])
  apply(simp_all add: OclIf_def OclUnion_def OclIncluding_def ss_lifting')
  done

lemma iterate_of_select:
  assumes undef_1_Q: "\<And> y. Q \<MathOclUndefined> y = \<MathOclUndefined>"
  and     undef_2_Q: "\<And> x. Q x \<MathOclUndefined>  = \<MathOclUndefined>"
  and         cp1_Q: "\<And> y. cp (\<lambda>x. Q x y)"
  and         cp2_Q: "\<And> x. cp (Q x)"
  and     undef_1_P: "P \<MathOclUndefined> = \<MathOclUndefined>"
  and          cp_P: "cp P"
  shows "OclIterate (OclSelect (S::('\<tau>,'\<alpha>::bot)Sequence) P) Q A = 
         OclIterate S (\<lambda> (x::('\<tau>,'\<alpha>)VAL) y. \<MathOclIf> (P x) \<MathOclThen> Q x y \<MathOclElse> y \<MathOclEndif>) A"
  apply(insert cp1_Q cp2_Q cp_P undef_1_P undef_2_Q undef_1_Q)
  apply(rule ext)
  apply(rule iterate_universal, simp_all add: select_by_iterate)
  apply(subst iterate_opt_of_including, simp_all)
  apply(rule_tac P="\<lambda> t. t x = ?Q" in subst)
  apply(rule sym[OF if_distrib_strict_alt])
  apply(simp_all add: iterate_opt_of_including)
  done

lemma select_of_select:
  assumes undef_1_P: "P \<MathOclUndefined> = \<MathOclUndefined>"
  and     undef_1_Q: "Q \<MathOclUndefined> = \<MathOclUndefined>"
  and          cp_P: "cp P"
  and          cp_Q: "cp Q"
  shows "OclSelect (OclSelect (S::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean)) (Q::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean) = 
         OclSelect S (\<lambda> x. (P x) \<MathOclSand> ((P x) \<MathOclImplies> (Q x)))"
  by(simp add: select_by_iterate iterate_of_select[simplified select_by_iterate] 
               undef_1_Q undef_1_P cp_P cp_Q if_contract)


section{* Reject *}

defs
   (* reject    :: "[('\<tau>,'c::bot Sequence) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] *)

   OclReject_def:  "OclReject \<equiv>  (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                       if    \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)
                       then  Abs_Sequence_0(\<HolOclLiftLeft>List.filter (\<lambda> x. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>) 
                                                         \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> \<HolOclLiftRight>)
                       else  \<bottom>))"

ocl_setup_op [OclReject]

lemma reject_by_select:
  "OclReject S P = (\<MathOclSelectLeft>x : (S::('\<tau>,'\<alpha>::bot)Sequence) | (\<MathOclNot> (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> '\<tau> Boolean) x)\<MathOclSelectRight>)"
  apply(rule ext, auto simp: OclReject_def OclNot_def OclSelect_def ss_lifting')
  apply(subgoal_tac "(\<lambda>xa. ((P (\<lambda>s. xa) x) = \<HolOclLiftLeft>False\<HolOclLiftRight>)) = 
                     (\<lambda>xa. (((P (\<lambda>s. xa) x) \<noteq> down) \<and> (((P (\<lambda>s. xa) x) \<noteq> down) \<longrightarrow> (\<not> \<HolOclDropLeft>(P (\<lambda>s. xa) x)\<HolOclDropRight>))))")
  apply(simp, rule ext)
  apply(rule_tac x="P (\<lambda>s. xa) x" in boolean_cases_sem, simp_all)
  done

lemma reject_by_iterate:
  "OclReject (source::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> '\<tau> Boolean) = 
   (source \<MathOclIterate>(iterator; result = \<MathOclMtSequence> | 
                             \<MathOclIf>  (\<MathOclNot> P iterator)\<MathOclThen> (OclIncluding result iterator) \<MathOclElse> (result) \<MathOclEndif> ))"
  by(simp add: reject_by_select select_by_iterate)


subsection{* Properties derived from the correspondence to iterate *}

lemmas reject_of_mtSequence [simp] = f1_by_iterate_mtSequence[OF reject_by_iterate]

lemma reject_of_union:
  "(OclReject ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean)) =
   (OclReject X P) \<MathOclUnion> (OclReject Y P)"
  by(simp add: reject_by_select)



section{* Excluding *}

defs
   (* excluding  :: "[('\<tau>,'\<alpha>::bot ) Sequence,('\<tau>, '\<alpha>) VAL] => ('\<tau>, '\<alpha>) Sequence" *)
   OclExcluding_def   :
   "OclExcluding  \<equiv> lift2(strictify(\<lambda>X. strictify (\<lambda>x. Abs_Sequence_0 
                      (\<HolOclLiftLeft> List.filter (\<lambda> e. x \<noteq> e) \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>))))"


ocl_setup_op [OclExcluding]

subsection{* Its computational characteristic by iterate *}

lemma excluding_by_iterate:
  "(OclExcluding (S::('\<tau>,'\<alpha>::bot)Sequence) (x::('\<tau>,'\<alpha>)VAL)) = 
   OclIterate S (\<lambda> (a::('\<tau>,'\<alpha>)VAL) y. \<MathOclIf> (a \<MathOclStrictEq> x) \<MathOclThen> y \<MathOclElse> y \<MathOclIncluding> a \<MathOclEndif>) 
                (\<MathOclIf> (\<MathOclIsDefined> x) \<MathOclThen> \<MathOclMtSequence> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif>)"
  apply(rule ext)
  apply(rule iterate_universal, simp_all add: localValidDefined2sem DEF_def)
  apply(simp add: OclMtSequence_def OclExcluding_def ss_lifting'
                  OclIf_def OclIsDefined_def OclUndefined_def)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsUndefined> x") 
  apply(ocl_hypsubst, simp_all add: localValidDefined2sem)
  apply(frule DEF_X_Sequence', clarify)
  apply(frule neq_commute[THEN iffD1]) 
  apply(rotate_tac 1, frule neq_commute[THEN iffD1]) 
  apply(simp add: OclExcluding_def OclIterate_def OclIf_def
                  OclIncluding_def OclStrictEq_def ss_lifting')
  done

lemmas excluding_empty [simp] = f1_by_iterate_mtSequence[OF excluding_by_iterate]
lemmas excluding_including [simp] = f1_by_iterate_opt_including[OF excluding_by_iterate, simplified]

lemma excluding_union [simp]:
  "(((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) \<MathOclExcluding> x) = 
   (X \<MathOclExcluding> x)  \<MathOclUnion> (Y \<MathOclExcluding> (x::('\<tau>,'\<alpha>)VAL))"
  apply(rule ext, case_tac "xa \<MathOclValid> \<MathOclIsUndefined> x", ocl_hypsubst, simp)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsUndefined> X", ocl_hypsubst, simp)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsUndefined> Y", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(rotate_tac 2, frule DEF_X_Sequence', clarify)
  apply(simp add: OclExcluding_def OclUnion_def ss_lifting')
  done

thm excluding_empty excluding_including excluding_union

subsection{* Further propierties of excluding *}

lemma includes_of_excluding:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(x::('\<tau>,'\<alpha>)VAL);  \<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,'\<alpha>::bot)Sequence) \<rbrakk> \<Longrightarrow>
   (x \<MathOclIncludes> (S \<MathOclExcluding> x)) \<tau> = \<MathOclFalse> \<tau>"
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(frule DEF_X_Sequence', clarify)
  apply(simp add: OclIncludes_def OclExcluding_def OclFalse_def
                  ss_lifting')
  done

lemma size_of_excluding:
  "\<MathOclSizeLeft>(S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclExcluding> (x::('\<tau>,'\<alpha>)VAL)\<MathOclSizeRight> =
   \<MathOclSizeLeft>S\<MathOclSizeRight> - OclCount S x"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(case_tac "x \<tau> \<noteq> \<bottom>")
  apply(frule neq_commute[THEN iffD1])
  apply(simp_all add: OclSize_def OclExcluding_def OclCount_def 
                      minus_def ss_lifting')
  apply(induct_tac c, auto)
  done

lemma excluding_by_select:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow>
   ( (S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclExcluding> (x::('\<tau>,'\<alpha>)VAL)) \<tau> = 
   (\<MathOclSelectLeft> a : S |  (\<MathOclNot> (a \<MathOclStrictEq> x)) \<MathOclSelectRight>) \<tau>"
  apply(rule_tac X="S" in Sequence_induct_including)
  apply simp
  apply(simp,ocl_subst,simp)
  apply(simp_all add:if_not)

  (* apply (ocl_subst) breaks here with loose-bound-var error in type inference *)
  apply(rule trans, rule_tac A="X \<MathOclExcluding> x" in cp_charn)
  apply simp_all
  done


section{* Sum *}

text{* Like collect sum is specified as mapping to other \OCL functions \ldots  this is because
       for every type supporting ${plus,zero}$ the sum function must be available and it
       make no sense to add a new definition for every one of these *}

defs
   (* sum          :: "('\<tau>,'\<alpha>::{bot,plus,zero})Sequence) => ('\<tau>,'\<alpha>::{bot,plus,zero})VAL" *)
   OclSum_def      :
   "OclSum S       \<equiv> OclIterate (S::('\<tau>, Integer_0)Sequence) (\<lambda> x y. x + y) 0"


(* TODO: Define the general operator if possible... 
         or add otherwise all needed definitions

         How would one do that\<dots> if it is possible?

         Add definitions for all supported types
*)

thm OclSum_def
thm OclIterate_def

lemma sum_by_iterate: "OclSum S = OclIterate (S::('\<tau>, Integer_0)Sequence) (\<lambda> x y. x + y) 0"
  by(simp add: OclSum_def)

lemma OCL_cp_OclSum [simp,intro!]:
  "cp P \<Longrightarrow> cp(\<lambda> S. (OclSum (P S::('\<tau>,Integer_0)Sequence)))"
  by(simp add: OclSum_def)

lemmas OCL_undef_1_OclSum [simp] = f1_by_iterate_undef_1[OF sum_by_iterate]
lemmas sum_empty [simp] = f1_by_iterate_mtSequence[OF sum_by_iterate]
thm f1_by_iterate_including[OF sum_by_iterate, simplified]
lemmas sum_including [simp] = f1_by_iterate_opt_including[OF sum_by_iterate, simplified]
lemmas sum_union [simp] = f1_by_iterate_distrib_union_opt[OF sum_by_iterate, of "op +", simplified plus_AC, simplified]

thm OCL_undef_1_OclSum sum_empty sum_including sum_union


section{* One *}

defs
   (* one         ::"[('\<tau>,'c::bot Sequence) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean" *)
   OclOne_def     :
   "OclOne \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                 if \<forall>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)
                 then \<HolOclLiftLeft>size (List.filter (\<lambda>x. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>) \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>) = 1\<HolOclLiftRight> 
                 else \<bottom>))"

ocl_setup_op [OclOne]

text{* Note that contrary to many other \OCL functions it is \emph{not possible}
  to represent the one operation directly by an iterate. The reason for it is, that
  the internal state space that is used has cardinality three whereas the state space
  of the output values has cardinality two. *}

lemma one_by_size_of_select:
  "(OclOne (S::('\<tau>,'\<alpha>::bot)Sequence) P) =
   \<MathOclSizeLeft>(\<MathOclSelectLeft>x : S | P (x::('\<tau>,'\<alpha>)VAL)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrictEq> 1"
  apply(rule_tac X="S" in Sequence_sem_cases_ext)
  apply(auto simp: OclSelect_def OclSize_def OclStrictEq_def cp_by_cpify
                   OclOne_def One_ocl_int_def ss_lifting')
  done

(* NOTE: if the computional rules for \<MathOclStrictEq> would be present this would be a simp *)
lemma one_of_mtSequence [simp]:
  "OclOne (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau> Boolean)) = \<MathOclFalse>"
  apply(simp add: one_by_size_of_select)
  apply(simp add: OclStrongEq_def OclFalse_def ss_lifting'
                  One_ocl_int_def Zero_ocl_int_def)
  done

lemma one_singleton [simp]:
  "\<lbrakk>  \<tau> \<MathOclValid> \<MathOclIsDefined> (x::('\<tau>,'\<alpha>)VAL); cp P \<rbrakk> \<Longrightarrow>
   OclOne (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence \<MathOclIncluding> x) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau> Boolean)) \<tau>
   = P x \<tau>"
  apply(simp only: one_by_size_of_select)
  apply(rule trans)
  apply(rule_tac A="\<MathOclSelectLeft>x : \<MathOclMtSequence> \<MathOclIncluding> x  | (P x)\<MathOclSelectRight>" in  cp_charn)
  apply(simp_all)
  apply(rule_tac x="P x \<tau>" in boolean_cases_sem)
  apply(simp_all add: OclSize_def OclStrictEq_def OclIncluding_def OclMtSequence_def 
                  OclIf_def One_ocl_int_def neq_commute
                  ss_lifting' localValidDefined2sem)
  done

lemma one_union [simp]:
  "OclOne ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) P =
   (OclOne X P) \<MathOclSxor> (OclOne Y (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean))"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> X ", ocl_hypsubst, simp)
  apply(simp add: OclSxor_def ss_lifting')
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> Y", ocl_hypsubst, simp)

  apply(simp add: one_by_size_of_select)
  oops
(* NEEED rules for strict logic *)





section{* IsUnique *}

defs
   (* isUnique        ::"[('\<tau>,'\<alpha>::bot)Sequence), ('\<tau>,'\<alpha>)VAL => ('\<tau>,'c::bot)VAL] => '\<tau> Boolean" *)
   OclIsUnique_def  :
   "OclIsUnique \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                   if   \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)
                   then \<HolOclLiftLeft>distinct (map (\<lambda>x. P x)  \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)\<HolOclLiftRight> 
                   else \<bottom>))"

ocl_setup_op [OclIsUnique]

(* TODO: finish if needed\ldots seems to be quite hard work\ldots especially without
         conditions concerning P *)
lemma isUnique_by_forall:
  "OclIsUnique (S::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) =
   OclIterate S (\<lambda> x y. ((OclCount (OclCollectNested S P) (P x)) \<MathOclStrictEq> 1) \<MathOclSand> y) \<MathOclTrue>"
  oops


section{* Any *}

defs
   OclAny_def : 
  "OclAny     \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P. 
                 if   (\<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)) \<and>
                      (\<exists> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>)
                 then hd (List.filter (\<lambda> x. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>) \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)
                 else \<bottom>))"

ocl_setup_op [OclAny]

lemma any_by_first_of_select:
  "OclAny (S::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Boolean) =
   OclFirst ((OclAsSequence (OclSelect S P)::('\<tau>,'\<alpha>)Sequence))"
  apply(rule_tac X="S" in Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclAny_def OclSelect_def OclFirst_def 
                   cp_by_cpify ss_lifting' filter_empty_conv)
  done

lemma any_of_mtSequence[simp]:
  "OclAny (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau> Boolean)) = \<MathOclUndefined>"
  by(simp add: any_by_first_of_select)





section{* collectNested *}

defs
   (* collectNested    :: "[('\<tau>,'\<alpha>::bot)Sequence, ('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL] \<Rightarrow> ('\<tau>,'c::bot)Sequence" *)
   OclCollectNested_def:
   "OclCollectNested \<equiv> (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                        if \<forall>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. DEF (P x)
                         then Abs_Sequence_0 \<HolOclLiftLeft>map (\<lambda>x. P x) \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>\<HolOclLiftRight> 
                         else \<bottom>))"

ocl_setup_op [OclCollectNested]


lemma collectNested_by_iterate:
  "((OclCollectNested (X::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL\<Rightarrow>('\<tau>,'c)VAL))::('\<tau>,'c::bot)Sequence) =
   (X \<MathOclIterate>(x; y=\<MathOclMtSequence> |  y \<MathOclIncluding> (P x)))"
  apply(rule ext, rule iterate_universal_0, simp_all)
  apply(simp add: OclCollectNested_def cp_by_cpify lift0_def)
  apply(simp_all add: OclUndefined_def OclMtSequence_def OclCollectNested_def
                      localValidDefined2sem ss_lifting')
  apply(frule DEF_X_Sequence', clarify)
  apply(frule neq_commute[THEN iffD1]) 
  apply(rotate_tac 1, frule neq_commute[THEN iffD1])
  apply(auto simp: OclIncluding_def ss_lifting')
  apply(subst (asm) Abs_Sequence_0_inject_absurd22, auto)
  apply(subst Abs_Sequence_0_inverse_charn2, auto)
  done


subsection{* Properties derived from the correspondence to iterate *}

lemmas collectNested_of_mtSequence [simp] = f1_by_iterate_mtSequence[OF collectNested_by_iterate]
thm f1_by_iterate_including[OF collectNested_by_iterate, simplified]
lemmas collectNested_of_including [simp] = f1_by_iterate_opt_including[OF collectNested_by_iterate, simplified]
lemmas collectNested_of_union [simp] = f1_by_iterate_distrib_union_opt[OF collectNested_by_iterate, of "OclUnion", simplified]

thm OCL_undef_1_OclCollectNested collectNested_of_mtSequence collectNested_of_including collectNested_of_union


section{* Checking a general property of iterate... on foldl *}


text{* This lemma could actually be made stronger because the two assumptions
       @{text "undef_1_P"} and @{text "cp_P"} are not needed. 
       But they make the proof much simpler. *}
lemma iterate_of_collectNested:
  assumes undef_1_Q: "\<And> y. Q \<MathOclUndefined> y = \<MathOclUndefined>"
  and     undef_2_Q: "\<And> x. Q x \<MathOclUndefined>  = \<MathOclUndefined>"
  and         cp1_Q: "\<And> y. cp (\<lambda>x. Q x y)"
  and         cp2_Q: "\<And> x. cp (Q x)"
  and     undef_1_P: "P \<MathOclUndefined> = \<MathOclUndefined>"
  and          cp_P: "cp P"
  shows "OclIterate ((OclCollectNested (S::('\<tau>,'\<alpha>::bot)Sequence) 
               (P::('\<tau>,'\<alpha>)VAL\<Rightarrow>('\<tau>,'c::bot)VAL))::('\<tau>,'c)Sequence) Q A = 
         OclIterate S (\<lambda> x y. Q (P x) y) A"
  apply(insert undef_1_Q undef_2_Q cp1_Q cp2_Q undef_1_P cp_P)
  apply(rule ext)
  apply(rule iterate_universal, simp_all)
  apply(rule_tac P'="P" in cp_compose2)
  apply(simp_all add: collectNested_by_iterate iterate_opt_of_including )
  done

(*
lemma iterate_of_collectNested_tha_real_one:
  assumes undef_1_Q: "\<And> y. Q \<MathOclUndefined> y = \<MathOclUndefined>"
  and     undef_2_Q: "\<And> x. Q x \<MathOclUndefined>  = \<MathOclUndefined>"
  and         cp1_Q: "\<And> y. cp (\<lambda>x. Q x y)"
  and         cp2_Q: "\<And> x. cp (Q x)"
  shows "OclIterate (OclCollectNested (S::('\<tau>,'\<alpha>::bot)Sequence) P) Q A = 
         OclIterate S (\<lambda> x y. Q (P x) y) A"
  apply(insert undef_1_Q undef_2_Q cp1_Q cp2_Q)
  apply(rule ext)
  apply(rule_tac X="S" in Sequence_induct_including)
  apply(simp_all add: iterate_of_including_0)
  apply(rule trans, rule_tac x="(OclCollectNested X \<MathOclIncluding> a  P)" in cp_subst)
  apply(simp_all add: collectNested_by_iterate iterate_of_including_0)
  apply(simp add: sym[OF collectNested_by_iterate])
  apply(rule trans, rule sym)
  apply(rule_tac x="lift0 (OclCollectNested X P x) \<MathOclIncluding> (P (lift0 (a x)))" in cp_subst)
  apply(simp_all)
  apply(subst iterate_of_including_0)
(* NOTE: to go on with the proof at this point some stronger results about the definedness of the
         iterators are needed. Because the assumptions undef_1_P is not needed, because if collectNested
         is undefined, then there is a x in S such that P(x) is undefined. And this x would also make
         the composed iterate undefined. This proof step has to be carried out before the induction. Such
         that during the induction an assumption of the form (\<MathOclForAll> x \<in> S. \<MathOclIsDefined> x) holds.
*)
  oops
*)

subsection{* Definedness rules *}

lemma undefCollectNestedI:
  assumes exUndef1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (C::('\<tau>,('\<alpha>::bot))Sequence)"
  and     exUndef2: "\<tau> \<MathOclValid> \<MathOclIsUndefined> (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,'c::bot)VAL) x"
  and         cp_P: "cp P"
  shows             "\<tau> \<MathOclValid> \<MathOclIsUndefined> ((OclCollectNested C P)::('\<tau>,'c::bot)Sequence)"
  apply(insert exUndef1 exUndef2)
  apply(frule isDefined_if_valid)
  apply(subst (asm) OCL_is_defopt_OclIncludes)
  apply(clarify, simp add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Sequence', clarify)
  apply(simp add: localValid2sem OclIsDefined_def OclNot_def OclCollectNested_def
                  OclIncludes_def ss_lifting')
  apply(clarify)
  apply(erule_tac x="x \<tau>" in ballE)
  apply(insert cp_P, simp_all add: cp_by_cpify lift0_def)
  done



section{* Flatten *}

text{* This definition is currently not complete as it deals only with the type
  sequence of sequences of some type. There must be written a tactic that defines
  all needed instances of flatten on the different types. See the semester thesis
  for further information about the problems around flatten. 
*}

(*  TODO: How would one define the sort collection of collection?
*)
defs
   (* flatten    ::"[('\<tau>,'\<alpha>::bot Sequence_0)Sequence \<Rightarrow> ('\<tau>,'\<alpha>::bot)Sequence] *)

   OclFlatten_def: "OclFlatten \<equiv>  lift1(strictify (\<lambda> S. 
                         Abs_Sequence_0 \<HolOclLiftLeft>concat (map (\<lambda> x. \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight>) 
                                                     \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)\<HolOclLiftRight>))"

ocl_setup_op [OclFlatten]

(*


lemma OCL_cp_OclFlatten [simp,intro!]:
  "cp F \<Longrightarrow> cp (\<lambda> X. (OclFlatten ((F X)::('\<tau>,'\<alpha>::bot Sequence_0)Sequence))::('\<tau>,'\<alpha>::bot)Sequence)"
  by(simp add: OclFlatten_def cp_by_cpify)

*)

lemma OCL_is_defopt_OclFlatten [simp]:
  "(\<tau> \<MathOclValid>  \<MathOclIsDefined> ((\<MathOclFlattenLeft>X\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence)) = (\<tau> \<MathOclValid> \<MathOclIsDefined> (X::('\<tau>,'\<alpha> Sequence_0)Sequence))"
  apply(rule_tac X=X in Sequence_sem_cases, simp_all)
  apply(simp_all add: localValidDefined2sem OclFlatten_def ss_lifting')
  apply(clarify)
  apply(subst (asm) Abs_Sequence_0_inject_absurd22, auto)
  apply(rule_tac x=a in Abs_Sequence_0_cases_charn, simp_all)
  done


subsection{* Its computational characteristic by iterate *}

lemma flatten_by_iterate:
  "((OclFlatten (X::('\<tau>,'\<alpha>::bot Sequence_0)Sequence))::('\<tau>,'\<alpha>)Sequence) =
   (X \<MathOclIterate>(x; y=\<MathOclMtSequence> |  y \<MathOclUnion> ((x::('\<tau>,'\<alpha>)Sequence))))"
  apply(rule ext, rule iterate_universal_0, simp_all)
  apply(simp_all add: OclUndefined_def OclMtSequence_def OclFlatten_def
                      localValidDefined2sem ss_lifting')
  apply(frule DEF_X_Sequence', clarify)
  apply(frule neq_commute[THEN iffD1]) 
  apply(rotate_tac 1, frule neq_commute[THEN iffD1])
  apply(frule DEF_X_Sequence', clarify)
  apply(simp add: OclUnion_def OclIncluding_def ss_lifting')
  apply(subgoal_tac "\<bottom> \<notin> set (concat (map (\<lambda>(x\<Colon>'\<alpha> Sequence_0). \<HolOclDropLeft>(Rep_Sequence_0 x)\<HolOclDropRight>) c))") 
  apply(auto)
  apply(rule_tac x=a in Abs_Sequence_0_cases_charn, auto)
  done

subsection{* Deriving the rules that follow from the correspondence to iterate *}

lemmas flatten_of_mtSequence [simp] = f1_by_iterate_mtSequence[OF flatten_by_iterate]
lemmas flatten_of_including [simp] = f1_by_iterate_opt_including[OF flatten_by_iterate, simplified]
lemmas flatten_of_union [simp] = f1_by_iterate_distrib_union_opt[OF flatten_by_iterate, of "OclUnion", simplified]

thm OCL_undef_1_OclFlatten flatten_of_mtSequence flatten_of_including flatten_of_union


subsection{* functions applied to flatten *}


lemma f_of_flatten:
  assumes g_by_iterate: "g f X \<tau> = OclIterate X (\<lambda> x y. h y (f x)) A \<tau>"
  and          f_union: "\<And> X Y. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> \<MathOclIsDefined> Y \<rbrakk> \<Longrightarrow> f (X \<MathOclUnion> Y) \<tau> = h (f X) (f Y) \<tau>"
  and          f_empty: "f \<MathOclMtSequence> \<tau> = (A::('\<tau>,'c::bot)VAL) \<tau>"
  and          f_undef: "f \<MathOclUndefined> \<tau> = \<MathOclUndefined> \<tau>"
  and             cp_f: "cp f"
  and            cp1_h: "\<And> y. cp (\<lambda> x. h x y)"
  and            cp2_h: "\<And> x. cp (h x)"
  shows "f ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) \<tau> 
         = ((g f X)::('\<tau>,'c::bot)VAL) \<tau>"
  apply(insert cp1_h cp2_h cp_f, subst g_by_iterate)
  apply(rule_tac X=X and \<tau>=\<tau> in  Sequence_induct_including)
  apply(simp_all add: f_undef f_empty f_union)
  apply(subst iterate_of_including, simp_all)
  apply(rule_tac P'=f in cp_compose2, simp_all)
  apply(rule_tac A="(f \<MathOclFlattenLeft>X\<MathOclFlattenRight>)" in cp_charn, simp_all)
  apply(rule cp_eq, rule_tac P=f in cp_compose2, simp_all) 
  done

lemma isEmpty_flatten:
  "(((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha>::bot Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence) \<MathOclStrictEq> \<MathOclMtSequence>) = 
   (\<MathOclForAll> x \<in> X \<bullet> ((x::('\<tau>,'\<alpha>)Sequence) \<MathOclStrictEq> \<MathOclMtSequence>))"
  apply(rule ext, rule_tac X=X and h="OclAnd" in f_of_flatten)
  apply(simp_all add: forall_by_iterate OCL_logic_ACI
                       and_equiv_sand[symmetric] OCL_is_defopt_OclStrictEq
                 del: OCL_is_def_OclStrictEq)
  done

lemma includes_flatten:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>::bot)VAL) \<rbrakk> \<Longrightarrow>
   (a \<MathOclIncludes> ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence)) \<tau> = 
   (\<MathOclExists> Y \<in> X \<bullet> (a \<MathOclIncludes> (Y::('\<tau>,'\<alpha>)Sequence))) \<tau>"
  apply(rule_tac X=X and h=OclOr in f_of_flatten)
  apply(simp_all add: includes_of_union exists_by_iterate)
  apply(ocl_subst, simp add: OCL_logic_ACI)
  done

lemma excludes_flatten:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('\<tau>,'\<alpha>::bot)VAL) \<rbrakk> \<Longrightarrow>
   (a \<MathOclExcludes> ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence)) \<tau> = 
   (\<MathOclForAll> Y \<in> X \<bullet> (a \<MathOclExcludes> (Y::('\<tau>,'\<alpha>)Sequence))) \<tau>"
  apply(rule_tac X=X and h=OclAnd in f_of_flatten)
  apply(simp_all add: includes_of_union forall_by_iterate)
  apply(ocl_subst, simp add: OCL_logic_ACI)
(* NEED cp-aware rewriter and stronger logic rules *)
  apply(rule trans, rule_tac A="(a \<MathOclIncludes> (X \<MathOclUnion> Y ))" in cp_charn)
  apply(rule includes_of_union, simp_all)
  apply(simp add: OclAnd_def OclOr_def OclNot_def ss_lifting')
  done

text{* The current formulation is NOT true. But what is an elegant solution to it? *}
lemma includesAll_flatten:
  "(OclIncludesAll X ((\<MathOclFlattenLeft>(Y::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence))= 
   (\<MathOclForAll> Z \<in> Y \<bullet> (OclIncludesAll X (Z::('\<tau>,'\<alpha>::bot)Sequence)))"
  oops

text{* The current formulation is NOT true. But what is an elegant solution to it? *}
lemma excludesAll_flatten:
  "(OclExcludesAll X ((\<MathOclFlattenLeft>(Y::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence))= 
   (\<MathOclForAll> Z \<in> Y \<bullet> (OclExcludesAll X (Z::('\<tau>,'\<alpha>::bot)Sequence)))"
  oops

lemma forall_flatten:
  "(\<MathOclForAll> y \<in> ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence) \<bullet> P y) = 
   (OclForAll X (\<lambda> Y::('\<tau>,'\<alpha>)Sequence. (\<MathOclForAll> y \<in> Y \<bullet> P (y::('\<tau>,'\<alpha>::bot)VAL))))"
  apply(rule ext, rule_tac X=X and h="OclAnd" in f_of_flatten)
  apply(simp_all, simp_all add: forall_by_iterate OCL_logic_ACI)
  done

lemma exists_flatten:
  "(\<MathOclExists> y \<in> ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>)Sequence) \<bullet> P y) = 
   (OclExists X (\<lambda> Y::('\<tau>,'\<alpha>)Sequence. (\<MathOclExists> y \<in> Y \<bullet> P (y::('\<tau>,'\<alpha>::bot)VAL))))"
  by(simp add: OclExists_def forall_flatten)

lemma select_flatten:
  "OclSelect ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) (P::(('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau> Boolean))) =
   \<MathOclFlattenLeft>((OclCollectNested X (\<lambda> (Y::('\<tau>,'\<alpha>)Sequence). OclSelect Y P))::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>"
  apply(rule ext)
  apply(rule_tac X=X and \<tau>=x in Sequence_induct_including, simp_all)
(* TODO: add select_of_union to the simpset *)
  apply(rule trans)
  apply(rule_tac A="OclSelect \<MathOclFlattenLeft>X\<MathOclFlattenRight> P" and \<tau>=x in cp_charn) 
  apply(simp_all)
  done

lemma reject_flatten:
  "OclReject ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) (P::(('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau> Boolean))) =
   \<MathOclFlattenLeft>((OclCollectNested X (\<lambda> (Y::('\<tau>,'\<alpha>)Sequence). OclReject Y P))::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>"
  by(simp add: reject_by_select select_flatten)

text{* By rewriting any to \verb|first_of_select it| is clear what the result of
  any on a flattened sequence will be. *}

 
text{* Currently there is no simplification rule for the equalities. 
  I just can't see a way how to represent the iterated comparison of 
a prefix of both sequences with its intertwined concatenations. *}

text{* By rewriting one to \verb|size_of_select| it is clear what the result of
  one on a flattened sequence will be. *}

(* TODO: isUnique *)
lemma size_flatten:
  "\<MathOclSizeLeft>((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence)\<MathOclSizeRight> = 
   OclSum ((OclCollectNested X (OclSize::('\<tau>,'\<alpha>)Sequence\<Rightarrow>'\<tau> Integer))::('\<tau>,Integer_0)Sequence)"
  apply(rule ext)
  apply(rule_tac X=X and \<tau>=x in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac A="\<MathOclSizeLeft>\<MathOclFlattenLeft>X\<MathOclFlattenRight>\<MathOclSizeRight>" and \<tau> =x in cp_charn)
  apply(simp_all add: plus_AC)
  done

lemma count_flatten:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> a \<rbrakk> \<Longrightarrow>
   (((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclCount> (a::('\<tau>,'\<alpha>)VAL)) \<tau> = 
   OclSum ((OclCollectNested X (\<lambda> (X::('\<tau>,'\<alpha>)Sequence). X \<MathOclCount> a))::('\<tau>,Integer_0)Sequence) \<tau>"
  apply(rule_tac X=X and \<tau>=\<tau> in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac A="(\<MathOclFlattenLeft>X\<MathOclFlattenRight> \<MathOclCount> a )" and \<tau>=\<tau> in cp_charn) 
  apply(simp_all add: plus_AC)
  done

lemma sum_flatten:
 "OclSum ((\<MathOclFlattenLeft>(X::('\<tau>,Integer_0 Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,Integer_0)Sequence) = 
  OclSum ((OclCollectNested X (OclSum::('\<tau>,Integer_0)Sequence\<Rightarrow>'\<tau> Integer))::('\<tau>,Integer_0)Sequence)"
  apply(rule ext)
  apply(rule_tac X=X and \<tau>=x in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac A="\<MathOclSum> \<MathOclFlattenLeft>X\<MathOclFlattenRight>" and \<tau>=x in cp_charn) 
  apply(simp_all add: plus_AC)
  done

lemma excluding_flatten:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow>
   OclExcluding ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) (x::('\<tau>,'\<alpha>)VAL) \<tau> =
   \<MathOclFlattenLeft>((OclCollectNested X (\<lambda> (Y::('\<tau>,'\<alpha>)Sequence). OclExcluding Y x))::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight> \<tau>"
  apply(rule_tac X=X and h="OclUnion" in f_of_flatten)
  apply(simp_all)
  apply(ocl_subst)
  apply(simp add: flatten_by_iterate iterate_of_collectNested)
  done

lemma collectNested_flatten:
  "((OclCollectNested ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence) 
                      (P::('\<tau>,'\<alpha>)VAL\<Rightarrow>('\<tau>,'c::bot)VAL))::('\<tau>,'c)Sequence) =
   \<MathOclFlattenLeft>((OclCollectNested X (\<lambda> (Y::('\<tau>,'\<alpha>)Sequence). (OclCollectNested Y P)::('\<tau>,'c)Sequence))::('\<tau>,'c Sequence_0)Sequence)\<MathOclFlattenRight>"
  apply(rule ext)
  apply(rule_tac X=X and \<tau>=x in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac A="OclCollectNested \<MathOclFlattenLeft>X\<MathOclFlattenRight> P" and \<tau>=x in cp_charn) 
  apply(simp_all)
  done

lemma test':
  "\<MathOclSizeLeft>((\<MathOclFlattenLeft>
       ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0 Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha> Sequence_0)Sequence)
      \<MathOclFlattenRight>)::(('\<tau>,'\<alpha>::bot)Sequence))\<MathOclSizeRight> = h"
  apply(simp add: size_flatten collectNested_flatten sum_flatten)
  oops

subsubsection{* Sequence specific functions *}

text{* These are: \inlineocl{first}, \inlineocl{last}, \inlineocl{at}, 
       \inlineocl{insertAt}, \inlineocl{indexOf}, \inlineocl{subSequence} *}

(* MOVE *)
lemma last_empty [simp]:
  "OclLast (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) = (\<MathOclUndefined>::('\<tau>,'\<alpha>)VAL)"
  by(simp add: last_at_UC)

lemma last_including [simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S\<rbrakk> \<Longrightarrow>
   OclLast ((S::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncluding> (x::('\<tau>,'\<alpha>)VAL)) \<tau> =
   ((x \<tau>)::'\<alpha>)"
  apply(rule_tac X="S" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(simp add: OclIncluding_def OclLast_def 
                  ss_lifting' neq_commute)
  done

lemma last_union:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> X; \<tau> \<MathOclValid> (Y `<>` \<MathOclMtSequence>) \<rbrakk> \<Longrightarrow> 
  OclLast ((X::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclUnion> Y) \<tau> = ((OclLast Y)::('\<tau>,'\<alpha>)VAL) \<tau> "
  apply(rule_tac X=X in Sequence_sem_cases)
  apply(rule_tac X=Y in Sequence_sem_cases, simp_all)
  apply(auto simp: localValid2sem OclNot_def OclStrictEq_def OclMtSequence_def
                  OclUnion_def OclLast_def OclIsDefined_def ss_lifting')
  done

(* MOVE *)
lemma ForAllD:
  assumes allP : "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> C \<bullet> P(x\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot))"
  and     cpP     : "cp P"
  shows   "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> (C::('\<tau>,('\<alpha>::bot))Sequence) \<Longrightarrow> \<tau> \<MathOclValid> P x" 
  apply (insert allP cpP)
  apply (simp add: localValid2sem OclForAll_def OclNot_def OclIncludes_def
                   ss_lifting')
  apply (simp add:   strictify_def
              split: split_if_asm)
  apply (erule_tac x="x \<tau>" in ballE)
  apply (simp_all add: cp_by_cpify lift0_def)
  done

lemma isdefSelectI:
  assumes localDef: "\<tau> \<MathOclValid> \<MathOclForAll> x \<in> C \<bullet> (\<MathOclIsDefined>(P (x::('\<tau>,'\<alpha>::bot)VAL)))"
  and         cp_P: "cp P"
  shows "\<tau> \<MathOclValid> \<MathOclIsDefined> (OclSelect (C::('\<tau>,'\<alpha>::bot)Sequence) P)"
  apply(insert localDef)
  apply(erule rev_mp)
  apply(rule_tac X=C in Sequence_induct_including)
  apply(simp_all, clarify)
  oops

lemma last_flatten:
  "((OclLast ((\<MathOclFlattenLeft>(X::('\<tau>,'\<alpha> Sequence_0)Sequence)\<MathOclFlattenRight>)::('\<tau>,'\<alpha>::bot)Sequence))::('\<tau>,'\<alpha>)VAL) =
   OclLast ((OclLast (OclSelect X (\<lambda> Y::('\<tau>,'\<alpha>)Sequence. Y `<>` \<MathOclMtSequence>)))::('\<tau>,'\<alpha>)Sequence)"
  apply(rule ext)
  apply(rule_tac X=X in Sequence_induct_including, simp_all)
  apply(case_tac "x \<MathOclValid> (a `<>` \<MathOclMtSequence>)")
(* TODO: ocl_subst does NOT work here :S *)
  apply(rule trans[symmetric])
  apply(rule_tac A="(a `<>` \<MathOclMtSequence>)" in cp_charn)

  apply(rotate_tac -1, subst (asm) is_TRUE_charn_local[symmetric])
  apply(simp_all)
  apply(simp add: last_union)
  apply(rule_tac B=a in cp_charn)
  apply(rule last_including)
  apply(simp_all)
 

  prefer 2
  apply(simp add: if_not)
  apply(subst (asm) localValidNot2not[symmetric])
  apply(simp add: weak_prop_LJE)
  apply(simp)
  apply(rotate_tac -1, ocl_subst)
  apply(simp add: weak_prop_LJE)
  apply(rule_tac  t="OclLast ((\<MathOclFlattenLeft>X\<MathOclFlattenRight>) \<MathOclUnion> a ) x" and 
                  s="OclLast (\<MathOclFlattenLeft>X\<MathOclFlattenRight>) x" in subst)

  apply(rotate_tac 1, thin_tac "?P")
  apply(rule trans[symmetric])
  apply(rule_tac A=a and B="\<MathOclMtSequence>" in cp_charn)
  apply(simp add: OclLast_def OclUnion_def localValid2sem OclStrictEq_def OclFlatten_def
                  OclIsDefined_def OclMtSequence_def ss_lifting')
  apply(simp_all)
(* WAITING FOR Definedness rules over iterateors, i.e., a cp-aware
               simplifier *)
  oops

section{* Collect *}
text{* Note: Currently collectNested is only defined here but it should actually
  be polymorphically defined in the theory \theory{OCL\_Collection}. Furthermore the
  collect operation could be defined right there because it should be defined on the \OCL
  level to avoid having to define collect over all different types like we have to do it
  for flatten. *}


subsection{* Defining collect based on collectNested according to the standard *}

defs
   (* collect      ::"[('\<tau>,'\<alpha>::bot)Sequence), ('\<tau>,'\<alpha>)VAL => ('\<tau>,'c::bot)VAL] => ('\<tau>,'c)Sequence" *)
   OclCollect_def   :
   "(OclCollect::[('\<tau>,'\<alpha>::bot)Sequence, (('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'c::bot)Sequence)] \<Rightarrow> ('\<tau>,'c::bot)Sequence) S P
                   \<equiv> \<MathOclFlattenLeft> ((OclCollectNested S P)::('\<tau>,'c Sequence_0)Sequence) \<MathOclFlattenRight>"


   OclCollect_Integer_def :
   "(OclCollect::[('\<tau>,'\<alpha>::bot)Sequence, (('\<tau>,'\<alpha>)VAL \<Rightarrow> '\<tau> Integer)] \<Rightarrow> ('\<tau>,Integer_0)Sequence) S P
                   \<equiv> (OclCollectNested S P)" 

(* TODO: 1. define flatten and collectNested on all types 
         2. remove assumptions when they are removed from the 
            iterate_of_collectNested lemma
         3. introduce alternative definition for ocl_setup_op
*)

lemma collect_by_iterate:
  "\<lbrakk> P \<MathOclUndefined> = \<MathOclUndefined>; cp P \<rbrakk> \<Longrightarrow>
   ((OclCollect (X::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'c::bot)Sequence))::('\<tau>,'c)Sequence) = 
   (X \<MathOclIterate>(x;y=\<MathOclMtSequence> | (y \<MathOclUnion> (P x) )))"
  by(simp add: OclCollect_def flatten_by_iterate iterate_of_collectNested)


lemma OCL_cp_OclCollect[simp,intro!]:
  "\<lbrakk> \<And> y. cp (\<lambda> x. P x y); cp S \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. (OclCollect ((S X)::('\<tau>,'\<alpha>::bot)Sequence) 
                         ((P X)::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'c)Sequence))::('\<tau>,'c::bot)Sequence)"
  by(simp add: OclCollect_def)

lemma OCL_undef_1_OclCollect[simp]:
  "OclCollect (\<MathOclUndefined>::('\<tau>,'\<alpha>::bot)Sequence) (P::('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'c)Sequence) =
   (\<MathOclUndefined>::('\<tau>,'c::bot)Sequence)"
  by(simp add: OclCollect_def)

(* TODO: undef_2_rules for the iterators *)

(*************************************************************************************

(* TODO: move to Set theory or to conversion theory


subsection{* Helper theorems about the relations of the Set and Sequence datatypes *}

lemma set_Sequence_eq_Set_set: 
  "(\<bottom> \<notin> set x) \<Longrightarrow>
   set \<HolOclDropLeft>(Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>))\<HolOclDropRight> = 
   \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set x \<HolOclLiftRight>))\<HolOclDropRight>  "
  by(simp)

lemma DEF_Set_if_DEF_Sequence:
  "(DEF X) \<Longrightarrow> (DEF (Abs_Set_0 \<HolOclLiftLeft>(set \<HolOclDropLeft>(Rep_Sequence_0 X)\<HolOclDropRight>)\<HolOclLiftRight>))"
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto)
thm DEF_Abs_Set
  apply(erule DEF_Abs_Set)
  done

lemma in_over_RepAbs:
"\<bottom> \<notin> S \<Longrightarrow> (x \<in> \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>S\<HolOclLiftRight>))\<HolOclDropRight>)= (x \<in> S)"
by(simp)


subsection{* Basic properties of OclAsSet *}

 
lemma OCL_is_defopt_OclAsSet: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('\<tau>,'\<alpha>::bot)Sequence) ->asSet())::(('\<tau>, '\<alpha>::bot)Set))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI) prefer 2
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp add: OclAsSet_def strictify_def lift1_def DEF_def)
  apply(fold DEF_def)
  apply(erule DEF_Set_if_DEF_Sequence)
  apply(auto simp: OclAsSet_def strictify_def lift1_def DEF_def)
done

lemma OCL_is_def_OclAsSet:
  "(\<MathOclIsDefined> (((self::('\<tau>,'\<alpha>::bot)Sequence) ->asSet())::(('\<tau>, '\<alpha>::bot)Set))) = (\<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule ext)
  apply(rule arg_cong) back
  apply(subgoal_tac "((St \<MathOclValid> (\<MathOclIsDefined> \<MathOclAsSet> self)) = (St \<MathOclValid> (\<MathOclIsDefined> self)))")
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule OCL_is_defopt_OclAsSet)
  done

lemma OclAsSet_finite:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('\<tau>,'\<alpha>::bot)Sequence)) =
   (\<tau> \<MathOclValid> \<MathOclIsDefined>(((OclAsSet S)::('\<tau>,'\<alpha>::bot)Set)->size()))"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclAsSet_def OCL_Set.OclSize_def
                      strictify_def lift1_def DEF_def)
  apply(auto)
  done

lemma OclAsSet_mtSequence_mtSet_conv[simp]:
  "((OclAsSet (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence))) = (\<MathOclMtSet>::('\<tau>,'\<alpha>::bot)Set)"
  apply(insert Abs_Sequence_Mt_DEF)
  apply(simp add: OclAsSet_def OclMtSequence_def OclMtSet_def
                  lift0_def lift1_def strictify_def DEF_def)
  done



section{* Relating the operations for which it is possible to the corresponding set operations *}
 
subsection{* Relating includes and excludes on Sequences to their counterparts on Sets *}

lemma includes_seq2set:
  "(x::('\<tau>,'\<alpha>)VAL) \<MathOclIncludes> (S::('\<tau>,'\<alpha>::bot)Sequence) =
   x \<MathOclIncludes> ((OclAsSet S)::('\<tau>,'\<alpha>::bot)Set)" 
  apply(rule ext)
  apply(case_tac "DEF (S xa)")
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIncludes_def OCL_Set.OclIncludes_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma excludes_seq2set:
  "(x::('\<tau>,'\<alpha>)VAL) \<MathOclExcludes> (S::('\<tau>,'\<alpha>::bot)Sequence) =
   x \<MathOclExcludes> ((OclAsSet S)::('\<tau>,'\<alpha>::bot)Set)"
  by(simp add: includes_seq2set)


lemma isEmpty_seq2set:
  "OclIsEmpty(S::('\<tau>, '\<alpha>::bot) Sequence) = OclIsEmpty ((OclAsSet S)::('\<tau>,('\<alpha>::bot))Set)"
  apply(rule ext)
  apply(case_tac "DEF (S x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIsEmpty_def OCL_Set.OclIsEmpty_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma isNotEmpty_seq2set:
  "OclNotEmpty(S::('\<tau>, '\<alpha>::bot) Sequence) = OclNotEmpty ((OclAsSet S)::('\<tau>,('\<alpha>::bot))Set)"
  by(simp only: notEmpty_eq_not_isEmpty OCL_Set.not_empty isEmpty_seq2set)

lemma forall_by_all_includes:
  "(\<tau> \<MathOclValid> \<MathOclForAll> x \<in> (S :: (('\<tau>,('\<alpha>::bot))Sequence)) \<bullet> P (x::('\<tau>,'\<alpha>::bot)VAL)) = 
   (\<forall> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<MathOclImplies> (P::((('\<tau>,'\<alpha>::bot)VAL)\<Rightarrow>('\<tau>,Boolean_0)VAL)) (x::('\<tau>,'\<alpha>::bot)VAL))"
  apply(case_tac "DEF (S \<tau>)")
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclLocalValid_def OclTrue_def OclImplies_def
                  OclForAll_def OclIncludes_def
                  strictify_def lift0_def lift1_def lift2_def DEF_def)
  oops

lemma forall_seq2set: 
  "(\<MathOclForAll> x \<in> (S :: (('\<tau>,('\<alpha>::bot))Sequence)) \<bullet> P (x\<Colon>'\<tau> \<Rightarrow> '\<alpha>)) = 
   (\<MathOclForAll> x \<in> ((OclAsSet S)::('\<tau>,('\<alpha>::bot))Set)  \<bullet> P x)"
  apply(simp add: OclForAll_def OCL_Set.OclForAll_def)
  apply(rule ext)
  apply(subgoal_tac "(DEF (S \<tau>)) = (DEF (((\<MathOclAsSet> S)::(('\<tau>,'\<alpha>::bot)Set)) \<tau>))")
  apply(case_tac "(DEF (S \<tau>))")
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(erule exE)
  apply(subgoal_tac "\<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set c \<HolOclLiftRight>))\<HolOclDropRight> = set c")
  apply(simp_all add: OclAsSet_def lift0_def lift1_def strictify_def in_over_RepAbs)
  apply(auto simp: in_over_RepAbs)
  apply(frule DEF_Set_if_DEF_Sequence)
  apply(simp_all add: DEF_def)
  done

lemma exists_seq2set:
  "(\<MathOclExists> x \<in> (S :: (('\<tau>,('\<alpha>::bot))Sequence)) \<bullet> P (x\<Colon>'\<tau> \<Rightarrow> '\<alpha>)) = 
   (\<MathOclExists> x \<in> ((OclAsSet S)::('\<tau>,('\<alpha>::bot))Set)  \<bullet> P x)"
  by(simp add: OclExists_def forall_seq2set)

lemma includesAll_seq2set:
  "OclIncludesAll (S1::('\<tau>, '\<alpha>::bot)Sequence) (S2::('\<tau>, '\<alpha>::bot)Sequence) =
   OclIncludesAll ((OclAsSet S1)::('\<tau>, '\<alpha>::bot)Set) ((OclAsSet S2)::('\<tau>, '\<alpha>::bot)Set)"
  apply(rule ext)
  apply(case_tac "DEF (S1 x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(case_tac "DEF (S2 x)", rotate_tac -1)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIncludesAll_def OCL_Set.OclIncludesAll_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma excludesAll_seq2set:
  "OclExcludesAll (S1::('\<tau>, '\<alpha>::bot)Sequence) (S2::('\<tau>, '\<alpha>::bot)Sequence) =
   OclExcludesAll ((OclAsSet S1)::('\<tau>, '\<alpha>::bot)Set) ((OclAsSet S2)::('\<tau>, '\<alpha>::bot)Set)"
  apply(rule ext)
  apply(case_tac "DEF (S1 x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(case_tac "DEF (S2 x)", rotate_tac -1)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclExcludesAll_def OCL_Set.OclExcludesAll_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

(* TODO: discuss the definition of OclAny with bu.

         It is not clear why Any should be strict and 
         the case of no element with property P is undefined is missing
         in the definition in the OCL_Set theory.

lemma any_seq2set:
  "(OclAny (S :: (('\<tau>,('\<alpha>::bot))Sequence)) (P::('\<tau>,'\<alpha>::bot)VAL \<Rightarrow> ('\<tau>,Boolean_0)VAL)) = 
   (OclAny ((OclAsSet S)::('\<tau>,('\<alpha>::bot))Set) P)" 
  apply(simp add: OclAny_def OCL_Set.OclAny_def)
  apply(rule ext)
  apply(subgoal_tac "(DEF (S \<tau>)) = (DEF (((\<MathOclAsSet> S)::(('\<tau>,'\<alpha>::bot)Set)) \<tau>))")
  apply(case_tac "(DEF (S \<tau>))")
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(erule exE)
  apply(subgoal_tac "\<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set c \<HolOclLiftRight>))\<HolOclDropRight> = set c")
  apply(simp_all add: OclAsSet_def lift0_def lift1_def strictify_def in_over_RepAbs)
  apply(auto simp: in_over_RepAbs)
  apply(frule DEF_Set_if_DEF_Sequence)
  apply(simp_all add: DEF_def)
  done

*)


section{* Characteristic Set of Sequences (parametric): *}

(* TODO: Shouldn't one take only the permutations of the list and all its sublists? *)

constdefs
   Sequence:: "('st, ('\<alpha>::bot) Sequence_0) VAL \<Rightarrow> ('st, '\<alpha> Sequence_0 Set_0) VAL"
  "Sequence \<equiv> lift1(strictify (\<lambda>X.
                   Abs_Set_0 \<HolOclLiftLeft>(Abs_Sequence_0 o lift) ` (lists (set  \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>))\<HolOclLiftRight> ))"

ocl_setup_op [Sequence]


lemma OCL_is_defopt_Sequence [simp]:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (Sequence X)) = (\<tau> \<MathOclValid> \<MathOclIsDefined> X)"
  apply(simp add: Sequence_def localValidDefined2sem
                  o_def image_def ss_lifting')
  apply(rule impI)
  apply(frule DEF_X_Sequence')
  apply(auto)
  done 

lemma mtSequence_in_Sequence :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('\<tau>, '\<alpha>::bot)Sequence) \<Longrightarrow> 
    \<tau> \<MathOclValid> (\<MathOclMtSequence>::('\<tau>,'\<alpha>::bot)Sequence) \<MathOclIncludes> (Sequence  S)"
  apply(frule Sequence_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Sequence_def OCL_Set.OclIncludes_def OclMtSequence_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Sequence_def UU_Set_def)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Sequence_def)
  done


lemma X_in_Sequence_X :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('\<tau>, '\<alpha>::bot)Sequence) \<Longrightarrow> 
    \<tau> \<MathOclValid> S \<MathOclIncludes> (Sequence S)"
  apply(frule Sequence_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Sequence_def OCL_Set.OclIncludes_def OclMtSequence_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Sequence_def UU_Set_def)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Sequence_def)
  done

*)
*)
end
   
